/**
***************************************************************************************
*  Copyright (C) 2008, Techwell
*  All rights reserved.
*
*  @file 		TWSDK.H
*  @brief		Techwell TW68XX capture sdk interface
*
*  @version		1.0.0.1
*  @author		jack
*  @date		2008/10/20
***************************************************************************************
*/

#ifndef __TWSDK_H_
#define __TWSDK_H_

#if(defined(TW68XX_EXPORTS) || defined(TW68XXC_EXPORTS))
#ifndef TW68XX_DEF
#define TW68XX_API  extern "C"__declspec(dllexport)
#else
#define TW68XX_API
#endif
#else
#define TW68XX_API  extern "C"__declspec(dllimport)
#endif

//error code
#define TWERR_SUCCESS			0x00000000
#define TWERR_FAILED			(ULONG)-1
#define TWERR_INVALID_PARAM		(ULONG)-2
#define TWERR_EXCEPTION			(ULONG)-3
#define TWERR_NOT_SUPPORT		(ULONG)-4
#define TWERR_NO_DEVICE			(ULONG)-5
#define TWERR_START_DEVICE		(ULONG)-6
#define TWERR_OUTOF_MEMORY		(ULONG)-7
#define TWERR_CREATE_THREAD		(ULONG)-8
#define TWERR_CREATE_DEVICE		(ULONG)-9
#define TWERR_TIMEOUT			(ULONG)-10

#define DEVICEID(hwid)		  ((hwid)>>16)
#define VENDERID(hwid)		  ((hwid)&0xFFFF)

#define COMPOSITE1_MASK		  (1<<0)
#define COMPOSITE2_MASK		  (1<<1)
#define COMPOSITE3_MASK		  (1<<2)
#define COMPOSITE4_MASK		  (1<<3)
#define INPUT_MASK(input)	  (1<<(input))
#define INPUT_ALL			  0x0F

typedef struct {
	int	    nBoardId;
	int		nSlot;
	int		nDeviceCount;
	int		nFirstDevice;
    ULONG	reserved1;
	ULONG	reserved2;
	ULONG	reserved3;
	ULONG	reserved4;
}TBoardDesc;

typedef struct {
	int		nDeviceId;
	int		nBoardId;
	int		nBus;
	ULONG	HardwareId;
	int		DeviceNumber;
	int		Function;
	int		nSubDevice;
	int		nFirstSubDevice;
	ULONG	reserved1;
	ULONG	reserved2;
	ULONG	reserved3;
	ULONG	reserved4;
}TDeviceDesc;

typedef struct {
	int		nDeviceId;
	int		nBoardId;
	int		nAudio;
	int		iSwitch;
	int		nChannels;
	int		nFirstChannel;
	BOOL	bField2;
	ULONG	VideoInputMask;
	ULONG	reserved1;
	ULONG	reserved2;
	ULONG	reserved3;
	ULONG	reserved4;
}TSubDeviceDesc;

typedef struct {
	int		nBoardId;
	int		nDeviceId;
	int		nSubDeviceId;
	int		nMaxFrameRate;
	int		nAudio;
	ULONG	reserved1;
	ULONG	reserved2;
	ULONG	reserved3;
	ULONG	reserved4;
}TChannelDesc;

typedef enum {
	viCompositeVideo1,
	viCompositeVideo2,
	viCompositeVideo3,
	viCompositeVideo4,
	viCompositeMax,
}TVideoInput;

typedef enum {
	vsStandardNone   = 0x80000000,
	vsStandardNTSC   = 0x00000001,
	vsStandardPAL    = 0x00000002,
	vsStandardSECAM  = 0x00000004,
}TVideoStandard;

typedef enum {
	vpfRGB15Alpha  = 0x00000004,
	vpfRGB16       = 0x00000008,
	vpfRGB24       = 0x00000010,
	vpfRGB24Alpha  = 0x00000020,

	vpfYUV420Planar         = 0x00000040,
	vpfYUV411Planar         = 0x00000100,
	vpfYUV420Interspersed   = 0x00000200,
	vpfYUV422Interspersed   = 0x00000400,
	vpfYUV411Interspersed   = 0x00000800,
}TPixelFormat;
		
typedef enum {
	vbtCIF,			//support all channels capture CIF at same time
	vbtHalfD1,		//support all channels capture half D1 at same time
	vbtD1,			//support all channels capture D1 at same time
	vbtD1_preview,	//support only one channel capture D1 at same time, others capture half D1
}TVideoBufferType;

typedef enum {
	vpBrightness,
	vpContrast,
	vpSaturation,
	vpHue,
	vpSharpness,
	
	vpSignal		= 0x00001000,
	vpBufferType,	//described by TVideoBufferType
	vpOptimizedCIF,	//enable DMA performance optimization over PCI when capture up to 10 CIF
	vpOptimizedCIF1,//enable DMA performance optimization over PCI when capture up to 10 CIF

	vpStandard		= 0x00002000,
	vpPixelFormat,
	vpFrameRate,
	vpWidth,
	vpHeight,
	vpField,

	vpLeft,			//for TW686x only
	vpRight,		//for TW686x only
	vpHDelay,
	vpVDelay,

	vpPixelFormat2	= 0x00003000,
	vpFrameRate2,
	vpWidth2,
	vpHeight2,
}TVideoPara;

typedef enum {
	apSamplesPerSecond,		//sample rate
	apBitsPerSample,	
	apChannel,				//mono, stereo
	apVolume,

	apPacketLength = 0x00001000,	//audio packet byte length
}TAudioPara;

TW68XX_API BOOL   __stdcall TwInitSDK();
TW68XX_API VOID   __stdcall TwUninitSDK();
TW68XX_API ULONG  __stdcall TwGetLastError();

TW68XX_API int    __stdcall TwGetBoardCount();
TW68XX_API HANDLE __stdcall TwGetBoardHandle(int nBoard);
TW68XX_API BOOL   __stdcall TwGetBoardDetail(HANDLE hBoard, TBoardDesc *pBoardDesc);
TW68XX_API BOOL   __stdcall TwGetDevicePciConfigSpaceEx(HANDLE hBoard, int nOffset, int nBytes, BYTE* pData);
TW68XX_API BOOL   __stdcall TwSetDevicePciConfigSpaceEx(HANDLE hBoard, int nOffset, int nBytes, BYTE* pData);

TW68XX_API int    __stdcall TwGetDeviceCount();
TW68XX_API HANDLE __stdcall TwGetDeviceHandle(int nDevice);
TW68XX_API BOOL   __stdcall TwGetDeviceDetail(HANDLE hDevice, TDeviceDesc *pDevDesc);

TW68XX_API HANDLE __stdcall TwGetDeviceAudioHandle(int nDevice);
TW68XX_API BOOL   __stdcall TwGetDeviceRegBYTE(HANDLE hDevice, ULONG uReg, BYTE* value);
TW68XX_API BOOL   __stdcall TwGetDeviceRegWORD(HANDLE hDevice, ULONG uReg, USHORT* value);
TW68XX_API BOOL   __stdcall TwGetDeviceRegDWORD(HANDLE hDevice, ULONG uReg, ULONG* value);
TW68XX_API BOOL   __stdcall TwSetDeviceRegBYTE(HANDLE hDevice, ULONG uReg, BYTE value);
TW68XX_API BOOL   __stdcall TwSetDeviceRegWORD(HANDLE hDevice, ULONG uReg, USHORT value);
TW68XX_API BOOL   __stdcall TwSetDeviceRegDWORD(HANDLE hDevice, ULONG uReg, ULONG value);
TW68XX_API BOOL   __stdcall TwGetDevicePciConfigSpace(HANDLE hDevice, int nOffset, int nBytes, BYTE* pData);
TW68XX_API BOOL   __stdcall TwSetDevicePciConfigSpace(HANDLE hDevice, int nOffset, int nBytes, BYTE* pData);

TW68XX_API HANDLE __stdcall TwI2COpen(HANDLE hDevice);
TW68XX_API VOID   __stdcall TwI2CClose(HANDLE hI2C);
TW68XX_API BOOL   __stdcall TwI2CWrite(HANDLE hI2C, int nBytes, BYTE* cValue);
TW68XX_API BOOL   __stdcall TwI2CRead(HANDLE hI2C, int nWriteBytes, BYTE* cValue, int nReadBytes, BYTE* pValue);
TW68XX_API BOOL   __stdcall TwI2CWriteNoStop(HANDLE hI2C, int nBytes, BYTE* cValue);
TW68XX_API BOOL   __stdcall TwI2CLock(HANDLE hI2C, ULONG uTimeout);
TW68XX_API VOID   __stdcall TwI2CUnlock(HANDLE hI2C);
TW68XX_API BOOL   __stdcall TwI2CSetSpeed(HANDLE hI2C, ULONG uSpeed);
TW68XX_API HANDLE __stdcall TwI2CCreate(HANDLE hDevice, int sclGpio, int sdaGpio, int nSpeed);
TW68XX_API VOID   __stdcall TwI2CFree(HANDLE hI2C);

TW68XX_API HANDLE __stdcall TwGPIOOpen(HANDLE hDevice);
TW68XX_API VOID   __stdcall TwGPIOClose(HANDLE hGpio);
TW68XX_API BOOL   __stdcall TwGPIOEnableOutput(HANDLE hGpio, int nGpio, int nBits, BOOL bEnable);
TW68XX_API BOOL   __stdcall TwGPIOSetBits(HANDLE hGpio, int nGpio, int nBits, ULONG* ulValue);
TW68XX_API BOOL   __stdcall TwGPIOGetBits(HANDLE hGpio, int nGpio, int nBits, ULONG* pValue);

TW68XX_API int    __stdcall TwGetSubDeviceCount();
TW68XX_API HANDLE __stdcall TwGetSubDeviceHandle(int nSubDevice);
TW68XX_API BOOL   __stdcall TwGetSubDeviceDetail(HANDLE hSubDevice, TSubDeviceDesc *pSubDevDesc);
TW68XX_API BOOL   __stdcall TwEnableVideoInput(HANDLE hSubDevice, TVideoInput vInput, BOOL bEnable);
TW68XX_API BOOL   __stdcall TwEnableVideoField2(HANDLE hSubDevice, BOOL bEnable);

TW68XX_API int    __stdcall TwGetChannelCount();
TW68XX_API BOOL   __stdcall TwChangeChannelId(int nChannel, int nId);
TW68XX_API HANDLE __stdcall TwChannelOpen(int nChannel);
TW68XX_API VOID   __stdcall TwChannelClose(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwGetChannelDetail(HANDLE hChannel, TChannelDesc *pChannelDesc);
TW68XX_API BOOL   __stdcall TwSetVideoPara(HANDLE hChannel, TVideoPara vPara, LONG lValue);
TW68XX_API BOOL   __stdcall TwGetVideoPara(HANDLE hChannel, TVideoPara vPara, LONG *pValue);
TW68XX_API BOOL   __stdcall TwStartVideoCapture(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwStopVideoCapture(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwSetAudioPara(HANDLE hChannel, TAudioPara aPara, LONG lValue);
TW68XX_API BOOL   __stdcall TwGetAudioPara(HANDLE hChannel, TAudioPara aPara, LONG *pValue);
TW68XX_API BOOL   __stdcall TwStartAudioCapture(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwStopAudioCapture(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwStartVideoField2(HANDLE hChannel);
TW68XX_API BOOL   __stdcall TwStopVideoField2(HANDLE hChannel);

typedef int (__stdcall *TwVideoStreamCallback)(int nChannel, VOID *DataBuf, int width, int height, BOOL bField2, void *context, __int64 pts);
TW68XX_API BOOL   __stdcall TwRegisterVideoStreamCallback(TwVideoStreamCallback VideoStreamCallback, VOID *Context);

typedef int (__stdcall *TwAudioStreamCallback)(int nChannel, VOID *DataBuf, LONG Length, void *Context, __int64 pts);
TW68XX_API BOOL   __stdcall TwRegisterAudioStreamCallback(TwAudioStreamCallback AudioStreamCallback, VOID *Context);

#endif