#include "StdAfx.h"
#include "TW68XX.h"
#include "TW68XXDlg.h"
#include "Channel.h"
#include "inifile.h"
#include <malloc.h>

#define AUDIO_SAMPLE_RATE	8000
#define AUDIO_SAMPLE_BIT	8

CVideoOut CChannel::m_vOut;
CPlayWave CChannel::m_aOut;

#ifndef _mm_malloc
#define _mm_malloc tw_malloc
#define _mm_free   tw_free

void *
tw_malloc(size_t size,
		   size_t alignment)
{
	BYTE *mem_ptr;
	
	if (!alignment) {
		
		/* We have not to satisfy any alignment */
		if ((mem_ptr = (BYTE *) malloc(size + 1)) != NULL) {
			
			/* Store (mem_ptr - "real allocated memory") in *(mem_ptr-1) */
			*mem_ptr = (BYTE)1;
			
			/* Return the mem_ptr pointer */
			return ((void *)(mem_ptr+1));
		}
	} else {
		BYTE *tmp;
		
		/* Allocate the required size memory + alignment so we
		* can realign the data if necessary */
		if ((tmp = (BYTE *) malloc(size + alignment)) != NULL) {
			
			/* Align the tmp pointer */
			mem_ptr =
				(BYTE *) ((long) (tmp + alignment - 1) &
				(~(long) (alignment - 1)));
			
				/* Special case where malloc have already satisfied the alignment
				* We must add alignment to mem_ptr because we must store
				* (mem_ptr - tmp) in *(mem_ptr-1)
				* If we do not add alignment to mem_ptr then *(mem_ptr-1) points
			* to a forbidden memory space */
			if (mem_ptr == tmp)
				mem_ptr += alignment;
			
				/* (mem_ptr - tmp) is stored in *(mem_ptr-1) so we are able to retrieve
			* the real malloc block allocated and free it in xvid_free */
			*(mem_ptr - 1) = (BYTE) (mem_ptr - tmp);
			
			/* Return the aligned pointer */
			return ((void *)mem_ptr);
		}
	}
	
	return(NULL);
}

/*!
********************************************************************************************
* \brief 
*	  Free a previously 'xvid_malloc' allocated block. Does not free NULL references.
*
********************************************************************************************
*/

void
tw_free(void *mem_ptr)
{
	
	BYTE *ptr;
	
	if (mem_ptr == NULL)
		return;
	
	/* Aligned pointer */
	ptr = (BYTE*)mem_ptr;
	
	/* *(ptr - 1) holds the offset to the real allocated block
	* we sub that offset os we free the real pointer */
	ptr -= *(ptr - 1);
	
	/* Free the memory */
	free(ptr);
}

#endif

CChannel::CChannel() :
m_bSound(FALSE),
m_bSaving(FALSE),
m_vDuration(40),
m_nSnap(0),
m_nPixelFormat(1),
m_bSignal(TRUE),
m_bDeinterlace(FALSE)
{
	ZeroMemory(&m_vFrameData, sizeof(m_vFrameData));
	ZeroMemory(&m_aFrameData, sizeof(m_aFrameData));
	ZeroMemory(m_bufTemp, sizeof(m_bufTemp));
}

CChannel::~CChannel(void)
{
	Close();
}

BOOL CChannel::Open(int width, int height, int pxlFormat, BOOL bY41P, BOOL bDeinterlace)
{
	if(FAILED(CSurface::Create(m_vOut.GetDDraw(), width, height, pxlFormat))) {
		//return FALSE;
	}

	m_nPixelFormat = bY41P ? PIX_FMT_Y41P : pxlFormat;
	m_bufTemp[0] = (PBYTE) _mm_malloc(width*height*3, 16);

	return TRUE;
}

void CChannel::Close()
{
	m_di.Uninit();
	Destroy();
	if(m_bufTemp[0]) {
		_mm_free(m_bufTemp[0]);
		m_bufTemp[0] = NULL;
	}
	if(m_bufTemp[1]) {
		_mm_free(m_bufTemp[1]);
		m_bufTemp[1] = NULL;
	}
}

void CChannel::EnableDeinterlace(BOOL bEnable)
{
	if(bEnable) {
		BOOL bY41P = (m_nPixelFormat==PIX_FMT_Y41P);
		if(bY41P) {
			bEnable = DISPLAY_Y41P_TO_YUY2;
		}
		else if(m_nPixelFormat!=PIX_FMT_YUV422) {
			bEnable = FALSE;
		}
		if(bEnable) {
			if(bY41P) {
				m_bufTemp[1] = (PBYTE) _mm_malloc(Width()*Height()*2, 16);
			}
			m_di.Init(Width(), Height());
		}
	}
	m_bDeinterlace = bEnable;
	if(!bEnable) {
		m_di.Uninit();
	}
}

void CChannel::OnSetCurrent()
{
	((CTW68XXDlg*)AfxGetMainWnd())->SetCurrentChannel(m_nChannel);
}

CVideoOut* CChannel::InitRender()
{
	if(m_vOut.InitDDraw() != 0) {
		return FALSE;
	}

	if(m_vOut.InitDibDraw() != 0) {
		return FALSE;
	}

	//m_aOut.Init(AUDIO_SAMPLE_RATE, 1, AUDIO_SAMPLE_BIT);
	
	return &m_vOut;
}

void CChannel::UninitRender()
{
	m_vOut.Uninit();
	//m_aOut.Uninit();
}

void CChannel::InitSound()
{
	m_aOut.Init(AUDIO_SAMPLE_RATE, 1, AUDIO_SAMPLE_BIT);
}

void CChannel::UninitSound()
{
	m_aOut.Uninit();
}

BOOL CChannel::OnVideoStream(VOID *DataBuf, int width, int height, __int64 pts, PVOID ctx)
{
	CTW68XXDlg* pMainDlg = (CTW68XXDlg*)ctx;
	if(m_nSnap > 0) {
		CString strFile, strRaw;
		CString  strPath, strTemp;
		CIniFile iniFile;
		CTime	 tNow;
		CFile	 rawFile;

		pMainDlg->GetModulePath(strPath.GetBuffer(MAX_PATH), MAX_PATH);
		strPath.ReleaseBuffer();
		iniFile.Create(strPath + OPTION_FILE);
		iniFile.ReadString(SECTION_DEFAULT, KEY_SAVEPATH, strPath, strTemp.GetBuffer(MAX_PATH), MAX_PATH);
		strTemp.ReleaseBuffer();
		tNow = CTime::GetCurrentTime();
		if(strTemp.Right(1) != TEXT('\\')) {
			strTemp += TEXT('\\');
		}
		strFile.Format(TEXT("%02d_"), m_nChannel);
		strFile+= tNow.FormatGmt(TEXT("%y-%m-%d-%H-%M-%S"));
		strRaw  = strFile;
		strFile = strTemp + strFile + TEXT(".bmp");
		strRaw  = strTemp + strRaw  + TEXT(".raw");

		/*
		 * write video raw data to file
		 */
		if(rawFile.Open(strRaw, CFile::modeCreate|CFile::modeWrite)) {
			try {
				rawFile.Write(DataBuf, width*height*BitCount()/8);
			}
			catch(...) {
				TRACE(TEXT("Write raw video data to file exception\n"));
			}
			rawFile.Close();
		}
		
		/*
		 * write video raw data to bmp file
		 */
		if((m_nPixelFormat==PIX_FMT_YUV422) || (m_nPixelFormat==PIX_FMT_RGB555)) {
			int   pitch = width*2;
			PBYTE bufTemp = m_bufTemp[0];
			PBYTE vdata = (PBYTE)DataBuf + pitch*(height-1);
			PBYTE vdest = bufTemp;
			for(long i=0; i<height; i++) {
				memcpy(vdest, vdata, pitch);
				vdata -= pitch;
				vdest += pitch;
			}
			if(m_nPixelFormat==PIX_FMT_YUV422) {
				m_vOut.SaveYUVToBmpFile(strFile, (PBYTE)bufTemp, width, height);
			}
			else {
				m_vOut.SaveRGBToBmpFile(strFile, (PBYTE)bufTemp, width, height, 16);
			}
		}

		m_nSnap--;
	}

	if(m_bVisible) {
		CRect rc;

		GetWindowRect(&rc);

		if(m_nPixelFormat==PIX_FMT_Y41P) {
#if(DISPLAY_Y41P_TO_YUY2)
			Y41PToYuy2((PBYTE)DataBuf, m_bufTemp[0], width, height);
			DataBuf = m_bufTemp[0];
			if(m_bDeinterlace) {
				m_di.Deinterlace((PBYTE)DataBuf, m_bufTemp[1]);
				DataBuf = m_bufTemp[1];
			}
#else
			Y41PToYuv420p((PBYTE)DataBuf, m_bufTemp[0], width, height);
			DataBuf = m_bufTemp[0];
#endif
		}
		else if(m_bDeinterlace) {
			m_di.Deinterlace((PBYTE)DataBuf, m_bufTemp[0]);
			DataBuf = m_bufTemp[0];
		}

		if(m_bSignal) {
			if(SUCCEEDED(Draw((PBYTE)DataBuf))) {
				//m_vOut.FuckSB(m_bufTemp[0], GetSafeHwnd());
				m_vOut.Draw(GetDDrawSurface(), GetSafeHwnd(), rc, NULL);
 			}
			else {
				//TRACE(TEXT("Copy data to offscreen surface failed\n"));
				if(m_nPixelFormat==PIX_FMT_YUV422) {
					SIZE sz={width, height};
					m_vOut.DrawYUY2((PBYTE)DataBuf, sz, GetSafeHwnd());
				}
				else if(m_nPixelFormat==PIX_FMT_YUV420P) {
					SIZE sz={width, height};
					Yuv420pToYuy2((PBYTE)DataBuf, m_bufTemp[0], width, height);
					m_vOut.DrawYUY2(m_bufTemp[0], sz, GetSafeHwnd());
				}
			}
		}
		else {
			m_vOut.Draw(pMainDlg->LogoSurface()->GetDDrawSurface(), GetSafeHwnd(), rc, NULL);
		}
	}

	if(m_bSaving) {
		//WriteVideo((PBYTE)DataBuf, width*height*2, (ULONGLONG)pts, m_vDuration);
		m_vFrameData.nLen = width*height*2;
		m_vFrameData.pts  = pts;
		m_vFrameData.pData= DataBuf;
		SetEvent(DataEvent());
	}

	m_nFrames++;

	return TRUE;
}

BOOL CChannel::OnAudioStream(VOID *DataBuf, LONG Length, __int64 pts)
{
	if(m_bSound) {
		m_aOut.Play((PBYTE)DataBuf, Length);
	}

	if(m_bSaving) {
		//WriteAudio((PBYTE)DataBuf, Length);
		m_aFrameData.nLen = Length;
		m_aFrameData.pts  = pts;
		m_aFrameData.pData= DataBuf;
		SetEvent(DataEvent());
	}
	
//	TRACE("audio buf %08X\n", DataBuf);

	return TRUE;
}

void CChannel::ShowVideo(BOOL bShow)
{
	ShowWindow(bShow?SW_SHOW:SW_HIDE); 
	ShowTitle(bShow);
	m_bVisible=bShow; 
	if(!bShow) {
		ShowBorder(bShow);
	}
}

void CChannel::UpdateStatus()
{
	CString strTitle;
	strTitle.Format(IDS_TITLE, m_nChannel+1, m_nFrames);
	UpdateTitle(strTitle);

	m_nFrames = 0;
}

BOOL CChannel::Save(BOOL bSave, ULONG nFrameRate)
{
	if(m_bSaving==bSave) {
		return TRUE;
	}

	if(bSave) {
		CString strFile;
		TVideoStreamInfo vsi;
		TAudioStreamInfo asi;
		CTW68XXDlg* pMainDlg = (CTW68XXDlg*)AfxGetMainWnd();
		CString  strPath, strTemp;
		CIniFile iniFile;
		CTime	 tNow;
		
		pMainDlg->GetModulePath(strPath.GetBuffer(MAX_PATH), MAX_PATH);
		strPath.ReleaseBuffer();
		iniFile.Create(strPath + OPTION_FILE);
		iniFile.ReadString(SECTION_DEFAULT, KEY_SAVEPATH, strPath, strTemp.GetBuffer(MAX_PATH), MAX_PATH);
		strTemp.ReleaseBuffer();
		tNow = CTime::GetCurrentTime();
		if(strTemp.Right(1) != TEXT('\\')) {
			strTemp += TEXT('\\');
		}
		strFile.Format(TEXT("%02d_"), m_nChannel);
		strFile+= tNow.FormatGmt(TEXT("%y-%m-%d-%H-%M-%S.avi"));
		strFile = strTemp + strFile;

		ZeroMemory(&vsi, sizeof(TVideoStreamInfo));
		ZeroMemory(&asi, sizeof(TAudioStreamInfo));

		vsi.fccHandler = mmioFOURCC('Y', 'U', 'Y', '2');
		vsi.frame_rate = nFrameRate==0 ? 25 : nFrameRate;
		vsi.width = Width();
		vsi.height= Height();
		m_vDuration = 1000/vsi.frame_rate;

		asi.fccHandler = 0;
		asi.wfx.cbSize = sizeof(WAVEFORMATEX);
		asi.wfx.nChannels = 1;
		asi.wfx.nBlockAlign = 1;
		asi.wfx.wFormatTag  = WAVE_FORMAT_PCM;
		asi.wfx.wBitsPerSample = AUDIO_SAMPLE_BIT;
		asi.wfx.nSamplesPerSec = AUDIO_SAMPLE_RATE;
		asi.wfx.nAvgBytesPerSec= asi.wfx.nSamplesPerSec*asi.wfx.wBitsPerSample/8;
		asi.sample_size = 1;

		if(!CAviWriter::Open(strFile, &vsi, &asi)) {
			CString strInfo;
			strInfo.Format(IDS_ERR_OPENFILE, m_nChannel+1, strFile);
			AfxMessageBox(strInfo);
			return FALSE;
		}

		CThread::Create();
	}
	else {
		Exit();
		CAviWriter::Close();
	}
	m_bSaving = bSave;
	
	return TRUE;
}

UINT CChannel::Run()
{
	TFrameData fd;
	if(m_vFrameData.pData != NULL) {
		fd = m_vFrameData;
		ZeroMemory(&m_vFrameData, sizeof(m_vFrameData));
		WriteVideo((PBYTE)fd.pData, fd.nLen, (ULONGLONG)fd.pts, m_vDuration);
	}
	if(m_aFrameData.pData != NULL) {
		fd = m_aFrameData;
		ZeroMemory(&m_aFrameData, sizeof(m_aFrameData));
		WriteAudio((PBYTE)fd.pData, fd.nLen);
	}

	return 0;
}

void CChannel::Snap()
{
	m_nSnap++;
}

void CChannel::CheckSignal()
{
	CTW68XXDlg* dlgMain = (CTW68XXDlg*)AfxGetMainWnd();

	TF_TwGetVideoPara fTwGetVideoPara = (TF_TwGetVideoPara)GetProcAddress(dlgMain->SdkHandle(), "TwGetVideoPara");
	if(fTwGetVideoPara == NULL) {
		return;
	}

	fTwGetVideoPara(m_hChannel, vpSignal, (long*)&m_bSignal);
}