// Field2.cpp: implementation of the CField2 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "tw68xx.h"
#include "TW68XXDlg.h"
#include "Field2.h"
#include "Field2Dlg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CVideoOut* CField2::m_vOut = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CField2::CField2() :
m_bRunning(FALSE),
m_bufTemp(NULL),
m_nPixelFormat(1)
{

}

CField2::~CField2()
{
	Close();
}

BOOL CField2::Open(int width, int height, int pxlFormat, BOOL bY41P)
{
	if(FAILED(CSurface::Create(m_vOut->GetDDraw(), width, height, pxlFormat))) {
		//return FALSE;
	}

	m_nPixelFormat = bY41P ? PIX_FMT_Y41P : pxlFormat;
	m_bufTemp = new BYTE[width*height*3];
	
	return TRUE;
}

void CField2::Close()
{
	Destroy();
	if(m_bufTemp) {
		delete[] m_bufTemp;
		m_bufTemp = NULL;
	}
}

void CField2::OnSetCurrent()
{
	((CField2Dlg*)GetParent())->SetCurrentChannel(m_nChannel);
}

void CField2::OnShowNormal()
{
	CRect rcPos, rcTemp;
	CField2Dlg* pParentDlg = (CField2Dlg*)GetParent();
	
	pParentDlg->ShowVideoWnds(-2, FALSE);
	pParentDlg->ShowVideoWnds(-1, TRUE);
}

void CField2::OnShowLarge()
{
	CRect rcPos, rcTemp;
	CField2Dlg* pParentDlg = (CField2Dlg*)GetParent();
	
	pParentDlg->ShowVideoWnds(-1, FALSE);
	pParentDlg->ShowVideoWnds(-2, TRUE);
}

void CField2::OnShowFull()
{
	CRect rcPos, rcTemp;
	CField2Dlg* pParentDlg = (CField2Dlg*)GetParent();
	
	pParentDlg->ShowVideoWnds(-3, m_bFull);
}

void CField2::ShowVideo(BOOL bShow)
{
	ShowWindow(bShow?SW_SHOW:SW_HIDE); 
	ShowTitle(bShow);
	m_bVisible=bShow; 
	if(!bShow) {
		ShowBorder(bShow);
	}
}

BOOL CField2::OnVideoStream(VOID *DataBuf, int width, int height, __int64 pts)
{
	m_nFrames++;
	
	if(m_bVisible) {
		CRect rc;
		
		GetWindowRect(&rc);
		
		if(m_nPixelFormat==PIX_FMT_Y41P) {
#if(DISPLAY_Y41P_TO_YUY2)
			Y41PToYuy2((PBYTE)DataBuf, m_bufTemp, width, height);
#else
			Y41PToYuv420p((PBYTE)DataBuf, m_bufTemp, width, height);
#endif
			DataBuf = m_bufTemp;
		}
		
		if(SUCCEEDED(Draw((PBYTE)DataBuf))) {
			m_vOut->Draw(GetDDrawSurface(), GetSafeHwnd(), rc, NULL);
		}
		else {
			//TRACE(TEXT("Field2: Copy data to offscreen surface failed\n"));
			if(m_nPixelFormat==PIX_FMT_YUV422) {
				SIZE sz={width, height};
				m_vOut->DrawYUY2((PBYTE)DataBuf, sz, GetSafeHwnd());
			}
			else if(m_nPixelFormat==PIX_FMT_YUV420P) {
				SIZE sz={width, height};
				Yuv420pToYuy2((PBYTE)DataBuf, m_bufTemp, width, height);
				m_vOut->DrawYUY2(m_bufTemp, sz, GetSafeHwnd());
			}
		}
	}

	return TRUE;
}

BOOL CField2::Run(BOOL bRun)
{
	CTW68XXDlg* pMainDlg = (CTW68XXDlg*)AfxGetMainWnd();
	HMODULE hSdk = pMainDlg->SdkHandle();

	if(bRun) {
		TF_TwStartVideoField2 fTwStartVideoField2 = (TF_TwStartVideoField2)GetProcAddress(hSdk, "TwStartVideoField2");
		if(fTwStartVideoField2 != NULL) {
			m_bRunning = fTwStartVideoField2(m_hChannel);
			return m_bRunning;
		}
		m_bRunning = FALSE;
	} else  {
		TF_TwStopVideoField2 fTwStopVideoField2 = (TF_TwStopVideoField2)GetProcAddress(hSdk, "TwStopVideoField2");
		if(fTwStopVideoField2 != NULL) {
			fTwStopVideoField2(m_hChannel);
			m_bRunning = FALSE;
		}
	}

	return FALSE;
}

void CField2::UpdateStatus()
{
	CString strTitle;
	strTitle.Format(IDS_TITLE, m_nChannel+1, m_nFrames);
	UpdateTitle(strTitle);
	
	m_nFrames = 0;
}