// PlayWave.cpp: implementation of the CPlayWave class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <windows.h>
#include <stdlib.h>
#include <iostream>
#include <WTYPES.H>
#include <MMSYSTEM.H>
#include <MMREG.H>
#include <process.h>
#include "ScopedLock.h"
#include "WavePacket.h"
#include "PlayWave.h"

#define BYTES_PER_FRAME		8800
#define MAX_DELAY_UNIT		1
#define NOTIFY_COUNT		10
#define MaxDelayBytes		(BYTES_PER_FRAME * MAX_DELAY_UNIT)
#define MaxAudioBuffer		(BYTES_PER_FRAME * NOTIFY_COUNT)
#define MAX_PLAYBUFFER		2

#pragma comment(lib,"Winmm.lib")
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPlayWave::CPlayWave() : 
m_hOut(0),
m_hThreadPlayWave(NULL),
m_uThreadPlayWaveId(0),
m_bPlaying(FALSE),
#ifdef USE_DSOUND
m_pDSnd(NULL),
m_pDSndBuff(NULL),
#endif
m_dwWritePos(0)
{
}

CPlayWave::~CPlayWave()
{
	Uninit();
}

/*------------------------------------------------------------------------------*/
//豸
BOOL CPlayWave::OpenDev(WAVEFORMATEX* pfmt,DWORD dwCallback,DWORD dwCallbackInstance,DWORD fdwOpen)
{
	if(m_hOut)
		return TRUE;
	
	BOOL bRet=FALSE;
	WAVEFORMATEX wf;
	if(!pfmt)//Ĭϸʽ
	{
		
		wf.wFormatTag	= WAVE_FORMAT_PCM;
		wf.cbSize		= 0;
		wf.wBitsPerSample=16;
		wf.nSamplesPerSec=8000;
		wf.nChannels=1;
		wf.nAvgBytesPerSec	= wf.nSamplesPerSec*(wf.wBitsPerSample/8);
		wf.nBlockAlign		= wf.nChannels     *(wf.wBitsPerSample/8);
		pfmt=&wf;
	}
	
	if(waveOutOpen(0,WAVE_MAPPER,pfmt,0,0,WAVE_FORMAT_QUERY))
		return FALSE;

	if(waveOutOpen(&m_hOut,WAVE_MAPPER,pfmt,dwCallback,dwCallbackInstance,fdwOpen))
		return FALSE;

	return TRUE;
}

VOID CPlayWave::CloseDev()
{
	if(!m_hOut)
		return;
	
	if(waveOutReset(m_hOut))
		return;
	
	if(waveOutClose(m_hOut))
		return;

	m_hOut=0;
}

BOOL CPlayWave::Init(DWORD sample_rate/* =48000 */, WORD channels/* =2*/, WORD bitsPersample/* =16*/)
{
#ifdef USE_DSOUND
	if(m_pDSnd) return FALSE;
#endif
	m_hThreadPlayWave = (HANDLE)_beginthreadex(NULL,
											  0,
											  Thread_PlayWave,
											  this,
											  0,
											  &m_uThreadPlayWaveId);
	if(!m_hThreadPlayWave)
		return FALSE;

	CWavePacket* wpkt = NULL;
	for(UINT i=0; i<MAX_PLAYBUFFER; i++) {
		wpkt = new CWavePacket();
		if(wpkt == NULL) {
			return FALSE;
		}
		if(!wpkt->Initialize(BYTES_PER_FRAME)) {
			delete wpkt;
			return FALSE;
		}
		m_hdrWave.push_back(wpkt);
	}
	
	m_iAudioBuf = 0;

	WAVEFORMATEX wf;
	ZeroMemory( &wf, sizeof(wf) );
	wf.cbSize = sizeof(wf);
	wf.nSamplesPerSec = sample_rate;
	wf.wBitsPerSample = bitsPersample;
	wf.nChannels = channels;
	wf.nBlockAlign = wf.wBitsPerSample / 8 * wf.nChannels;
	wf.nAvgBytesPerSec = wf.nBlockAlign * wf.nSamplesPerSec;
	wf.wFormatTag = WAVE_FORMAT_PCM;
	
	return OpenDev(&wf,m_uThreadPlayWaveId,(DWORD)this,CALLBACK_THREAD);
}

void CPlayWave::Uninit()
{
	if( m_hThreadPlayWave )
	{
		int t=50;
		DWORD ExitCode;
		BOOL bEnd=FALSE;
		PostThreadMessage(m_uThreadPlayWaveId,WM_QUIT,0,0);
		while(t)
		{
			GetExitCodeThread(m_hThreadPlayWave,&ExitCode);
			if(ExitCode!= STILL_ACTIVE)
			{
				bEnd=TRUE;
				break;
			}
			else
				Sleep(10);
			t--;
		}
		if(!bEnd)
			TerminateThread(m_hThreadPlayWave,0);
		CloseHandle(m_hThreadPlayWave);
		m_hThreadPlayWave = NULL;
	}
#ifdef USE_DSOUND
	SAFE_RELEASE(m_pDSndBuff);
	SAFE_RELEASE(m_pDSnd);
#endif

	CloseDev();

	while(m_hdrWave.size() > 0) {
		delete m_hdrWave[m_hdrWave.size()-1];
		m_hdrWave.pop_back();
	}
}

void CPlayWave::Play()
{
#ifdef USE_DSOUND
	if(m_pDSnd) {
		if(m_pDSndBuff)
			m_pDSndBuff->Play(0,0,DSBPLAY_LOOPING);
	}
	else 
#endif
	if(m_hOut)
	{
		waveOutReset(m_hOut);
		m_bPlaying = TRUE;
	}
}

void CPlayWave::Stop()
{
#ifdef USE_DSOUND
	if(m_pDSnd) {
		if(m_pDSndBuff)
			m_pDSndBuff->Stop();
	}
	else 
#endif
	if(m_hOut)
	{
		m_bPlaying = FALSE;
		waveOutPause(m_hOut);
	}
	m_dwWritePos = 0;
}

/*------------------------------------------------------------------------------*/
//һ
BOOL CPlayWave::Play(BYTE *buf, UINT uSize)
{
#ifdef USE_DSOUND
	if(m_pDSnd) {
		LPVOID pSoundBlock = NULL,pSecondBlock = NULL;
		DWORD  dwBlockSize = 0,dwSecondSize = 0;
		HRESULT	hr = m_pDSndBuff->Lock(m_dwWritePos,uSize,
									   &pSoundBlock,&dwBlockSize,
									   &pSecondBlock,&dwSecondSize,0L);
		if(hr == DSERR_BUFFERLOST)
		{
			m_pDSndBuff->Restore();
			hr = m_pDSndBuff->Lock(m_dwWritePos,uSize,
								   &pSoundBlock,&dwBlockSize,
								   &pSecondBlock,&dwSecondSize,0L);
			if(FAILED(hr)) 
				return FALSE;
		}

		UINT uWrite = min(uSize, dwBlockSize);
		memcpy(pSoundBlock, buf, uWrite);
		if(uSize > uWrite) {
			memcpy(pSecondBlock, buf+uWrite, min(uSize-uWrite, dwSecondSize));
		}
		m_pDSndBuff->Unlock(pSoundBlock,dwBlockSize,
							pSecondBlock,dwSecondSize);
		m_dwWritePos = (m_dwWritePos+uSize) % MaxAudioBuffer;
	}
	else 
#endif
	{
		UINT uMore = uSize % BYTES_PER_FRAME;
		UINT num = min(MAX_PLAYBUFFER, uSize/BYTES_PER_FRAME + (uMore ? 1 : 0));
		UINT i, j;
		UINT uLen  = 0;
		UINT uTotal= 0;

		for(i=0; i<num; i++) {

			if(m_iAudioBuf >= MAX_PLAYBUFFER)
				return FALSE;

			if(num > 1) {
				uLen = ((num-1 == i) && uMore) ? uMore : BYTES_PER_FRAME;
			}
			else {
				uLen = uSize;
			}

			LPWAVEHDR pwh=NULL;
			{
				CrScopedLock autoLock(&m_lock);
				for(j=0; j<m_hdrWave.size(); j++) {
					if(!m_hdrWave[j]->IsBusy()) {
						pwh = m_hdrWave[j]->WaveHdr();
						break;
					}
				}
				if(!pwh) {
					return FALSE;
				}
				m_hdrWave[j]->SetBusy(TRUE);
			}

			ZeroMemory(pwh,sizeof(WAVEHDR));
			pwh->dwBufferLength=uLen;
			pwh->lpData = (LPSTR)m_hdrWave[j]->Buffer();
			CopyMemory(pwh->lpData, buf+uTotal, uLen);
			if(waveOutPrepareHeader(m_hOut,pwh,sizeof(WAVEHDR))) {
				m_hdrWave[j]->SetBusy(FALSE);
    			return FALSE;
			}
			if(waveOutWrite(m_hOut,pwh,sizeof(WAVEHDR))) {
				m_hdrWave[j]->SetBusy(FALSE);
				return FALSE;
			}
			m_iAudioBuf++;
			uTotal += uLen;
		}
		if(m_bPlaying == FALSE) {
			Play();
		}
	}

	return TRUE;
}
/*------------------------------------------------------------------------------*/

//------------------------------------------------------------------------------
// Name: CAudioCapture::ProcessAudioData()
// Desc: The thread function, processing the wave play message
//------------------------------------------------------------------------------
UINT CPlayWave::Thread_PlayWave(LPVOID pContext)
{
	CPlayWave* pSelf=(CPlayWave*)pContext;
	MSG  msg;

	while(GetMessage(&msg,0,0,0))
	{
		switch(msg.message)
		{
		case WOM_DONE:
			{
				WAVEHDR* pwh=(WAVEHDR*)msg.lParam;
				waveOutUnprepareHeader((HWAVEOUT)msg.wParam,pwh,sizeof(WAVEHDR));
				pSelf->OnFramePlayDone(pwh);
			}
			break;
		case WM_QUIT:
			return (UINT)msg.wParam;
			break;
		}
	}
	return (UINT)msg.wParam;
}

void CPlayWave::OnFramePlayDone(LPWAVEHDR hdr)
{
	UINT i;
	CrScopedLock autoLock(&m_lock);
	m_iAudioBuf--;
	for(i=0; i<m_hdrWave.size(); i++) {
		if(m_hdrWave[i]->WaveHdr() == hdr) {
			m_hdrWave[i]->SetBusy(FALSE);
			break;
		}
	}
}

#ifdef USE_DSOUND
BOOL CPlayWave::Init(DWORD sample_rate, WORD channels, HWND hWnd)
{
	if(m_hThreadPlayWave) return FALSE;
	
    if( FAILED( DirectSoundCreate8( NULL, &m_pDSnd, NULL ) ) )
        return  FALSE;
	
    if( FAILED( m_pDSnd->SetCooperativeLevel( hWnd, DSSCL_PRIORITY )) )
        return  FALSE;
    
    if( FAILED( SetPrimaryBufferFormat( channels, sample_rate, 16 ))) //ʽ
		return  FALSE;
	
	if( FAILED( CreateAssistantSoundBuffer(channels, sample_rate, 16)))		//ø
		return  FALSE;

	/*Clear null all all buffer*/
	long lSize = MaxAudioBuffer;
	LPVOID pSoundBlock;
	DWORD  pSoundBlockSize;

	if(FAILED(m_pDSndBuff->SetCurrentPosition(0L)))
		return FALSE;
	if(FAILED(m_pDSndBuff->Lock(0,lSize,&pSoundBlock,&pSoundBlockSize,NULL,NULL,0L)))
		return FALSE;
	ZeroMemory(pSoundBlock,pSoundBlockSize);
	m_pDSndBuff->Unlock(pSoundBlock,pSoundBlockSize,NULL,0);

	return TRUE;
}

HRESULT CPlayWave::SetPrimaryBufferFormat(DWORD dwPrimaryChannels, 
                                          DWORD dwPrimaryFreq, 
                                          DWORD dwPrimaryBitRate)
{
	HRESULT             hr;
    LPDIRECTSOUNDBUFFER pDSBPrimary = NULL;

    if( m_pDSnd == NULL )
        return CO_E_NOTINITIALIZED;

    // Get the primary buffer 
    DSBUFFERDESC dsbd;
    ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize        = sizeof(DSBUFFERDESC);
    dsbd.dwFlags       = DSBCAPS_PRIMARYBUFFER;
    dsbd.dwBufferBytes = 0;						//must be for primary
    dsbd.lpwfxFormat   = NULL;					//must be for primary
       
    if( FAILED( hr = m_pDSnd->CreateSoundBuffer( &dsbd, &pDSBPrimary, NULL ) ) )
        return  hr;

    WAVEFORMATEX wfx;
    ZeroMemory( &wfx, sizeof(WAVEFORMATEX) ); 
    wfx.wFormatTag      = WAVE_FORMAT_PCM; 
    wfx.nChannels       = (WORD) dwPrimaryChannels; 
    wfx.nSamplesPerSec  = dwPrimaryFreq; 
    wfx.wBitsPerSample  = (WORD) dwPrimaryBitRate; 
    wfx.nBlockAlign     = wfx.wBitsPerSample / 8 * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

    if( FAILED( hr = pDSBPrimary->SetFormat(&wfx) ) )
        return  hr;

    SAFE_RELEASE( pDSBPrimary );

    return S_OK;
}

/**/

HRESULT CPlayWave::CreateAssistantSoundBuffer(WORD  Channels, 
											  DWORD sample_rate, 
											  WORD  bitCount)
{
    HRESULT hr;

    DSBPOSITIONNOTIFY*  aPosNotify = NULL; 
    LPDIRECTSOUNDNOTIFY pDSNotify  = NULL;

    if( m_pDSnd == NULL )
        return CO_E_NOTINITIALIZED;

	WAVEFORMATEX wfm;
	ZeroMemory(&wfm, sizeof(WAVEFORMATEX)); 
    wfm.wFormatTag      = WAVE_FORMAT_PCM; 
    wfm.nChannels       = Channels; 
    wfm.nSamplesPerSec  = sample_rate;
    wfm.wBitsPerSample  = bitCount; 
    wfm.nBlockAlign     = wfm.wBitsPerSample / 8 * wfm.nChannels;
    wfm.nAvgBytesPerSec = wfm.nSamplesPerSec * wfm.nBlockAlign;
	
    DSBUFFERDESC dsbd;
    ZeroMemory(&dsbd, sizeof(DSBUFFERDESC));
    dsbd.dwSize         = sizeof(DSBUFFERDESC);
    dsbd.dwFlags        = DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME |
						  DSBCAPS_CTRLFREQUENCY| DSBCAPS_CTRLPOSITIONNOTIFY | 
                          DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_GLOBALFOCUS;

//    dsbd.dwBufferBytes   = 2 * wfm.nAvgBytesPerSec;		//2s pcm audio data len
	dsbd.dwBufferBytes	 = MaxAudioBuffer;
    dsbd.lpwfxFormat     = (LPWAVEFORMATEX)&wfm;

	LPDIRECTSOUNDBUFFER pSndBuff = NULL;
    if(FAILED( hr = m_pDSnd->CreateSoundBuffer( &dsbd, &pSndBuff, NULL ) ) ) {
        return hr;
    }
	if(FAILED( hr = pSndBuff->QueryInterface(IID_IDirectSoundBuffer8, (void**)&m_pDSndBuff)))
		return hr;

	if(FAILED( hr = m_pDSndBuff->SetVolume(DSBVOLUME_MAX)))
		return hr;

    // Create the notification events, so that we know when to fill
    // the buffer as the sound plays. 

	return S_OK;
}
#endif