//////////////////////////////////////////////////////////////////////
// ScopedLock.h: interface for the CThread class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__SCOPEDLOCK_H_)
#define __SCOPEDLOCK_H_

//------------------
//߳:ʹٽ
//------------------
class CrLock
{
public:
	CrLock() : m_bAttach(FALSE) {};
	CrLock(BOOL bAttach) : m_bAttach(bAttach) {};
	virtual ~CrLock(){};

	virtual void Lock() = 0;
	virtual BOOL Lock(ULONG time_out) = 0;
	virtual void Unlock() = 0;

protected:
	BOOL m_bAttach;
};

class CrLockCS : public CrLock
{
public:
	CrLockCS(CRITICAL_SECTION cs) : CrLock(TRUE), m_cs(cs) {}
	CrLockCS() : CrLock(FALSE) {InitializeCriticalSection(&m_cs);}
	virtual ~CrLockCS(){if(!m_bAttach) DeleteCriticalSection(&m_cs);}

	virtual void Lock(){EnterCriticalSection(&m_cs);}
	virtual BOOL Lock(ULONG time_out){EnterCriticalSection(&m_cs);return TRUE;}
	virtual void Unlock(){LeaveCriticalSection(&m_cs);}
	CRITICAL_SECTION& Handle(){return m_cs;}

protected:
	CRITICAL_SECTION m_cs;
};

class CrLockEvt : public CrLock
{
public:
	CrLockEvt(HANDLE evt) : CrLock(TRUE), m_cs(evt) {}
	CrLockEvt() : CrLock(FALSE) {m_cs = CreateEvent(NULL, FALSE, TRUE, NULL);}
	virtual ~CrLockEvt(){if(!m_bAttach) CloseHandle(m_cs);}

	virtual void Lock(){WaitForSingleObject(m_cs, INFINITE);}
	virtual BOOL Lock(ULONG time_out){return WaitForSingleObject(m_cs, time_out)==WAIT_OBJECT_0;}
	virtual void Unlock(){SetEvent(m_cs);}
	HANDLE&		 Handle(){return m_cs;}

protected:
	HANDLE m_cs;
};

class CrScopedLock
{
public:
	CrScopedLock(CrLock* lock) : m_lock(lock),m_status(TRUE) {if(m_lock) m_lock->Lock();}
	CrScopedLock(CrLock* lock, ULONG time_out) : m_lock(lock),m_status(FALSE) {if(m_lock) m_status = m_lock->Lock(time_out);}
	~CrScopedLock(){if(m_status && m_lock) m_lock->Unlock();}
	BOOL Status(){return m_status;}

private:
	CrLock* m_lock;
	BOOL   m_status;
};

#endif // !defined(__SCOPEDLOCK_H_)
