// VideoOut.h: interface for the CVideoOut class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VIDEOOUT_H_)
#define __VIDEOOUT_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <ddraw.h>
#include <VFW.h>
#include "ScopedLock.h"

enum PixelFormat {
    PIX_FMT_YUV420P,   ///< Planar YUV 4:2:0 (1 Cr & Cb sample per 2x2 Y samples)
	PIX_FMT_YUV422,    ///< Packed pixel, Y0 Cb Y1 Cr 
	PIX_FMT_RGB24,     ///< Packed pixel, 3 bytes per pixel, RGBRGB...
	PIX_FMT_BGR24,     ///< Packed pixel, 3 bytes per pixel, BGRBGR...
	PIX_FMT_YUV422P,   ///< Planar YUV 4:2:2 (1 Cr & Cb sample per 2x1 Y samples)
	PIX_FMT_YUV444P,   ///< Planar YUV 4:4:4 (1 Cr & Cb sample per 1x1 Y samples)
	PIX_FMT_RGBA32,    ///< Packed pixel, 4 bytes per pixel, BGRABGRA..., stored in cpu endianness
	PIX_FMT_YUV410P,   ///< Planar YUV 4:1:0 (1 Cr & Cb sample per 4x4 Y samples)
	PIX_FMT_YUV411P,   ///< Planar YUV 4:1:1 (1 Cr & Cb sample per 4x1 Y samples)
	PIX_FMT_RGB565,    ///< always stored in cpu endianness 
	PIX_FMT_RGB555,    ///< always stored in cpu endianness, most significant bit to 1 
	PIX_FMT_GRAY8,
	PIX_FMT_MONOWHITE, ///< 0 is white 
	PIX_FMT_MONOBLACK, ///< 0 is black 
	PIX_FMT_PAL8,      ///< 8 bit with RGBA palette 
	PIX_FMT_YUVJ420P,  ///< Planar YUV 4:2:0 full scale (jpeg)
	PIX_FMT_YUVJ422P,  ///< Planar YUV 4:2:2 full scale (jpeg)
	PIX_FMT_YUVJ444P,  ///< Planar YUV 4:4:4 full scale (jpeg)
	PIX_FMT_XVMC_MPEG2_MC,///< XVideo Motion Acceleration via common packet passing(xvmc_render.h)
	PIX_FMT_XVMC_MPEG2_IDCT,
	PIX_FMT_UYVY422,   ///< Packed pixel, Cb Y0 Cr Y1 
	PIX_FMT_Y411,	   ///< Packed pixel, Cb Y0 Y1 Cr Y2 Y3 Cb Y4 Y5 Cr Y6 Y7
	PIX_FMT_Y41P,      ///< Packed pixel, Cb Y0 Cr Y1 Cb Y2 Cr Y3 Y4 Y5 Y6 Y7
	PIX_FMT_NB,
};

class CVideoOut
{
protected:
	LPDIRECTDRAW7		 m_pDDraw;
	LPDIRECTDRAWSURFACE7 m_pPrimary;
	LPDIRECTDRAWSURFACE7 m_pOverlay;
	LPDIRECTDRAWCLIPPER  m_pClipper;
	DDSURFACEDESC2		 m_ddover;
	CrLockCS			 m_csLock;
	COLORREF			 m_clrKey;
	DWORD				 m_DDsColor;
	PBYTE				 m_pBuff[2];
	static BOOL			 m_bBltSystemToVideo;

	HDRAWDIB			 m_hDDib;

public:
	CVideoOut();
	virtual ~CVideoOut();

	int  InitDDraw(HWND hWnd=NULL);
	int  InitDibDraw();
	void Uninit();
	int  Draw(LPDIRECTDRAWSURFACE7 pdds, RECT rcDest, HWND hWnd);
	int  Draw(LPDIRECTDRAWSURFACE7 pdds, HWND hWnd, RECT rcDest, RECT* rcSrc=NULL);
	int  Draw(PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, LONG lPitch=0);
	int	 DrawRGB24(PBYTE data, SIZE srcSize, HWND hWnd, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	int	 DrawYUY2(PBYTE data, SIZE srcSize, HWND hWnd, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	LPDIRECTDRAW7 GetDDraw(){return m_pDDraw;}
	int  CopyBitmapToYUVSurface(HBITMAP hbm);
	BOOL CopyData(PBYTE buff, RECT* rect);
	BOOL Restore();
	BOOL HideOverlay();
	BOOL CanBltYuv2Primary();

	int	  SetOverlayColorKey(COLORREF clrKey);
	ULONG GetOverlayColorKey(){ return m_clrKey; }
	int   EnableOverlay(BOOL bEnable, int pixel_format=1, 
						COLORREF clrKey=RGB(10,10,10),
						int  width=0, int height=0);
	int   EnableOverlayEx(BOOL bEnable, int pixel_format, int  width, int height);
	BOOL  IsOverlay(){return m_pOverlay!=NULL;}
	PVOID GetOverlayBuffer(POINT pt, DWORD* pitch);
	void  RefreshOverlay(RECT* rcSrc, RECT* rcDst);
	void  FillOverlayRect(RECT* rc, BYTE val);
	BOOL  TestBltSystemToVideo(LPDIRECTDRAWSURFACE7 pdds, HWND hWnd);
	static void ReverseBGR888(PBYTE src, PBYTE dst, UINT widthbytes, UINT height);
	DWORD ConvertGDIColor(COLORREF dwGDIColor) { return ConvertGDIColor(m_pPrimary, dwGDIColor); }
	static BOOL  SaveYUVToBmpFile(LPCTSTR FileName, PBYTE yuv, int Width, int Height);
	static BOOL  SaveRGBToBmpFile(LPCTSTR FileName, PBYTE rgb, int Width, int Height, int bitCount);

	void   FuckSB(PBYTE data, HWND hWnd);

protected:
	DWORD ConvertGDIColor(LPDIRECTDRAWSURFACE7 pdds, COLORREF dwGDIColor);

#ifndef WIN64
	void  memcpyMMX(void *Dest, void *Src, size_t nBytes);
	void  MMXRectCpy(RECT dsr, RECT ddr, BYTE *Dest, BYTE *Src);
#endif
};

#endif // __VIDEOOUT_H_
