// VideoWnd.cpp : implementation file
//

#include "stdafx.h"
#include "TW68XX.h"
#include "TW68XXDlg.h"
#include "VideoWnd.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define MOTION_TIME		10		//10s
#define MOTION_TIMER	1001
#define SIGNAL_TIMER	1002

#define ENABLE_BORDER	1
#define ENABLE_TITLE	1

BOOL CVideoWnd::m_bFull = FALSE;

/////////////////////////////////////////////////////////////////////////////
// CVideoWnd

CVideoWnd::CVideoWnd() :
m_bVisible(TRUE),
m_hChannel(NULL),
m_nChannel(0),
m_nFrames(0)
{
	m_nState = VShow_normal;
	m_rcNormal.SetRect(0, 0, 352, 288);
	m_bTitle = FALSE;
}

CVideoWnd::~CVideoWnd()
{
}


BEGIN_MESSAGE_MAP(CVideoWnd, CStatic)
	//{{AFX_MSG_MAP(CVideoWnd)
	ON_WM_LBUTTONDBLCLK()
	ON_WM_LBUTTONDOWN()
	ON_WM_CTLCOLOR_REFLECT()
	ON_WM_RBUTTONDBLCLK()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CVideoWnd message handlers

BOOL CVideoWnd::CreateWnd(CRect& rc, CWnd* pParentWnd, int nIndex)
{
	m_rcNormal = rc;
	m_bkBrush.CreateSolidBrush(RGB(10,10,10));
	BOOL ret = CStatic::CreateEx( /*WS_EX_STATICEDGE|*/WS_EX_NOPARENTNOTIFY, 
							  _T("STATIC"), 
							  NULL, 
							  WS_CHILD|WS_VISIBLE|SS_NOTIFY, 
							  rc, 
							  pParentWnd,
							  IDC_VIDEO+nIndex, 
							  NULL);

#if(ENABLE_BORDER)	
	m_bdLeft.CreateWnd(pParentWnd, CBorder::bsLeft, RGB(0,0,255));
	m_bdRight.CreateWnd(pParentWnd, CBorder::bsRight, RGB(0,0,255));
	m_bdTop.CreateWnd(pParentWnd, CBorder::bsTop, RGB(0,0,255));
	m_bdBottom.CreateWnd(pParentWnd, CBorder::bsBottom, RGB(0,0,255));
	MoveBorder();
#endif

#if(ENABLE_TITLE)
	m_Title.CreateWnd(pParentWnd, RGB(255,255,255));
	MoveTitle();
	if(m_bTitle) {
		ShowTitle(TRUE);
	}
#endif
	
	return ret;
}

void CVideoWnd::MoveVideo(CRect rc)
{
	MoveWindow(rc);
	MoveBorder();
	MoveTitle();
	if(m_nState==VShow_normal) {
		m_rcNormal = rc;
	}
}

void CVideoWnd::MoveTitle()
{
#if(ENABLE_TITLE)	
	CRect rcClient, rcTitle;
	GetWindowRect(&rcClient);
	m_Title.GetWindowRect(&rcTitle);

	rcClient.left += 6;
	rcClient.top  += 6;
	rcClient.right = rcClient.left+rcTitle.Width();
	rcClient.bottom= rcClient.top +rcTitle.Height();
	m_Title.MoveWindow(&rcClient);
#endif
}

void CVideoWnd::UpdateTitle(CString strTitle) 
{ 
#if(ENABLE_TITLE)
	m_Title.UpdateTitle(strTitle); 
#endif
}

void CVideoWnd::ShowTitle(BOOL bShow, BOOL bForce) 
{ 
#if(ENABLE_TITLE)
	if(m_bTitle || bForce) {
		m_Title.ShowWindow(bShow?SW_SHOW:SW_HIDE); 
	}
#endif
}

void CVideoWnd::MoveBorder()
{
#if(ENABLE_BORDER)	
	CRect rcClient, rcBorder;
	GetWindowRect(&rcClient);
	m_bdLeft.MoveTo(rcClient);
	m_bdTop.MoveTo(rcClient);
	rcBorder = rcClient;
	rcBorder.left = rcClient.right - BORDER_BOLD;
	m_bdRight.MoveTo(rcBorder);
	rcBorder = rcClient;
	rcBorder.top  = rcClient.bottom - BORDER_BOLD;
	m_bdBottom.MoveTo(rcBorder);
#endif
}

void CVideoWnd::ShowBorder(BOOL bShow)
{
#if(ENABLE_BORDER)	
	int nShow = bShow ? SW_SHOW : SW_HIDE;
	m_bdBottom.ShowWindow(nShow);
	m_bdTop.ShowWindow(nShow);
	m_bdRight.ShowWindow(nShow);
	m_bdLeft.ShowWindow(nShow);
#endif
}

void CVideoWnd::OnShowNormal()
{
	CRect rcPos, rcTemp;
	CTW68XXDlg* pMainDlg = (CTW68XXDlg*)AfxGetMainWnd();
	
	pMainDlg->ShowVideoWnds(-2, FALSE);
	pMainDlg->ShowVideoWnds(-1, TRUE);
}

void CVideoWnd::OnShowLarge()
{
	CRect rcPos, rcTemp;
	CTW68XXDlg* pMainDlg = (CTW68XXDlg*)AfxGetMainWnd();
	
 	pMainDlg->ShowVideoWnds(-1, FALSE);
	pMainDlg->ShowVideoWnds(ChannelId(), TRUE);
}

void CVideoWnd::OnShowFull()
{
	CRect rcPos, rcTemp;
	CTW68XXDlg* pMainDlg = (CTW68XXDlg*)AfxGetMainWnd();
	
	pMainDlg->ShowVideoWnds(-3, m_bFull);
}

void CVideoWnd::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	m_nState = (m_nState+1) % VShow_full;
	switch(m_nState)		
	{
	case VShow_normal:
		OnShowNormal();
		break;
	case VShow_large:
		OnShowLarge();
		break;
	default:
		break;
	}
	
	CStatic::OnLButtonDblClk(nFlags, point);
}

void CVideoWnd::OnRButtonDblClk(UINT nFlags, CPoint point) 
{
	m_bFull = !m_bFull;

	OnShowFull();

	CStatic::OnRButtonDblClk(nFlags, point);
}

void CVideoWnd::OnLButtonDown(UINT nFlags, CPoint point) 
{
	OnSetCurrent();

	CStatic::OnLButtonDown(nFlags, point);
}

HBRUSH CVideoWnd::CtlColor(CDC* pDC, UINT nCtlColor) 
{
	return m_bkBrush;
}

void CVideoWnd::Y41PToYuv420p(PBYTE y41p, PBYTE yuv420p, int w, int h)
{
	PBYTE y = yuv420p;
	PBYTE u = NULL;
	PBYTE v = NULL;
	int   i, j, pitch = 0;
	
	u = yuv420p + (h*w);
	v = u + (h*w/4);
	pitch = w*3/2;
	
	for(i=0; i<h/2; i++) {
		for(j=0; j<pitch/12; j++) {
			*u++ = *y41p;
			*u++ = *y41p++;
			*y++ = *y41p++;
			*v++ = *y41p;
			*v++ = *y41p++;
			*y++ = *y41p++;
			
			*u++ = *y41p;
			*u++ = *y41p++;
			*y++ = *y41p++;
			*v++ = *y41p;
			*v++ = *y41p++;
			*y++ = *y41p++;
			
			*y++ = *y41p++;
			*y++ = *y41p++;
			*y++ = *y41p++;
			*y++ = *y41p++;
		}
		
		for(j=0; j<pitch/12; j++) {
			y41p++;
			*y++ = *y41p++;
			y41p++;
			*y++ = *y41p++;
			
			y41p++;
			*y++ = *y41p++;
			y41p++;
			*y++ = *y41p++;
			
			*y++ = *y41p++;
			*y++ = *y41p++;
			*y++ = *y41p++;
			*y++ = *y41p++;
		}
	}
}

void CVideoWnd::Y41PToYuy2(PBYTE y41p, PBYTE yuy2, int w, int h)
{
	int i, j;
	int pitch = w*3/2;
	
	for(i=0; i<h; i++) {
		for(j=0; j<pitch/12; j++) {
			yuy2[0] = y41p[1];	//Y0
			yuy2[1] = y41p[0];	//U0
			yuy2[2] = y41p[3];	//Y1
			yuy2[3] = y41p[2];	//V0

			yuy2[4] = y41p[5];	
			yuy2[5] = y41p[0];	
			yuy2[6] = y41p[7];
			yuy2[7] = y41p[2];

			yuy2[8] = y41p[8];
			yuy2[9] = y41p[4];
			yuy2[10]= y41p[9];
			yuy2[11]= y41p[6];

			yuy2[12]= y41p[10];
			yuy2[13]= y41p[4];
			yuy2[14]= y41p[11];
			yuy2[15]= y41p[6];

			yuy2 += 16;
			y41p += 12;
		}
	}
}

void CVideoWnd::Yuv420pToYuy2(PBYTE yuv420p, PBYTE yuy2, int w, int h)
{
	int i, j;
	PBYTE y = yuv420p;
	PBYTE u = yuv420p + w*h;
	PBYTE v = u + w*h/4;
	
	for(i=0; i<h; i++) {
		for(j=0; j<w/2; j++) {
			yuy2[0]	= y[0];			//Y
			yuy2[1] = u[j];	//U0
			yuy2[2] = y[1];			//Y1
			yuy2[3] = v[j];	//V0
			
			yuy2 += 4;
			y += 2;
		}
		if((i>0) && !(i%2)) {
			u += w/2;
			v += w/2;
		}
	}
}
