// TW68XXDevice.cpp
// Implementation of CTW68XXDevice device class
//
// Generated by DriverWizard version DriverStudio 3.1.0 (Build 1722)
// Requires Compuware's DriverWorks classes
//

#pragma warning(disable:4065) // Allow switch statement with no cases

#include "TW68XX.h"
#include "TW68XXDevice.h"
#include "I2C6864.h"

#define MAKE_TIMERID(deviceId, idTimer)	(((deviceId) << 16) | (idTimer))
#define GET_TIMERID(val)				((val) & 0xFFFF)
#define GET_DEVICEID(val)				((val) >> 16)

extern KDebugOnlyTrace t;	// Global driver trace object	

GUID TW68XXDevice_Guid = TW68XXDevice_CLASS_GUID;

KArray<KMemoryToProcessMap*> CTW68XXDevice::m_Memory;

class CAutoSpinLock
{
public:
	CAutoSpinLock(KSpinLock * pLock, BOOLEAN AtDispatchLevel = FALSE) 
	{
		m_pLock = pLock;
		if(m_pLock)	m_pLock->Lock(AtDispatchLevel);
	}
	~CAutoSpinLock() 
	{
		if(m_pLock)	m_pLock->Unlock();
	}
	
protected:
	KSpinLock * m_pLock;
};

////////////////////////////////////////////////////////////////////////
//  class CContiguousMemory
////////////////////////////////////////////////////////////////////////

CContiguousMemory::CContiguousMemory() :
m_pCommBuffer(NULL),
m_Size(0),
m_pVirtualBase(NULL),
m_pMappedBase(NULL)
{
	m_LogicalAddress.QuadPart = 0;
}

CContiguousMemory::CContiguousMemory(KDmaAdapter* pAdapter, ULONG size, BOOLEAN CacheEnabled) :
m_pCommBuffer(NULL),
m_Size(size),
m_pVirtualBase(NULL),
m_pMappedBase(NULL)
{
	Initialize(pAdapter, size, CacheEnabled);
}

VOID CContiguousMemory::Initialize(KDmaAdapter* pAdapter, ULONG size, BOOLEAN CacheEnabled)
{
	m_pCommBuffer = new (NonPagedPool) KCommonDmaBuffer(pAdapter, size, CacheEnabled);
	if(m_pCommBuffer != NULL) {
		m_Size = size;
		m_pVirtualBase  = m_pCommBuffer->VirtualAddress();
		m_LogicalAddress= m_pCommBuffer->LogicalAddress(); 
		m_Memory = m_pCommBuffer->Mdl();
	}
	else {
		m_Size = 0;
		m_LogicalAddress.QuadPart = 0;
	}
}

VOID CContiguousMemory::Invalidate(void)
{
	if(m_pMappedBase != NULL) {
		m_Memory.Unmap(m_pMappedBase);
		m_pMappedBase = NULL;
	}
	if(m_pCommBuffer != NULL) {
		m_Memory.Invalidate();
		m_pCommBuffer->Invalidate();
		delete m_pCommBuffer;
		m_pCommBuffer = NULL;
	}
	m_pVirtualBase = NULL;
	m_LogicalAddress.QuadPart = 0;
	m_Size = 0;
}

PVOID CContiguousMemory::MapToUserSpace()
{
	if(!m_Memory.IsValid()) {
		return NULL;
	}

	m_pMappedBase = m_Memory.MapToUserSpaceEx(0, MmCached, HighPagePriority, 0);
	return m_pMappedBase;
}

VOID CContiguousMemory::Unmap(PVOID BaseAddress)
{
	if(m_Memory.IsValid() && (BaseAddress!=NULL)) {
		if(m_pMappedBase == BaseAddress) {
			m_pMappedBase = NULL;
		}
		m_Memory.Unmap(BaseAddress);
	}
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::CTW68XXDevice
//
//	Routine Description:
//		This is the constructor for the Functional Device Object, or FDO.
//		It is derived from KPnpDevice, which builds in automatic
//	    dispatching of subfunctions of IRP_MJ_POWER and IRP_MJ_PNP to
//		virtual member functions.
//
//	Parameters:
//		Pdo - Physical Device Object - this is a pointer to a system
//			device object that represents the physical device.
//
//		Unit - Unit number. This is a number to append to the device's
//			base device name to form the Logical Device Object's name
//
//	Return Value:
//		None   
//
//	Comments:
//		The object being constructed contains a data member (m_Lower) of type
//		KPnpLowerDevice. By initializing it, the driver binds the FDO to the
//		PDO and creates an interface to the upper edge of the system class driver.
//
CTW68XXDevice::CTW68XXDevice(PDEVICE_OBJECT Pdo, ULONG Unit) :
	KPnpDevice(Pdo, &TW68XXDevice_Guid), KPciConfiguration(Pdo)
{
	// Check constructor status
    if ( ! NT_SUCCESS(m_ConstructorStatus) )
	{
	    return;
	}

	// Remember our unit number
	m_Unit = Unit;

	// Initialize the lower device
	m_Lower.Initialize(this, Pdo);

    // Inform the base class of the lower edge device object
	SetLowerDevice(&m_Lower);

	// Initialize the PnP Policy settings to the "standard" policy
	SetPnpPolicy();

// TODO:	Customize the PnP Policy for this device by setting
//			flags in m_Policies.

	// Initialize the Power Policy settings to the "standard" policy
	SetPowerPolicy();

	m_isOpened = FALSE;
	m_isIoCtrl = FALSE;
	m_i2c = NULL;
	m_uInIsr = 0;
	m_bProcess32 = FALSE;
}


////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::~CTW68XXDevice
//
//	Routine Description:
//		This is the destructor for the Functional Device Object, or FDO.
//
//	Parameters:
//		None
//
//	Return Value:
//		None
//
//	Comments:
//		None
//

CTW68XXDevice::~CTW68XXDevice()
{
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::DefaultPnp
//
//	Routine Description:
//		Default handler for IRP_MJ_PNP
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result returned from lower device
//
//	Comments:
//		This routine just passes the IRP through to the lower device. It is 
//		the default handler for IRP_MJ_PNP. IRPs that correspond to
//		any virtual members of KpnpDevice that handle minor functions of
//		IRP_MJ_PNP and that are not overridden get passed to this routine.
//

NTSTATUS CTW68XXDevice::DefaultPnp(KIrp I) 
{
	I.ForceReuseOfCurrentStackLocationInCalldown();
	return m_Lower.PnpCall(this, I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::DefaultPower
//
//	Routine Description:
//		Default handler for IRP_MJ_POWER 
//  
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result returned from lower device
//
//	Comments:
//		This routine just passes the IRP through to the lower device. It is 
//		the default handler for IRP_MJ_POWER.
//

NTSTATUS CTW68XXDevice::DefaultPower(KIrp I) 
{
	I.IndicatePowerIrpProcessed();
	I.CopyParametersDown();
	return m_Lower.PnpPowerCall(this, I);
}

////////////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::SystemControl
//
//	Routine Description:
//		Default handler for IRP_MJ_SYSTEM_CONTROL
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result returned from lower device
//
//	Comments:
//		This routine just passes the IRP through to the next device since this driver
//		is not a WMI provider.
//

NTSTATUS CTW68XXDevice::SystemControl(KIrp I) 
{
	I.ForceReuseOfCurrentStackLocationInCalldown();
	return m_Lower.PnpCall(this, I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::Invalidate
//
//	Routine Description:
//		Calls Invalidate methods for system resources
//
//	Parameters:
//		None
//
//	Return Value:
//		None
//
//	Comments:
//		This function is called from OnStopDevice, OnRemoveDevice and
//		OnStartDevice (in error conditions).  It calls the Invalidate
//		member funcitons for each resource to free the underlying system
//		resource if allocated.  It is safe to call Invalidate more than
//		once for a resource, or for an uninitialized resource.

VOID CTW68XXDevice::Invalidate()
{
//	CAutoSpinLock autoLock(&m_syncLock);	//ʹܻᵼͣ豸ʱ

	m_isIoCtrl = TRUE;
	m_Irq.Invalidate();

	Uninit(TRUE);

	//m_Memory.Invalidate();	//豸ʱﲻܵãͣһ豸ٵOnCmdInitϵͳ

	//ɾӳPCIڴ
	if(m_memPci.IsValid()) {
		m_memPci.Invalidate();
	}

	m_isIoCtrl = FALSE;
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::OnStartDevice
//
//	Routine Description:
//		Handler for IRP_MJ_PNP subfcn IRP_MN_START_DEVICE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//		Initialize the physical device. Typically, the driver initializes
//		physical resources here.  Call I.AllocatedResources() for a list
//		of the raw resources that the system has assigned to the device,
//		or I.TranslatedResources() for the translated resource list.
//

NTSTATUS CTW68XXDevice::OnStartDevice(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

	I.Information() = 0;

	// Get the list of raw resources from the IRP
	// Get the list of translated resources from the IRP
	PCM_RESOURCE_LIST pResListTranslated = I.TranslatedResources();
	ULONG i;
	TMemNode node;

	if(!m_Memory.IsValid()) {
		status = m_Memory.Initialize(pResListTranslated->Count, MAX_PCIMEM_NUM, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			return status;
		}
	}
	status = m_memPci.Initialize(pResListTranslated->Count, MAX_PCIMEM_NUM, NonPagedPool);
	if(!NT_SUCCESS(status)) {
		return status;
	}
	for(i=0; i<pResListTranslated->Count; i++) {
		node.addr = (PVOID)pResListTranslated->List[i].PartialResourceList.PartialDescriptors->u.Memory.Start.LowPart;
		node.len  = pResListTranslated->List[i].PartialResourceList.PartialDescriptors->u.Memory.Length;
		status = m_memPci.InsertNext(node, NULL);
		if(!NT_SUCCESS(status)) {
			return status;
		}
	}

#define MAX_DMA_LENGTH	((ULONG) -1)	// 0x100000 is 1 MB

	// Initialize the device descriptor for the DMA object using the assigned resource
	DEVICE_DESCRIPTION dd;
	RtlZeroMemory(&dd, sizeof(dd));
	dd.Version = DEVICE_DESCRIPTION_VERSION;
	dd.Master = TRUE;
	dd.ScatterGather = TRUE;
	dd.DemandMode = FALSE;
	dd.AutoInitialize = FALSE;
	dd.Dma32BitAddresses = TRUE;
	dd.IgnoreCount = FALSE;
	dd.DmaChannel = ((ULONG) ~0);
	dd.InterfaceType = PCIBus;
	dd.DmaWidth = Width32Bits;	// PCI default width
	dd.DmaSpeed = Compatible;
	dd.MaximumLength = MAX_DMA_LENGTH;

	// Initialize the DMA adapter object
	if(!m_Dma.Initialize(&dd, m_Lower.PDO())) {
		return STATUS_UNSUCCESSFUL;
	}

	// For each memory mapped region, initialize the memory mapped range
	// using the resources provided by NT. Once initialized, each memory
	// range's base virtual address in system space can be obtained by calling
	// member Base(). Each memory range's physical address in CPU space can
	// obtained by calling CpuPhysicalAddress(). To access the memory mapped
	// range use member functions such as inb/outb, or the array element operator. 

	// Initialize and connect the interrupt
	status = m_Irq.InitializeAndConnect(
		pResListTranslated, 
		LinkTo(Isr_Irq), 
		this
		);
	if (!NT_SUCCESS(status))
	{
		Invalidate();
		return status;		
	}

	// Setup the DPC to be used for interrupt processing
	m_DpcFor_Irq.Setup(LinkTo(DpcFor_Irq), this);

	return status;
}


////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::OnStopDevice
//
//	Routine Description:
//		Handler for IRP_MJ_PNP subfcn IRP_MN_STOP_DEVICE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//		The system calls this when the device is stopped.
//		The driver should release any hardware resources
//		in this routine.
//
//		The base class passes the irp to the lower device.
//

NTSTATUS CTW68XXDevice::OnStopDevice(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

	// Device stopped, release the system resources.
	Invalidate();

	return status;
	
	// The following macro simply allows compilation at Warning Level 4
	// If you reference this parameter in the function simply remove the macro.
	UNREFERENCED_PARAMETER(I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::OnRemoveDevice
//
//	Routine Description:
//		Handler for IRP_MJ_PNP subfcn IRP_MN_REMOVE_DEVICE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//		The system calls this when the device is removed.
//		Our PnP policy will take care of 
//			(1) giving the IRP to the lower device
//			(2) detaching the PDO
//			(3) deleting the device object
//

NTSTATUS CTW68XXDevice::OnRemoveDevice(KIrp I)
{
	// Device removed, release the system resources.
	Invalidate();

	return STATUS_SUCCESS;

	// The following macro simply allows compilation at Warning Level 4
	// If you reference this parameter in the function simply remove the macro.
	UNREFERENCED_PARAMETER(I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::OnDevicePowerUp
//
//	Routine Description:
//		Handler for IRP_MJ_POWER with minor function IRP_MN_SET_POWER
//		for a request to go to power on state from low power state
//
//	Parameters:
//		I - IRP containing POWER request
//
//	Return Value:
//		NTSTATUS - Status code indicating success or failure
//
//	Comments:
//		This routine implements the OnDevicePowerUp function.
//		This function was called by the framework from the completion
//		routine of the IRP_MJ_POWER dispatch handler in KPnpDevice.
//		The bus driver has completed the IRP and this driver can now
//		access the hardware device.  
//		This routine runs at dispatch level.
//	

NTSTATUS CTW68XXDevice::OnDevicePowerUp(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

// TODO:	Service the device.
//			Restore any context to the hardware device that
//			was saved during the handling of a power down request.
//			See the OnDeviceSleep function.
//			Do NOT complete this IRP.
//
	return status;

	// The following macro simply allows compilation at Warning Level 4
	// If you reference this parameter in the function simply remove the macro.
	UNREFERENCED_PARAMETER(I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::OnDeviceSleep
//
//	Routine Description:
//		Handler for IRP_MJ_POWER with minor function IRP_MN_SET_POWER
//		for a request to go to a low power state from a high power state
//
//	Parameters:
//		I - IRP containing POWER request
//
//	Return Value:
//		NTSTATUS - Status code indicating success or failure
//
//	Comments:
//		This routine implements the OnDeviceSleep function.
//		This function was called by the framework from the IRP_MJ_POWER 
//		dispatch handler in KPnpDevice prior to forwarding to the PDO.
//		The hardware has yet to be powered down and this driver can now
//		access the hardware device.  
//		This routine runs at passive level.
//	

NTSTATUS CTW68XXDevice::OnDeviceSleep(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

// TODO:	Service the device.
//			Save any context to the hardware device that will be required 
//			during a power up request. See the OnDevicePowerUp function.
//			Do NOT complete this IRP.  The base class handles forwarding
//			this IRP to the PDO.
//

#if(0)
	KMemoryRange* regMem = NULL;
	PHYSICAL_ADDRESS memBase;
	int		 i, j;
	CAutoSpinLock autoLock(&m_syncLock);

	m_isIoCtrl = TRUE;

	if(m_devList.IsValid()) {
		for(j=0; j<m_devList.MaxInserted()+1; j++) {
			if(m_devList[j]->tmList.IsValid()) {
				for(i=0; i<m_devList[j]->tmList.MaxInserted()+1; i++) {
					if(m_devList[j]->tmList[i].tmc != NULL) {
						m_devList[j]->tmList[i].tmc->Cancel();
					}
				}
			}
		}
	}

	//Crash
	if(m_crsList.IsValid()) {
		for(i=0; i<m_crsList.MaxInserted()+1; i++) {
			memBase.QuadPart = (LONGLONG)m_crsList[i].regBase;
			regMem = FindMemRage(memBase);
			if(regMem && regMem->IsValid()) {
				if(m_crsList[i].regCrash != (ULONG)-1) {
					ULONG value = regMem->ind(m_crsList[i].regCrash);
					value |= m_crsList[i].higMask;
					value &= ~m_crsList[i].lowMask;
					regMem->outd(m_crsList[i].regCrash, value);
				}
			}
		}
	}
	
	m_isIoCtrl = FALSE;
#endif

	return status;

	// The following macro simply allows compilation at Warning Level 4
	// If you reference this parameter in the function simply remove the macro.
	UNREFERENCED_PARAMETER(I);
}

NTSTATUS CTW68XXDevice::OnSystemSleep(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

	KMemoryRange* regMem = NULL;
	PHYSICAL_ADDRESS memBase;
	int		 i, j;
	CAutoSpinLock autoLock(&m_syncLock);

	m_isIoCtrl = TRUE;

	if(m_devList.IsValid()) {
		for(j=0; j<m_devList.MaxInserted()+1; j++) {
			if(m_devList[j]->tmList.IsValid()) {
				for(i=0; i<m_devList[j]->tmList.MaxInserted()+1; i++) {
					if(m_devList[j]->tmList[i].tmc != NULL) {
						m_devList[j]->tmList[i].tmc->Cancel();
					}
				}
			}
		}
	}

	//Crash
	if(m_crsList.IsValid()) {
		for(i=0; i<m_crsList.MaxInserted()+1; i++) {
			memBase.QuadPart = (LONGLONG)m_crsList[i].regBase;
			regMem = FindMemRage(memBase);
			if(regMem && regMem->IsValid()) {
				if(m_crsList[i].regCrash != (ULONG)-1) {
					ULONG value = regMem->ind(m_crsList[i].regCrash);
					value |= m_crsList[i].higMask;
					value &= ~m_crsList[i].lowMask;
					regMem->outd(m_crsList[i].regCrash, value);
				}
			}
		}
	}
	
	m_isIoCtrl = FALSE;

	return status;

	// The following macro simply allows compilation at Warning Level 4
	// If you reference this parameter in the function simply remove the macro.
	UNREFERENCED_PARAMETER(I);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::Create
//
//	Routine Description:
//		Handler for IRP_MJ_CREATE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//

NTSTATUS CTW68XXDevice::Create(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

	if(m_isOpened) {
		status = STATUS_ACCESS_DENIED;
	}
	m_isOpened = TRUE;

	status = I.PnpComplete(this, status, IO_NO_INCREMENT);

	return status;
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::Close
//
//	Routine Description:
//		Handler for IRP_MJ_CLOSE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//

NTSTATUS CTW68XXDevice::Close(KIrp I)
{
	NTSTATUS status;
	KMemoryRange* regMem = NULL;
	PHYSICAL_ADDRESS memBase;
	int		 i, j, iPos;
	CAutoSpinLock autoLock(&m_syncLock);

	m_isIoCtrl = TRUE;

	if(m_devList.IsValid()) {
		for(j=0; j<m_devList.MaxInserted()+1; j++) {
			if(m_devList[j]->tmList.IsValid()) {
				for(i=0; i<m_devList[j]->tmList.MaxInserted()+1; i++) {
					if(m_devList[j]->tmList[i].tmc != NULL) {
						m_devList[j]->tmList[i].tmc->Cancel();
					}
				}
			}
		}
	}

	//Crash
	if(m_crsList.IsValid()) {
		for(i=0; i<m_crsList.MaxInserted()+1; i++) {
			memBase.QuadPart = (LONGLONG)m_crsList[i].regBase;
			regMem = FindMemRage(memBase);
			if(regMem && regMem->IsValid()) {
				if(m_crsList[i].regCrash != (ULONG)-1) {
					ULONG value = regMem->ind(m_crsList[i].regCrash);
					value |= m_crsList[i].higMask;
					value &= ~m_crsList[i].lowMask;
					regMem->outd(m_crsList[i].regCrash, value);
				}
			}
		}
		m_crsList.Invalidate();
	}
	
	Uninit(FALSE);

	status = I.PnpComplete(this, STATUS_SUCCESS, IO_NO_INCREMENT);

	m_isOpened = FALSE;
	m_isIoCtrl = FALSE;

    return status;
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::Cleanup
//
//	Routine Description:
//		Handler for IRP_MJ_CLEANUP	
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//

NTSTATUS CTW68XXDevice::CleanUp(KIrp I)
{
// TODO:	Insert your code to respond to the CLEANUP message.
//			This code cleans up the single Wizard created queue.  If you
//			have created additional queues,	or have any outstanding Irps
//			stored in some other fashion in your driver, you should clean
//			these up as well for the file object specified in the cleanup Irp.

	m_DriverManagedQueue.PnpCleanUp(this, I.FileObject());
	return I.PnpComplete(this, STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::DeviceControl
//
//	Routine Description:
//		Handler for IRP_MJ_DEVICE_CONTROL
//
//	Parameters:
//		I - Current IRP
// 
//	Return Value:
//		None
//
//	Comments:
//		This routine is the first handler for Device Control requests.
//		Some function codes may be handled immediately, 
//		while others may be serialized through the StartIo routine.
//		
//		The KPnpDevice class handles restricting IRP flow
//		if the device is stopping or being removed.
//

NTSTATUS CTW68XXDevice::DeviceControl(KIrp I) 
{
	NTSTATUS status;

	switch (I.IoctlCode())
	{
		case IOCTL_COMMAND:
			status = IOCTL_COMMAND_Handler(I);
			break;

		default:
			// Unrecognized IOCTL request
			status = STATUS_INVALID_PARAMETER;
			break;
	}

	// If the IRP was queued, or its IOCTL handler deferred processing using some
	// driver specific scheme, the status variable is set to STATUS_PENDING.
	// In this case we simply return that status, and the IRP will be completed
	// later.  Otherwise, complete the IRP using the status returned by the
	// IOCTL handler.
	if (status == STATUS_PENDING)
	{
		return status;
	}
	else
	{
		return I.PnpComplete(this, status);
	}
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::IOCTL_COMMAND_Handler
//
//	Routine Description:
//		Handler for IO Control Code IOCTL_COMMAND
//
//	Parameters:
//		I - IRP containing IOCTL request
//
//	Return Value:
//		NTSTATUS - Status code indicating success or failure
//
//	Comments:
//		This routine implements the IOCTL_COMMAND function.
//		This routine runs at passive level.
//

NTSTATUS CTW68XXDevice::IOCTL_COMMAND_Handler(KIrp I)
{
	PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
	NTSTATUS status = STATUS_SUCCESS;

	I.Information() = 0;
	I.Status() = STATUS_SUCCESS;
	m_isIoCtrl = TRUE;

	switch(pInfo->cmd) {
	case cmdInit:
		status = OnCmdInit(I);
		break;
	case cmdUninit:
		status = OnCmdUninit(I);
		break;
	case cmdInitDevice:
		status = OnCmdInitDevice(I);
		break;
	case cmdUninitDevice:
		status = OnCmdUninitDevice(I);
		break;
	case cmdAppendCrash:
		status = OnCmdAppendCrash(I);
		break;
	case cmdAppendIrq:
		status = OnCmdAppendIrq(I);
		break;
	case cmdModifyIrq:
		status = OnCmdModifyIrq(I);
		break;
	case cmdRemoveIrq:
		status = OnCmdRemoveIrq(I);
		break;
	case cmdModifyEvt:
		status = OnCmdModifyEvt(I);
		break;
	case cmdAppendTimer:
		status = OnCmdAppendTimer(I);
		break;
	case cmdRemoveTimer:
		status = OnCmdRemoveTimer(I);
		break;
	case cmdModifyTimer:
		status = OnCmdModifyTimer(I);
		break;
	case cmdCancelTimer:
		status = OnCmdCancelTimer(I);
		break;
	case cmdAppendI2C:
		status = OnCmdAppendI2C(I);
		break;
	case cmdRemoveI2C:
		status = OnCmdRemoveI2C(I);
		break;
	case cmdReadMemBYTEs:
		status = OnCmdReadMemBYTEs(I);
		break;
	case cmdReadMemWORDs:
		status = OnCmdReadMemWORDs(I);
		break;
	case cmdReadMemDWORDs:
		status = OnCmdReadMemDWORDs(I);
		break;
	case cmdWriteMemBYTEs:
		status = OnCmdWriteMemBYTEs(I);
		break;
	case cmdWriteMemWORDs:
		status = OnCmdWriteMemWORDs(I);
		break;
	case cmdWriteMemDWORDs:
		status = OnCmdWriteMemDWORDs(I);
		break;
	case cmdReadPciConfigSpace:
		status = OnCmdReadPciConfigSpace(I);
		break;
	case cmdWritePciConfigSpace:
		status = OnCmdWritePciConfigSpace(I);
		break;
	case cmdReadPciConfigSpaceEx:
		status = OnCmdReadPciConfigSpaceEx(I);
		break;
	case cmdWritePciConfigSpaceEx:
		status = OnCmdWritePciConfigSpaceEx(I);
		break;
	case cmdAllocMemory:
		status = OnCmdAllocMemory(I);
		break;
	case cmdFreeMemory:
		status = OnCmdFreeMemory(I);
		break;
	case cmdMapMemory:
		status = OnCmdMapMemory(I);
		break;
	case cmdUnmapMemory:
		status = OnCmdUnmapMemory(I);
		break;
	case cmdMapPciMemory:
		status = OnCmdMapPciMemory(I);
		break;
	case cmdUnmapPciMemory:
		status = OnCmdUnmapPciMemory(I);
		break;
	default:
		status = STATUS_INVALID_PARAMETER;
		I.Status() = status;
		break;
	}

	m_isIoCtrl = FALSE;

	return status;
}

////////////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice_DriverManagedQueue::StartIo
//	
//	Routine Description:
//		This routine is called when an IRP is taken off
//		the Driver Managed Queue (used for serializing I/O) and
//		presented for processing.
//
//	Parameters:
//		I - IRP removed from queue
//
//	Return Value:
//		None
//
//	Comments:
//

VOID CTW68XXDevice_DriverManagedQueue::StartIo(KIrp I)
{
	// The KDriverManagedQueueEx class gives us the Irp in a non-cancelable state
	// (cancel routine set to NULL) so we can process it without having to worry
	// about clearing the cancel routine first, as is the case with system queuing,
	// or the legacy class KDriverManagedQueue. You may want to set a different cancel
	// routine here, or at other points during the processing of this Irp.

	// Find the device class so we can call the serialized
	// routines in the device class.  The handlers can be
	// moved to the DriverManagedQueue class if it is more
	// convenient.

	// Start processing request.

	// Switch on the IRP's function:
	switch (I.MajorFunction())
	{
		case IRP_MJ_DEVICE_CONTROL:
			switch (I.IoctlCode())
			{
				default:
					// We queued a request that shouldn't have been queued
					// (should never get here)
					ASSERT(FALSE);
					break;
			}
			break;

		default:
			// Error - unexpected IRP received
			// NextIrp completes this IRP and starts processing
			// for the next IRP in the queue.
			ASSERT(FALSE);
			I.Status() = STATUS_INVALID_PARAMETER;
			PnpNextIrp(I);
			break;
	}
}

VOID CTW68XXDevice::DpcOnTimer(KDPC *Dpc, PVOID DeferredContext, PVOID Arg1, PVOID Arg2)
{
	((CTW68XXDevice*)DeferredContext)->OnTimer((LONG)Dpc->SystemArgument1, (LONG)Dpc->SystemArgument2);
	
	// The following macros simply allows compilation at Warning Level 4
	// If you reference these parameters in the function simply remove the macro.
	UNREFERENCED_PARAMETER(Arg1);
	UNREFERENCED_PARAMETER(Arg2);
}

VOID CTW68XXDevice::OnTimer(LONG iDevice, LONG iTimer)
{
	KMemoryRange*		regValue = NULL;
	PHYSICAL_ADDRESS	memBase;
	ULONG val;
	LONG  i, tid;
	
	if(m_isIoCtrl || (iDevice<0) || (iDevice>m_devList.MaxInserted())) {
		return;
	}

	if((iTimer<0) || (iTimer>m_devList[iDevice]->tmList.MaxInserted())) {
		return;
	}

	if(m_devList[iDevice]->tmList[iTimer].irq) {
		i  = (m_devList[iDevice]->tmList[iTimer].irqPos >> 16);
		val= (m_devList[iDevice]->tmList[iTimer].irqPos & 0xFFFF);
		m_devList[iDevice]->irqList[i].status[val] = m_devList[iDevice]->tmList[iTimer].irqMask;
		OnIrq();
	}
	else {
		memBase.QuadPart = (LONGLONG)m_devList[iDevice]->tmList[iTimer].tmi.regBase;
		regValue = FindMemRage(memBase);

		//ҪдļĴд
		if(regValue && regValue->IsValid()) {
			tid = m_devList[iDevice]->tmList[iTimer].timerId;
			for(i=0; i<MAX_RW_TIMER; i++) {
				if(m_devList[iDevice]->tmList[iTimer].tmi.regRead[i] != (ULONG)-1) {
					m_devList[iDevice]->datShare.irqData[tid].regReadTimer[i] = regValue->ind(m_devList[iDevice]->tmList[iTimer].tmi.regRead[i]);
				}
				else {
					m_devList[iDevice]->datShare.irqData[tid].regReadTimer[i] = 0;
				}
				if(m_devList[iDevice]->tmList[iTimer].tmi.regWrite[i] != (ULONG)-1) {
					CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
					val = regValue->ind(m_devList[iDevice]->tmList[iTimer].tmi.regWrite[i]);
					val &= ~m_devList[iDevice]->tmList[iTimer].tmi.mskWrite[i];
					val |= (m_devList[iDevice]->datShare.irqData[tid].regWriteTimer[i]&m_devList[iDevice]->tmList[iTimer].tmi.mskWrite[i]);
					regValue->outd(m_devList[iDevice]->tmList[iTimer].tmi.regWrite[i], val);
				}
			}
		}
		else {
			for(i=0; i<MAX_RW_TIMER; i++) {
				m_devList[iDevice]->datShare.irqData[m_devList[iDevice]->tmList[iTimer].timerId].regReadTimer[i] = 0;
			}
		}
	}
}

//========================================================================
void ke_wait_tick( IN ULONG tick_cnt)
//========================================================================
{
	ULONG			tick_diff;
    LARGE_INTEGER	Time0, TimeRun;

	Time0 =  KeQueryPerformanceCounter(NULL);
	do {
		TimeRun		=  KeQueryPerformanceCounter(NULL);
		tick_diff   =  TimeRun.LowPart - Time0.LowPart;
	}while(tick_diff < tick_cnt);
}

void CTW68XXDevice::OnIrq()
{
	int	  i, j, k, ik, iDev=-1;
	ULONG id, val;
	UCHAR data[3] = {0, 0, 0};
	bool  bSame=false, bWrite=false, bIrq;
	KMemoryRange*		regValue = NULL;
	LARGE_INTEGER		tkNow, tkFreq;
	PHYSICAL_ADDRESS	memBase;
	LONGLONG			dueTime = 0;
	ULONG				intStatus[MAX_INTSTATUS];
	
	tkNow = KeQueryPerformanceCounter(&tkFreq);

#if defined(_WIN64)
	if(m_bProcess32) {
		PShareData32 datShare = NULL;
		PShareData32 datShareSame = NULL;

		for(k=0; k<m_devList.MaxInserted()+1; k++) {
			if(m_devList[k] == NULL) { continue; }

			m_devList[k]->hasIrq = FALSE;
			for(i=0; i<m_devList[k]->irqList.MaxInserted()+1; i++) {
				if(!m_devList[k]->irqList.IsValid()) { continue; }

				id = m_devList[k]->irqList[i].irqId;
				bIrq = FALSE;
				datShare = &m_devList[k]->datShare32;
				for(ik=0; ik<MAX_INTSTATUS; ik++) {
					if(m_devList[k]->irqList[i].regStatus[ik] == (ULONG)-1) { continue; }
					intStatus[ik] = m_devList[k]->irqList[i].status[ik] & m_devList[k]->irqList[i].mask[ik];
					m_devList[k]->irqList[i].status[ik] = 0;
					if(intStatus[ik] > 0) {
						datShare->irqData[id].status[ik] |= intStatus[ik];
					}
				}
				bIrq = (intStatus[0] > 0);
				if(!bIrq) { continue; }

				m_devList[k]->hasIrq = TRUE;
 				memBase.QuadPart = (LONGLONG)m_devList[k]->irqList[i].regRWBase;
 				regValue = FindMemRage(memBase);

				datShare->irqData[id].irqId  = id;
				datShare->irqData[id].tkCount= tkNow;
				datShare->irqData[id].tkFreq = tkFreq;
				datShare->irqData[id].irqCount++;
				//datShare->irqData[id].evtId  = evt_DataIrq;
				if(m_devList[k]->irqList[i].freqWrite) {
					bWrite = (datShare->irqData[id].irqCount % datShare->irqData[id].freqWrite)==0;
				}
				else {
					bWrite = true;
				}
			
				//ҪдļĴд
				if(regValue && regValue->IsValid()) {
					for(j=0; j<MAX_RW_NUM/2; j++) {
						if(m_devList[k]->irqList[i].regRead[j] != (ULONG)-1) {
							if(iDev == -1) iDev = k;
							bSame = (k != iDev) &&
									(m_devList[k]->irqList[i].regRWBase  == m_devList[iDev]->irqList[i].regRWBase) &&
									(m_devList[k]->irqList[i].regRead[j] == m_devList[iDev]->irqList[i].regRead[j]);
							datShareSame = &m_devList[iDev]->datShare32;
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regRead[j])) {
								datShare->irqData[id].regRead[j] = bSame ? datShareSame->irqData[id].regRead[j] :
									regValue->ind(m_devList[k]->irqList[i].regRead[j]);
							}
							else if(bWrite && ((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C)) {
								if(bSame) {	datShare->irqData[id].regRead[j] = datShareSame->irqData[id].regRead[j]; }
								else if(m_i2c != NULL) {
									data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regRead[j]);
									data[1] = m_devList[k]->irqList[i].regRead[j] & 0xFF;
									m_i2c->i2cReadData(data, 2, (UCHAR*)&datShare->irqData[id].regRead[j], 1);
								}
							}
						}
						if(bWrite && (m_devList[k]->irqList[i].regWrite[j] != (ULONG)-1)) {
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regWrite[j])) {
								//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
								val = regValue->ind(m_devList[k]->irqList[i].regWrite[j]);
								val &= ~m_devList[k]->irqList[i].mskWrite[j];
								val |= (datShare->irqData[id].regWrite[j]&m_devList[k]->irqList[i].mskWrite[j]);
								regValue->outd(m_devList[k]->irqList[i].regWrite[j], val);
							}
							else if((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C) {
								data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regWrite[j]);
								data[1] = m_devList[k]->irqList[i].regWrite[j] & 0xFF;
								m_i2c->i2cWriteByte(data[0], data[1], (UCHAR)m_devList[k]->irqList[i].mskWrite[j], (UCHAR)datShare->irqData[id].regWrite[j]);
							}
						}
					}
				}
				else {
					for(j=0; j<MAX_RW_NUM/2; j++) {
						datShare->irqData[id].regRead[j] = 0;
					}
				}

				//Copyڴ鷽ʽﵽDMAĿbufferĿ
				if((m_devList[k]->irqList[i].mskCopy > 0) && (m_devList[k]->irqList[i].devCopy == 0x6864)) {
					PUCHAR cpyFrom=NULL, cpyDest=NULL;
					int    n = 0;

					if(m_devList[k]->irqList[i].mskCopy & (1<<m_devList[k]->deviceId)) {
						n = (datShare->irqData[id].regRead[1] >> m_devList[k]->deviceId) & 1;
						for(j=0; j<m_bufList.MaxInserted()+1; j++) {
							if(cpyFrom==NULL) {
								if((ULONG)m_bufList[j].buf->LogicalAddress().LowPart == datShare->cpyItems[n].idFrom) {
									cpyFrom = (PUCHAR)m_bufList[j].buf->VirtualAddress() + datShare->cpyItems[n].fromOffset;
								}
							}
							if(cpyDest==NULL) {
								if((ULONG)m_bufList[j].buf->LogicalAddress().LowPart == datShare->cpyItems[n].idDest) {
									cpyDest = (PUCHAR)m_bufList[j].buf->VirtualAddress() + datShare->cpyItems[n].destOffset;
								}
							}
							if(cpyDest && cpyFrom) {
								if(datShare->cpyItems[n].maxCount > 0) {
									cpyFrom += datShare->cpyItems[n].cpyStep * datShare->cpyItems[n].cpyCount;
									datShare->cpyItems[n].cpyCount = (datShare->cpyItems[n].cpyCount+1) % datShare->cpyItems[n].maxCount;
									datShare->cpyItems[!n].cpyCount= datShare->cpyItems[n].cpyCount;
								}
								RtlCopyMemory(cpyDest, cpyFrom, datShare->cpyItems[n].cpyLen);
								break;
							}
						}
					}
				}

				//ҪдļĴд
				if(regValue && regValue->IsValid()) {
					for(j=MAX_RW_NUM/2; j<MAX_RW_NUM; j++) {
						if(m_devList[k]->irqList[i].regRead[j] != (ULONG)-1) {
							if(iDev == -1) iDev = k;
							bSame = (k != iDev) &&
									(m_devList[k]->irqList[i].regRWBase  == m_devList[iDev]->irqList[i].regRWBase) &&
									(m_devList[k]->irqList[i].regRead[j] == m_devList[iDev]->irqList[i].regRead[j]);
							datShareSame = &m_devList[iDev]->datShare32;
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regRead[j])) {
								datShare->irqData[id].regRead[j] = bSame ? datShareSame->irqData[id].regRead[j] :
									regValue->ind(m_devList[k]->irqList[i].regRead[j]);
							}
							else if(bWrite && ((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C)) {
								if(bSame) {	datShare->irqData[id].regRead[j] = datShareSame->irqData[id].regRead[j]; }
								else if(m_i2c != NULL) {
									data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regRead[j]);
									data[1] = m_devList[k]->irqList[i].regRead[j] & 0xFF;
									m_i2c->i2cReadData(data, 2, (UCHAR*)&datShare->irqData[id].regRead[j], 1);
								}
							}
						}
						if(bWrite && (m_devList[k]->irqList[i].regWrite[j] != (ULONG)-1)) {
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regWrite[j])) {
								//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
								val = regValue->ind(m_devList[k]->irqList[i].regWrite[j]);
								val &= ~m_devList[k]->irqList[i].mskWrite[j];
								val |= (datShare->irqData[id].regWrite[j]&m_devList[k]->irqList[i].mskWrite[j]);
								regValue->outd(m_devList[k]->irqList[i].regWrite[j], val);
							}
							else if((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C) {
								data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regWrite[j]);
								data[1] = m_devList[k]->irqList[i].regWrite[j] & 0xFF;
								m_i2c->i2cWriteByte(data[0], data[1], (UCHAR)m_devList[k]->irqList[i].mskWrite[j], (UCHAR)datShare->irqData[id].regWrite[j]);
							}
						}
					}
				}
				else {
					for(j=MAX_RW_NUM/2; j<MAX_RW_NUM; j++) {
						datShare->irqData[id].regRead[j] = 0;
					}
				}

				//Timer
				if(bWrite && (m_devList[k]->irqList[i].timerId != (ULONG)-1)) {
					int tid = -1;
					for(int m=0; m<m_devList[k]->tmList.MaxInserted()+1; m++) {
						if(m_devList[k]->tmList[m].timerId == m_devList[k]->irqList[i].timerId) {
							tid = m;
							break;
						}
					}
					if(tid != -1) {
						if(datShare->irqData[id].dueTime != (ULONG)-1) {
							dueTime = datShare->irqData[id].dueTime;
							dueTime = (dueTime==0) ? m_devList[k]->tmList[tid].tmi.dueTime : (-10000 * dueTime);
							m_devList[k]->tmList[tid].irq = datShare->irqData[id].timerIrq;
							m_devList[k]->tmList[tid].irqPos  = datShare->irqData[id].timerIrqPos;
							m_devList[k]->tmList[tid].irqMask = datShare->irqData[id].timerIrqMask;
							m_devList[k]->tmList[tid].tmc->Set(dueTime, DpcOnTimer, this, (PVOID)k, (PVOID)tid);
						}
						else if(regValue && regValue->IsValid()) {
							for(j=0; j<MAX_RW_TIMER; j++) {
								if(m_devList[k]->tmList[tid].tmi.regRead[j] != (ULONG)-1) {
									datShare->irqData[id].regReadTimer[j] = regValue->ind(m_devList[k]->tmList[tid].tmi.regRead[j]);
								}
								if(m_devList[k]->tmList[tid].tmi.regWrite[j] != (ULONG)-1) {
									//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
									val = regValue->ind(m_devList[k]->tmList[tid].tmi.regWrite[j]);
									val &= ~m_devList[k]->tmList[tid].tmi.mskWrite[j];
									val |= (datShare->irqData[id].regWriteTimer[j]&m_devList[k]->tmList[tid].tmi.mskWrite[j]);
									regValue->outd(m_devList[k]->tmList[tid].tmi.regWrite[j], val);
								}
							}
						}
					}
				}

				datShare->irqData[id].evtId  = evt_DataIrq;
			}

			if(!m_uInIsr) {
				if(m_devList[k]->hasIrq && m_devList[k]->evtData.IsValid()) {
					m_devList[k]->evtData.Set();
				}
			}
		}
	}
	else
#endif
	{
		PShareData datShare = NULL;
		PShareData datShareSame = NULL;
//bool bStopped = false;

		for(k=0; k<m_devList.MaxInserted()+1; k++) {
			if(m_devList[k] == NULL) { continue; }

			m_devList[k]->hasIrq = FALSE;
			for(i=0; i<m_devList[k]->irqList.MaxInserted()+1; i++) {
				if(!m_devList[k]->irqList.IsValid()) { continue; }

				id = m_devList[k]->irqList[i].irqId;
				bIrq = FALSE;
				datShare = &m_devList[k]->datShare;
				for(ik=0; ik<MAX_INTSTATUS; ik++) {
					if(m_devList[k]->irqList[i].regStatus[ik] == (ULONG)-1) { continue; }
					intStatus[ik] = m_devList[k]->irqList[i].status[ik] & m_devList[k]->irqList[i].mask[ik];
					m_devList[k]->irqList[i].status[ik] = 0;
					if(intStatus[ik] > 0) {
						datShare->irqData[id].status[ik] |= intStatus[ik];
					}
				}
				bIrq = (intStatus[0] > 0);
				if(!bIrq) { continue; }

				m_devList[k]->hasIrq = TRUE;
 				memBase.QuadPart = (LONGLONG)m_devList[k]->irqList[i].regRWBase;
 				regValue = FindMemRage(memBase);

#if(0)
				//liran for test
				if((intStatus[0] & 0x20000) && !bStopped) 
				{
					regValue->outd(0x28, 0);
					bStopped = true;
					KdPrint(("time out\n"));
					if(!m_uInIsr) 
					{
						ke_wait_tick(80000L);
					}
					regValue->outd(0x08, 0);
				}
#endif

				datShare->irqData[id].irqId  = id;
				datShare->irqData[id].tkCount= tkNow;
				datShare->irqData[id].tkFreq = tkFreq;
				datShare->irqData[id].irqCount++;
				//datShare->irqData[id].evtId  = evt_DataIrq;
				if(m_devList[k]->irqList[i].freqWrite) {
					bWrite = (datShare->irqData[id].irqCount % datShare->irqData[id].freqWrite)==0;
				}
				else {
					bWrite = true;
				}
			
				//ҪдļĴд
				if(regValue && regValue->IsValid()) {
					for(j=0; j<MAX_RW_NUM/2; j++) {
						if(m_devList[k]->irqList[i].regRead[j] != (ULONG)-1) {
							if(iDev == -1) iDev = k;
							bSame = (k != iDev) &&
									(m_devList[k]->irqList[i].regRWBase  == m_devList[iDev]->irqList[i].regRWBase) &&
									(m_devList[k]->irqList[i].regRead[j] == m_devList[iDev]->irqList[i].regRead[j]);
							datShareSame = &m_devList[iDev]->datShare;
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regRead[j])) {
								datShare->irqData[id].regRead[j] = bSame ? datShareSame->irqData[id].regRead[j] :
									regValue->ind(m_devList[k]->irqList[i].regRead[j]);
							}
							else if(bWrite && ((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C)) {
								if(bSame) {	datShare->irqData[id].regRead[j] = m_devList[iDev]->datShare.irqData[id].regRead[j]; }
								else if(m_i2c != NULL) {
									data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regRead[j]);
									data[1] = m_devList[k]->irqList[i].regRead[j] & 0xFF;
									m_i2c->i2cReadData(data, 2, (UCHAR*)&datShare->irqData[id].regRead[j], 1);
								}
							}
						}
						if(bWrite && (m_devList[k]->irqList[i].regWrite[j] != (ULONG)-1)) {
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regWrite[j])) {
								//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
								val = regValue->ind(m_devList[k]->irqList[i].regWrite[j]);
								val &= ~m_devList[k]->irqList[i].mskWrite[j];
								val |= (datShare->irqData[id].regWrite[j]&m_devList[k]->irqList[i].mskWrite[j]);
								regValue->outd(m_devList[k]->irqList[i].regWrite[j], val);
							}
							else if((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C) {
								data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regWrite[j]);
								data[1] = m_devList[k]->irqList[i].regWrite[j] & 0xFF;
								m_i2c->i2cWriteByte(data[0], data[1], (UCHAR)m_devList[k]->irqList[i].mskWrite[j], (UCHAR)datShare->irqData[id].regWrite[j]);
							}
						}
					}
				}
				else {
					for(j=0; j<MAX_RW_NUM/2; j++) {
						datShare->irqData[id].regRead[j] = 0;
					}
				}

				//Copyڴ鷽ʽﵽDMAĿbufferĿ
				if((m_devList[k]->irqList[i].mskCopy > 0) && (m_devList[k]->irqList[i].devCopy == 0x6864)) {
					PUCHAR cpyFrom=NULL, cpyDest=NULL;
					int    n = 0;

					if(m_devList[k]->irqList[i].mskCopy & (1<<m_devList[k]->deviceId)) {
						n = (datShare->irqData[id].regRead[1] >> m_devList[k]->deviceId) & 1;
						for(j=0; j<m_bufList.MaxInserted()+1; j++) {
							if(cpyFrom==NULL) {
								if((ULONG)m_bufList[j].buf->LogicalAddress().LowPart == datShare->cpyItems[n].idFrom) {
									cpyFrom = (PUCHAR)m_bufList[j].buf->VirtualAddress() + datShare->cpyItems[n].fromOffset;
								}
							}
							if(cpyDest==NULL) {
								if((ULONG)m_bufList[j].buf->LogicalAddress().LowPart == datShare->cpyItems[n].idDest) {
									cpyDest = (PUCHAR)m_bufList[j].buf->VirtualAddress() + datShare->cpyItems[n].destOffset;
								}
							}
							if(cpyDest && cpyFrom) {
								if(datShare->cpyItems[n].maxCount > 0) {
									cpyFrom += datShare->cpyItems[n].cpyStep * datShare->cpyItems[n].cpyCount;
									datShare->cpyItems[n].cpyCount = (datShare->cpyItems[n].cpyCount+1) % datShare->cpyItems[n].maxCount;
									datShare->cpyItems[!n].cpyCount= datShare->cpyItems[n].cpyCount;
								}
								RtlCopyMemory(cpyDest, cpyFrom, datShare->cpyItems[n].cpyLen);
								break;
							}
						}
					}
				}

				//ҪдļĴд
				if(regValue && regValue->IsValid()) {
					for(j=MAX_RW_NUM/2; j<MAX_RW_NUM; j++) {
						if(m_devList[k]->irqList[i].regRead[j] != (ULONG)-1) {
							if(iDev == -1) iDev = k;
							bSame = (k != iDev) &&
									(m_devList[k]->irqList[i].regRWBase  == m_devList[iDev]->irqList[i].regRWBase) &&
									(m_devList[k]->irqList[i].regRead[j] == m_devList[iDev]->irqList[i].regRead[j]);
							datShareSame = &m_devList[iDev]->datShare;
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regRead[j])) {
								datShare->irqData[id].regRead[j] = bSame ? datShareSame->irqData[id].regRead[j] :
									regValue->ind(m_devList[k]->irqList[i].regRead[j]);
							}
							else if(bWrite && ((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C)) {
								if(bSame) {	datShare->irqData[id].regRead[j] = datShareSame->irqData[id].regRead[j]; }
								else if(m_i2c != NULL) {
									data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regRead[j]);
									data[1] = m_devList[k]->irqList[i].regRead[j] & 0xFF;
									m_i2c->i2cReadData(data, 2, (UCHAR*)&datShare->irqData[id].regRead[j], 1);
								}
							}
						}
						if(bWrite && (m_devList[k]->irqList[i].regWrite[j] != (ULONG)-1)) {
							if(!IS_PCI2I2C(m_devList[k]->irqList[i].regWrite[j])) {
								//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
								val = regValue->ind(m_devList[k]->irqList[i].regWrite[j]);
								val &= ~m_devList[k]->irqList[i].mskWrite[j];
								val |= (datShare->irqData[id].regWrite[j]&m_devList[k]->irqList[i].mskWrite[j]);
								regValue->outd(m_devList[k]->irqList[i].regWrite[j], val);
							}
							else if((ULONG)m_i2c == m_devList[k]->irqList[i].regI2C) {
								data[0] = (UCHAR)ADDR_PCI2I2C(m_devList[k]->irqList[i].regWrite[j]);
								data[1] = m_devList[k]->irqList[i].regWrite[j] & 0xFF;
								m_i2c->i2cWriteByte(data[0], data[1], (UCHAR)m_devList[k]->irqList[i].mskWrite[j], (UCHAR)datShare->irqData[id].regWrite[j]);
							}
						}
					}
				}
				else {
					for(j=MAX_RW_NUM/2; j<MAX_RW_NUM; j++) {
						datShare->irqData[id].regRead[j] = 0;
					}
				}

				//Timer
				if(bWrite && (m_devList[k]->irqList[i].timerId != (ULONG)-1)) {
					int tid = -1;
					for(int m=0; m<m_devList[k]->tmList.MaxInserted()+1; m++) {
						if(m_devList[k]->tmList[m].timerId == m_devList[k]->irqList[i].timerId) {
							tid = m;
							break;
						}
					}
					if(tid != -1) {
						if(datShare->irqData[id].dueTime != (ULONG)-1) {
							dueTime = datShare->irqData[id].dueTime;
							dueTime = (dueTime==0) ? m_devList[k]->tmList[tid].tmi.dueTime : (-10000 * dueTime);
							m_devList[k]->tmList[tid].irq = datShare->irqData[id].timerIrq;
							m_devList[k]->tmList[tid].irqPos  = datShare->irqData[id].timerIrqPos;
							m_devList[k]->tmList[tid].irqMask = datShare->irqData[id].timerIrqMask;
							m_devList[k]->tmList[tid].tmc->Set(dueTime, DpcOnTimer, this, (PVOID)k, (PVOID)tid);
						}
						else if(regValue && regValue->IsValid()) {
							for(j=0; j<MAX_RW_TIMER; j++) {
								if(m_devList[k]->tmList[tid].tmi.regRead[j] != (ULONG)-1) {
									datShare->irqData[id].regReadTimer[j] = regValue->ind(m_devList[k]->tmList[tid].tmi.regRead[j]);
								}
								if(m_devList[k]->tmList[tid].tmi.regWrite[j] != (ULONG)-1) {
									//CAutoSpinLock autoLock(&m_syncLock);	//ͬTW6864豸DPC_IrqDPC_TimerжԼĴTW6864_R_DMA_CMDTW6864_R_DMA_CHANNEL_ENABLEĶд
									val = regValue->ind(m_devList[k]->tmList[tid].tmi.regWrite[j]);
									val &= ~m_devList[k]->tmList[tid].tmi.mskWrite[j];
									val |= (datShare->irqData[id].regWriteTimer[j]&m_devList[k]->tmList[tid].tmi.mskWrite[j]);
									regValue->outd(m_devList[k]->tmList[tid].tmi.regWrite[j], val);
								}
							}
						}
					}
				}

				datShare->irqData[id].evtId  = evt_DataIrq;
			}

			if(!m_uInIsr) {
				if(m_devList[k]->hasIrq && m_devList[k]->evtData.IsValid()) {
					m_devList[k]->evtData.Set();
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::DpcFor_Irq
//
//	Routine Description:
//		Deferred Procedure Call (DPC) for Irq
//
//	Parameters:
//		Arg1 - User-defined context variable
//		Arg2 - User-defined context variable
//
//	Return Value:
//		None
//
//	Comments:
//		This function is called for secondary processing of an interrupt.
//		Most code that runs at elevated IRQL should run here rather than
//		in the ISR, so that other interrupt handlers can continue to run.
//

VOID CTW68XXDevice::DpcFor_Irq(PVOID Arg1, PVOID Arg2)
{
	CAutoSpinLock autoLock(&m_syncLock);
	if(m_isIoCtrl) {
		KIrp I(m_DriverManagedQueue.CurrentIrp());
		if (!I.IsNull()) {
			
			I.Status() = STATUS_SUCCESS;
			I.Information() = 0;
			
			// PnpNextIrp completes this IRP and starts processing 
			// for the next IRP in the driver managed queue.
			m_DriverManagedQueue.PnpNextIrp(I);
		}
		return;
	}

	if(!m_uInIsr) {
		OnIrq();
	} 
	else {
		for(int k=0; k<m_devList.MaxInserted()+1; k++) {
			if(m_devList[k] == NULL) { continue; }
			if(m_devList[k]->hasIrq && m_devList[k]->evtData.IsValid()) {
				m_devList[k]->evtData.Set();
			}
		}
	}

	// The following macros simply allows compilation at Warning Level 4
	// If you reference these parameters in the function simply remove the macro.
	UNREFERENCED_PARAMETER(Arg1);
	UNREFERENCED_PARAMETER(Arg2);
}

////////////////////////////////////////////////////////////////////////
//  CTW68XXDevice::Isr_Irq
//
//	Routine Description:
//		Interrupt Service Routine (ISR) for IRQ Irq
//
//	Parameters:
//		None
// 
//	Return Value:
//		BOOLEAN		True if this is our interrupt
//
//	Comments:
//

BOOLEAN CTW68XXDevice::Isr_Irq(void)
{
	if(m_isIoCtrl) { return FALSE; }
	if(!IsInterrupt()) { return FALSE; }

	if(m_uInIsr) {
		OnIrq();
	}

	// Request deferred procedure call
	// The arguments to Request may be any values that you choose
	if (!m_DpcFor_Irq.Request(NULL, NULL))
	{
	}

	// Return TRUE to indicate that our device caused the interrupt
	return TRUE;
}

//========================================
// жϼ顪ǷǱ豸ж
// ж״̬
//========================================
bool CTW68XXDevice::IsInterrupt()
{
	KMemoryRange		*regMem;
	PHYSICAL_ADDRESS	memBase;
	ULONG	intMask[MAX_INTSTATUS];
	ULONG	intStatus[MAX_INTSTATUS];
	int		iDev[MAX_INTSTATUS];
	bool	bSame = false;
	bool	bRet  = false;
	int		i, j, ik;

	if(!m_devList.IsValid()) { return 0; }

	for(ik=0; ik<MAX_INTSTATUS; ik++) {
		iDev[ik] = -1;
	}

	for(i=0; i<m_devList.MaxInserted()+1; i++) {
		if(m_devList[i] == NULL)  { continue; }
		if(!m_devList[i]->irqList.IsValid()) { continue; }

		for(j=0; j<m_devList[i]->irqList.MaxInserted()+1; j++) {
			memBase.QuadPart = (LONGLONG)m_devList[i]->irqList[j].regBase;
			regMem = FindMemRage(memBase);
			if(regMem && regMem->IsValid()) {
				for(ik=0; ik<MAX_INTSTATUS; ik++) {
					intMask[ik] = intStatus[ik] = 0;
					if(m_devList[i]->irqList[j].regStatus[ik] == (ULONG)-1) { continue; }
					if(iDev[ik] == -1) iDev[ik] = i;
					if(m_devList[i]->irqList[j].regMask[ik] != (ULONG)-1) {
						intMask[ik] = regMem->ind(m_devList[i]->irqList[j].regMask[ik]);
					}
					else {
						intMask[ik] = m_devList[i]->irqList[j].mask[ik];
					}

					bSame = (i != iDev[ik]) &&
							(m_devList[i]->irqList[j].regBase  == m_devList[iDev[ik]]->irqList[j].regBase) &&
							(m_devList[i]->irqList[j].regStatus[ik]== m_devList[iDev[ik]]->irqList[j].regStatus[ik]);
					if(bSame) {
						intStatus[ik] = m_devList[iDev[ik]]->irqList[j].status[ik];
					}
					else {
						intStatus[ik] = regMem->ind(m_devList[i]->irqList[j].regStatus[ik]);
					}
					if(intStatus[ik]) {
						m_devList[i]->irqList[j].status[ik] = intStatus[ik];// & intMask[ik];
					}
					bRet = bRet || ((intStatus[ik] & intMask[ik]) > 0);
				}
			}
		}
	}

	//ж
	for(i=0; i<m_devList.MaxInserted()+1; i++) {
		if(m_devList[i] == NULL)  { continue; }
		if(!m_devList[i]->irqList.IsValid()) { continue; }

		for(j=0; j<m_devList[i]->irqList.MaxInserted()+1; j++) {
			memBase.QuadPart = (LONGLONG)m_devList[i]->irqList[j].regBase;
			regMem = FindMemRage(memBase);
			if(regMem && regMem->IsValid()) {
				for(ik=0; ik<MAX_INTSTATUS; ik++) {
					if(m_devList[i]->irqList[j].regClear[ik] == (ULONG)-1) { continue; }
					if(m_devList[i]->irqList[j].status[ik]) {
						if(iDev[ik] == -1) iDev[ik] = i;
						bSame = (i != iDev[ik]) &&
								(m_devList[i]->irqList[j].regBase  == m_devList[iDev[ik]]->irqList[j].regBase) &&
								(m_devList[i]->irqList[j].regClear[ik] == m_devList[iDev[ik]]->irqList[j].regClear[ik]);
						if(!bSame) {
							regMem->outd(m_devList[i]->irqList[j].regClear[ik], m_devList[i]->irqList[j].status[ik]);
						}
					}
				}
			}
		}
	}

	return bRet;
}

//=========================================
// ʼ
//=========================================
void CTW68XXDevice::Uninit(UCHAR isRemove) 
{
	int		 i=0, j=0;

	if(m_i2c != NULL) {
		delete m_i2c;
		m_i2c = NULL;
	}

	//ɾCrash
	if(m_crsList.IsValid()) {
		m_crsList.Invalidate();
	}

	//ɾTimerб
	if(m_devList.IsValid()) {
		for(i=0; i<m_devList.MaxInserted()+1; i++) {
			if(m_devList[i] == NULL) { continue; }
			if(m_devList[i]->tmList.IsValid()) {
				for(j=0; j<m_devList[i]->tmList.MaxInserted()+1; j++) {
					if(m_devList[i]->tmList[j].tmc != NULL) {
						m_devList[i]->tmList[j].tmc->Invalidate();
						delete m_devList[i]->tmList[j].tmc;
					}
				}
				m_devList[i]->tmList.Invalidate();
			}
		}
	}

	//ͷӳڴ
	if(m_Memory.IsValid()) {
		for(i=0; i<m_Memory.MaxInserted()+1; i++) {
			if(m_Memory[i]) {
				bool bFind = false;
				for(int j=0; j<m_memPci.MaxInserted()+1; j++) {
					if((PVOID)m_Memory[i]->CpuPhysicalAddress().QuadPart == m_memPci[j].addr) {
						bFind = true;
						break;
					}
				}
				if(bFind) {
					if(m_Memory[i]->IsValid()) {
						m_Memory[i]->Invalidate();
					}
					delete m_Memory[i];
					m_Memory[i] = NULL;
				}
			}
		}
	}

	if(m_bufList.IsValid()) {
		if(isRemove) {
			//ɾڴ
			for(i=0; i<m_bufList.MaxInserted()+1; i++) {
				if(m_bufList[i].buf) {
					delete m_bufList[i].buf;
					m_bufList[i].buf = NULL;
					m_bufList[i].isUsed = FALSE;
				}
			}
			m_bufList.Invalidate();
		}
		else {
			//ͷռõڴ
			for(i=0; i<m_bufList.MaxInserted()+1; i++) {
				if(m_bufList[i].isUsed) {
					m_bufList[i].buf->Unmap();
					m_bufList[i].isUsed = FALSE;
				}
			}
		}
	}

	//ɾԭӳڴ
	if(m_mapList.IsValid()) {
		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i]) {
				if(m_mapList[i]->IsValid()) {
					m_mapList[i]->UnlockPages();
					m_mapList[i]->Invalidate();
				}
				delete m_mapList[i];
				m_mapList[i] = NULL;
			}
		}
		m_mapList.Invalidate();
	}

	//ɾӳPCIڴռ
	if(m_pciMap.IsValid()) {
		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i]) {
				if(m_pciMap[i]->IsValid()) {
					m_pciMap[i]->Invalidate();
				}
				delete m_pciMap[i];
				m_pciMap[i] = NULL;
			}
		}
		m_pciMap.Invalidate();
	}

	if(m_devList.IsValid()) {
		for(i=0; i<m_devList.MaxInserted()+1; i++) {
			if(m_devList[i] == NULL) { continue; }

			if(m_devList[i]->irqList.IsValid()) {
				m_devList[i]->irqList.Invalidate();
			}
			if(m_devList[i]->basShare != NULL) {
				m_devList[i]->mapShare.Unmap(m_devList[i]->basShare);
				m_devList[i]->mapShare.Invalidate();
				m_devList[i]->basShare = NULL;
			}
			if(m_devList[i]->evtData.IsValid()) {
				m_devList[i]->evtData.Invalidate();
			}
			delete m_devList[i];
		}
		m_devList.Invalidate();
	}

	ExSetTimerResolution(0, FALSE);
}

//========================================
// ʼ
// PCIڴԼַ
//========================================
NTSTATUS CTW68XXDevice::OnCmdInit(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	KMemoryToProcessMap* pciMap;
	int		 i=0, j=0;

#if defined(_WIN64)
	m_bProcess32 = IoIs32bitProcess(I);
#endif

	//ɾTimerб
	if(m_devList.IsValid()) {
		for(i=0; i<m_devList.MaxInserted()+1; i++) {
			if(m_devList[i] == NULL) { continue; }
			if(m_devList[i]->tmList.IsValid()) {
				for(j=0; j<m_devList[i]->tmList.MaxInserted()+1; j++) {
					if(m_devList[i]->tmList[j].tmc != NULL) {
						m_devList[i]->tmList[j].tmc->Invalidate();
						delete m_devList[i]->tmList[j].tmc;
					}
				}
				m_devList[i]->tmList.Invalidate();
			}
		}
	}

	//ͷԭռõڴ
	if(!m_bufList.IsValid()) {
		status = m_bufList.Initialize(MAX_PCIMEM_NUM, MAX_PCIMEM_NUM, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			return status;
		}
	}
	else {
		for(i=0; i<m_bufList.MaxInserted()+1; i++) {
			if(m_bufList[i].isUsed) {
				m_bufList[i].buf->Unmap();
				m_bufList[i].isUsed = FALSE;
			}
		}
	}

	//ɾԭӳڴ
	if(!m_mapList.IsValid()) {
		status = m_mapList.Initialize(MAX_PCIMEM_NUM, MAX_PCIMEM_NUM, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			return status;
		}
	}
	else {
		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i]) {
				if(m_mapList[i]->IsValid()) {
					m_mapList[i]->UnlockPages();
					m_mapList[i]->Invalidate();
				}
				delete m_mapList[i];
				m_mapList[i] = NULL;
			}
		}
	}
		
	//ͷӳڴ
	for(i=0; i<m_Memory.MaxInserted()+1; i++) {
		if(m_Memory[i]) {
			bool bFind = false;
			for(int j=0; j<m_memPci.MaxInserted()+1; j++) {
				if((PVOID)m_Memory[i]->CpuPhysicalAddress().QuadPart == m_memPci[j].addr) {
					bFind = true;
					break;
				}
			}
			if(bFind) {
				if(m_Memory[i]->IsValid()) {
					m_Memory[i]->Invalidate();
				}
				delete m_Memory[i];
				m_Memory[i] = NULL;
			}
		}
	}

	//ɾԭӳڴ
	if(!m_pciMap.IsValid()) {
		status = m_pciMap.Initialize(MAX_PCIMEM_NUM, MAX_PCIMEM_NUM, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			return status;
		}
	}
	else {
		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i]) {
				if(m_pciMap[i]->IsValid()) {
					m_pciMap[i]->Invalidate();
				}
				delete m_pciMap[i];
				m_pciMap[i] = NULL;
			}
		}
	}

	//ɾCrash
	if(m_crsList.IsValid()) {
		m_crsList.Invalidate();
	}
	status = m_crsList.Initialize(MAX_IRQ_NUM, MAX_IRQ_NUM, NonPagedPool);
	if(!NT_SUCCESS(status)) {
		return status;
	}

#if defined(_WIN64)
	if(m_bProcess32) {
		//32λ̵64λ
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdInitInfo32		pCmd  = (PCmdInitInfo32)pInfo->context;

		m_uInIsr = (pCmd->toSystem>>16)!=0;

		//PCIĴڴַ
		for(i=0; i<m_memPci.MaxInserted()+1; i++) {
			pciMap = new (NonPagedPool) KMemoryToProcessMap((ULONG)m_memPci[i].addr, m_memPci[i].len, pInfo->hProcess, (BOOLEAN)(pCmd->toSystem&TRUE));
			if(pciMap) {
				if(pciMap->IsValid()) {
					for(j=0; j<m_Memory.MaxInserted()+1; j++) {
						if(m_Memory[i]==NULL) {
							break;
						}
					}
					if(j <= m_Memory.MaxInserted()) {
						m_Memory[i] = pciMap;
					}
					else {
						status = m_Memory.InsertNext(pciMap, NULL);
						if(!NT_SUCCESS(status)) {
							//ڴʧ
							delete pciMap;
							pciMap = NULL;
							return status;
						}
					}
					//t << "pciMap = " << pciMap->ProcessAddress() << " from " << m_memPci[i].addr << EOL;
					pCmd->mem[i].procBase = static_cast <PVOID32> (pciMap->ProcessAddress());
					pCmd->mem[i].addrBase = static_cast <PVOID32> (m_memPci[i].addr);
					pCmd->mem[i].len	  = static_cast <ULONG32> (pciMap->Count());
				}
				else {
					delete pciMap;
					pciMap = NULL;
				}
			}
			if(pciMap == NULL) {
				pCmd->mem[i].procBase = NULL;
				pCmd->mem[i].addrBase = static_cast <PVOID32> (m_memPci[i].addr);
				pCmd->mem[i].len	  = static_cast <ULONG32> (m_memPci[i].len);
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdInitInfo	  pCmd  = (PCmdInitInfo)pInfo->context;

		m_uInIsr = (pCmd->toSystem>>16)!=0;

		//PCIĴڴַ
		for(i=0; i<m_memPci.MaxInserted()+1; i++) {
			pciMap = new (NonPagedPool) KMemoryToProcessMap((ULONG)m_memPci[i].addr, m_memPci[i].len, pInfo->hProcess, (BOOLEAN)(pCmd->toSystem&TRUE));
			if(pciMap) {
				if(pciMap->IsValid()) {
					for(j=0; j<m_Memory.MaxInserted()+1; j++) {
						if(m_Memory[i]==NULL) {
							break;
						}
					}
					if(j <= m_Memory.MaxInserted()) {
						m_Memory[i] = pciMap;
					}
					else {
						status = m_Memory.InsertNext(pciMap, NULL);
						if(!NT_SUCCESS(status)) {
							//ڴʧ
							delete pciMap;
							pciMap = NULL;
							return status;
						}
					}
					pCmd->mem[i].procBase = pciMap->ProcessAddress();
					pCmd->mem[i].addrBase = m_memPci[i].addr;
					pCmd->mem[i].len	  = pciMap->Count();
				}
				else {
					delete pciMap;
					pciMap = NULL;
				}
			}
			if(pciMap == NULL) {
				pCmd->mem[i].procBase = NULL;
				pCmd->mem[i].addrBase = m_memPci[i].addr;
				pCmd->mem[i].len	  = m_memPci[i].len;
			}
		}
	}

	//ɾжϢ
	if(m_devList.IsValid()) {
		for(i=0; i<m_devList.MaxInserted()+1; i++) {
			if(m_devList[i]->irqList.IsValid()) {
				m_devList[i]->irqList.Invalidate();
			}
			if(m_devList[i]->basShare != NULL) {
				m_devList[i]->mapShare.Unmap(m_devList[i]->basShare);
				m_devList[i]->mapShare.Invalidate();
				m_devList[i]->basShare = NULL;
			}
			delete m_devList[i];
		}
		m_devList.Invalidate();
	}

	status = m_devList.Initialize(1, 1, NonPagedPool);
	if(!NT_SUCCESS(status)) {
		return status;
	}
	
	I.Information() = IOBUFF_LEN;
	I.Status() = status;

	//Windows 7 32bitϵͳϣCloseʱϵͳ
	//ExSetTimerResolution(10000, TRUE);

	return status;
}

//=========================================
// ʼ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdUninit(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

	Uninit(FALSE);
	
	I.Information() = 0;
	I.Status() = status;

	return status;
}

//========================================
// ʼ
// PCIڴԼַ
//========================================
NTSTATUS CTW68XXDevice::OnCmdInitDevice(KIrp I)
{
	NTSTATUS			 status	 = STATUS_SUCCESS;
	PDeviceNode			 devNode = NULL;
	int					 i=0, iFree=-1;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdInitDeviceInfo32 pCmd  = (PCmdInitDeviceInfo32)pInfo->context;

		if(m_devList.IsValid()) {
			for(i=0; i<m_devList.MaxInserted()+1; i++) {
				if(m_devList[i] != NULL){
					if(m_devList[i]->deviceId == pCmd->deviceId) {
						return STATUS_INVALID_PARAMETER;
					}
				}
				else {
					iFree = i;
					break;
				}
			}
		}

		devNode = new (NonPagedPool) TDeviceNode;
		if(devNode == NULL) {
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		status = devNode->irqList.Initialize(MAX_IRQ_NUM, 1, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			delete devNode;
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		
		RtlZeroMemory(&devNode->datShare, sizeof(TShareData));
		devNode->mapShare.Initialize(&devNode->datShare, sizeof(TShareData));
		if(!devNode->mapShare.IsValid()) {
			delete devNode;
			return STATUS_UNSUCCESSFUL;
		}

		//ж֪ͨ¼
		if(pCmd->evtData) {
			devNode->evtData.Initialize(pCmd->evtData);
			if(!devNode->evtData.IsValid()) {
				return STATUS_UNSUCCESSFUL;
			}
		}
		
		devNode->deviceId = pCmd->deviceId;
		devNode->basShare = devNode->mapShare.MapToUserSpaceEx();
		if(devNode->basShare == NULL) {
			devNode->mapShare.Invalidate();
			delete devNode;
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		if(iFree < 0) {
			m_devList.InsertNext(devNode, NULL);
		}
		else {
			m_devList[iFree] = devNode;
		}

		//t << "datShare = " << devNode->basShare << " len = " << devNode->mapShare.Size() << EOL;
		pCmd->datShare = static_cast <PVOID32> (devNode->basShare);
		pCmd->lenShare = devNode->mapShare.Size();
	}
	else 
#endif	
	{
		PIoctlCommandInfo	 pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdInitDeviceInfo	 pCmd  = (PCmdInitDeviceInfo)pInfo->context;

		if(m_devList.IsValid()) {
			for(i=0; i<m_devList.MaxInserted()+1; i++) {
				if(m_devList[i] != NULL){
					if(m_devList[i]->deviceId == pCmd->deviceId) {
						return STATUS_INVALID_PARAMETER;
					}
				}
				else {
					iFree = i;
					break;
				}
			}
		}

		devNode = new (NonPagedPool) TDeviceNode;
		if(devNode == NULL) {
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		status = devNode->irqList.Initialize(MAX_IRQ_NUM, 1, NonPagedPool);
		if(!NT_SUCCESS(status)) {
			delete devNode;
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		
		RtlZeroMemory(&devNode->datShare, sizeof(TShareData));
		devNode->mapShare.Initialize(&devNode->datShare, sizeof(TShareData));
		if(!devNode->mapShare.IsValid()) {
			delete devNode;
			return STATUS_UNSUCCESSFUL;
		}

		//ж֪ͨ¼
		if(pCmd->evtData) {
			devNode->evtData.Initialize(pCmd->evtData);
			if(!devNode->evtData.IsValid()) {
				return STATUS_UNSUCCESSFUL;
			}
		}
		
		devNode->deviceId = pCmd->deviceId;
		devNode->basShare = devNode->mapShare.MapToUserSpaceEx();
		if(devNode->basShare == NULL) {
			devNode->mapShare.Invalidate();
			delete devNode;
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		if(iFree < 0) {
			m_devList.InsertNext(devNode, NULL);
		}
		else {
			m_devList[iFree] = devNode;
		}

		pCmd->datShare = devNode->basShare;
		pCmd->lenShare = devNode->mapShare.Size();
	}
	
	I.Information() = IOBUFF_LEN;
	I.Status() = STATUS_SUCCESS;

	return STATUS_SUCCESS;
}

//=========================================
// ʼ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdUninitDevice(KIrp I)
{
	int i=0, j;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdInitDeviceInfo32 pCmd  = (PCmdInitDeviceInfo32)pInfo->context;

		if(m_devList.IsValid()) {
			for(i=0; i<m_devList.MaxInserted()+1; i++) {
				if((m_devList[i] != NULL) && (m_devList[i]->deviceId == pCmd->deviceId)) {
					if(m_devList[i]->tmList.IsValid()) {
						for(j=0; j<m_devList[i]->tmList.MaxInserted()+1; j++) {
							if(m_devList[i]->tmList[j].tmc != NULL) {
								m_devList[i]->tmList[j].tmc->Invalidate();
								delete m_devList[i]->tmList[j].tmc;
							}
						}
						m_devList[i]->tmList.Invalidate();
					}
					if(m_devList[i]->irqList.IsValid()) {
						m_devList[i]->irqList.Invalidate();
					}
					if(m_devList[i]->basShare != NULL) {
						m_devList[i]->mapShare.Unmap(m_devList[i]->basShare);
						m_devList[i]->mapShare.Invalidate();
						m_devList[i]->basShare = NULL;
					}
					if(m_devList[i]->evtData.IsValid()) {
						m_devList[i]->evtData.Invalidate();
					}
					delete m_devList[i];
					m_devList[i] = NULL;
					return STATUS_SUCCESS;
				}
			}
		}
	}
	else
#endif
	{
		PIoctlCommandInfo	 pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdInitDeviceInfo	 pCmd  = (PCmdInitDeviceInfo)pInfo->context;

		if(m_devList.IsValid()) {
			for(i=0; i<m_devList.MaxInserted()+1; i++) {
				if((m_devList[i] != NULL) && (m_devList[i]->deviceId == pCmd->deviceId)) {
					if(m_devList[i]->tmList.IsValid()) {
						for(j=0; j<m_devList[i]->tmList.MaxInserted()+1; j++) {
							if(m_devList[i]->tmList[j].tmc != NULL) {
								m_devList[i]->tmList[j].tmc->Invalidate();
								delete m_devList[i]->tmList[j].tmc;
							}
						}
						m_devList[i]->tmList.Invalidate();
					}
					if(m_devList[i]->irqList.IsValid()) {
						m_devList[i]->irqList.Invalidate();
					}
					if(m_devList[i]->basShare != NULL) {
						m_devList[i]->mapShare.Unmap(m_devList[i]->basShare);
						m_devList[i]->mapShare.Invalidate();
						m_devList[i]->basShare = NULL;
					}
					if(m_devList[i]->evtData.IsValid()) {
						m_devList[i]->evtData.Invalidate();
					}
					delete m_devList[i];
					m_devList[i] = NULL;
					return STATUS_SUCCESS;
				}
			}
		}
	}
		
	return STATUS_INVALID_PARAMETER;
}

//=========================================
// ûCrashʱϢ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdAppendCrash(KIrp I)
{
	TCrashInfo		  crash;
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdCrashInfo32		pCmd  = (PCmdCrashInfo32)pInfo->context;

		//ǷѾӹCrashϢ
		for(i=0; i<m_crsList.MaxInserted()+1; i++) {
			if(m_crsList[i].regBase==pCmd->regBase) {
				if(m_crsList[i].regCrash == pCmd->regCrash) {
					break;
				}
			}
		}

		//ûӹ
		if(i>=m_crsList.MaxInserted()+1) {
			crash.regBase = pCmd->regBase;
			crash.regCrash= pCmd->regCrash;
			crash.higMask = pCmd->higMask;
			crash.lowMask = pCmd->lowMask;
			status = m_crsList.InsertNext(crash, NULL);
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdCrashInfo	  pCmd  = (PCmdCrashInfo)pInfo->context;

		//ǷѾӹCrashϢ
		for(i=0; i<m_crsList.MaxInserted()+1; i++) {
			if(m_crsList[i].regBase==pCmd->regBase) {
				if(m_crsList[i].regCrash == pCmd->regCrash) {
					break;
				}
			}
		}

		//ûӹ
		if(i>=m_crsList.MaxInserted()+1) {
			crash = *pCmd;
			status = m_crsList.InsertNext(crash, NULL);
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// жϢ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdAppendIrq(KIrp I)
{
	TIrqInfo		  irq;
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0, j=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdIrqInfo32		pCmd  = (PCmdIrqInfo32)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}

		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}

		//ǷѾӹжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				for(j=0; j<MAX_INTSTATUS; j++) {
					if((m_devList[iPos]->irqList[i].regMask[j]==pCmd->irqInfo.regMask[j]) && 
					   (m_devList[iPos]->irqList[i].regStatus[j]==pCmd->irqInfo.regStatus[j])) {
						if(m_devList[iPos]->irqList[i].mask[j]==pCmd->irqInfo.mask[j]) {
							break;
						}
					}
				}
			}
		}

		//ûӹ
		if(i >= m_devList[iPos]->irqList.MaxInserted()+1) {
			ToIrqInfo(&irq, &pCmd->irqInfo);
			status = m_devList[iPos]->irqList.InsertNext(irq, NULL);
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdIrqInfo		  pCmd  = (PCmdIrqInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}

		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}

		//ǷѾӹжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				for(j=0; j<MAX_INTSTATUS; j++) {
					if((m_devList[iPos]->irqList[i].regMask[j]==pCmd->irqInfo.regMask[j]) && 
					   (m_devList[iPos]->irqList[i].regStatus[j]==pCmd->irqInfo.regStatus[j])) {
						if(m_devList[iPos]->irqList[i].mask[j]==pCmd->irqInfo.mask[j]) {
							break;
						}
					}
				}
			}
		}

		//ûӹ
		if(i >= m_devList[iPos]->irqList.MaxInserted()+1) {
			irq = pCmd->irqInfo;
			status = m_devList[iPos]->irqList.InsertNext(irq, NULL);
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// жϢ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdModifyIrq(KIrp I)
{
	PIrqInfo		  irq;
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdIrqInfo32		pCmd  = (PCmdIrqInfo32)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}

		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		//ǷѾӹжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				break;
			}
		}

		//ûӹ
		if(i >= m_devList[iPos]->irqList.MaxInserted()+1) {
			return STATUS_INVALID_PARAMETER;
		}

		irq	 = &m_devList[iPos]->irqList[i];
		ToIrqInfo(irq, &pCmd->irqInfo);
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdIrqInfo		  pCmd  = (PCmdIrqInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}

		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		//ǷѾӹжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				break;
			}
		}

		//ûӹ
		if(i >= m_devList[iPos]->irqList.MaxInserted()+1) {
			return STATUS_INVALID_PARAMETER;
		}

		irq	 = &m_devList[iPos]->irqList[i];
		*irq = pCmd->irqInfo;
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// ɾжϢ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdRemoveIrq(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdIrqInfo32		pCmd  = (PCmdIrqInfo32)pInfo->context;
		
		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		//ɾѾӵĴжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				RtlZeroMemory(&m_devList[iPos]->irqList[i], sizeof(TIrqInfo));
				break;
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdIrqInfo		  pCmd  = (PCmdIrqInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->irqInfo.irqId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		//ɾѾӵĴжϢ
		for(i=0; i<m_devList[iPos]->irqList.MaxInserted()+1; i++) {
			if((m_devList[iPos]->irqList[i].irqId==pCmd->irqInfo.irqId) && 
			   (m_devList[iPos]->irqList[i].regBase==pCmd->irqInfo.regBase)) {
				RtlZeroMemory(&m_devList[iPos]->irqList[i], sizeof(TIrqInfo));
				break;
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// TimerϢ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdAppendTimer(KIrp I)
{
	TTimerNode		  tmn;
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdTimerInfo32	    pCmd  = (PCmdTimerInfo32)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(!m_devList[iPos]->tmList.IsValid()) {
			status = m_devList[iPos]->tmList.Initialize(1, 1, NonPagedPool);
			if(!NT_SUCCESS(status)) {
				return status;
			}
		}

		//ǷѾӹTimer
		for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
			if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
				break;
			}
		}
		
		//ûӹ
		if(i >= m_devList[iPos]->tmList.MaxInserted()+1) {
			tmn.timerId = pCmd->timerId;
			tmn.tmi.regBase = pCmd->tmiInfo.regBase;
			tmn.tmi.dueTime = pCmd->tmiInfo.dueTime;
			for(i=0; i<MAX_RW_TIMER; i++) {
				tmn.tmi.regRead[i] = pCmd->tmiInfo.regRead[i];
				tmn.tmi.regWrite[i]= pCmd->tmiInfo.regWrite[i];
				tmn.tmi.mskWrite[i]= pCmd->tmiInfo.mskWrite[i];
			}
			tmn.tmc	= new (NonPagedPool) CTimedCallback();
			if(tmn.tmc == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				status = m_devList[iPos]->tmList.InsertNext(tmn, NULL);
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdTimerInfo	  pCmd  = (PCmdTimerInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(!m_devList[iPos]->tmList.IsValid()) {
			status = m_devList[iPos]->tmList.Initialize(1, 1, NonPagedPool);
			if(!NT_SUCCESS(status)) {
				return status;
			}
		}

		//ǷѾӹTimer
		for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
			if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
				break;
			}
		}
		
		//ûӹ
		if(i >= m_devList[iPos]->tmList.MaxInserted()+1) {
			tmn.timerId = pCmd->timerId;
			tmn.tmi = pCmd->tmiInfo;
			tmn.tmc	= new (NonPagedPool) CTimedCallback();
			if(tmn.tmc == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				status = m_devList[iPos]->tmList.InsertNext(tmn, NULL);
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//=========================================
// ɾTimer
//=========================================
NTSTATUS CTW68XXDevice::OnCmdRemoveTimer(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdTimerInfo32	    pCmd  = (PCmdTimerInfo32)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
			
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}

		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					if(m_devList[iPos]->tmList[i].tmc != NULL) {
						m_devList[iPos]->tmList[i].tmc->Invalidate();
						delete m_devList[iPos]->tmList[i].tmc;
						m_devList[iPos]->tmList[i].tmc = NULL;
					}
					RtlZeroMemory(&m_devList[iPos]->tmList[i].tmi, sizeof(TTimerInfo));
					m_devList[iPos]->tmList[i].timerId = (ULONG)-1;
					break;
				}
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdTimerInfo	  pCmd  = (PCmdTimerInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
			
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}

		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					if(m_devList[iPos]->tmList[i].tmc != NULL) {
						m_devList[iPos]->tmList[i].tmc->Invalidate();
						delete m_devList[iPos]->tmList[i].tmc;
						m_devList[iPos]->tmList[i].tmc = NULL;
					}
					RtlZeroMemory(&m_devList[iPos]->tmList[i].tmi, sizeof(TTimerInfo));
					m_devList[iPos]->tmList[i].timerId = (ULONG)-1;
					break;
				}
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//=========================================
// ޸Timer
//=========================================
NTSTATUS CTW68XXDevice::OnCmdModifyTimer(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	PTimerInfo tmi	= NULL;
	LONG	 i=0, j;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdTimerInfo32	    pCmd  = (PCmdTimerInfo32)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					tmi = &m_devList[iPos]->tmList[i].tmi;
					tmi->regBase = pCmd->tmiInfo.regBase;
					tmi->dueTime = pCmd->tmiInfo.dueTime;
					for(j=0; j<MAX_RW_TIMER; j++) {
						tmi->regRead[j] = pCmd->tmiInfo.regRead[j];
						tmi->regWrite[j]= pCmd->tmiInfo.regWrite[j];
						tmi->mskWrite[j]= pCmd->tmiInfo.mskWrite[j];
					}
					break;
				}
			}
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdTimerInfo	  pCmd  = (PCmdTimerInfo)pInfo->context;

		iPos = GetDevicePos(pCmd->deviceId);

		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		if(pCmd->timerId >= MAX_IRQ_NUM) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					m_devList[iPos]->tmList[i].tmi = pCmd->tmiInfo;
					break;
				}
			}
		}
	}
	
	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//=========================================
// ɾTimer
//=========================================
NTSTATUS CTW68XXDevice::OnCmdCancelTimer(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	LONG	 i=0;
	LONG	 iPos = -1;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdTimerInfo32	    pCmd  = (PCmdTimerInfo32)pInfo->context;
		
		iPos = GetDevicePos(pCmd->deviceId);
	
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
			
		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					if(m_devList[iPos]->tmList[i].tmc != NULL) {
						m_devList[iPos]->tmList[i].tmc->Cancel();
					}
					break;
				}
			}
		}
	}
	else
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdTimerInfo	  pCmd  = (PCmdTimerInfo)pInfo->context;
		
		iPos = GetDevicePos(pCmd->deviceId);
	
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
			
		if(m_devList[iPos]->tmList.IsValid()) {
			//ɾѾӵĴжϢ
			for(i=0; i<m_devList[iPos]->tmList.MaxInserted()+1; i++) {
				if(m_devList[iPos]->tmList[i].timerId==pCmd->timerId) {
					if(m_devList[iPos]->tmList[i].tmc != NULL) {
						m_devList[iPos]->tmList[i].tmc->Cancel();
					}
					break;
				}
			}
		}
	}
	
	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//=========================================
// I2C
//=========================================
NTSTATUS CTW68XXDevice::OnCmdAppendI2C(KIrp I)
{
	NTSTATUS	  status= STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdI2CInfo32		pCmd  = (PCmdI2CInfo32)pInfo->context;
		PHYSICAL_ADDRESS	addr  = {pCmd->memBase, 0};
		KMemoryRange*		reg	  = FindMemRage(addr);

		if((pCmd->hwid!=0x6864) || (reg==NULL)) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(m_i2c != NULL) {
			return STATUS_UNSUCCESSFUL;
		}

		m_i2c = new (NonPagedPool) CI2C6864();
		if(m_i2c == NULL) {
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		m_i2c->i2cInit(reg, pCmd->hLock);
		m_i2c->I2CInitSWMode(2000000);
		pCmd->i2cHandle = (ULONG32)m_i2c;
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdI2CInfo		  pCmd  = (PCmdI2CInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange*	  reg	= FindMemRage(addr);

		if((pCmd->hwid!=0x6864) || (reg==NULL)) {
			return STATUS_INVALID_PARAMETER;
		}
		
		if(m_i2c != NULL) {
			return STATUS_UNSUCCESSFUL;
		}

		m_i2c = new (NonPagedPool) CI2C6864();
		if(m_i2c == NULL) {
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		m_i2c->i2cInit(reg, pCmd->hLock);
		m_i2c->I2CInitSWMode(2000000);
		pCmd->i2cHandle = (ULONG)m_i2c;
	}

	I.Information() = IOBUFF_LEN;
	I.Status() = status;
	
	return status;
}

//=========================================
// I2C
//=========================================
NTSTATUS CTW68XXDevice::OnCmdRemoveI2C(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	
	if(m_i2c != NULL) {
		delete m_i2c;
		m_i2c = NULL;
	}
	
	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//=========================================
// ȡPCIĴBYTEs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdReadMemBYTEs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemByteInfo32 pCmd  = (PCmdRwMemByteInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->inb(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->inb(pCmd->memOffset);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemByteInfo pCmd  = (PCmdRwMemByteInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->inb(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->inb(pCmd->memOffset);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = IOBUFF_LEN;
	I.Status() = status;

	return status;
}

//=========================================
// ȡPCIĴWORDs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdReadMemWORDs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemWordInfo32 pCmd  = (PCmdRwMemWordInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->inw(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->inw(pCmd->memOffset);
			}
			pCmd->len *= sizeof(USHORT);
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemWordInfo pCmd  = (PCmdRwMemWordInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->inw(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->inw(pCmd->memOffset);
			}
			pCmd->len *= sizeof(USHORT);
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = IOBUFF_LEN;
	I.Status() = status;

	return status;
}

//=========================================
// ȡPCIĴDWORDs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdReadMemDWORDs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemDwordInfo32 pCmd = (PCmdRwMemDwordInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg = FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->ind(pCmd->memOffset, (PULONG)pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->ind(pCmd->memOffset);
			}
			pCmd->len *= sizeof(ULONG32);
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemDwordInfo pCmd  = (PCmdRwMemDwordInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg = FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->ind(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				pCmd->data[0] = reg->ind(pCmd->memOffset);
			}
			pCmd->len *= sizeof(ULONG);
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = IOBUFF_LEN;
	I.Status() = status;

	return status;
}

//=========================================
// дPCIĴBYTEs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdWriteMemBYTEs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemByteInfo32 pCmd  = (PCmdRwMemByteInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outb(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				reg->outb(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemByteInfo pCmd  = (PCmdRwMemByteInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outb(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				reg->outb(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// дPCIĴWORDs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdWriteMemWORDs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemWordInfo32 pCmd  = (PCmdRwMemWordInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outw(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				reg->outw(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemWordInfo pCmd  = (PCmdRwMemWordInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg= FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outw(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				reg->outw(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// дPCIĴDWORDs
//=========================================
NTSTATUS CTW68XXDevice::OnCmdWriteMemDWORDs(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemDwordInfo32 pCmd = (PCmdRwMemDwordInfo32)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg = FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outd(pCmd->memOffset, (PULONG)pCmd->data, pCmd->len);
			}
			else {
				reg->outd(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemDwordInfo pCmd  = (PCmdRwMemDwordInfo)pInfo->context;
		PHYSICAL_ADDRESS  addr  = {pCmd->memBase, 0};
		KMemoryRange* reg = FindMemRage(addr);

		if(reg) {
			if(pCmd->len > 1) {
				reg->outd(pCmd->memOffset, pCmd->data, pCmd->len);
			}
			else {
				reg->outd(pCmd->memOffset, pCmd->data[0]);
			}
		}
		else {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//=========================================
// PCIÿռ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdReadPciConfigSpace(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemByteInfo32 pCmd  = (PCmdRwMemByteInfo32)pInfo->context;

		if(KPciConfiguration::IsValid()) {
			ReadConfigSpace(pCmd->data, pCmd->memOffset, pCmd->len);
			I.Information() = IOBUFF_LEN;
			I.Status() = status;
		}
		else {
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemByteInfo pCmd  = (PCmdRwMemByteInfo)pInfo->context;

		if(KPciConfiguration::IsValid()) {
			ReadConfigSpace(pCmd->data, pCmd->memOffset, pCmd->len);
			I.Information() = IOBUFF_LEN;
			I.Status() = status;
		}
		else {
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
		}
	}

	return status;
}

//=========================================
// дPCIÿռ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdWritePciConfigSpace(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwMemByteInfo32 pCmd  = (PCmdRwMemByteInfo32)pInfo->context;

		if(KPciConfiguration::IsValid()) {
			WriteConfigSpace(pCmd->data, pCmd->memOffset, pCmd->len);
			I.Information() = 0;
			I.Status() = status;
		}
		else {
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
		}
	}
	else
#endif
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwMemByteInfo pCmd  = (PCmdRwMemByteInfo)pInfo->context;

		if(KPciConfiguration::IsValid()) {
			WriteConfigSpace(pCmd->data, pCmd->memOffset, pCmd->len);
			I.Information() = 0;
			I.Status() = status;
		}
		else {
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
		}
	}

	return status;
}

//=========================================
// PCIÿռ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdReadPciConfigSpaceEx(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	ULONG ulAddr=0, ulData=0;
	PDEVICE_OBJECT pdo = PDO()->NextDevice;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32	pInfo32 = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwPciConfigSpaceInfo32 pCmd  = (PCmdRwPciConfigSpaceInfo32)pInfo32->context;
		PBridgeDesc32	brg = &pCmd->bridge;

		if(brg->bus > 255) {
			//find the bridge chipset

			while(pdo) {
				KPciConfiguration pciCfg(pdo);
				if(pciCfg.IsValid()) {
					ulAddr = 0;
					pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

					if(ulAddr == 1) {
						//header type1,Ƭ
						pciCfg.ReadHeader(&ulAddr, 0x19, 1);

						if(ulAddr == pCmd->slot) {
							ulData = 0;
							pciCfg.ReadHeader(&ulData, 0, 4);
							brg->hwid = ulData;
							ulData = 0;
							pciCfg.ReadHeader(&ulData, 0x18, 1);
							brg->bus = ulData;
							pCmd->len = 0;

							ulData = ulAddr = 0;
							if(NT_SUCCESS(IoGetDeviceProperty(
									pdo,
									DevicePropertyAddress,
									4,
									&ulData,
									&ulAddr
									))) {
								brg->device = (ulData>>16) & 0xFFFF;
								brg->function = ulData & 0xFFFF;
							}
							else {
								brg->device = 0;
								brg->function = 0;
							}

							I.Information() = IOBUFF_LEN;
							I.Status() = status;
							return status;
						}
					}
				}
				pdo=pdo->NextDevice;
			}

			//no bridge chip find
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
			return status;
		}

		status = STATUS_UNSUCCESSFUL;
		while(pdo) {
			KPciConfiguration pciCfg(pdo);
			if(pciCfg.IsValid()) {
				ulAddr = 0;
				pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

				if(ulAddr == 1) {
					//header type1,Ƭ

					ulData = ulAddr = 0;
					pciCfg.ReadHeader(&ulData, 0, 4);
					pciCfg.ReadHeader(&ulAddr, 0x18, 2);

					if((((ulAddr>>8)&0xFF) == pCmd->slot) && ((ulAddr&0xFF) == brg->bus) && (ulData == brg->hwid)) {
						status = STATUS_SUCCESS;
						break;
					}
				}
			}
			pdo=pdo->NextDevice;
		}
		
		if(NT_SUCCESS(status)) {
			KPciConfiguration pciCfg(pdo);
			pciCfg.ReadHeader(pCmd->data, pCmd->offset, pCmd->len);
			I.Information() = IOBUFF_LEN;
		}
		I.Status() = status;
	}
	else 
#endif	
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwPciConfigSpaceInfo pCmd  = (PCmdRwPciConfigSpaceInfo)pInfo->context;
		PBridgeDesc		brg  = &pCmd->bridge;

		if(brg->bus > 255) {
			//find the bridge chipset

			while(pdo) {
				KPciConfiguration pciCfg(pdo);
				if(pciCfg.IsValid()) {
					ulAddr = 0;
					pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

					if(ulAddr == 1) {
						//header type1,Ƭ
						pciCfg.ReadHeader(&ulAddr, 0x19, 1);

						if(ulAddr == pCmd->slot) {
							ulData = 0;
							pciCfg.ReadHeader(&ulData, 0, 4);
							brg->hwid = ulData;
							ulData = 0;
							pciCfg.ReadHeader(&ulData, 0x18, 1);
							brg->bus = ulData;
							pCmd->len = 0;

							ulData = ulAddr = 0;
							if(NT_SUCCESS(IoGetDeviceProperty(
									pdo,
									DevicePropertyAddress,
									4,
									&ulData,
									&ulAddr
									))) {
								brg->device = (ulData>>16) & 0xFFFF;
								brg->function = ulData & 0xFFFF;
							}
							else {
								brg->device = 0;
								brg->function = 0;
							}

							I.Information() = IOBUFF_LEN;
							I.Status() = status;
							return status;
						}
					}
				}
				pdo=pdo->NextDevice;
			}

			//no bridge chip find
			status = STATUS_UNSUCCESSFUL;
			I.Information() = 0;
			I.Status() = status;
			return status;
		}

		status = STATUS_UNSUCCESSFUL;
		while(pdo) {
			KPciConfiguration pciCfg(pdo);
			if(pciCfg.IsValid()) {
				ulAddr = 0;
				pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

				if(ulAddr == 1) {
					//header type1,Ƭ

					ulData = ulAddr = 0;
					pciCfg.ReadHeader(&ulData, 0, 4);
					pciCfg.ReadHeader(&ulAddr, 0x18, 2);

					if((((ulAddr>>8)&0xFF) == pCmd->slot) && ((ulAddr&0xFF) == brg->bus) && (ulData == brg->hwid)) {
						status = STATUS_SUCCESS;
						break;
					}
				}
			}
			pdo=pdo->NextDevice;
		}
		
		if(NT_SUCCESS(status)) {
			KPciConfiguration pciCfg(pdo);
			pciCfg.ReadHeader(pCmd->data, pCmd->offset, pCmd->len);
			I.Information() = IOBUFF_LEN;
		}
		I.Status() = status;
	}
	
	return status;
}

//=========================================
// дPCIÿռ
//=========================================
NTSTATUS CTW68XXDevice::OnCmdWritePciConfigSpaceEx(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	ULONG ulAddr, ulData;
	PDEVICE_OBJECT pdo = PDO()->NextDevice;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32  pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdRwPciConfigSpaceInfo32 pCmd  = (PCmdRwPciConfigSpaceInfo32)pInfo->context;
		PBridgeDesc32	brg = &pCmd->bridge;

		status = STATUS_UNSUCCESSFUL;
		while(pdo) {
			KPciConfiguration pciCfg(pdo);
			if(pciCfg.IsValid()) {
				ulAddr = 0;
				pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

				if(ulAddr == 1) {
					//header type1,Ƭ

					ulData = ulAddr = 0;
					pciCfg.ReadHeader(&ulData, 0, 4);
					pciCfg.ReadHeader(&ulAddr, 0x18, 2);

					if((((ulAddr>>8)&0xFF) == pCmd->slot) && ((ulAddr&0xFF) == brg->bus) && (ulData == brg->hwid)) {
						status = STATUS_SUCCESS;
						break;
					}
				}
			}
			pdo=pdo->NextDevice;
		}
		
		if(NT_SUCCESS(status)) {
			KPciConfiguration pciCfg(pdo);
			pciCfg.WriteHeader(pCmd->data, pCmd->offset, pCmd->len);
			I.Information() = 0;
		}
		I.Status() = status;
	}
	else 
#endif	
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdRwPciConfigSpaceInfo pCmd  = (PCmdRwPciConfigSpaceInfo)pInfo->context;
		PBridgeDesc	brg = &pCmd->bridge;

		status = STATUS_UNSUCCESSFUL;
		while(pdo) {
			KPciConfiguration pciCfg(pdo);
			if(pciCfg.IsValid()) {
				ulAddr = 0;
				pciCfg.ReadHeader(&ulAddr, 0x0E, 1);

				if(ulAddr == 1) {
					//header type1,Ƭ

					ulData = ulAddr = 0;
					pciCfg.ReadHeader(&ulData, 0, 4);
					pciCfg.ReadHeader(&ulAddr, 0x18, 2);

					if((((ulAddr>>8)&0xFF) == pCmd->slot) && ((ulAddr&0xFF) == brg->bus) && (ulData == brg->hwid)) {
						status = STATUS_SUCCESS;
						break;
					}
				}
			}
			pdo=pdo->NextDevice;
		}
		
		if(NT_SUCCESS(status)) {
			KPciConfiguration pciCfg(pdo);
			pciCfg.WriteHeader(pCmd->data, pCmd->offset, pCmd->len);
			I.Information() = 0;
		}
		I.Status() = status;
	}
	
	return status;
}

//===================================================================
// ڴ䡪Ҫڷڴ棬ʹDmaCommonBuffer
// ķΪȲѾ䵫δʹõĴСbuffer
// Ͳٷ䣬ʹڴ棬·
//===================================================================
NTSTATUS CTW68XXDevice::OnCmdAllocMemory(KIrp I)
{
	CContiguousMemory* pCgm	 = NULL;
	NTSTATUS status = STATUS_SUCCESS;
	int		 i = 0;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32  pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	 pCmd  = (PCmdMemoryInfo32)pInfo->context;

		//Ѿ䵫δʹõڴ
		for(i=0; i<m_bufList.MaxInserted()+1; i++) {
			if(!m_bufList[i].isUsed && m_bufList[i].buf) {
				if((m_bufList[i].buf->Size() == pCmd->len) && !m_bufList[i].buf->IsContiguous()) {
					pCmd->nodes[0].addr = static_cast<PVOID32> (ULongToPtr(m_bufList[i].buf->LogicalAddress().LowPart));
					pCmd->nodes[0].len  = m_bufList[i].buf->Size();
					pCmd->vaddr = static_cast<PVOID32> (m_bufList[i].buf->MapToUserSpace());
					pCmd->inode = 1;
					m_bufList[i].isUsed = TRUE;
					I.Information() = IOBUFF_LEN;
					I.Status() = status;
					return status;
				}
			}
		}
		
		//ûѾbuffer·һ
		pCgm = new (NonPagedPool) CContiguousMemory(&m_Dma, pCmd->len);
		if(pCgm == NULL) {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
		else if(!pCgm->IsValid()) {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
		else {
			TUserMemoryBlock umb;
			PMemNode32 node = pCmd->nodes;
			umb.buf = pCgm;
			umb.isUsed = TRUE;
			//PVOID userAddr = pCgm->MapToUserSpace();
			//pCmd->vaddr = static_cast<PVOID32> (userAddr);
			//t << "alloc = " << userAddr << " from " << umb.buf->LogicalAddress().QuadPart << EOL;
			pCmd->vaddr = static_cast<PVOID32> (pCgm->MapToUserSpace());
			if(pCmd->vaddr != NULL) {
				node->addr = static_cast<PVOID32> (ULongToPtr(umb.buf->LogicalAddress().LowPart));
				node->len  = umb.buf->Size();
				pCmd->inode = 1;
				for(i=0; i<m_bufList.MaxInserted()+1; i++) {
					if(m_bufList[i].buf == NULL) {
						m_bufList[i] = umb;
						break;
					}
				}
				if(i >= m_bufList.MaxInserted()+1) {
					status = m_bufList.InsertNext(umb, NULL);
					if(!NT_SUCCESS(status)) {
						if(pCmd->vaddr != NULL) {
							umb.buf->Unmap(pCmd->vaddr);
							delete umb.buf;
						}
					}
				}
			}
			else {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			if(!NT_SUCCESS(status)) {
				delete pCgm;
			}
		}
		
		if(!NT_SUCCESS(status)){
			pCmd->vaddr = NULL;
			pCmd->inode = 0;
			pCmd->nodes[0].addr= NULL;
			pCmd->nodes[0].len = 0;
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo  pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	   pCmd  = (PCmdMemoryInfo)pInfo->context;

		//Ѿ䵫δʹõڴ
		for(i=0; i<m_bufList.MaxInserted()+1; i++) {
			if(!m_bufList[i].isUsed && m_bufList[i].buf) {
				if((m_bufList[i].buf->Size() == pCmd->len) && !m_bufList[i].buf->IsContiguous()) {
					pCmd->nodes->addr = (PVOID)m_bufList[i].buf->LogicalAddress().LowPart;
					pCmd->nodes->len  = m_bufList[i].buf->Size();
					pCmd->vaddr = m_bufList[i].buf->MapToUserSpace();
					pCmd->inode = 1;
					m_bufList[i].isUsed = TRUE;
					I.Information() = IOBUFF_LEN;
					I.Status() = status;
					return status;
				}
			}
		}
		
		//ûѾbuffer·һ
		pCgm = new (NonPagedPool) CContiguousMemory(&m_Dma, pCmd->len);
		if(pCgm == NULL) {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
		else if(!pCgm->IsValid()) {
			status = STATUS_MEMORY_NOT_ALLOCATED;
		}
		else {
			TUserMemoryBlock umb;
			PMemNode node = pCmd->nodes;
			umb.buf = pCgm;
			umb.isUsed = TRUE;
			pCmd->vaddr = pCgm->MapToUserSpace();
			if(pCmd->vaddr != NULL) {
				node->addr = (PVOID)umb.buf->LogicalAddress().LowPart;
				node->len  = umb.buf->Size();
				pCmd->inode = 1;
				for(i=0; i<m_bufList.MaxInserted()+1; i++) {
					if(m_bufList[i].buf == NULL) {
						m_bufList[i] = umb;
						break;
					}
				}
				if(i >= m_bufList.MaxInserted()+1) {
					status = m_bufList.InsertNext(umb, NULL);
					if(!NT_SUCCESS(status)) {
						if(pCmd->vaddr != NULL) {
							umb.buf->Unmap(pCmd->vaddr);
							delete umb.buf;
						}
					}
				}
			}
			else {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			if(!NT_SUCCESS(status)) {
				delete pCgm;
			}
		}
		
		if(!NT_SUCCESS(status)){
			pCmd->vaddr = NULL;
			pCmd->inode = 0;
			pCmd->nodes[0].addr= NULL;
			pCmd->nodes[0].len = 0;
		}
	}

	I.Information() = IOBUFF_LEN;
	I.Status() = status;
	
	return status;
}

//===========================================
// ڴͷšҪͷŷڴ
//===========================================
NTSTATUS CTW68XXDevice::OnCmdFreeMemory(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	
#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	pCmd  = (PCmdMemoryInfo32)pInfo->context;
		PMemNode32			node	= pCmd->nodes;

		for(int i=0; i<m_bufList.MaxInserted()+1; i++) {
			if(m_bufList[i].isUsed && m_bufList[i].buf) {
				if((PVOID)m_bufList[i].buf->LogicalAddress().LowPart==node->addr) {
					m_bufList[i].buf->Unmap(pCmd->vaddr);
					m_bufList[i].isUsed = FALSE;
					break;
				}
			}
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	  pCmd  = (PCmdMemoryInfo)pInfo->context;
		PMemNode		  node	= pCmd->nodes;

		for(int i=0; i<m_bufList.MaxInserted()+1; i++) {
			if(m_bufList[i].isUsed && m_bufList[i].buf) {
				if((PVOID)m_bufList[i].buf->LogicalAddress().LowPart==node->addr) {
					m_bufList[i].buf->Unmap(pCmd->vaddr);
					m_bufList[i].isUsed = FALSE;
					break;
				}
			}
		}
	}
	
	I.Information() = 0;
	I.Status() = status;
	
	return status;
}

//===============================================
// ڴӳ䡪Ҫӳַַ
//===============================================
NTSTATUS CTW68XXDevice::OnCmdMapMemory(KIrp I)
{
	KMemory* memMap = NULL;
	NTSTATUS status = STATUS_SUCCESS;
	int		 i=0, idx=m_mapList.MaxInserted()+2;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	pCmd  = (PCmdMemoryInfo32)pInfo->context;

		//ȼûӳӳĻֱӷѾӳļ
		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i] && m_mapList[i]->IsValid()) {
				if(m_mapList[i]->VirtualAddress() == pCmd->vaddr) {
					memMap = m_mapList[i];
					break;
				}
			}
			else if(idx > m_mapList.MaxInserted()+1){
				//õһλõ
				idx = i;
			}
		}

		//ûӳĻ²һ
		if(i >= m_mapList.MaxInserted()+1) {
			memMap = new (NonPagedPool) KMemory(pCmd->vaddr, pCmd->len);
			if(memMap == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				if(!memMap->IsValid()) {
					status = STATUS_MEMORY_NOT_ALLOCATED;
				}
				else {
					status = memMap->LockPages(IoModifyAccess, KernelMode);
					if(NT_SUCCESS(status)) {
						if(idx > m_mapList.MaxInserted()+1) {
							status = m_mapList.InsertNext(memMap, NULL);
						}
						else {
							m_mapList[idx] = memMap;
						}
					}
				}
				if(!NT_SUCCESS(status)) {
					memMap->Invalidate();
					delete memMap;
					memMap = NULL;
				}
			}
		}

		//ȡҳڴ
		if(memMap) {
			int total_pages = (memMap->RequiredMdlStorage(memMap->VirtualAddress(), memMap->Size())-sizeof(MDL)) / sizeof(ULONG);
			int pages		= memMap->Size()/PAGE_SIZE;
			ULONG	pageOffset	= 0;
			PULONG64  page	= (PULONG64)memMap->GetPageArray();
			bool	isContiguous = true;
			bool	isFragment   = ((memMap->Size() % PAGE_SIZE) > 0);

			if(isFragment)
				pages++;
			if(pages > (IOBUFF_LEN_DW_MAX/2 - sizeof(TCmdMemoryInfo)/4))
				pages = IOBUFF_LEN_DW_MAX/2 - sizeof(TCmdMemoryInfo)/4;
			if(pages > total_pages)
				pages = total_pages;
			pCmd->inode = pages;
			for(int i=0; i<pages; i++) {
				pCmd->nodes[i].addr = (PVOID32)(page[i] << 12);
				pCmd->nodes[i].len	= PAGE_SIZE;
				if(i == 0) {
					//ڴʼַҳַͷ:Ҳҳ
					pageOffset = ((ULONG)pCmd->vaddr) & (PAGE_SIZE-1);
					if(pageOffset > 0) {
						pCmd->nodes[i].addr = (PUCHAR)pCmd->nodes[i].addr + pageOffset;
						pCmd->nodes[i].len -= pageOffset;
					}
				}
				else if(isContiguous) {
					isContiguous = (((PUCHAR)pCmd->nodes[i-1].addr + pCmd->nodes[i-1].len) == (PUCHAR)pCmd->nodes[i].addr);
				}
			}

			//ڴ
			if((pages > 1) && isContiguous) {
				pCmd->inode = 1;
				pCmd->nodes[0].len = memMap->Size();
			}
		}
		
		if(!NT_SUCCESS(status)){
			pCmd->inode = 0;
			pCmd->nodes[0].addr = NULL;
			pCmd->nodes[0].len  = 0;
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	  pCmd  = (PCmdMemoryInfo)pInfo->context;

		//ȼûӳӳĻֱӷѾӳļ
		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i] && m_mapList[i]->IsValid()) {
				if(m_mapList[i]->VirtualAddress() == pCmd->vaddr) {
					memMap = m_mapList[i];
					break;
				}
			}
			else if(idx > m_mapList.MaxInserted()+1){
				//õһλõ
				idx = i;
			}
		}

		//ûӳĻ²һ
		if(i >= m_mapList.MaxInserted()+1) {
			memMap = new (NonPagedPool) KMemory(pCmd->vaddr, pCmd->len);
			if(memMap == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				if(!memMap->IsValid()) {
					status = STATUS_MEMORY_NOT_ALLOCATED;
				}
				else {
					status = memMap->LockPages(IoModifyAccess, KernelMode);
					if(NT_SUCCESS(status)) {
						if(idx > m_mapList.MaxInserted()+1) {
							status = m_mapList.InsertNext(memMap, NULL);
						}
						else {
							m_mapList[idx] = memMap;
						}
					}
				}
				if(!NT_SUCCESS(status)) {
					memMap->Invalidate();
					delete memMap;
					memMap = NULL;
				}
			}
		}

		//ȡҳڴ
		if(memMap) {
			int total_pages = (memMap->RequiredMdlStorage(memMap->VirtualAddress(), memMap->Size())-sizeof(MDL)) / sizeof(ULONG);
			int pages		= memMap->Size()/PAGE_SIZE;
			ULONG	pageOffset	= 0;
#if defined(_WIN64)
			PULONG64  page	= (PULONG64)memMap->GetPageArray();
#else
			PULONG	  page	= memMap->GetPageArray();
#endif
			bool	isContiguous = true;
			bool	isFragment   = ((memMap->Size() % PAGE_SIZE) > 0);

			if(isFragment)
				pages++;
			if(pages > (IOBUFF_LEN_DW_MAX/2 - sizeof(TCmdMemoryInfo)/4))
				pages = IOBUFF_LEN_DW_MAX/2 - sizeof(TCmdMemoryInfo)/4;
			if(pages > total_pages)
				pages = total_pages;
			pCmd->inode = pages;
			for(int i=0; i<pages; i++) {
				pCmd->nodes[i].addr = (PVOID)(page[i] << 12);
				pCmd->nodes[i].len	= PAGE_SIZE;
				if(i == 0) {
					//ڴʼַҳַͷ:Ҳҳ
					pageOffset = ((ULONG)pCmd->vaddr) & (PAGE_SIZE-1);
					if(pageOffset > 0) {
						pCmd->nodes[i].addr = (PUCHAR)pCmd->nodes[i].addr + pageOffset;
						pCmd->nodes[i].len -= pageOffset;
					}
				}
				else if(isContiguous) {
					isContiguous = (((PUCHAR)pCmd->nodes[i-1].addr + pCmd->nodes[i-1].len) == (PUCHAR)pCmd->nodes[i].addr);
				}
			}

			//ڴ
			if((pages > 1) && isContiguous) {
				pCmd->inode = 1;
				pCmd->nodes[0].len = memMap->Size();
			}
		}
		
		if(!NT_SUCCESS(status)){
			pCmd->inode = 0;
			pCmd->nodes[0].addr = NULL;
			pCmd->nodes[0].len  = 0;
		}
	}

	I.Information() = IOBUFF_LEN_MAX;
	I.Status() = status;

	return status;
}

//===============================================
// ڴӳ
//===============================================
NTSTATUS CTW68XXDevice::OnCmdUnmapMemory(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	int		 i = 0;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	pCmd  = (PCmdMemoryInfo32)pInfo->context;

		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i] && m_mapList[i]->IsValid()) {
				if(m_mapList[i]->VirtualAddress() == pCmd->vaddr) {
					m_mapList[i]->UnlockPages();
					m_mapList[i]->Invalidate();
					delete m_mapList[i];
					m_mapList[i] = NULL;
					break;
				}
			}
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	  pCmd  = (PCmdMemoryInfo)pInfo->context;

		for(i=0; i<m_mapList.MaxInserted()+1; i++) {
			if(m_mapList[i] && m_mapList[i]->IsValid()) {
				if(m_mapList[i]->VirtualAddress() == pCmd->vaddr) {
					m_mapList[i]->UnlockPages();
					m_mapList[i]->Invalidate();
					delete m_mapList[i];
					m_mapList[i] = NULL;
					break;
				}
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

//===============================================
// PCIڴӳ䵽ָ
//===============================================
NTSTATUS CTW68XXDevice::OnCmdMapPciMemory(KIrp I)
{
	NTSTATUS status = STATUS_UNSUCCESSFUL;
	int		 i=0, idx=m_pciMap.MaxInserted()+2;
	CMemoryToProcessMap* pciMap = NULL;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	pCmd  = (PCmdMemoryInfo32)pInfo->context;

		//ȼûӳӳĻֱӷѾӳļ
		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i] && m_pciMap[i]->IsValid()) {
				if(m_pciMap[i]->CpuPhysicalAddress().LowPart == (ULONG)pCmd->vaddr) {
					if(m_pciMap[i]->Process() == pInfo->hProcess) {
						pciMap = m_pciMap[i];
						break;
					}
				}
			}
			else if(idx > m_pciMap.MaxInserted()+1){
				//õһλõ
				idx = i;
			}
		}

		//ûӳĻ²һ
		if(i >= m_pciMap.MaxInserted()+1) {
			pciMap = new (NonPagedPool) CMemoryToProcessMap((ULONG)pCmd->vaddr, pCmd->len, pInfo->hProcess);
			if(pciMap == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				if(!pciMap->IsValid()) {
					status = STATUS_MEMORY_NOT_ALLOCATED;
					delete pciMap;
				}
				else {
					if(idx <= m_pciMap.MaxInserted()) {
						m_pciMap[idx] = pciMap;
					}
					else {
						status = m_pciMap.InsertNext(pciMap, NULL);
						if(!NT_SUCCESS(status)) {
							delete pciMap;
						}
					}
				}
			}
		}

		if(NT_SUCCESS(status)) {
			pCmd->vaddr= static_cast<PVOID32> (pciMap->ProcessAddress());
			pCmd->len  = pciMap->Count();
		}
		else {
			pCmd->vaddr= NULL;
			pCmd->len  = 0;
		}
	}
	else 
#endif	
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	  pCmd  = (PCmdMemoryInfo)pInfo->context;

		//ȼûӳӳĻֱӷѾӳļ
		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i] && m_pciMap[i]->IsValid()) {
				if(m_pciMap[i]->CpuPhysicalAddress().LowPart == (ULONG)pCmd->vaddr) {
					if(m_pciMap[i]->Process() == pInfo->hProcess) {
						pciMap = m_pciMap[i];
						break;
					}
				}
			}
			else if(idx > m_pciMap.MaxInserted()+1){
				//õһλõ
				idx = i;
			}
		}

		//ûӳĻ²һ
		if(i >= m_pciMap.MaxInserted()+1) {
			pciMap = new (NonPagedPool) CMemoryToProcessMap((ULONG)pCmd->vaddr, pCmd->len, pInfo->hProcess);
			if(pciMap == NULL) {
				status = STATUS_MEMORY_NOT_ALLOCATED;
			}
			else {
				if(!pciMap->IsValid()) {
					status = STATUS_MEMORY_NOT_ALLOCATED;
					delete pciMap;
				}
				else {
					if(idx <= m_pciMap.MaxInserted()) {
						m_pciMap[idx] = pciMap;
					}
					else {
						status = m_pciMap.InsertNext(pciMap, NULL);
						if(!NT_SUCCESS(status)) {
							delete pciMap;
						}
					}
				}
			}
		}

		if(NT_SUCCESS(status)) {
			pCmd->vaddr= pciMap->ProcessAddress();
			pCmd->len  = pciMap->Count();
		}
		else {
			pCmd->vaddr= NULL;
			pCmd->len  = 0;
		}
	}

	//ȡҳڴ
	I.Information() = IOBUFF_LEN_MAX;
	I.Status() = status;

	return status;
}

//===============================================
// PCIڴӳ
//===============================================
NTSTATUS CTW68XXDevice::OnCmdUnmapPciMemory(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;
	int		 i = 0;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32 pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdMemoryInfo32	pCmd  = (PCmdMemoryInfo32)pInfo->context;

		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i] && m_pciMap[i]->IsValid()) {
				if((m_pciMap[i]->CpuPhysicalAddress().LowPart == (ULONG)pCmd->vaddr) &&
				   (m_pciMap[i]->Process() == pInfo->hProcess)) {
					m_pciMap[i]->Invalidate();
					delete m_pciMap[i];
					m_pciMap[i] = NULL;
					break;
				}
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdMemoryInfo	  pCmd  = (PCmdMemoryInfo)pInfo->context;

		for(i=0; i<m_pciMap.MaxInserted()+1; i++) {
			if(m_pciMap[i] && m_pciMap[i]->IsValid()) {
				if((m_pciMap[i]->CpuPhysicalAddress().LowPart == (ULONG)pCmd->vaddr) &&
				   (m_pciMap[i]->Process() == pInfo->hProcess)) {
					m_pciMap[i]->Invalidate();
					delete m_pciMap[i];
					m_pciMap[i] = NULL;
					break;
				}
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}

/** 
***************************************************************************************
*  SDKݵ¼
*  @param[in] IKIrp
*
*  @return ɹSTATUS_SUCCESS---0򷵻ش״ֵ̬
***************************************************************************************
*/
NTSTATUS CTW68XXDevice::OnCmdModifyEvt(KIrp I)
{
	NTSTATUS status = STATUS_SUCCESS;

#if defined(_WIN64)
	if(m_bProcess32) {
		PIoctlCommandInfo32		pInfo = (PIoctlCommandInfo32) I.IoctlBuffer();
		PCmdNotifyEventInfo32	pCmd  = (PCmdNotifyEventInfo32)pInfo->context;
		LONG	 iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		//ж֪ͨ¼
		if(pCmd->evtData) {
			if(m_devList[iPos]->evtData.IsValid()) {
				m_devList[iPos]->evtData.Invalidate();
			}
			if(pCmd->evtData) {
				m_devList[iPos]->evtData.Initialize(pCmd->evtData);
				if(!m_devList[iPos]->evtData.IsValid()) {
					return STATUS_UNSUCCESSFUL;
				}
			}
		}
	}
	else 
#endif
	{
		PIoctlCommandInfo	pInfo = (PIoctlCommandInfo) I.IoctlBuffer();
		PCmdNotifyEventInfo	pCmd  = (PCmdNotifyEventInfo)pInfo->context;
		LONG	 iPos = GetDevicePos(pCmd->deviceId);
		
		if(iPos < 0) {
			return STATUS_UNSUCCESSFUL;
		}
		
		//ж֪ͨ¼
		if(pCmd->evtData) {
			if(m_devList[iPos]->evtData.IsValid()) {
				m_devList[iPos]->evtData.Invalidate();
			}
			if(pCmd->evtData) {
				m_devList[iPos]->evtData.Initialize(pCmd->evtData);
				if(!m_devList[iPos]->evtData.IsValid()) {
					return STATUS_UNSUCCESSFUL;
				}
			}
		}
	}

	I.Information() = 0;
	I.Status() = status;

	return status;
}