//////////////////////////////////////////////////////////////////////////
// TW68XXDevice.h
// ʵ־豸
//////////////////////////////////////////////////////////////////////////

#ifndef __TW68XXDevice_h__
#define __TW68XXDevice_h__

#include <vdw.h>
#include "TW68XXioctl.h"
#include "TW68XXioctl32.h"
#include "TimedCallback.h"
#include "I2C.h"

class CMemoryToProcessMap : public KMemoryToProcessMap  
{
public:
	CMemoryToProcessMap(ULONGLONG CpuPhysicalAddress,
						ULONG Size,
						HANDLE hProcess) :	
	KMemoryToProcessMap(CpuPhysicalAddress,	Size, hProcess) {};
	virtual ~CMemoryToProcessMap() {};
	
	HANDLE Process(){ return m_hProcess; }
};

class CContiguousMemory
{
public:
	CContiguousMemory();
	CContiguousMemory(KDmaAdapter* pAdapter, ULONG size, BOOLEAN CacheEnabled = FALSE);
	virtual ~CContiguousMemory(){ if (IsValid()) Invalidate(); }

	VOID Initialize(KDmaAdapter* pAdapter, ULONG size, BOOLEAN CacheEnabled = FALSE);

	VOID Invalidate(void);
	BOOLEAN IsValid(void){ return (m_pVirtualBase != NULL); }

	ULONG				Size(void)				{ return m_Size; }
	PVOID				VirtualAddress(void)	{ return m_pVirtualBase; }
	PHYSICAL_ADDRESS	LogicalAddress(void)	{ return m_LogicalAddress; }
	PVOID				MapToUserSpace(void);
	VOID				Unmap(PVOID BaseAddress);
	VOID				Unmap() { Unmap(m_pMappedBase); }
	BOOLEAN				IsContiguous() { return (m_pCommBuffer==NULL); }

protected:
	KCommonDmaBuffer* m_pCommBuffer;
	KMemory			  m_Memory;
	PHYSICAL_ADDRESS  m_LogicalAddress;
	PVOID	m_pVirtualBase;
	PVOID	m_pMappedBase;		//Ӱ䵽̵ĵַ
	ULONG	m_Size;
};

class CTW68XXDevice_DriverManagedQueue : public KDriverManagedQueueEx
{
public:
	SAFE_DESTRUCTORS;
	VOID StartIo(KIrp I);
};

class CTW68XXDevice : public KPnpDevice, KPciConfiguration
{
	// Constructors
public:
	SAFE_DESTRUCTORS;
	CTW68XXDevice(PDEVICE_OBJECT Pdo, ULONG Unit);
	~CTW68XXDevice();

	// Member Functions
public:
	DEVMEMBER_DISPATCHERS
		// The ISR (interrupt service routine)
	MEMBER_ISR(CTW68XXDevice, Isr_Irq);
		// The DPC (deferred procedure call) for the ISR
	MEMBER_DPC(CTW68XXDevice, DpcFor_Irq);
	virtual NTSTATUS OnStartDevice(KIrp I);
	virtual NTSTATUS OnStopDevice(KIrp I);
	virtual NTSTATUS OnRemoveDevice(KIrp I);
	virtual NTSTATUS DefaultPnp(KIrp I);
	virtual NTSTATUS DefaultPower(KIrp I);
	virtual NTSTATUS OnDevicePowerUp(KIrp I);
	virtual NTSTATUS OnDeviceSleep(KIrp I);
	virtual NTSTATUS OnSystemSleep(KIrp I);

	NTSTATUS IOCTL_COMMAND_Handler(KIrp I);
	VOID Invalidate(void);

	static void __stdcall DpcOnTimer(KDPC *Dpc, PVOID Context, PVOID Argument1, PVOID Argument2);

	void __stdcall OnTimer(LONG iDevice, LONG iTimer);
	
#ifdef _COMMENT_ONLY

		// The following member functions are actually defined by the
		// a DEVMEMBER_xxx or MEMBER_xxx macro (such as DEVMEMBER_DISPATCHERS).
		// The macro __COMMENT_ONLY never gets defined.  These comment-only
		// definitions simply allow easy navigation to the functions within
		// the Developer Studio using the class browser.

	BOOLEAN Isr_Irq(void); // COMMENT_ONLY
	VOID DpcFor_Irq(PVOID Arg1, PVOID Arg2);		// COMMENT_ONLY
	virtual NTSTATUS CleanUp(KIrp I);				// COMMENT_ONLY
	virtual NTSTATUS Create(KIrp I);				// COMMENT_ONLY
	virtual NTSTATUS Close(KIrp I);					// COMMENT_ONLY
	virtual NTSTATUS DeviceControl(KIrp I);			// COMMENT_ONLY
	virtual NTSTATUS SystemControl(KIrp I);			// COMMENT_ONLY
#endif

	//Ӧý̷ڴ
	typedef struct tagUserMemroyBlock{
		CContiguousMemory*	 buf;
		ULONG				 isUsed;
	}TUserMemoryBlock, *PUserMemoryBlock;

	typedef struct tagTimerNode{
		ULONG			timerId;
		TTimerInfo		tmi;
		CTimedCallback* tmc;
		ULONG			irq;
		ULONG			irqPos;
		ULONG			irqMask;
	}TTimerNode, *PTimerNode;

	typedef struct tagDeviceNode{
		ULONG				deviceId;
		union {
			TShareData			datShare;
			TShareData32		datShare32;
		};
		PVOID				basShare;
		BOOLEAN				hasIrq;
		KMemory				mapShare;
		KEvent				evtData;
		KArray<TIrqInfo>	irqList;			//ûӵж
		KArray<TTimerNode>	tmList;				//Timerб
	}TDeviceNode, *PDeviceNode;

protected:

	bool	 IsInterrupt();
	void	 Uninit(UCHAR isRemove);
	void	 OnIrq();

	//ʼ-ʼ
	NTSTATUS OnCmdInit(KIrp I);
	NTSTATUS OnCmdUninit(KIrp I);
	NTSTATUS OnCmdInitDevice(KIrp I);
	NTSTATUS OnCmdUninitDevice(KIrp I);
	NTSTATUS OnCmdAppendCrash(KIrp I);

	//-޸-ɾ豸ж
	NTSTATUS OnCmdAppendIrq(KIrp I);
	NTSTATUS OnCmdModifyIrq(KIrp I);
	NTSTATUS OnCmdRemoveIrq(KIrp I);

	NTSTATUS OnCmdModifyEvt(KIrp I);

	NTSTATUS OnCmdAppendTimer(KIrp I);
	NTSTATUS OnCmdRemoveTimer(KIrp I);
	NTSTATUS OnCmdModifyTimer(KIrp I);
	NTSTATUS OnCmdCancelTimer(KIrp I);

	NTSTATUS OnCmdAppendI2C(KIrp I);
	NTSTATUS OnCmdRemoveI2C(KIrp I);

	//Ĵд
	NTSTATUS OnCmdReadMemBYTEs(KIrp I);
	NTSTATUS OnCmdReadMemWORDs(KIrp I);
	NTSTATUS OnCmdReadMemDWORDs(KIrp I);
	NTSTATUS OnCmdWriteMemBYTEs(KIrp I);
	NTSTATUS OnCmdWriteMemWORDs(KIrp I);
	NTSTATUS OnCmdWriteMemDWORDs(KIrp I);
	NTSTATUS OnCmdReadPciConfigSpace(KIrp I);
	NTSTATUS OnCmdWritePciConfigSpace(KIrp I);
	NTSTATUS OnCmdReadPciConfigSpaceEx(KIrp I);
	NTSTATUS OnCmdWritePciConfigSpaceEx(KIrp I);

	//ڴ䡢ͷšӳ
	NTSTATUS OnCmdAllocMemory(KIrp I);
	NTSTATUS OnCmdFreeMemory(KIrp I);
	NTSTATUS OnCmdMapMemory(KIrp I);
	NTSTATUS OnCmdUnmapMemory(KIrp I);
	NTSTATUS OnCmdMapPciMemory(KIrp I);
	NTSTATUS OnCmdUnmapPciMemory(KIrp I);

	//========================================
	// ַPCIڴӳ
	// PCIڴָ
	//========================================
	KMemoryRange* FindMemRage(PHYSICAL_ADDRESS addr) {
		for(int i=0; i<m_Memory.MaxInserted()+1; i++) {
			if(m_Memory[i] && m_Memory[i]->IsValid()) {
				if(m_Memory[i]->CpuPhysicalAddress().QuadPart == addr.QuadPart) {
					return m_Memory[i];
				}
			}
		}
		return NULL;
	}
	
	LONG	 GetDevicePos(ULONG deviceId) {
		if(!m_devList.IsValid()) { -1; }
		for(LONG i=0; i<m_devList.MaxInserted()+1; i++) {
			if((m_devList[i] != NULL) && (m_devList[i]->deviceId == deviceId)) {
				return i;
			}
		}
		return -1;
	}

	void ToIrqInfo32(PIrqInfo32 irqInfo32, PIrqInfo irqInfo);
	void ToIrqInfo(PIrqInfo irqInfo, PIrqInfo32 irqInfo32);
	
protected:
	ULONG				m_Unit;
	KPnpLowerDevice		m_Lower;
	CTW68XXDevice_DriverManagedQueue m_DriverManagedQueue;

	static KArray<KMemoryToProcessMap*> m_Memory;	//pciڴռӳ
	KArray<CMemoryToProcessMap*> m_pciMap;	//pciڴռӳ
	KArray<TUserMemoryBlock>	 m_bufList;	//ûڴ
	KArray<TCrashInfo>	m_crsList;			//ûӵcrash
	KArray<TMemNode>	m_memPci;			//PCIڴ
	KArray<KMemory*>	m_mapList;			//ûӳڴ
	KArray<PDeviceNode> m_devList;
	KDmaAdapter			m_Dma;
	KInterrupt			m_Irq;
	KDeferredCall		m_DpcFor_Irq;
	KSpinLock			m_syncLock;
	CI2C*				m_i2c;

	UCHAR				m_isOpened;
	UCHAR				m_isIoCtrl;
	ULONG				m_uInIsr;
	BOOLEAN				m_bProcess32;
};

inline void CTW68XXDevice::ToIrqInfo(PIrqInfo irqInfo, PIrqInfo32 irqInfo32)
{
	int i = 0;
	irqInfo->irqId		= irqInfo32->irqId;
	irqInfo->regBase	= irqInfo32->regBase;
	irqInfo->regRWBase	= irqInfo32->regRWBase;
	irqInfo->regI2C		= irqInfo32->regI2C;
	irqInfo->timerId	= irqInfo32->timerId;
	irqInfo->devCopy	= irqInfo32->devCopy;
	irqInfo->mskCopy	= irqInfo32->mskCopy;
	irqInfo->freqWrite	= irqInfo32->freqWrite;
	for(i=0; i<MAX_INTSTATUS; i++) {
		irqInfo->mask[i]	 = irqInfo32->mask[i];
		irqInfo->regMask[i]	 = irqInfo32->regMask[i];
		irqInfo->regStatus[i]= irqInfo32->regStatus[i];
		irqInfo->regClear[i] = irqInfo32->regClear[i];
		irqInfo->status[i]	 = irqInfo32->status[i];
	}
	for(i=0; i<MAX_RW_NUM; i++) {
		irqInfo->regRead[i]	 = irqInfo32->regRead[i];
		irqInfo->regWrite[i] = irqInfo32->regWrite[i];
		irqInfo->mskWrite[i] = irqInfo32->mskWrite[i];
	}
}

inline void CTW68XXDevice::ToIrqInfo32(PIrqInfo32 irqInfo32, PIrqInfo irqInfo)
{
	int i = 0;
	irqInfo32->irqId	= static_cast <ULONG32> (irqInfo->irqId);
	irqInfo32->regBase	= static_cast <ULONG32> (irqInfo->regBase);
	irqInfo32->regRWBase= static_cast <ULONG32> (irqInfo->regRWBase);
	irqInfo32->regI2C	= static_cast <ULONG32> (irqInfo->regI2C);
	irqInfo32->timerId	= static_cast <ULONG32> (irqInfo->timerId);
	irqInfo32->devCopy	= static_cast <ULONG32> (irqInfo->devCopy);
	irqInfo32->mskCopy	= static_cast <ULONG32> (irqInfo->mskCopy);
	irqInfo32->freqWrite= static_cast <ULONG32> (irqInfo->freqWrite);
	for(i=0; i<MAX_INTSTATUS; i++) {
		irqInfo32->mask[i]	   = static_cast <ULONG32> (irqInfo->mask[i]);
		irqInfo32->regMask[i]  = static_cast <ULONG32> (irqInfo->regMask[i]);
		irqInfo32->regStatus[i]= static_cast <ULONG32> (irqInfo->regStatus[i]);
		irqInfo32->regClear[i] = static_cast <ULONG32> (irqInfo->regClear[i]);
		irqInfo32->status[i]   = static_cast <ULONG32> (irqInfo->status[i]);
	}
	for(i=0; i<MAX_RW_NUM; i++) {
		irqInfo32->regRead[i]  = static_cast <ULONG32> (irqInfo->regRead[i]);
		irqInfo32->regWrite[i] = static_cast <ULONG32> (irqInfo->regWrite[i]);
		irqInfo32->mskWrite[i] = static_cast <ULONG32> (irqInfo->mskWrite[i]);
	}
}

#endif
