/**
 ***************************************************************************************
 *  Copyright (C) 2008, Sharebit
 *  All rights reserved.
 *
 *  @file 		AviWriter.cpp
 *  @brief		дAVIļCAviWriterʵ
 *
 *  @version	0.1
 *  @author		liran
 *	@brief		created
 *  @date		2008/03/06
 ***************************************************************************************
 */

#include <windows.h>
#include "ScopedLock.h"
#include "AviWriter.h"

#pragma comment(lib, "vfw32.lib")

int CAviWriter::m_iAviRef = 0;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/** 
 ***************************************************************************************
 *  캯
 ***************************************************************************************
 */
CAviWriter::CAviWriter() :
m_pAviFile(NULL),
m_pStreamV(NULL),m_pStreamA(NULL),
m_iFrameRate(25),m_bKeyFramed(FALSE),
m_iLastPts(0),m_iLastPtsV(0),
m_iFramesCnt(0),m_iLength(0)
{
	ZeroMemory(m_szFileName, MAX_PATH*sizeof(TCHAR));
	if(m_iAviRef==0) {
		AVIFileInit();
	}
	m_iAviRef++;
}

/** 
 ***************************************************************************************
 *  
 ***************************************************************************************
 */
CAviWriter::~CAviWriter()
{
	Close();
	m_iAviRef--;
	if(m_iAviRef<=0) {
		AVIFileExit();
		m_iAviRef=0;
	}
}

/** 
 ***************************************************************************************
 *  ļ
 *  @param[in] szFileName ļ
 *  @param[in] vsi ƵϢΪNULLʾûƵ
 *  @param[in] asi ƵϢΪNULLʾûƵ
 *  @return ɹTRUEFALSE
 *
 *  @pre \e szFileNameЧ·+ļΪNULL
 *  @note ļ򿪳ɹ󣬲ʱҪClose
 *  @see ::WriteVideo ::WriteAudio ::Close
 ***************************************************************************************
 */
BOOL CAviWriter::Open(LPCTSTR szFileName, PVideoStreamInfo vsi, PAudioStreamInfo asi)
{
	CrScopedLock autoLock(&m_csWrite);

	if(!Open(szFileName)) {
		return FALSE;
	}

	if((vsi!=NULL) && !CreateVideoStream(vsi->width, vsi->height, vsi->frame_rate, vsi->fccHandler)) {
		return FALSE;
	}
	
	if((asi!=NULL) && !CreateAudioStream(asi->wfx, asi->sample_size, asi->fccHandler)) {
		return FALSE;
	}

	return TRUE;
}

/** 
 ***************************************************************************************
 *  ļ
 *  @param[in] szFileName ļ
 *  @return ɹTRUEFALSE
 *
 *  @pre \e szFileNameЧ·+ļΪNULL
 *  @note ļ򿪳ɹ󣬲ʱҪClose
 *  @see ::WriteVideo ::WriteAudio ::Close
 ***************************************************************************************
 */
inline BOOL CAviWriter::Open(LPCTSTR szFileName)
{
	if(m_pAviFile != NULL) {
		return FALSE;
	}
	
	m_iLastPtsV = 0;
	m_iSamplesA= m_iSamplesV = 0;
	m_pStreamA = m_pStreamV  =NULL;
	m_iLength  = m_iFramesCnt= 0;
	m_bKeyFramed = FALSE;
	
	/*ͬʼ*/
	ZeroMemory(m_szFileName, MAX_PATH*sizeof(TCHAR));
	lstrcpy(m_szFileName, szFileName);
	
	return (AVIFileOpen(&m_pAviFile, szFileName, OF_CREATE|OF_WRITE|OF_SHARE_DENY_WRITE, NULL)==0);
}

/** 
 ***************************************************************************************
 *  رļ
 *
 *  @see ::WriteVideo ::WriteAudio ::Open
 ***************************************************************************************
 */
VOID CAviWriter::Close()
{
	CrScopedLock autoLock(&m_csWrite);
	
	if(m_pStreamV) {
		AVIStreamRelease(m_pStreamV);
		m_pStreamV = NULL;
	}
	if(m_pStreamA) {
		AVIStreamRelease(m_pStreamA);
		m_pStreamA = NULL;
	}
	if(m_pAviFile) {
		AVIFileRelease(m_pAviFile);
		m_pAviFile = NULL;
	}
}

/** 
 ***************************************************************************************
 *  Ƶ
 *  @param[in] dwWidthƵͼ
 *  @param[in] dwHeightƵͼ߶
 *  @param[in] dwFrmRateƵͼ֡
 *  @param[in] fccHandlerƵFOURCC
 *  @return ɹTRUEFALSE
 *
 *  @pre \e ļ򿪳ɹܵøú
 *  @see ::Open ::CreateAudioStream ::WriteVideo ::WriteAudio ::Close
 ***************************************************************************************
 */
inline BOOL CAviWriter::CreateVideoStream(ULONG dwWidth, ULONG dwHeight, ULONG dwFrmRate, ULONG fccHandler)
{
	if((m_pAviFile==NULL) || (m_pStreamV!=NULL)) {
		return FALSE;
	}
	if(dwFrmRate != 0) m_iFrameRate=dwFrmRate;
	
	AVISTREAMINFO aviStreamInfo;	/*AVIStreamHeader aviStreamInfo;*/
	BITMAPINFOHEADER bmiHeader;
	ZeroMemory(&aviStreamInfo,sizeof(AVISTREAMINFO));
	ZeroMemory(&bmiHeader,sizeof(BITMAPINFOHEADER));
	
	aviStreamInfo.fccType				= streamtypeVIDEO;//mmioFOURCC('v','i','d','s');
	aviStreamInfo.fccHandler			= fccHandler;
	aviStreamInfo.dwRate				= m_iFrameRate;
	aviStreamInfo.dwScale				= 1;
	aviStreamInfo.dwSuggestedBufferSize = (dwWidth*dwHeight*2);
	aviStreamInfo.rcFrame.left			= 0;
	aviStreamInfo.rcFrame.top			= 0;
	aviStreamInfo.rcFrame.right			= dwWidth;
	aviStreamInfo.rcFrame.bottom		= dwHeight;
	aviStreamInfo.dwQuality				= -1;
	
	if(AVIFileCreateStream(m_pAviFile, &m_pStreamV, &aviStreamInfo) != 0) {
		return FALSE;
	}

	bmiHeader.biSize		= sizeof(BITMAPINFOHEADER);
	bmiHeader.biCompression = fccHandler;
	bmiHeader.biWidth		= dwWidth;
	bmiHeader.biHeight		= dwHeight;
	bmiHeader.biBitCount	= 16;
	bmiHeader.biPlanes		= 1;
	bmiHeader.biSizeImage	= dwWidth * dwHeight * (bmiHeader.biBitCount/8);
	if(AVIStreamSetFormat(m_pStreamV, 0, &bmiHeader, bmiHeader.biSize) != 0) {
		AVIStreamRelease(m_pStreamV);
		m_pStreamV = NULL;
		return FALSE;
	}
	
	return TRUE;
}

/** 
 ***************************************************************************************
 *  Ƶ
 *  @param[in] wfxƵʽ
 *  @param[in] nSampleSizeһƵС
 *  @param[in] fccHandlerƵFOURCC
 *  @return ɹTRUEFALSE
 *
 *  @pre \e ļ򿪳ɹܵøú
 *  @see ::Open ::CreateVideoStream ::WriteVideo ::WriteAudio ::Close
 ***************************************************************************************
 */
inline BOOL CAviWriter::CreateAudioStream(WAVEFORMATEX wfx, ULONG nSampleSize, ULONG fccHandler)
{
	if((m_pAviFile==NULL) || (m_pStreamA!=NULL)) {
		return FALSE;
	}
	
	AVISTREAMINFO aviStreamInfo;	/*AVIStreamHeader aviStreamInfo;*/
	
	ZeroMemory(&aviStreamInfo,sizeof(AVISTREAMINFO));
	aviStreamInfo.fccType		= streamtypeAUDIO;
	aviStreamInfo.fccHandler	= fccHandler;
	aviStreamInfo.dwQuality		= 0;
	aviStreamInfo.dwRate		= wfx.nAvgBytesPerSec;
	aviStreamInfo.dwScale		= wfx.nBlockAlign;
	aviStreamInfo.dwSampleSize	= nSampleSize;
	if(AVIFileCreateStream(m_pAviFile, &m_pStreamA, &aviStreamInfo) != 0) {
		return FALSE;
	}

	if(wfx.wFormatTag==WAVE_FORMAT_IMA_ADPCM) {
 		IMAADPCMWAVEFORMAT adpcmHeader;
		adpcmHeader.wfx = wfx;
		adpcmHeader.wSamplesPerBlock = (WORD)((wfx.nAvgBytesPerSec / 4) / (WORD)nSampleSize);
		if(AVIStreamSetFormat(m_pStreamA, 0, &adpcmHeader, sizeof(IMAADPCMWAVEFORMAT)) != 0) {
			AVIStreamRelease(m_pStreamA);
			m_pStreamA = NULL;
		}
	}
	else {
		if(AVIStreamSetFormat(m_pStreamA, 0, &wfx, sizeof(WAVEFORMATEX)) != 0) {
			AVIStreamRelease(m_pStreamA);
			m_pStreamA = NULL;
		}
	}
	
	return TRUE;	
}

/** 
 ***************************************************************************************
 *  дƵ֡ļ
 *  @param[in] pDataָƵͼݻָ
 *  @param[in] dwSizeͼݴС
 *  @param[in] isKeyFrmǷǹؼ֡(I֡)
 *  @param[in] ptsǰ֡ʱͬ
 *  @param[in] durationǰ֡ʱ䳤
 *  @return ɹdwSize򷵻0 -- ļڵȴI֡֡I֡
 *          -1 -- Ƶδ-2 -- дհ֡ļʧܣ
 *			-3 -- д֡ݵļʧ
 *
 *  @pre \e Ƶɹſɵøú
 *  @see ::CreateVideoStream ::WriteAudio
 ***************************************************************************************
 */
LONG CAviWriter::WriteVideo(BYTE *pData, ULONG dwSize, ULONGLONG pts, ULONG duration)
{
	CrScopedLock autoLock(&m_csWrite);

	if(m_pStreamV == NULL) {
		return -1;
	}

	if(!m_bKeyFramed) {
		m_bKeyFramed = TRUE;
		m_iLastPts = m_iLastPtsV = pts;
	}

//	if(pts <= m_iLastPts) {
		//ᵼӹĿ֡
//		return -4;
//	}

	//ļ֡Ϊȫ֡ʱʱмҪҪϿյƵ֡
	ULONG iDiff = (ULONG)(pts - m_iLastPts);
	ULONG iEmpty, iSamples=m_iSamplesV;
	ULONG iSecDiff = (ULONG)(pts - m_iLastPtsV);
		
	duration = (duration-1) * 10000;	//Ϊ,ΧΪ1ms
	iEmpty = iDiff / duration;
	if(iEmpty > 1) {
		m_iFramesCnt += iEmpty-1;
	}
	if(iSecDiff >= 10000000) {
		//ʱѾһ
		if(m_iFramesCnt < m_iFrameRate) {
			//ȫ֡ʱһ֡
			iEmpty += m_iFrameRate - m_iFramesCnt;
		}
		iDiff = iSecDiff % 10000000;
		if(iDiff >= duration) {
			iEmpty += iDiff / duration;
			if(iEmpty > 10) {
				//ֹ
				iEmpty = min(iEmpty, m_iFrameRate);
			}
		}
		m_iFramesCnt= 0;
		m_iLastPtsV+= iSecDiff;
	}
	if(iEmpty>1) {
		//ط̫ĻܿܵAVIFileReleaseúӶ
		//ӵĻ֡ʲʵʱʱPlay sdkõʱ䳤ʵʳȲһ£׼ȷ
		//Ӧþ취Ҫд֡aviļܸĲųԴﵽĿ
		for(iDiff=1; iDiff<iEmpty; iDiff++) {
			if(AVIStreamWrite(m_pStreamV, m_iSamplesV++, 1, pData, 0, 0, NULL, NULL)) {
				return -2;
			}
		}
	}

	if(AVIStreamWrite(m_pStreamV, m_iSamplesV++, 1,	pData, dwSize, AVIIF_KEYFRAME,	NULL, NULL)!=0) {
		return -3;
	}
	m_iLastPts = pts;
	m_iFramesCnt++;
	m_iLength += dwSize;

	return dwSize;
}

/** 
 ***************************************************************************************
 *  дƵļ
 *  @param[in] pDataָƵݻָ
 *  @param[in] dwSizeݴС
 *  @param[in] ptsǰʱͬ
 *  @param[in] durationǰʱ䳤
 *  @return ɹdwSize
 *          -1 -- Ƶδ-3 -- дݵļʧܣ
 *
 *  @pre \e Ƶɹſɵøú
 *  @see ::CreateAudioStream ::WriteVideo
 ***************************************************************************************
 */
LONG CAviWriter::WriteAudio(BYTE *pData, ULONG dwSize)
{
	CrScopedLock autoLock(&m_csWrite);

	if(m_pStreamA == NULL) {
		return -1;
	}

	if(AVIStreamWrite(m_pStreamA, m_iSamplesA++, 1,	pData, dwSize, AVIIF_KEYFRAME,	NULL, NULL)!=0) {
		return -3;
	}

	m_iLength += dwSize;

	return dwSize;
}

/** 
 ***************************************************************************************
 *  ȡļǰС
 *  @return ɹǰļСʧܷ-1
 *
 *  @pre \e ļſɵøú
 *  @see ::Open
 ***************************************************************************************
 */
LONG CAviWriter::GetSize()
{
	if(m_pAviFile == NULL) {
		return -1;
	}

	return m_iLength;
}
