// D3DDisplay.cpp: implementation of the CD3DDisplay class.
//
//////////////////////////////////////////////////////////////////////

#include "D3DDisplay.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CD3DDisplay::CD3DDisplay()
{
	m_pVB = NULL;
	m_pTextures = NULL;
	m_pd3dDevice = NULL;
	m_pD3D = NULL;
    
	m_hWnd = NULL;
	m_Format = (D3DFORMAT)0;
}

CD3DDisplay::~CD3DDisplay()
{
	ClearD3D();	
}

HRESULT CD3DDisplay::InitD3D(HWND hWnd, int format, RECT rect)
{
    D3DPRESENT_PARAMETERS d3dpp;

	if(hWnd == NULL)
		return E_FAIL;
	m_hWnd = hWnd;
	
	if(format ==D3D_ARGB24)
		m_Format = D3DFMT_A8R8G8B8;
	else if(format == D3D_YUV422)
		m_Format = D3DFMT_YUY2;
	else
		return E_FAIL;

	m_Rect = rect;

    if( NULL == ( m_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	
    if( FAILED( m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, m_hWnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING | D3DCREATE_MULTITHREADED,
		&d3dpp, &m_pd3dDevice ) ) )
    {
        return E_FAIL;
    }
	
	m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE ); 
	
	m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, D3DZB_FALSE );
	
    return S_OK;
}

void CD3DDisplay::ClearD3D()
{
	if(m_pVB != NULL)
	    m_pVB->Release();
	if(m_pTextures != NULL)
	    m_pTextures->Release();
	if(m_pd3dDevice != NULL)
	    m_pd3dDevice->Release();
	if(m_pD3D != NULL)
	    m_pD3D->Release(); 
    
	m_hWnd = NULL;
	m_Format = (D3DFORMAT)0;
}

HRESULT CD3DDisplay::CreateTexture(int width, int height)
{
    HRESULT hr = S_OK;
    D3DSURFACE_DESC ddsd;
	
	hr = m_pd3dDevice->CreateTexture( width, height, 1, 0, m_Format, 
		D3DPOOL_MANAGED, &m_pTextures, NULL);
	if( FAILED(hr))
	{
		return hr;
	}
	
	hr = m_pTextures->GetLevelDesc( 0, &ddsd );
	if( FAILED(hr))
	{
		return hr;
	}
	
	if (ddsd.Format != D3DFMT_A8R8G8B8 || ddsd.Format != D3DFMT_YUY2) 
	{
		return -1;
	}
	
    return hr;
}

HRESULT CD3DDisplay::InitGeometry()
{
    HRESULT hr = S_OK;
    TCustomVertex* pVertices = NULL;
    RECT rect;
	
	rect.top = 0;			
	rect.left = 0;
	rect.right = 750;
	rect.bottom = 550;
	
	if(m_Format == D3DFMT_A8R8G8B8)
	{
		TCustomVertex vertices[4] =  
		{
			{ (float)rect.left, (float)rect.top, 0.0f, 1.0f, D3DCOLOR_XRGB(255,255,255), 0.0f, 0.0f }, 
			{ (float)rect.right, (float)rect.top, 0.0f, 1.0f, D3DCOLOR_XRGB(255,255,255), 1.0f, 0.0f }, 
			{ (float)rect.left, (float)rect.bottom, 0.0f, 1.0f, D3DCOLOR_XRGB(255,255,255), 0.0f, 1.0f }, 
			{ (float)rect.right, (float)rect.bottom, 0.0f, 1.0f, D3DCOLOR_XRGB(255,255,255), 1.0f, 1.0f }
		};
		
		m_pd3dDevice->CreateVertexBuffer(sizeof(vertices), 
			0, 
			D3DFVF_CUSTOMVERTEX, 
			D3DPOOL_DEFAULT, 
			&m_pVB, NULL ); 
		
		m_pVB->Lock( 0, sizeof(vertices), (void**)&pVertices, 0 ); 
		
		memcpy( pVertices, vertices, sizeof(vertices)); 
		m_pVB->Unlock(); 
	}
	else
	{
		TCustomVertex vertices[4] =  
		{
			{ (float)rect.left, (float)rect.top, 0.0f, 1.0f, D3DCOLOR_XYUV(255,255,255), 0.0f, 0.0f }, 
			{ (float)rect.right, (float)rect.top, 0.0f, 1.0f, D3DCOLOR_XYUV(255,255,255), 1.0f, 0.0f }, 
			{ (float)rect.left, (float)rect.bottom, 0.0f, 1.0f, D3DCOLOR_XYUV(255,255,255), 0.0f, 0.5f }, 
			{ (float)rect.right, (float)rect.bottom, 0.0f, 1.0f, D3DCOLOR_XYUV(255,255,255), 1.0f, 0.5f }
		};
		
		m_pd3dDevice->CreateVertexBuffer( sizeof(vertices), 
			0, 
			D3DFVF_CUSTOMVERTEX, 
			D3DPOOL_DEFAULT, 
			&m_pVB, NULL ); 
		
		m_pVB->Lock( 0, sizeof(vertices), (void**)&pVertices, 0 ); 
		
		memcpy( pVertices, vertices, sizeof(vertices)); 
		m_pVB->Unlock(); 
	}
	
    return hr;
}

HRESULT CD3DDisplay::Setup2DCamera(int width, int height)
{
    D3DXMATRIX matOrtho; 
    D3DXMATRIX matIdentity;
    
	//Setup the orthogonal projection matrix and the default world/view matrix
    D3DXMatrixOrthoLH(&matOrtho, (float)width, (float)height, 
                      0.0f, 1.0f);
    D3DXMatrixIdentity(&matIdentity);

    m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matOrtho);
    m_pd3dDevice->SetTransform(D3DTS_WORLD, &matIdentity);
    m_pd3dDevice->SetTransform(D3DTS_VIEW, &matIdentity);

    //Make sure that the z-buffer and lighting are disabled
    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);

	return 0;
}


HRESULT CD3DDisplay::RenderARGB(BYTE *pSampleBuffer)
{
    HRESULT hr = S_OK;
	
    BYTE * pTextureBuffer = NULL;

    D3DLOCKED_RECT d3dlr;
	D3DSURFACE_DESC ddsd;
    LONG  lTexturePitch;     
	
	hr = m_pTextures->LockRect( 0, &d3dlr, 0, 0 );
	if( FAILED(hr))
	{
		return E_FAIL;
	}
	if( FAILED(m_pTextures->GetLevelDesc(0, &ddsd)) ) {
		return E_FAIL;
	}
	
	lTexturePitch = d3dlr.Pitch;
	pTextureBuffer = static_cast<byte *>(d3dlr.pBits) + lTexturePitch*(ddsd.Width - 1);
	
	for(int i = 0; i < (int)ddsd.Height; i++ ) 
	{
		BYTE *pBmpBufferOld = pSampleBuffer;
		BYTE *pTxtBufferOld = pTextureBuffer;   
		
		for (int j = 0; j < (int)ddsd.Width; j++) 
		{
			pTextureBuffer[0] = pSampleBuffer[0];
			pTextureBuffer[1] = pSampleBuffer[1];
			pTextureBuffer[2] = pSampleBuffer[2];
			pTextureBuffer[3] = 0xFF;
			
			pTextureBuffer += 4;
			pSampleBuffer  += 3;
		}
		
		pSampleBuffer  = pBmpBufferOld + ddsd.Width *3;
		pTextureBuffer = pTxtBufferOld - lTexturePitch;
	}
	
	
	hr = m_pTextures->UnlockRect(0);
	
    return hr;
}

HRESULT CD3DDisplay::RenderYUYV(BYTE* pSampleBuffer)
{
	
	HRESULT hr = S_OK;
	
    BYTE * pTextureBuffer = NULL;
	
    D3DLOCKED_RECT d3dlr;
	D3DSURFACE_DESC ddsd;
    LONG  lTexturePitch;     
	
	hr = m_pTextures->LockRect( 0, &d3dlr, 0, 0 );
	if( FAILED(hr))
	{
		return E_FAIL;
	}
	if( FAILED(m_pTextures->GetLevelDesc(0, &ddsd)) ) {
		return E_FAIL;
	}
	
	pTextureBuffer = static_cast<byte *>(d3dlr.pBits);
	lTexturePitch = d3dlr.Pitch;
	
	for(int i = 0; i < (int)ddsd.Height; i++)
	{
		CopyMemory(pTextureBuffer + (lTexturePitch * i), pSampleBuffer + (ddsd.Width*i<<1), ddsd.Width<<1);
	}
	hr = m_pTextures->UnlockRect(0);
	
    return hr;
}

HRESULT CD3DDisplay::RenderTOSrceen()
{
    HRESULT hr = S_OK;

	if(m_Format == D3DFMT_A8R8G8B8)
	{
		hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0 );
	    if( FAILED(hr))
		{
		    return hr;
		}
	}
	else
	{
	    hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XYUV(0, 0, 0), 1.0f, 0 );
	    if( FAILED(hr))
		{
		    return hr;
		}
	}

	hr = m_pd3dDevice->BeginScene();
	if( FAILED(hr))
	{
		return hr;
	}
	
	hr = m_pd3dDevice->SetTexture( 0, m_pTextures );
	m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX ); 
	m_pd3dDevice->SetStreamSource( 0, m_pVB, 0, sizeof(TCustomVertex) ); 
	m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ); 
	m_pd3dDevice->SetTexture( 0, NULL ); 
	m_pd3dDevice->EndScene();
	
	m_pd3dDevice->Present( NULL, &m_Rect, NULL, NULL );

	return hr;
}
