// Dinterlace.cpp: implementation of the CDinterlace class.
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include "Dinterlace.h"
#include "cpummx.h"
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDinterlace::CDinterlace()
{
}

CDinterlace::~CDinterlace()
{
	Uninit();
}

UINT CDinterlace::TestCpu()
{
#ifndef WIN64

	UINT CpuFeatureFlags = 0;
	UINT signature = 0;
    char vendor[13]        = "UnknownVendr";  // Needs to be exactly 12 chars 

	// Define known vendor strings here 

	char vendorAMD[13]     = "AuthenticAMD";  // Needs to be exactly 12 chars 
	char vendorIntel[13]   = "GenuineIntel";  // Needs to be exactly 12 chars 

    try 
    {
        __asm xor    eax, eax
        __asm xor    ebx, ebx
        __asm xor    ecx, ecx
        __asm xor    edx, edx
        __asm cpuid
    }
    catch(...)
    {
        return 0;
    }
    
    CpuFeatureFlags |= FEATURE_CPUID;

    _asm 
    {
         // Step 2: Check if CPUID supports function 1 (signature/std features)
         xor     eax, eax                      // CPUID function #0
         cpuid                                 // largest std func/vendor string
         mov     dword ptr [vendor], ebx       // save 
         mov     dword ptr [vendor+4], edx     //  vendor
         mov     dword ptr [vendor+8], ecx     //   string
         test    eax, eax                      // largest standard function==0?
         jz      $no_standard_features         // yes, no standard features func
         or      [CpuFeatureFlags], FEATURE_STD_FEATURES// does have standard features

         // Step 3: Get standard feature flags and signature
         mov     eax, 1                        // CPUID function #1 
         cpuid                                 // get signature/std feature flgs
         mov     [signature], eax              // save processor signature

         // Step 4: Extract desired features from standard feature flags
         // Check for time stamp counter support
         mov     ecx, CPUID_STD_TSC            // bit 4 indicates TSC support
         and     ecx, edx                      // supports TSC ? CPUID_STD_TSC:0
         neg     ecx                           // supports TSC ? CY : NC
         sbb     ecx, ecx                      // supports TSC ? 0xffffffff:0
         and     ecx, FEATURE_TSC              // supports TSC ? FEATURE_TSC:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Check for MMX support
         mov     ecx, CPUID_STD_MMX            // bit 23 indicates MMX support
         and     ecx, edx                      // supports MMX ? CPUID_STD_MMX:0
         neg     ecx                           // supports MMX ? CY : NC
         sbb     ecx, ecx                      // supports MMX ? 0xffffffff:0  
         and     ecx, FEATURE_MMX              // supports MMX ? FEATURE_MMX:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Check for CMOV support
         mov     ecx, CPUID_STD_CMOV           // bit 15 indicates CMOV support
         and     ecx, edx                      // supports CMOV?CPUID_STD_CMOV:0
         neg     ecx                           // supports CMOV ? CY : NC
         sbb     ecx, ecx                      // supports CMOV ? 0xffffffff:0
         and     ecx, FEATURE_CMOV             // supports CMOV ? FEATURE_CMOV:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags
         
         // Check support for P6-style MTRRs
         mov     ecx, CPUID_STD_MTRR           // bit 12 indicates MTRR support
         and     ecx, edx                      // supports MTRR?CPUID_STD_MTRR:0
         neg     ecx                           // supports MTRR ? CY : NC
         sbb     ecx, ecx                      // supports MTRR ? 0xffffffff:0
         and     ecx, FEATURE_P6_MTRR          // supports MTRR ? FEATURE_MTRR:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Check for initial SSE support. There can still be partial SSE
         // support. Step 9 will check for partial support.
         mov     ecx, CPUID_STD_SSE            // bit 25 indicates SSE support
         and     ecx, edx                      // supports SSE ? CPUID_STD_SSE:0
         neg     ecx                           // supports SSE ? CY : NC
         sbb     ecx, ecx                      // supports SSE ? 0xffffffff:0
         and     ecx, (FEATURE_MMXEXT+FEATURE_SSEFP+FEATURE_SSE) // supports SSE ? 
                                               // FEATURE_MMXEXT+FEATURE_SSEFP:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Check for SSE2 support. (TRB - Was not part of sample code)
         mov     ecx, CPUID_STD_SSE2            // bit 26 indicates SSE2 support
         and     ecx, edx                      // supports SSE2 ? CPUID_STD_SSE2:0
         neg     ecx                           // supports SSE2 ? CY : NC
         sbb     ecx, ecx                      // supports SSE2 ? 0xffffffff:0
         and     ecx, (FEATURE_SSE2)           // supports SSE2 ? 
                                               // FEATURE_SSE2:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Step 5: Check for CPUID extended functions
         mov     eax, 0x80000000               // extended function 0x80000000
         cpuid                                 // largest extended function
         cmp     eax, 0x80000000               // no function > 0x80000000 ?
         jbe     $no_extended_features         // yes, no extended feature flags
         or      [CpuFeatureFlags], FEATURE_EXT_FEATURES// does have ext. feature flags

         // Step 6: Get extended feature flags 
         mov     eax, 0x80000001               // CPUID ext. function 0x80000001
         cpuid                                 // EDX = extended feature flags
         
         // Step 7: Extract vendor independent features from extended flags 
         // Check for 3DNow! support (vendor independent)
         mov     ecx, CPUID_EXT_3DNOW          // bit 31 indicates 3DNow! supprt
         and     ecx, edx                      // supp 3DNow! ?CPUID_EXT_3DNOW:0
         neg     ecx                           // supports 3DNow! ? CY : NC
         sbb     ecx, ecx                      // supports 3DNow! ? 0xffffffff:0
         and     ecx, FEATURE_3DNOW            // support 3DNow!?FEATURE_3DNOW:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         // Step 8: Determine CPU vendor
         lea     esi, vendorAMD                // AMD's vendor string
         lea     edi, vendor                   // this CPU's vendor string
         mov     ecx, 12                       // strings are 12 characters
         cld                                   // compare lowest to highest
         repe    cmpsb                         // current vendor str == AMD's ?
         jnz     $not_AMD                      // no, CPU vendor is not AMD

         // Step 9: Check AMD specific extended features
         mov     ecx, CPUID_EXT_AMD_3DNOWEXT   // bit 30 indicates 3DNow! ext.
         and     ecx, edx                      // 3DNow! ext? 
         neg     ecx                           // 3DNow! ext ? CY : NC
         sbb     ecx, ecx                      // 3DNow! ext ? 0xffffffff : 0
         and     ecx, FEATURE_3DNOWEXT         // 3DNow! ext?FEATURE_3DNOWEXT:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags

         test    [CpuFeatureFlags], FEATURE_MMXEXT      // determined SSE MMX support?
         jnz     $has_mmxext                   // yes, don't need to check again

         // Check support for AMD's multimedia instruction set additions 

         mov     ecx, CPUID_EXT_AMD_MMXEXT     // bit 22 indicates MMX extension
         and     ecx, edx                      // MMX ext?CPUID_EXT_AMD_MMXEXT:0
         neg     ecx                           // MMX ext? CY : NC
         sbb     ecx, ecx                      // MMX ext? 0xffffffff : 0
         and     ecx, FEATURE_MMXEXT           // MMX ext ? FEATURE_MMXEXT:0
         or      [CpuFeatureFlags], ecx                 // merge into feature flags
        
$has_mmxext:
 
         // Step 10: Check AMD-specific features not reported by CPUID
         // Check support for AMD-K6 processor-style MTRRs          
         mov     eax, [signature]   // get processor signature
         and     eax, 0xFFF         // extract family/model/stepping
         cmp     eax, 0x588         // CPU < AMD-K6-2/CXT ? CY : NC
         sbb     edx, edx       // CPU < AMD-K6-2/CXT ? 0xffffffff:0
         not     edx            // CPU < AMD-K6-2/CXT ? 0:0xffffffff
         cmp     eax, 0x600         // CPU < AMD Athlon ? CY : NC
         sbb     ecx, ecx       // CPU < AMD-K6 ? 0xffffffff:0
         and     ecx, edx       // (CPU>=AMD-K6-2/CXT)&& 
                    // (CPU<AMD Athlon) ? 0xffffffff:0
         and     ecx, FEATURE_K6_MTRR   // (CPU>=AMD-K6-2/CXT)&& 
                    // (CPU<AMD Athlon) ? FEATURE_K6_MTRR:0
         or      [CpuFeatureFlags], ecx         // merge into feature flags

         jmp     $all_done      // desired features determined

$not_AMD:
         // Extract features specific to non AMD CPUs (except SSE2 already done)
$no_extended_features:
$no_standard_features:
$all_done:
   }
	return CpuFeatureFlags;

#else	//WIN64
	return 0;
#endif
}

void CDinterlace::memcpySSE(void* Dest, void* Src, size_t nBytes)
{
#ifndef WIN64
    __asm
    {

		mov     esi, dword ptr[Src]
		mov     edi, dword ptr[Dest]
		mov     ecx, nBytes
		shr     ecx, 7
	align 4
	MemcpySSE_Loop:
		//movaps should be slightly more efficient as the data is 16 bit aligned
		movaps  xmm0, [esi]
		movaps  xmm1, [esi+16*1]
		movaps  xmm2, [esi+16*2]
		movaps  xmm3, [esi+16*3]
		movaps  xmm4, [esi+16*4]
		movaps  xmm5, [esi+16*5]
		movaps  xmm6, [esi+16*6]
		movaps  xmm7, [esi+16*7]
		movntps [edi], xmm0
		movntps [edi+16*1], xmm1
		movntps [edi+16*2], xmm2
		movntps [edi+16*3], xmm3
		movntps [edi+16*4], xmm4
		movntps [edi+16*5], xmm5
		movntps [edi+16*6], xmm6
		movntps [edi+16*7], xmm7
		add     esi, 128
		add     edi, 128
		dec ecx
		jnz MemcpySSE_Loop

		mov     ecx, nBytes
		and     ecx, 127
		cmp     ecx, 0
		je MemcpySSE_End
align 8
CopyLoop2:
        mov dl, BYTE ptr[esi] 
        mov BYTE ptr[edi], dl
        inc esi
        inc edi
        dec ecx
        jne near CopyLoop2
	MemcpySSE_End:
	}
#endif
}

void CDinterlace::memcpyMMX(void *Dest, void *Src, size_t nBytes)
{
#ifndef WIN64
    __asm {
        mov     esi, dword ptr[Src]
        mov     edi, dword ptr[Dest]
        mov     ecx, nBytes
        shr     ecx, 6                      // nBytes / 64
align 8
CopyLoop:
        movq    mm0, qword ptr[esi]
        movq    mm1, qword ptr[esi+8*1]
        movq    mm2, qword ptr[esi+8*2]
        movq    mm3, qword ptr[esi+8*3]
        movq    mm4, qword ptr[esi+8*4]
        movq    mm5, qword ptr[esi+8*5]
        movq    mm6, qword ptr[esi+8*6]
        movq    mm7, qword ptr[esi+8*7]
        movq    qword ptr[edi], mm0
        movq    qword ptr[edi+8*1], mm1
        movq    qword ptr[edi+8*2], mm2
        movq    qword ptr[edi+8*3], mm3
        movq    qword ptr[edi+8*4], mm4
        movq    qword ptr[edi+8*5], mm5
        movq    qword ptr[edi+8*6], mm6
        movq    qword ptr[edi+8*7], mm7
        add     esi, 64
        add     edi, 64
        loop CopyLoop
        mov     ecx, nBytes
        and     ecx, 63
        cmp     ecx, 0
        je EndCopyLoop
align 8
CopyLoop2:
        mov dl, BYTE ptr[esi] 
        mov BYTE ptr[edi], dl
        inc esi
        inc edi
        dec ecx
        jne near CopyLoop2
EndCopyLoop:
		emms
    }
#else
	memcpy(Dest, Src, nBytes);
#endif
}

//
//ͨȡƽֵˮƽ
//
//ȱ㣬ĸƣ
void CDinterlace::ABCD_SSE(BYTE *pIn, BYTE *pOut, LPVOID pInfo)
{
#ifndef WIN64
	DeinterlaceInfo *pinfoDI = (DeinterlaceInfo *)pInfo;
	int LineLength = pinfoDI->LineLength;
	int Height = pinfoDI->FrameHeight;
	int Width = pinfoDI->FrameWidth;
	int OverLength = pinfoDI->OverlayPitch;
	__asm
	{
		MOV EBX,Height
		MOV EDX,1
		DEC EBX
		CMP EBX,1
		JLE near EndCode
Loop1:
		MOV EDI,DWORD PTR SS:[pOut]
		MOV ESI,DWORD PTR SS:[pIn]
		MOV ECX,Width
		MOV EAX,LineLength
		SHR ECX,4
align 8
CopyLoop:
		PREFETCHNTA QWORD PTR DS:[ESI+EAX*2+0x20]
		MOVQ MM1,QWORD PTR DS:[ESI]				//MM1 = pByte
		MOVQ MM2,QWORD PTR DS:[ESI+EAX]			//MM2 = pByte + LineLength
		MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]		//MM3 = pByte + LineLength * 2
		MOVQ MM4,QWORD PTR DS:[ESI+8]				;MM4 = pByte;
		MOVQ MM5,QWORD PTR DS:[ESI+8+EAX]			;MM5 = pByte + LineLength;
		MOVQ MM6,QWORD PTR DS:[ESI+8+EAX*2]			;MM6 = pByte + LineLength * 2;
		PAVGB MM1,MM3
		PAVGB MM4,MM6
		PAVGB MM1,MM2
		PAVGB MM4,MM5
		MOVQ  MM7,MM1
		MOVQ  MM0,MM4
		MOVQ MM1,QWORD PTR DS:[ESI+0x10]
		MOVQ MM2,QWORD PTR DS:[ESI+0x10+EAX]
		MOVQ MM3,QWORD PTR DS:[ESI+0x10+EAX*2]
		MOVQ MM4,QWORD PTR DS:[ESI+0x18]
		MOVQ MM5,QWORD PTR DS:[ESI+0x18+EAX]
		MOVQ MM6,QWORD PTR DS:[ESI+0x18+EAX*2]
		PAVGB MM1,MM3
		PAVGB MM4,MM6
		PAVGB MM1,MM2
		PAVGB MM4,MM5
		MOVQ QWORD PTR DS:[EDI],MM7
		MOVQ QWORD PTR DS:[EDI+8],MM0
		MOVQ QWORD PTR DS:[EDI+0x10],MM1
		MOVQ QWORD PTR DS:[EDI+0x18],MM4
		ADD ESI,0x20
		ADD EDI,0x20
		DEC ECX								;i--
		JNZ near CopyLoop

		MOV ECX,Width							;n = Width % 8
		AND ECX,8								;if(n)
		je near NoNeedAVG
		MOVQ MM1,QWORD PTR DS:[ESI]
		MOVQ MM2,QWORD PTR DS:[ESI+EAX]
		MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
		PAVGB MM1,MM3
		PAVGB MM1,MM2
		MOVQ QWORD PTR DS:[EDI],MM1
		MOVQ MM4,QWORD PTR DS:[ESI+8]
		MOVQ MM5,QWORD PTR DS:[ESI+EAX+8]
		MOVQ MM6,QWORD PTR DS:[ESI+EAX*2+8]
		PAVGB MM4,MM6
		PAVGB MM4,MM5
		MOVQ QWORD PTR DS:[EDI+8],MM4

NoNeedAVG:	
		MOV EAX,OverLength
		ADD DWORD PTR SS:[pOut],EAX
		MOV ECX,LineLength
		ADD DWORD PTR SS:[pIn],ECX
		INC EDX									;edx++
		CMP EDX,EBX
		JL near Loop1
EndCode:
	EMMS
	}
#endif //WIN64
}

void CDinterlace::ABCD_3DNOW(BYTE *pIn, BYTE *pOut, LPVOID pInfo)
{
#ifndef WIN64
	DeinterlaceInfo *pinfoDI = (DeinterlaceInfo *)pInfo;
	int LineLength = pinfoDI->LineLength;
	int Height = pinfoDI->FrameHeight;
	int Width = pinfoDI->FrameWidth;
	int OverLength = pinfoDI->OverlayPitch;
	int n = Height;
	__asm
	{
		MOV EBX,Height
		DEC EBX
		CMP EBX,1
		DEC EBX
		JLE EndCode
Loop1:
		MOV EDI,DWORD PTR SS:[pOut]
		MOV ESI,DWORD PTR SS:[pIn]
		MOV ECX,Width					//	EBP-C
		MOV EAX,LineLength				//	EBP-8
		SHR ECX,4
Loop2:
			PREFETCH QWORD PTR DS:[ESI+EAX*2+0x20]
			MOVQ MM1,QWORD PTR DS:[ESI]
			MOVQ MM2,QWORD PTR DS:[ESI+EAX]
			MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
			PAVGB MM1,MM3
			PAVGB MM1,MM2
			MOVQ QWORD PTR DS:[EDI],MM1
			ADD ESI,8
			ADD EDI,8
			MOVQ MM1,QWORD PTR DS:[ESI]
			MOVQ MM2,QWORD PTR DS:[ESI+EAX]
			MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
			PAVGB MM1,MM3
			PAVGB MM1,MM2
			MOVQ QWORD PTR DS:[EDI],MM1
			ADD ESI,8
			ADD EDI,8
			MOVQ MM1,QWORD PTR DS:[ESI]
			MOVQ MM2,QWORD PTR DS:[ESI+EAX]
			MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
			PAVGB MM1,MM3
			PAVGB MM1,MM2
			MOVQ QWORD PTR DS:[EDI],MM1
			ADD ESI,8
			ADD EDI,8
			MOVQ MM1,QWORD PTR DS:[ESI]
			MOVQ MM2,QWORD PTR DS:[ESI+EAX]
			MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
			PAVGB MM1,MM3
			PAVGB MM1,MM2
			MOVQ QWORD PTR DS:[EDI],MM1
			ADD ESI,8
			ADD EDI,8
			DEC ECX
			JNZ SHORT Loop2
		MOV ECX,Width
		AND ECX,8
		JE SHORT Loop3
		MOVQ MM1,QWORD PTR DS:[ESI]
		MOVQ MM2,QWORD PTR DS:[ESI+EAX]
		MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
		PAVGB MM1,MM3
		PAVGB MM1,MM2
		MOVQ QWORD PTR DS:[EDI],MM1
		MOVQ MM4,QWORD PTR DS:[ESI+8]
		MOVQ MM5,QWORD PTR DS:[ESI+EAX+8]
		MOVQ MM6,QWORD PTR DS:[ESI+EAX*2+8]
		PAVGB MM4,MM6
		PAVGB MM4,MM5
		MOVQ QWORD PTR DS:[pOut],MM4
Loop3:
		MOV ESI,DWORD PTR SS:[pOut]
		MOV ECX,DWORD PTR SS:[pIn]
		ADD ESI,OverLength
		ADD ECX,EAX
		MOV DWORD PTR SS:[pOut],ESI
		MOV DWORD PTR SS:[pIn],ECX
		DEC EBX
		JNZ Loop1
EndCode:
	FEMMS
	}
#endif	//WIN64
}

void CDinterlace::ABCD_MMX(BYTE *pIn, BYTE *pOut, LPVOID pInfo)
{
#ifndef WIN64
	DeinterlaceInfo *pinfoDI = (DeinterlaceInfo *)pInfo;
	int LineLength = pinfoDI->LineLength;
	int Height = pinfoDI->FrameHeight;
	int Width = pinfoDI->FrameWidth;
	int OverLength = pinfoDI->OverlayPitch;
	int Cycles = LineLength / 8;
	const __int64 qwNoLowBitsMask   = 0xFEFEFEFEFEFEFEFE;
	__asm
	{
		MOVQ MM7,qwNoLowBitsMask
		MOV EBX,Height
		DEC EBX
		DEC EBX
		CMP EBX,1
		JLE near EndCode
Loop1:
			MOV EDI,DWORD PTR SS:[pOut]
			MOV ESI,DWORD PTR SS:[pIn]
			MOV ECX,Width
			MOV EAX,LineLength
			SHR ECX,2
Loop2:
				MOVQ MM1,QWORD PTR DS:[ESI]
				MOVQ MM2,QWORD PTR DS:[ESI+EAX]
				MOVQ MM3,QWORD PTR DS:[ESI+EAX*2]
				PAND MM1,MM7
				PAND MM3,MM7
				PSRLQ MM1,1
				PSRLQ MM3,1
				PADDB MM1,MM3
				PAND MM1,MM7
				PAND MM2,MM7
				PSRLQ MM1,1
				PSRLQ MM2,1
				PADDB MM1,MM2
				MOVQ QWORD PTR DS:[EDI],MM1
				ADD ESI,8
				ADD EDI,8
				DEC ECX
				JNZ SHORT Loop2
			MOV ESI,DWORD PTR SS:[pIn]
			MOV ECX,DWORD PTR SS:[pOut]
			ADD ESI,EAX
			ADD ECX,OverLength
			MOV DWORD PTR SS:[pIn],ESI
			MOV DWORD PTR SS:[pOut],ECX
			DEC EBX
			JNZ SHORT Loop1
EndCode:
		EMMS
	}
#endif //WIN64
}

void CDinterlace::Init(int nWidth, int nHeight, int nBitCount, int nDestPitch)
{
    m_dtInfo.CpuFeatureFlags = TestCpu();
	m_dtInfo.FrameWidth = nWidth;
	m_dtInfo.FrameHeight= nHeight;
	m_dtInfo.FieldHeight= nHeight/2;
	m_dtInfo.LineLength = nWidth*nBitCount/8;
	m_dtInfo.InputPitch = m_dtInfo.LineLength;
    m_dtInfo.OverlayPitch = nDestPitch ? nDestPitch : m_dtInfo.InputPitch;
	m_dtInfo.Overlay = NULL;
    if (m_dtInfo.FieldHeight > MAX_INPUT_LINES_PER_FIELD) 
    {
        m_dtInfo.FieldHeight = MAX_INPUT_LINES_PER_FIELD;
    }
	if(m_dtInfo.CpuFeatureFlags & FEATURE_SSE)
	{
		m_dtInfo.pMemcpy = memcpySSE;
		m_dtInfo.pABCD = ABCD_SSE;
	}
	else if(m_dtInfo.CpuFeatureFlags & FEATURE_3DNOW)
	{
		m_dtInfo.pMemcpy = memcpyMMX;
		m_dtInfo.pABCD = ABCD_3DNOW;
	}
	else
    {
        m_dtInfo.pMemcpy = memcpyMMX;
		m_dtInfo.pABCD = ABCD_MMX;
    }
}

void CDinterlace::Uninit()
{
}

BOOL CDinterlace::Deinterlace(PBYTE inData, PBYTE outData)
{
	m_dtInfo.pABCD(inData,outData,&m_dtInfo);
	m_dtInfo.pMemcpy(outData + m_dtInfo.OverlayPitch * (m_dtInfo.FrameHeight - 2),
		inData + m_dtInfo.LineLength * (m_dtInfo.FrameHeight - 2),
		m_dtInfo.LineLength);	//
	m_dtInfo.pMemcpy(outData + m_dtInfo.OverlayPitch * (m_dtInfo.FrameHeight - 1),
		inData + m_dtInfo.LineLength * (m_dtInfo.FrameHeight - 1),
		m_dtInfo.LineLength);	//
	
	return TRUE;
}