// Surface.h: interface for the CSurface class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__SUFACE_H_)
#define __SUFACE_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <ddraw.h>

class CSurface
{
protected:
    LPDIRECTDRAWSURFACE7 m_pdds;
    DDSURFACEDESC2       m_ddsd;
    BOOL                 m_bColorKeyed;

public:
    CSurface();
    virtual ~CSurface();

    LPDIRECTDRAWSURFACE7 GetDDrawSurface() { return m_pdds; }
    BOOL                 IsColorKeyed()    { return m_bColorKeyed; }

    HRESULT DrawBitmap( HBITMAP hBMP, DWORD dwBMPOriginX = 0, DWORD dwBMPOriginY = 0, 
		                DWORD dwBMPWidth = 0, DWORD dwBMPHeight = 0 );
    HRESULT DrawBitmap( TCHAR* strBMP, DWORD dwDesiredWidth, DWORD dwDesiredHeight );
    HRESULT DrawText( HFONT hFont, TCHAR* strText, DWORD dwOriginX, DWORD dwOriginY,
		              COLORREF crBackground, COLORREF crForeground );

    HRESULT SetColorKey( DWORD dwColorKey );
    DWORD   ConvertGDIColor( COLORREF dwGDIColor );
    static HRESULT GetBitMaskInfo( DWORD dwBitMask, DWORD* pdwShift, DWORD* pdwBits );

    HRESULT Create( LPDIRECTDRAW7 pDD, DDSURFACEDESC2* pddsd );
    HRESULT Create( LPDIRECTDRAWSURFACE7 pdds );
	HRESULT Create( LPDIRECTDRAW7 pDD, DWORD dwWidth, DWORD dwHeight, 
					int pixel_format=1, PVOID lpSurface=NULL);
    HRESULT Destroy();

	HDC		GetSurfaceDC();
	HRESULT Draw(PBYTE data, BOOL isMirror=FALSE, LONG lPitch=0);
	HRESULT Draw(LPDIRECTDRAWSURFACE7 pdds, BOOL isMirror=FALSE);
	HRESULT Draw(LPDIRECTDRAWSURFACE7 pdds, RECT* rcSrc, RECT* rcDst, BOOL bSrcClkKey);
	BOOL	CopyData(PBYTE buff);
	DWORD	Width(){return m_ddsd.dwWidth;}
	DWORD	Height(){return m_ddsd.dwHeight;}
	DWORD	BitCount(){return m_ddsd.ddpfPixelFormat.dwYUVBitCount;}
	DWORD	Pitch(){return (DWORD)m_ddsd.lPitch;}
	void	SetWidth(DWORD dwWidth){m_ddsd.dwWidth=dwWidth;}
	void	SetHeight(DWORD dwHeight){m_ddsd.dwHeight=dwHeight;}
	PVOID	GetBuffer(DWORD *pitch);
	void	Restore();
	HRESULT SetSurface(PVOID lpSurface, int width, int height);

//	virtual IsReady(){return TRUE;}
//	virtual HWND GetWnd(){return NULL;}
//	virtual RECT GetDestRect(){RECT rc={0,0,0,0}; return rc;}

	static HRESULT DrawYuy2(DDSURFACEDESC2* pddsd, PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	static HRESULT DrawYv12(DDSURFACEDESC2* pddsd, PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	static HRESULT DrawYuy2(LPDIRECTDRAWSURFACE7 pSurface, PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	static HRESULT DrawYv12(LPDIRECTDRAWSURFACE7 pSurface, PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
	static HRESULT DrawRgb(DDSURFACEDESC2* pddsd, PBYTE data, SIZE srcSize, RECT* rcDest=NULL, RECT* rcSrc=NULL, BOOL isMirror=FALSE, LONG lPitch=0);
};

inline HDC	CSurface::GetSurfaceDC()
{
	if(!m_pdds)
		return NULL;

	HDC	hDC;
	m_pdds->GetDC(&hDC);
	return hDC;
}

#endif // !defined(__SUFACE_H_)
