// VideoOut.cpp: implementation of the CVideoOut class.
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include "ScopedLock.h"
#include "VideoOut.h"
#include "Surface.h"
#include <stdio.h>

#include "ConvertYuvRgb.h"

#pragma comment(lib,"ddraw.lib")
#pragma comment(lib,"dxguid.lib")
#pragma comment(lib,"vfw32.lib")

#define SAFE_RELEASE(p) {if(p) {p->Release();p=NULL;}}

#ifdef _VC6
extern "C" void asm_convert_yuy2_bgr24(void *dst, void *src, int w, int h, long dstmod, long srcmod);
#endif

DDPIXELFORMAT	g_pixelformat[] =
{
    {sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('Y','V','1','2'),12,0,0,0,0},	// YV12
	{sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('Y','U','Y','2'),16,0,0,0,0},	// YUY2
    {sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16,  0xF800, 0x07e0, 0x001F, 0},			// 16-bit RGB 5:6:5
	{sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16,  0x7C00, 0x03e0, 0x001F, 0},			// 16-bit RGB 5:5:5
	{sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('U','Y','V','Y'),16,0,0,0,0},	// UYVY
	{sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16,0,0,0,0},								// primary surface pixel format
	{sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('Y','4','1','1'),12,0,0,0,0},	// Y411/IYU1
	{sizeof(DDPIXELFORMAT), DDPF_FOURCC, MAKEFOURCC('Y','4','1','P'),12,0,0,0,0},	// Y41P
    {sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 24,  0xFF0000, 0x00FF00, 0x0000FF, 0},		// 24-bit RGB
    {sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 24,  0x0000FF, 0x00FF00, 0xFF0000, 0},		// 24-bit BGR
};

const DDPIXELFORMAT* g_pixelformatindex[]=
{
	&g_pixelformat[0],	//PIX_FMT_YUV420P,   ///< Planar YUV 4:2:0 (1 Cr & Cb sample per 2x2 Y samples)
	&g_pixelformat[1],	//PIX_FMT_YUV422,    ///< Packed pixel, Y0 Cb Y1 Cr 
	&g_pixelformat[8],	//PIX_FMT_RGB24,     ///< Packed pixel, 3 bytes per pixel, RGBRGB...
	&g_pixelformat[9],	//PIX_FMT_BGR24,     ///< Packed pixel, 3 bytes per pixel, BGRBGR...
	NULL,				//PIX_FMT_YUV422P,   ///< Planar YUV 4:2:2 (1 Cr & Cb sample per 2x1 Y samples)
	NULL,				//PIX_FMT_YUV444P,   ///< Planar YUV 4:4:4 (1 Cr & Cb sample per 1x1 Y samples)
	&g_pixelformat[5],	//PIX_FMT_RGBA32,    ///< Packed pixel, 4 bytes per pixel, BGRABGRA..., stored in cpu endianness
	NULL,				//PIX_FMT_YUV410P,   ///< Planar YUV 4:1:0 (1 Cr & Cb sample per 4x4 Y samples)
	&g_pixelformat[0],	//PIX_FMT_YUV411P,   ///< Planar YUV 4:1:1 (1 Cr & Cb sample per 4x1 Y samples)
	&g_pixelformat[2],	//PIX_FMT_RGB565,    ///< always stored in cpu endianness 
	&g_pixelformat[3],	//PIX_FMT_RGB555,    ///< always stored in cpu endianness, most significant bit to 1 
	NULL,				//PIX_FMT_GRAY8,
	NULL,				//PIX_FMT_MONOWHITE, ///< 0 is white 
	NULL,				//PIX_FMT_MONOBLACK, ///< 0 is black 
	NULL,				//PIX_FMT_PAL8,      ///< 8 bit with RGBA palette 
	NULL,				//PIX_FMT_YUVJ420P,  ///< Planar YUV 4:2:0 full scale (jpeg)
	NULL,				//PIX_FMT_YUVJ422P,  ///< Planar YUV 4:2:2 full scale (jpeg)
	NULL,				//PIX_FMT_YUVJ444P,  ///< Planar YUV 4:4:4 full scale (jpeg)
	NULL,				//PIX_FMT_XVMC_MPEG2_MC,///< XVideo Motion Acceleration via common packet passing(xvmc_render.h)
	NULL,				//PIX_FMT_XVMC_MPEG2_IDCT,
	&g_pixelformat[4],	//PIX_FMT_UYVY422,   ///< Packed pixel, Cb Y0 Cr Y1 
	&g_pixelformat[6],	//PIX_FMT_UYVY411,   ///< Packed pixel, Cb Y0 Y1 Cr Y2 Y3 Cb Y4 Y5 Cr Y6 Y7
	&g_pixelformat[7],	//PIX_FMT_UYVY41P,	 ///< Packed pixel, Cb Y0 Cr Y1 Cb Y2 Cr Y3 Y4 Y5 Y6 Y7
	&g_pixelformat[5]	//PIX_FMT_NB,
};

BOOL CVideoOut::m_bBltSystemToVideo = TRUE;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVideoOut::CVideoOut() :
m_hDDib(NULL),
m_pDDraw(NULL),
m_pPrimary(NULL),
m_pOverlay(NULL),
m_pClipper(NULL),
m_clrKey(RGB(10,10,10)),
m_DDsColor(0)
{
	ZeroMemory(m_pBuff, sizeof(m_pBuff));
}

CVideoOut::~CVideoOut()
{
	Uninit();
}

void CVideoOut::Uninit()
{
	if(m_pDDraw) {
		m_pDDraw->RestoreDisplayMode();
	}

	if(m_hDDib) {
		DrawDibClose(m_hDDib);
		m_hDDib = NULL;
	}

	if(m_pBuff[0] != NULL) {
		delete[] m_pBuff[0];
		m_pBuff[0] = NULL;
	}

	if(m_pBuff[1] != NULL) {
		delete[] m_pBuff[1];
		m_pBuff[1] = NULL;
	}

	CrScopedLock autoLock(&m_csLock);
	SAFE_RELEASE(m_pClipper);
	SAFE_RELEASE(m_pOverlay);
	SAFE_RELEASE(m_pPrimary);
	SAFE_RELEASE(m_pDDraw);
}

int CVideoOut::InitDibDraw() 
{
	if(m_hDDib == NULL) {
		m_hDDib = DrawDibOpen();
	}

	if(m_pBuff[0] == NULL) {
		m_pBuff[0] = new BYTE[720*576*3];
	}
	
	if(m_pBuff[1] == NULL) {
		m_pBuff[1] = new BYTE[720*576*3];
	}
	
	return (m_hDDib!=NULL) ? S_OK : S_FALSE;
}

int CVideoOut::InitDDraw(HWND hWnd)
{
	if(m_pDDraw) {
		return S_OK;
	}
	
	HRESULT        ddrval;
	DDSURFACEDESC2 ddsd;

	// Create the main DirectDraw Object
	ddrval = DirectDrawCreateEx(NULL, (void **) &m_pDDraw, IID_IDirectDraw7, NULL);
	if(FAILED(ddrval)) 
		return ddrval;
	ddrval = m_pDDraw->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
	if(FAILED(ddrval)) 
	{
		ddrval = DDERR_INVALIDOBJECT;
		Uninit();
		return ddrval;
	}
	//Create the primary surface
	memset( &ddsd, 0, sizeof(ddsd) );
	ddsd.dwSize     = sizeof( ddsd );
	ddsd.dwFlags           = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE;// | DDSCAPS_VIDEOMEMORY;
	ddrval = m_pDDraw->CreateSurface(&ddsd, &m_pPrimary, NULL);
	if(FAILED(ddrval)) 
	{
		Uninit();
		return ddrval;
	}

	//ظʽ
	m_pPrimary->GetSurfaceDesc(&ddsd);
	g_pixelformat[5] = ddsd.ddpfPixelFormat;

	// Now set Clipping
	ddrval = m_pDDraw->CreateClipper(0, &m_pClipper, NULL);
	if(FAILED(ddrval)) {
		Uninit();
		return ddrval;
	}

//	ddrval = m_pClipper->SetHWnd(0, hWnd);
//	if(FAILED(ddrval)) {
//	Uninit();
//	return EEoReturn(ERR_V_SET_DDRAW);
//	}
	ddrval = m_pPrimary->SetClipper(m_pClipper);
	if(ddrval != DD_OK) {
		Uninit();
		return ddrval;
	}

	return S_OK;
}

int CVideoOut::Draw(LPDIRECTDRAWSURFACE7 pdds, RECT rcDest, HWND hWnd)
{
	HRESULT hr;
	int nCount = 0;
	DDSURFACEDESC2 ddsd;
	RECT rcSrc;

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	if(pdds && FAILED(hr=pdds->GetSurfaceDesc(&ddsd)))
		return hr;

	SetRect(&rcSrc, 0, 0, ddsd.dwWidth, ddsd.dwHeight);

	while( pdds )
	{
		CrScopedLock autoLock(&m_csLock);
		if(m_pOverlay)
		{
//			rcDest.right  += 2;
			if(rcDest.right > (LONG)m_ddover.dwWidth)
				rcDest.right = m_ddover.dwWidth;
			if(rcDest.bottom > (LONG)m_ddover.dwHeight)
				rcDest.bottom = m_ddover.dwHeight;

			hr = m_pOverlay->Blt( &rcDest, pdds, &rcSrc, /*DDBLT_ASYNC |*/ DDBLT_WAIT, NULL);
			
/*	IDirectDrawColorControl* pCCtrl = NULL;
	m_pOverlay->QueryInterface(IID_IDirectDrawColorControl, (void**)&pCCtrl);
	if(pCCtrl)
	{
		DDCOLORCONTROL cc;
		ZeroMemory(&cc, sizeof(cc));
		cc.dwSize = sizeof(cc);

		hr = pCCtrl->GetColorControls(&cc);
		cc.dwFlags= DDCOLOR_SHARPNESS;
		cc.lSharpness = 4;
//		cc.lBrightness= 5000;

		hr = pCCtrl->SetColorControls(&cc);
		pCCtrl->Release();
	}
*/		}
		else
		{
			m_pClipper->SetHWnd(0, hWnd);
//			m_pDDraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
			hr = m_pPrimary->Blt( &rcDest, pdds, &rcSrc, /*DDBLT_ASYNC |*/ DDBLT_WAIT, NULL);

//			HDC hdc;
//			if(SUCCEEDED(m_pPrimary->GetDC(&hdc)))
//			{
//				char* sz = "liran liran";
//				//SetBkMode(hdc, TRANSPARENT);
//				SetBkColor(hdc, RGB(255,0,0));
//				SetTextColor(hdc, RGB(255, 255, 255));
//
//				DWORD DDsColor = ConvertGDIColor( m_pPrimary, RGB(255,0,0) );
//				DDCOLORKEY dd = {DDsColor, DDsColor};
//				m_pPrimary->SetColorKey(DDCKEY_COLORSPACE, &dd);
//
//				TextOut( hdc, rcDest.left, rcDest.top, sz, lstrlen(sz) );
//			    m_pPrimary->ReleaseDC( hdc );
//			}
		}
		if( hr == DD_OK )
		{
			break;
		}
		else if( hr == DDERR_SURFACELOST )
		{
			nCount++;
			if(FAILED(hr=m_pDDraw->RestoreAllSurfaces()))
			{
				return hr;
			}
			else if(nCount > 1)
				break;
		}
		else if( hr != DDERR_WASSTILLDRAWING )
		{
			return hr;
		}
	}
	
	return S_OK;
}

int CVideoOut::Draw(LPDIRECTDRAWSURFACE7 pdds, HWND hWnd, RECT rcDest, RECT* rcSrc)
{
	HRESULT hr;
	int nCount = 0;
	DDSURFACEDESC2 ddsd;
	RECT rect;

	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	if(pdds && FAILED(hr=pdds->GetSurfaceDesc(&ddsd)))
		return hr;

	if(rcSrc) {
		rect = *rcSrc;
	}
	else {
		SetRect(&rect, 0, 0, ddsd.dwWidth, ddsd.dwHeight);
	}

	while( pdds )
	{
		CrScopedLock autoLock(&m_csLock);
		if(m_pOverlay)
		{
			if(rcDest.right > (LONG)m_ddover.dwWidth)
				rcDest.right = m_ddover.dwWidth;
			if(rcDest.bottom > (LONG)m_ddover.dwHeight)
				rcDest.bottom = m_ddover.dwHeight;
			hr = m_pOverlay->Blt( &rcDest, pdds, &rect, DDBLT_WAIT, NULL);
		}
		else
		{
			m_pClipper->SetHWnd(0, hWnd);
			hr = m_pPrimary->Blt( &rcDest, pdds, &rect, DDBLT_WAIT, NULL);
		}
		if( hr == DD_OK )
		{
			break;
		}
		else if( hr == DDERR_SURFACELOST )
		{
			nCount++;
			if(FAILED(hr=m_pDDraw->RestoreAllSurfaces()))
			{
				return hr;
			}
			else if(nCount > 1)
				break;
		}
		else if( hr != DDERR_WASSTILLDRAWING )
		{
			return hr;
		}
	}
	
	return S_OK;
}

DWORD CVideoOut::ConvertGDIColor(LPDIRECTDRAWSURFACE7 pdds, COLORREF dwGDIColor)
{
    if(pdds == NULL)
	    return 0x00000000;

    COLORREF       rgbT;
    HDC            hdc;
    DWORD          dw = CLR_INVALID;
    DDSURFACEDESC2 ddsd;
    HRESULT        hr;

    //  Use GDI SetPixel to color match for us
    if( dwGDIColor != CLR_INVALID && pdds->GetDC(&hdc) == DD_OK)
    {
        rgbT = GetPixel(hdc, 0, 0);     // Save current pixel value
        SetPixel(hdc, 0, 0, dwGDIColor);       // Set our value
        pdds->ReleaseDC(hdc);
    }

    // Now lock the surface so we can read back the converted color
    ddsd.dwSize = sizeof(ddsd);
    hr = pdds->Lock( NULL, &ddsd, DDLOCK_WAIT, NULL );
    if( hr == DD_OK)
    {
        dw = *(DWORD *) ddsd.lpSurface; 
        if( ddsd.ddpfPixelFormat.dwRGBBitCount < 32 ) // Mask it to bpp
            dw &= ( 1 << ddsd.ddpfPixelFormat.dwRGBBitCount ) - 1;  
        pdds->Unlock(NULL);
    }

    //  Now put the color that was there back.
    if( dwGDIColor != CLR_INVALID && pdds->GetDC(&hdc) == DD_OK )
    {
        SetPixel( hdc, 0, 0, rgbT );
        pdds->ReleaseDC(hdc);
    }
    
    return dw;    
}

#ifndef WIN64

void CVideoOut::memcpyMMX(void *Dest, void *Src, size_t nBytes)
{
    __asm {
        mov     esi, dword ptr[Src]
        mov     edi, dword ptr[Dest]
        mov     ecx, nBytes
        shr     ecx, 6                      // nBytes / 64
align 8
CopyLoop:
        movq    mm0, qword ptr[esi]
        movq    mm1, qword ptr[esi+8*1]
        movq    mm2, qword ptr[esi+8*2]
        movq    mm3, qword ptr[esi+8*3]
        movq    mm4, qword ptr[esi+8*4]
        movq    mm5, qword ptr[esi+8*5]
        movq    mm6, qword ptr[esi+8*6]
        movq    mm7, qword ptr[esi+8*7]
        movq    qword ptr[edi], mm0
        movq    qword ptr[edi+8*1], mm1
        movq    qword ptr[edi+8*2], mm2
        movq    qword ptr[edi+8*3], mm3
        movq    qword ptr[edi+8*4], mm4
        movq    qword ptr[edi+8*5], mm5
        movq    qword ptr[edi+8*6], mm6
        movq    qword ptr[edi+8*7], mm7
        add     esi, 64
        add     edi, 64
        loop CopyLoop
        mov     ecx, nBytes
        and     ecx, 63
        cmp     ecx, 0
        je EndCopyLoop
EndCopyLoop:
		emms
    }
}

void CVideoOut::MMXRectCpy(RECT dsr, RECT ddr, BYTE *Dest, BYTE *Src)
{
	DWORD dswidth= dsr.right -dsr.left;
	DWORD dsheight= dsr.bottom - dsr.top;
	DWORD ddwidth= ddr.right - ddr.left;
	DWORD ddheight= ddr.bottom - ddr.top;
	BYTE *pD= NULL;
	BYTE *pS= NULL;

	pD= Dest + ddr.top*m_ddover.lPitch + (ddr.left << 1);
	pS= Src;

	for(UINT i=0; i<dsheight; i++)
	{
		memcpyMMX(pD, pS, dswidth*2);
		pD+= m_ddover.lPitch;
		pS+= dswidth * 2;
	}
}

#endif

int CVideoOut::Draw(PBYTE data, SIZE srcSize, RECT* rcDest, RECT* rcSrc, LONG lPitch)
{
	if(!m_pOverlay)
		return S_FALSE;
	
	HRESULT hr = E_FAIL;
	CrScopedLock autoLock(&m_csLock);

	if(m_pOverlay->IsLost() == DDERR_SURFACELOST) {
		Restore();
	}
	switch(m_ddover.ddpfPixelFormat.dwFourCC)
	{
	case MAKEFOURCC('Y','U','Y','2'):
		hr = CSurface::DrawYuy2(&m_ddover, data, srcSize, rcDest, rcSrc, FALSE, lPitch);
		break;
	case MAKEFOURCC('Y','V','1','2'):
		hr = CSurface::DrawYv12(&m_ddover, data, srcSize, rcDest, rcSrc, FALSE, lPitch);
		break;
	default:
		break;
	}

	return hr;
}

int CVideoOut::SetOverlayColorKey(COLORREF clrKey)
{
	HRESULT hr = 0;
	DWORD OverlayFlags= DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;
	DDOVERLAYFX	OverlayFX;

	ZeroMemory(&OverlayFX, sizeof(OverlayFX));
	m_clrKey = clrKey;
	m_DDsColor = ConvertGDIColor( m_pPrimary, clrKey );
	OverlayFX.dckDestColorkey.dwColorSpaceLowValue  = m_DDsColor;
	OverlayFX.dckDestColorkey.dwColorSpaceHighValue = m_DDsColor;

	OverlayFX.dwSize = sizeof(DDOVERLAYFX);

	if(m_pOverlay) {
		hr = m_pOverlay->UpdateOverlay(NULL, m_pPrimary, NULL, OverlayFlags, &OverlayFX);
	}

	return hr;
}

int CVideoOut::EnableOverlayEx(BOOL bEnable, int pixel_format, int width, int height)
{
	return EnableOverlay(bEnable, pixel_format, m_clrKey, width, height);
}

int CVideoOut::EnableOverlay(BOOL bEnable, int pixel_format, COLORREF clrKey, int width, int height)
{
	if(!bEnable)
	{
		CrScopedLock autoLock(&m_csLock);
		SAFE_RELEASE(m_pOverlay);
		return S_OK;
	}
	else if(m_pOverlay)
		return S_OK;

	if(!g_pixelformatindex[pixel_format])
		return S_FALSE;

	HRESULT hr;
	int i=0;
	DDCAPS capsDrv;
	DWORD dwWidth = (width==0) ? GetSystemMetrics(SM_CXSCREEN) : width;
	DWORD dwHeight= (height==0)? GetSystemMetrics(SM_CYSCREEN) : height;

	ZeroMemory(&capsDrv, sizeof(capsDrv));
	capsDrv.dwSize = sizeof(DDCAPS);
	hr = m_pDDraw->GetCaps(&capsDrv, NULL);

	if(FAILED(hr))
		return hr;

	if(!(capsDrv.dwCaps & DDCAPS_OVERLAY))
		return S_FALSE;
	if(!(capsDrv.dwCaps & DDCAPS_OVERLAYSTRETCH))
		return S_FALSE;
 
	memset(&m_ddover,0,sizeof(DDSURFACEDESC2));
	m_ddover.dwSize = sizeof(DDSURFACEDESC2);
	m_ddover.dwFlags           = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	m_ddover.ddsCaps.dwCaps    = DDSCAPS_OVERLAY|DDSCAPS_VIDEOMEMORY/*|DDSCAPS_LIVEVIDEO|DDSCAPS_LOCALVIDMEM*/;
//|DDSCAPS_VIDEOMEMORY | DDSCAPS_OVERLAY |DDSCAPS_COMPLEX| DDSCAPS_FLIP 
	m_ddover.dwWidth           = dwWidth;
	m_ddover.dwHeight          = dwHeight;
	memcpy(&m_ddover.ddpfPixelFormat, g_pixelformatindex[pixel_format], sizeof(DDPIXELFORMAT));

	if( FAILED( hr = m_pDDraw->CreateSurface( &m_ddover, &m_pOverlay, NULL ) ) )
		return hr;

	RECT rcc= {0,0,dwWidth,dwHeight};
	int  oversize = 0;
	void*pData = NULL;

	m_pOverlay->AddOverlayDirtyRect(&rcc);
	if((hr=m_pOverlay->Lock(NULL, &m_ddover, DDLOCK_WAIT, NULL))!=NOERROR )
		return hr;
	pData = (BYTE*) m_ddover.lpSurface;
	oversize  = m_ddover.dwWidth * m_ddover.dwHeight * m_ddover.ddpfPixelFormat.dwRGBBitCount / 8;
	
	if(pData)
		//ZeroMemory(pData,oversize);
		memset(pData, 128, oversize);
	m_pOverlay->Unlock(NULL);

	DWORD OverlayFlags= DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;
	DDOVERLAYFX	OverlayFX;

	ZeroMemory(&OverlayFX, sizeof(OverlayFX));
	m_clrKey = clrKey;
	m_DDsColor = ConvertGDIColor( m_pPrimary, clrKey );
	OverlayFX.dckDestColorkey.dwColorSpaceLowValue  = m_DDsColor;
	OverlayFX.dckDestColorkey.dwColorSpaceHighValue = m_DDsColor;

	OverlayFX.dwSize = sizeof(DDOVERLAYFX);

	hr = m_pOverlay->UpdateOverlay(&rcc, m_pPrimary, &rcc, OverlayFlags, &OverlayFX);

	return hr;
}

#define INIT_DIRECTDRAW_STRUCT(x) (ZeroMemory(&x, sizeof(x)), x.dwSize=sizeof(x))
int CVideoOut::CopyBitmapToYUVSurface(HBITMAP hbm)
{
    HDC                 hdcImage;
    HRESULT             ddrval;
    DDSURFACEDESC2      ddsd;
    DWORD               x, y, dwWidth, dwHeight;
    LONG                lPitch;
    LPBYTE              pSurf;
    DWORD               dwBytesInRow;
    COLORREF            color;
    BYTE                R,G,B, Y0,Y1,U,V;
    BOOL                bRet = FALSE;

    if (hbm == NULL || m_pOverlay == NULL)
        return FALSE;

    //
    //  select bitmap into a memoryDC so we can use it.
    //
    hdcImage = CreateCompatibleDC(NULL);
    SelectObject(hdcImage, hbm);


    INIT_DIRECTDRAW_STRUCT(ddsd);
    // Lock down the surface so we can modify it's contents.
    ddrval=m_pOverlay->Lock( NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL);
    if (FAILED(ddrval))
	    goto CleanUp;

	//SIZE dm;
	//GetBitmapDimensionEx(hbm, &dm);
	BITMAP bm;
	GetObject(hbm, sizeof(BITMAP), &bm);

    dwWidth=bm.bmWidth;//ddsd.dwWidth;
    dwHeight=bm.bmHeight;//ddsd.dwHeight;
    lPitch=ddsd.lPitch;
    pSurf=(LPBYTE)ddsd.lpSurface;
	dwBytesInRow=0;//ddsd.dwWidth*2;

    // Go through the image 2 pixels at a time and convert to YUV
    for(y=0; y<dwHeight; y++)
    {
    	for(x=0; x<dwWidth; x+=2)
		{
            // The equations for color conversion used here, probably aren't 
            // exact, but they seem to do an OK job.
			color=GetPixel(hdcImage, x,y);
			R=GetRValue(color);
			G=GetGValue(color);
			B=GetBValue(color);
	        Y0= (BYTE)(0.29*R + 0.59*G + 0.14*B);
			U= (BYTE)(128.0 - 0.14*R - 0.29*G + 0.43*B);

			color=GetPixel(hdcImage, x+1,y);
			R=GetRValue(color);
			G=GetGValue(color);
			B=GetBValue(color);
			Y1= (BYTE)(0.29*R + 0.57*G + 0.14*B);
			V= (BYTE)(128.0 + 0.36*R - 0.29*G - 0.07*B);

			switch (ddsd.ddpfPixelFormat.dwFourCC)
			{
				case MAKEFOURCC('Y','U','Y','2'): 
					*(pSurf++) = Y0;
					*(pSurf++) = U;
					*(pSurf++) = Y1;
					*(pSurf++) = V;
					break;
				case MAKEFOURCC('U','Y','V','Y'): 
					*(pSurf++) = U;
					*(pSurf++) = Y0;
					*(pSurf++) = V;
					*(pSurf++) = Y1;
					break;
			}                       
		}
		pSurf+=(lPitch-dwBytesInRow);
    }

    m_pOverlay->Unlock(NULL);     

CleanUp:
    if(hdcImage)
        DeleteDC(hdcImage);

	return TRUE;
}

BOOL CVideoOut::Restore()
{
	BOOL restored = FALSE;
	if(m_pPrimary->IsLost() == DDERR_SURFACELOST) {
		m_pDDraw->RestoreAllSurfaces();
		m_pPrimary->Restore();
		restored = TRUE;
	}
	if(m_pOverlay)
	{
		if(restored || m_pOverlay->IsLost() == DDERR_SURFACELOST) {
			m_pOverlay->Restore();
			m_pOverlay->Lock(NULL, &m_ddover, DDLOCK_WAIT, NULL);
			m_pOverlay->Unlock(NULL);
			restored = TRUE;
		}

//		DWORD OverlayFlags= DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;
//		DDOVERLAYFX	OverlayFX;
//		RECT  rcc= {0,0,GetSystemMetrics(SM_CXSCREEN),GetSystemMetrics(SM_CYSCREEN)};
//		DWORD DDsColor = ConvertGDIColor( m_pPrimary, m_clrKey );
//
//		ZeroMemory(&OverlayFX, sizeof(OverlayFX));
//		OverlayFX.dckDestColorkey.dwColorSpaceLowValue  = DDsColor;
//		OverlayFX.dckDestColorkey.dwColorSpaceHighValue = DDsColor;
//
//		OverlayFX.dwSize = sizeof(DDOVERLAYFX);
//
//		m_pOverlay->UpdateOverlay(&rcc, m_pPrimary, &rcc, OverlayFlags, &OverlayFX);
	}
	return restored;
}

PVOID CVideoOut::GetOverlayBuffer(POINT pt, DWORD* pitch)
{
	PBYTE pBuffer = NULL;
	if(m_pOverlay)
	{
		if ((m_pPrimary->IsLost() == DDERR_SURFACELOST) || (m_pOverlay->IsLost() == DDERR_SURFACELOST)) {
			Restore();
			if ((m_pPrimary->IsLost() == DDERR_SURFACELOST) || (m_pOverlay->IsLost() == DDERR_SURFACELOST)) {
				return NULL;
			}
		}

		if(SUCCEEDED(m_pOverlay->Lock(NULL, &m_ddover, DDLOCK_WAIT, NULL))) {
			pBuffer = (PBYTE)m_ddover.lpSurface;
			pBuffer+= m_ddover.lPitch*pt.y + pt.x*2;

			if(pitch)
				*pitch = m_ddover.lPitch;
			m_pOverlay->Unlock(NULL);
		}
	}

	return pBuffer;
}

void CVideoOut::RefreshOverlay(RECT* rcSrc, RECT* rcDst)
{
	if(m_pOverlay)
	{
		DWORD OverlayFlags= DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;
		DDOVERLAYFX	OverlayFX;

		ZeroMemory(&OverlayFX, sizeof(OverlayFX));
		OverlayFX.dckDestColorkey.dwColorSpaceLowValue  = m_DDsColor;
		OverlayFX.dckDestColorkey.dwColorSpaceHighValue = m_DDsColor;

		OverlayFX.dwSize = sizeof(DDOVERLAYFX);

		m_pOverlay->UpdateOverlay(rcSrc, m_pPrimary, rcDst, OverlayFlags, &OverlayFX);
	}
}

BOOL CVideoOut::HideOverlay()
{
	if(m_pOverlay) {
		DWORD OverlayFlags= DDOVER_HIDE | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;
		DDOVERLAYFX	OverlayFX;

		ZeroMemory(&OverlayFX, sizeof(OverlayFX));
		OverlayFX.dckDestColorkey.dwColorSpaceLowValue  = m_DDsColor;
		OverlayFX.dckDestColorkey.dwColorSpaceHighValue = m_DDsColor;

		OverlayFX.dwSize = sizeof(DDOVERLAYFX);

		HRESULT hr = m_pOverlay->UpdateOverlay(NULL, m_pPrimary, NULL, OverlayFlags, &OverlayFX);
		return (hr==DD_OK);
	}
	return FALSE;
}

void CVideoOut::FillOverlayRect(RECT* rc, BYTE val)
{
	if(m_pOverlay) {
		if(m_pOverlay->Lock(NULL, &m_ddover, DDLOCK_WAIT, NULL)!=NOERROR) return;
		try {
			if(rc==NULL) {
				memset(m_ddover.lpSurface, val, m_ddover.dwHeight*m_ddover.lPitch);
			}
			else if(IsRectEmpty(rc)) {
				m_pOverlay->Unlock(NULL);
				return;
			}
			else {
				if((rc->top<0) || (rc->left<0) || (rc->right<0) || (rc->bottom<0)) return;
				if((rc->left>(int)m_ddover.dwWidth) || (rc->right>(int)m_ddover.dwWidth)) return;
				if((rc->top>(int)m_ddover.dwHeight) || (rc->bottom>(int)m_ddover.dwHeight)) return;

				int i = 0;
				int stride = (rc->right-rc->left) << 1;
				PBYTE pBuffer = (PBYTE)m_ddover.lpSurface;
				pBuffer+= m_ddover.lPitch*rc->top + (rc->left<<1);
				for(i=0; i<rc->bottom-rc->top; i++) {
					memset(pBuffer, val, stride);
					pBuffer += m_ddover.lPitch;
				}
			}
		}
		catch (...) {
			m_pOverlay->Unlock(NULL);
			throw;
		}
		m_pOverlay->Unlock(NULL);
	}
}

BOOL CVideoOut::TestBltSystemToVideo(LPDIRECTDRAWSURFACE7 pdds, HWND hWnd)
{
	if(pdds) {
		HRESULT hr = DDERR_UNSUPPORTED;
		DDSURFACEDESC2 ddsd;
		RECT rcDest;

		ddsd.dwSize = sizeof(DDSURFACEDESC2);
		if(pdds && FAILED(pdds->GetSurfaceDesc(&ddsd)))
		{
			m_bBltSystemToVideo = FALSE;
			return FALSE;
		}

		SetRect(&rcDest, 0, 0, ddsd.dwWidth, ddsd.dwHeight);

		CrScopedLock autoLock(&m_csLock);
		if(m_pOverlay)
		{
			hr = m_pOverlay->Blt( &rcDest, pdds, NULL, DDBLT_ASYNC | DDBLT_WAIT, NULL);
		}
		else
		{
			m_pClipper->SetHWnd(0, hWnd);
			hr = m_pPrimary->Blt( &rcDest, pdds, NULL, DDBLT_ASYNC | DDBLT_WAIT, NULL);
		}

		m_bBltSystemToVideo = SUCCEEDED(hr);
	}
	else
	{
		return m_bBltSystemToVideo;
	}

	return m_bBltSystemToVideo;
}


//now only for yuy2
BOOL CVideoOut::CopyData(PBYTE buff, RECT* rect)
{
	if(m_pOverlay==NULL) 
		return FALSE;

	int   w,h; 
	RECT  rc;
	PBYTE pSrc = (PBYTE)m_ddover.lpSurface;
	DWORD dwLine= 0;

	if(rect)
		rc = *rect;
	else 
		SetRect(&rc, 0, 0, m_ddover.dwWidth, m_ddover.dwHeight);
	w = rc.right-rc.left;
	h = rc.bottom-rc.top;
	dwLine = w*2;

	pSrc += rc.top*m_ddover.lPitch + rc.left * 2;
	buff += (h-1) * dwLine;
	for(int i=0; i<h; i++)
	{
		try {
			memcpy(buff, pSrc, dwLine);
		}
		catch(...){
			throw;
		}
		pSrc += m_ddover.lPitch;
		buff -= dwLine;
	}
        
	return TRUE;
}

int CVideoOut::DrawRGB24(PBYTE data, SIZE srcSize, HWND hWnd, RECT* rcDest/* =NULL */, RECT* rcSrc/* =NULL */, BOOL isMIrror, LONG lPitch)
{
	if(m_hDDib == NULL)
		return -1;

	RECT rcFrom, rcTo;
	if(rcSrc) {
		rcFrom = *rcSrc;
	} else {
		SetRect(&rcFrom, 0, 0, srcSize.cx, srcSize.cy);
	}
	if(rcDest) {
		rcTo = *rcDest;
	} else if(hWnd != NULL){
		::GetWindowRect(hWnd, &rcTo);
	}
	else {
		return -1;
	}

	HDC hDC = GetDC(hWnd);
	if(hDC == NULL) {
		return FALSE;
	}

	BITMAPINFOHEADER bmih;
	ZeroMemory(&bmih, sizeof(bmih));
	bmih.biSize = sizeof(bmih);
	bmih.biBitCount = 24;
	bmih.biWidth = srcSize.cx;
	bmih.biHeight = srcSize.cy;
	bmih.biPlanes = 1;
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = srcSize.cx * srcSize.cy * 3;
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;

	if(isMIrror) {
		BYTE  pixel;
		PBYTE rgb=data;
		PBYTE rgbMirror;

		for(int i=0; i<srcSize.cy; i++) {
			rgbMirror = rgb + (srcSize.cx-1)*3;
			for(int j=0; j<srcSize.cx/2; j++) {
				for(int k=0; k<3; k++) {
					pixel = *rgb;
					*rgb  = *rgbMirror;
					*rgbMirror = pixel;
					rgb++;
					rgbMirror++;
				}
				rgbMirror -= 6;
			}
			rgb += srcSize.cx/2 * 3;
		}
	}

	CrScopedLock autoLock(&m_csLock);
	DrawDibBegin(m_hDDib, hDC, rcTo.right-rcTo.left, rcTo.bottom-rcTo.top, &bmih, 
				 rcFrom.right-rcFrom.left, rcFrom.bottom-rcFrom.top, DDF_SAME_HDC);
	DrawDibDraw(m_hDDib, hDC, rcTo.left, rcTo.top, rcTo.right-rcTo.left,
				rcTo.bottom-rcTo.top, &bmih, data, rcFrom.left, rcFrom.top, 
				rcFrom.right-rcFrom.left, rcFrom.bottom-rcFrom.top, DDF_SAME_HDC|DDF_SAME_DRAW);
	DrawDibEnd(m_hDDib);

	ReleaseDC(hWnd, hDC);

	return 0;
}

int CVideoOut::DrawYUY2(PBYTE data, SIZE srcSize, HWND hWnd, RECT* rcDest/* =NULL */, RECT* rcSrc/* =NULL */, BOOL isMIrror, LONG lPitch)
{
	if(m_hDDib == NULL)
		return -1;

	RECT rcFrom, rcTo;
	if(rcSrc) {
		rcFrom = *rcSrc;
	} else {
		SetRect(&rcFrom, 0, 0, srcSize.cx, srcSize.cy);
	}
	if(rcDest) {
		rcTo = *rcDest;
	} else if(hWnd != NULL){
		::GetClientRect(hWnd, &rcTo);
	}
	else {
		return -1;
	}

	HDC hDC = GetDC(hWnd);
	if(hDC == NULL) {
		return FALSE;
	}

	BITMAPINFOHEADER bmih;
	ZeroMemory(&bmih, sizeof(bmih));
	bmih.biSize = sizeof(bmih);
	bmih.biBitCount = 24;
	bmih.biWidth = srcSize.cx;
	bmih.biHeight = srcSize.cy;
	bmih.biPlanes = 1;
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = srcSize.cx * srcSize.cy * 3;
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;

	CrScopedLock autoLock(&m_csLock);

#ifdef _VC6
	asm_convert_yuy2_bgr24(m_pBuff[0], data, srcSize.cx/2, srcSize.cy, 0, 0);
#else
	int stride = srcSize.cx*2;
	CConvertYuvRgb::yuy2_to_rgb24(data,m_pBuff[0],
		data+srcSize.cy*stride,stride,srcSize.cx*3,srcSize.cx, srcSize.cy);
#endif

	ReverseBGR888(m_pBuff[0], m_pBuff[1], srcSize.cx*3, srcSize.cy);
	data = m_pBuff[1];

	if(isMIrror) {
		BYTE  pixel;
		PBYTE rgb=data;
		PBYTE rgbMirror;

		for(int i=0; i<srcSize.cy; i++) {
			rgbMirror = rgb + (srcSize.cx-1)*3;
			for(int j=0; j<srcSize.cx/2; j++) {
				for(int k=0; k<3; k++) {
					pixel = *rgb;
					*rgb  = *rgbMirror;
					*rgbMirror = pixel;
					rgb++;
					rgbMirror++;
				}
				rgbMirror -= 6;
			}
			rgb += srcSize.cx/2 * 3;
		}
	}

	DrawDibBegin(m_hDDib, hDC, rcTo.right-rcTo.left, rcTo.bottom-rcTo.top, &bmih, 
				 rcFrom.right-rcFrom.left, rcFrom.bottom-rcFrom.top, DDF_SAME_HDC);
	DrawDibDraw(m_hDDib, hDC, rcTo.left, rcTo.top, rcTo.right-rcTo.left,
				rcTo.bottom-rcTo.top, &bmih, data, rcFrom.left, rcFrom.top, 
				rcFrom.right-rcFrom.left, rcFrom.bottom-rcFrom.top, DDF_SAME_HDC|DDF_SAME_DRAW);
	DrawDibEnd(m_hDDib);

	ReleaseDC(hWnd, hDC);

	return 0;
}

void CVideoOut::ReverseBGR888(PBYTE src, PBYTE dst, UINT widthbytes, UINT height)
{
	src = src + widthbytes*(height-1);
	for(UINT y=0; y<height; y++)
	{
		memcpy(dst, src, widthbytes);
		src -= widthbytes;
		dst += widthbytes;
	}
}

BOOL CVideoOut::CanBltYuv2Primary()
{
	DDCAPS DDDriverCaps, DDHELCaps;

	if(m_pDDraw==NULL) {
		return FALSE;
	}

	ZeroMemory(&DDDriverCaps, sizeof(DDCAPS));
	ZeroMemory(&DDHELCaps, sizeof(DDCAPS));
	DDDriverCaps.dwSize = sizeof(DDCAPS);
	DDHELCaps.dwSize = sizeof(DDCAPS);

	if(FAILED(m_pDDraw->GetCaps(&DDDriverCaps, &DDHELCaps))) {
		return FALSE;
	}

	return ((DDHELCaps.dwCaps&DDCAPS_BLTFOURCC) || (DDDriverCaps.dwCaps&DDCAPS_BLTFOURCC));
}

static BOOL BmpFromDib(LPCTSTR szFile, HANDLE hDIB, long size)
{
	BITMAPFILEHEADER hdr;
	LPBITMAPINFOHEADER lpbi;
	if(!hDIB)
		return FALSE;
	DWORD  dwWriten;
	HANDLE hFile = CreateFile(szFile, GENERIC_WRITE|GENERIC_READ, FILE_SHARE_WRITE, 
			NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	lpbi = (LPBITMAPINFOHEADER)hDIB;
	
	int nColors = 1 << lpbi->biBitCount; //  note!!!!! 1<< 
	if(nColors > 256)
		nColors = 0;
	
	// Fill header
	hdr.bfType = ((WORD)('M'<<8)|'B');
	hdr.bfSize = size + sizeof(hdr);
	hdr.bfReserved1 = 0;
	hdr.bfReserved2 = 0;
	hdr.bfOffBits = (DWORD)(sizeof(hdr) + lpbi->biSize + nColors*sizeof(RGBQUAD));
	
	WriteFile(hFile, &hdr, sizeof(hdr), &dwWriten, NULL);
	WriteFile(hFile, lpbi, (DWORD)size, &dwWriten, NULL);
	CloseHandle(hFile);
	return TRUE;
}

//------------------------------------------------------------------------
// yuvʽݱΪbmpļ
//------------------------------------------------------------------------
BOOL CVideoOut::SaveYUVToBmpFile(LPCTSTR FileName, PBYTE yuv, int Width, int Height)
{
	if(NULL==FileName || NULL==yuv) 
		return FALSE;
	
	DWORD dwLen = Width*Height*3 + sizeof(BITMAPINFOHEADER);
	PBYTE buff = new BYTE[dwLen];
	int	  stride = (Width*2);
	
	ZeroMemory(buff, dwLen);
#ifdef _VC6
	asm_convert_yuy2_bgr24(buff+sizeof(BITMAPINFOHEADER), yuv, Width/2, Height, 0, 0);
#else
	CConvertYuvRgb::yuy2_to_rgb24(yuv,buff+sizeof(BITMAPINFOHEADER),
		yuv+Height*stride,stride,Width*3,Width, Height);
#endif

	BITMAPINFOHEADER * pHeader = (BITMAPINFOHEADER *) buff;
	pHeader->biSize = sizeof(BITMAPINFOHEADER);
	pHeader->biBitCount = 24;
	pHeader->biWidth = Width;
	pHeader->biHeight = Height;
	pHeader->biPlanes = 1;
	pHeader->biCompression = BI_RGB;
	pHeader->biSizeImage = Width * Height * 3;
	pHeader->biXPelsPerMeter = 0;
	pHeader->biYPelsPerMeter = 0;
	pHeader->biClrUsed = 0;
	pHeader->biClrImportant = 0;
	
	BmpFromDib(FileName, buff, dwLen);
	delete []buff;

	return TRUE;
}

//------------------------------------------------------------------------
// RGBʽݱΪbmpļ
//------------------------------------------------------------------------
BOOL CVideoOut::SaveRGBToBmpFile(LPCTSTR FileName, PBYTE rgb, int Width, int Height, int bitCount)
{
	if(NULL==FileName || NULL==rgb) 
		return FALSE;
	
	DWORD dwLen = Width*Height*bitCount/8 + sizeof(BITMAPINFOHEADER);
	PBYTE buff = new BYTE[dwLen];
	BITMAPINFOHEADER * pHeader = (BITMAPINFOHEADER *) buff;
	
	ZeroMemory(buff, dwLen);
	memcpy(buff+sizeof(BITMAPINFOHEADER), rgb, Width*Height*bitCount/8);
	
	pHeader->biSize = sizeof(BITMAPINFOHEADER);
	pHeader->biBitCount = bitCount;
	pHeader->biWidth = Width;
	pHeader->biHeight = Height;
	pHeader->biPlanes = 1;
	pHeader->biCompression = BI_RGB;
	pHeader->biSizeImage = Width * Height * bitCount/8;
	pHeader->biXPelsPerMeter = 0;
	pHeader->biYPelsPerMeter = 0;
	pHeader->biClrUsed = 0;
	pHeader->biClrImportant = 0;
	
	BmpFromDib(FileName, buff, dwLen);
	delete []buff;
	
	return TRUE;
}

