//////////////////////////////////////////////////////////////////////
// Channel.cpp: implementation of the CChannel class.
// ͨ
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Channel.h"
#include "PlgvSource.h"
#include "PlgaSource.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CChannel::CChannel() :
m_index(0),
m_plgList(NULL),
m_plgAdded(0),
m_plgEnabled(0),
m_pData(NULL),
m_nDataLen(0)
{
	int numPlugin = pluginIdx(PLG_end);
	m_plgList = new CPlugin*[numPlugin];
	ZeroMemory(m_plgList, sizeof(CPlugin*)*numPlugin);
}

CChannel::~CChannel()
{
	SAFE_DELETE_ARRAY(m_plgList);
}

//----------------------------------------------------------
// Ӳͨ
//----------------------------------------------------------
BOOL CChannel::AppendPlugin(CPlugin* plg)
{
	CScopedLock autoLock(&m_lock);
	if(plg == NULL) {
		return FALSE;
	}
	
	int i = pluginIdx(plg->PluginId());
	if(m_plgList[i] == plg) {
		return TRUE;
	}
	
	m_plgList[i] = plg;
	m_plgAdded |= plg->PluginId();
	plg->Init(this);

	return TRUE;
}

//----------------------------------------------------------
// Ӳͨ,ûоʵֵĲPLG_venccbk
//----------------------------------------------------------
BOOL CChannel::AppendPlugin(ULONG plgId)
{
	CScopedLock autoLock(&m_lock);
	plgId &= PLG_virtual;
	m_plgAdded |= plgId;

	return TRUE;
}

//----------------------------------------------------------
// ֹ
//----------------------------------------------------------
BOOL CChannel::EnablePlugin(ULONG plgId, BOOL enable)
{
	CDevBase*	dev;
	CPlugin*	plg;
	ULONG		enabled;
	
	if((m_plgAdded & plgId) == 0) {
		return FALSE;
	}

	if(enable) {
		{
			//
			CScopedLock autoLock(&m_lock);
			enabled = m_plgEnabled | plgId;
			m_plgEnabled = enabled;
		}

		//Ƶ豸
		if(plgId & PLG_RunVideoDevice) {
			//ҪƵ豸
			plg = Plugin(PLG_vsource);
			if(plg == NULL) {
				_trace(TEXT("[]CChannel::EnablePluginͨ%dƵԴ\n"), m_index);
				return FALSE;
			}
			dev = ((CPlgvSource*)plg)->Device();
			if(dev == NULL) {
				_trace(TEXT("[]CChannel::EnablePluginͨ%dԴƵ豸\n"), m_index);
				return FALSE;
			}

			dev->SetStateTo(DEVSTAT_readyrun);
			_trace(TEXT("[Ϣ]CChannel::EnablePluginͨ%dƵ豸׼\n"), m_index);
		}

		//Ƶ豸
		if(plgId & PLG_RunAudioDevice) {
			plg = Plugin(PLG_asource);
			if(plg == NULL) { _trace(TEXT("[]CChannel::EnablePluginͨ%dƵԴ\n"), m_index); }
			else {
				dev = ((CPlgaSource*)plg)->Device();
				if(dev) {
					dev->SetStateTo(DEVSTAT_readyrun);
					_trace(TEXT("[Ϣ]CChannel::EnablePluginͨ%dƵ豸׼\n"), m_index);
				}
			}
		}
	}
	else {
		{
			//
			CScopedLock autoLock(&m_lock);
			enabled = m_plgEnabled;
			enabled &= ~plgId;
			m_plgEnabled = enabled;
		}

		if(!(enabled & PLG_RunVideoDevice)) {
			//ҪֹͣƵ豸ֹͣ
			plg = Plugin(PLG_vsource);
			if(plg != NULL) {
				dev = ((CPlgvSource*)plg)->Device();
				if(dev != NULL) {
					if(!dev->HasChannelRunning(PLG_RunVideoDevice)) {
						dev->SetStateTo(DEVSTAT_readystop);
						_trace(TEXT("[Ϣ]CChannel::EnablePluginͨ%dƵ豸׼ֹͣ\n"), m_index);
					}
				}
			}
		}

		//ֹͣƵ豸
		if(!(enabled & PLG_RunAudioDevice)) {
			//ҪֹͣƵ豸ֹͣ
			plg = Plugin(PLG_asource);
			if(plg == NULL) {
				_trace(TEXT("[]CChannel::EnablePluginͨ%dƵԴ\n"), m_index);
			}
			else {
				dev = ((CPlgaSource*)plg)->Device();
				if(dev) {
					if(!dev->HasChannelRunning(PLG_RunAudioDevice)) {
						dev->SetStateTo(DEVSTAT_readystop);
						_trace(TEXT("[Ϣ]CChannel::EnablePluginͨ%dƵ豸׼ֹͣ\n"), m_index);
					}
				}
			}
		}
	}

	return TRUE;
}

/** 
 ***************************************************************************************
 *  ͨɾָ
 *
 *  @param[in] plgIdָҪɾĲID
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CChannel::RemovePlugin(ULONG plgId)
{
	CScopedLock autoLock(&m_lock);
	if((m_plgAdded&plgId) == 0) {
		return FALSE;
	}

	m_plgEnabled &= ~plgId;
	m_plgAdded	 &= ~plgId;

	//ɾ
	int i = pluginIdx(plgId);
	CPlugin* plg = m_plgList[i];
	if(plg) {
		m_plgList[i] = NULL;
		plg->Uninit();
		delete plg;
	}

	return (plg!=NULL);
}

/** 
 ***************************************************************************************
 *  ʼɾв
 ***************************************************************************************
 */
VOID CChannel::Uninit()
{
	CScopedLock autoLock(&m_lock);
	CPlugin* plg = NULL;
	int		 i = 0;

	m_plgEnabled = 0;
	m_plgAdded &= (PLG_end-1);

	//ɾв
	do {
		if((m_plgAdded & 1) && m_plgList[i]) {
			m_plgList[i]->Uninit();
			delete m_plgList[i];
			m_plgList[i] = NULL;
		}
		m_plgAdded >>= 1;
		i++;
	}while(m_plgAdded > 0);
}
