/**
 ***************************************************************************************
 *  Copyright (C) 2009, Techwell
 *  All rights reserved.
 *
 *  @file 		Chipset.h
 *  @brief		оƬCChipsetʵ
 *
 *  @version	0.0.0.1
 *  @author		liran
 *  @date		2009/06/25
 ***************************************************************************************
 */

#include "stdafx.h"
#include "Chipset.h"
#include "PlgvSource.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CChipset::CChipset() :
m_hDevice(NULL),
m_cardInfo(NULL)
{
	ZeroMemory(&m_devInfo, sizeof(TDeviceDesc));
}

CChipset::~CChipset()
{
}

//-------------------------------------------
// 豸ʼ
//-------------------------------------------
BOOL CChipset::Open(HANDLE hDevice, TBoardDesc* cardInfo, TDeviceDesc devInfo)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdInitInfo	  cmdInfo   = (PCmdInitInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	int   i=0;
	BOOL  mapSystem = TRUE;

	if(m_hDevice || (hDevice==NULL) || (hDevice==INVALID_HANDLE_VALUE)) {
		return FALSE;
	}

	m_hDevice	= hDevice;
	m_devInfo	= devInfo;
	m_cardInfo	= cardInfo;

	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdInit;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	cmdInfo->toSystem	= mapSystem;

#ifdef ON_ISR
	bytesRet = DEVICEID(devInfo.HardwareId);
	cmdInfo->toSystem |= (bytesRet<<16);
#endif

	if(!DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN,
						 ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	
	m_devInfo.reserved1 = (ULONG)cmdInfo->mem[0].procBase;
	for(i=0; i<MAX_PCIMEM_NUM; i++) {
		if(cmdInfo->mem[i].addrBase) {
			m_Mem[i].init(m_hDevice, cmdInfo->mem[i].addrBase, cmdInfo->mem[i].procBase, cmdInfo->mem[i].len, TRUE);
		}
	}

	if(!OnInit()) {
		return FALSE;
	}

	if(!OnOpen()) {
		return FALSE;
	}

	if(!OnOpen(cardInfo, &devInfo)) {
		return FALSE;
	}
	
	return TRUE;
}

BOOL CChipset::OnOpen(TBoardDesc* cardInfo, TDeviceDesc* devInfo) 
{
	TDeviceCapability devCaps, *pdevCaps;
	int i = 0;
	ULONG devId = DEVICEID(devInfo->HardwareId);

	ZeroMemory(&devCaps, sizeof(TDeviceCapability));
	devCaps.ddesc.nBoardId	= devInfo->nBoardId;
	devCaps.ddesc.nDeviceId	= devInfo->nDeviceId;

	if((devInfo->Function == 0) || ((devId>=0x6810) && (devId<0x6814))) {	//video
		int nAudio = Devices(DEV_audio);
		m_devInfo.nFirstSubDevice = (int)g_capsListV.size();
		m_devInfo.nSubDevice	  = Devices(DEV_video);
		for(i=0; i<m_devInfo.nSubDevice; i++) {
			pdevCaps = new TDeviceCapability;
			if(pdevCaps == NULL) { return FALSE; }
			*pdevCaps = devCaps;
			OnOpen(pdevCaps, i, DEV_video);
			g_capsListV.push_back(pdevCaps);
			m_capsListV.push_back(pdevCaps);
		}

		for(i=0; i<nAudio; i++) {
			pdevCaps = new TDeviceCapability;
			if(pdevCaps == NULL) { return FALSE; }
			*pdevCaps = devCaps;
			OnOpen(pdevCaps, i, DEV_audio);
			g_capsListA.push_back(pdevCaps);
			m_capsListA.push_back(pdevCaps);
		}
	}
	else {	//audio
		m_devInfo.nFirstSubDevice = (int)g_capsListA.size();
		m_devInfo.nSubDevice	  = Devices(DEV_audio);
		for(i=0; i<m_devInfo.nSubDevice; i++) {
			pdevCaps = new TDeviceCapability;
			if(pdevCaps == NULL) { return FALSE; }
			*pdevCaps = devCaps;
			OnOpen(pdevCaps, i, DEV_audio);
			g_capsListA.push_back(pdevCaps);
			m_capsListA.push_back(pdevCaps);
		}
	}

	return TRUE;
}

//-------------------------------------------
// ʼ豸ر
//-------------------------------------------
void CChipset::Close()
{
	int i=0;
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) { return; }
	
	OnClose();
	OnUninit();

	for(i=0; i<MAX_PCIMEM_NUM; i++) {
		m_Mem[i].uninit();
	}
	m_capsListV.clear();
	m_capsListA.clear();
	m_devList.clear();

	ioctlInfo->cmd = cmdUninit;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	ioctlInfo->context[0] = 0;
	DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL);

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	SAFE_CLOSE_HANDLE(m_hDevice);
}

PShareData CChipset::OpenDevice(ULONG deviceId, PULONG lenShare, HANDLE hEvent)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo  ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdInitDeviceInfo cmdInfo   = (PCmdInitDeviceInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(m_hDevice == NULL) {
		return FALSE;
	}
	
	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdInitDevice;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	cmdInfo->deviceId = deviceId;
	cmdInfo->evtData  = hEvent;
	if(!DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN,
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return NULL;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	if(lenShare != NULL) {
		*lenShare = cmdInfo->lenShare;
	}

	return (PShareData)cmdInfo->datShare;
}

void CChipset::CloseDevice(ULONG deviceId)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo  ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdInitDeviceInfo cmdInfo   = (PCmdInitDeviceInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(m_hDevice == NULL) {	return; }
	
	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdUninitDevice;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	cmdInfo->deviceId = deviceId;
	DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL);
	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
}

//-------------------------------------------
// ȡPCIÿռ
//-------------------------------------------
BOOL CChipset::ReadPciConfigSpace(int nOffset, int nLen, PBYTE pData)
{
	ULONG ioBuff[IOBUFF_LEN];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdRwMemByteInfo cmdInfo   = (PCmdRwMemByteInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd	= cmdReadPciConfigSpace;
	cmdInfo->len	= nLen;
	cmdInfo->memBase= 0;
	cmdInfo->memOffset = nOffset;
	memcpy(cmdInfo->data, pData, nLen);
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		ZeroMemory(pData, nLen);
		return FALSE;
	}
	
	memcpy(pData, cmdInfo->data, nLen);

	return TRUE;
}

//-------------------------------------------
// дPCIÿռ
//-------------------------------------------
BOOL CChipset::WritePciConfigSpace(int nOffset, int nLen, PBYTE pData)
{
	ULONG ioBuff[IOBUFF_LEN];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdRwMemByteInfo cmdInfo   = (PCmdRwMemByteInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd	= cmdWritePciConfigSpace;
	cmdInfo->len	= nLen;
	cmdInfo->memBase= 0;
	cmdInfo->memOffset = nOffset;
	memcpy(cmdInfo->data, pData, nLen);
	return DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL);
}

//-------------------------------------------
// ȡPCIÿռ
//-------------------------------------------
BOOL CChipset::ReadPciConfigSpaceEx(int nOffset, int nLen, PBYTE pData)
{
	ULONG ioBuff[IOBUFF_LEN];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdRwPciConfigSpaceInfo cmdInfo   = (PCmdRwPciConfigSpaceInfo)ioctlInfo->context;
	PBridgeDesc brg = (PBridgeDesc)m_cardInfo->reserved1;
	ULONG bytesRet = 0;
	
	if(!m_hDevice || !brg) {
		return FALSE;
	}
	
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdReadPciConfigSpaceEx;
	cmdInfo->slot  = m_devInfo.nBus;
	cmdInfo->bridge= *brg;
	cmdInfo->len   = nLen;
	cmdInfo->offset= nOffset;
	memcpy(cmdInfo->data, pData, nLen);
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		ZeroMemory(pData, nLen);
		return FALSE;
	}
	
	memcpy(pData, cmdInfo->data, nLen);
	
	return TRUE;
}

//-------------------------------------------
// дPCIÿռ
//-------------------------------------------
BOOL CChipset::WritePciConfigSpaceEx(int nOffset, int nLen, PBYTE pData)
{
	ULONG ioBuff[IOBUFF_LEN];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdRwPciConfigSpaceInfo cmdInfo   = (PCmdRwPciConfigSpaceInfo)ioctlInfo->context;
	PBridgeDesc brg = (PBridgeDesc)m_cardInfo->reserved1;
	ULONG bytesRet = 0;
	
	if(!m_hDevice || !brg) {
		return FALSE;
	}
	
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdWritePciConfigSpaceEx;
	cmdInfo->slot  = m_devInfo.nBus;
	cmdInfo->bridge= *brg;
	cmdInfo->len   = nLen;
	cmdInfo->offset= nOffset;
	memcpy(cmdInfo->data, pData, nLen);
	return DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL);
}

//-------------------------------------------
// ҵƬÿ忨һμ
//-------------------------------------------
BOOL CChipset::FindBridge(PBridgeDesc brg)
{
	ULONG ioBuff[IOBUFF_LEN];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdRwPciConfigSpaceInfo cmdInfo   = (PCmdRwPciConfigSpaceInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd	= cmdReadPciConfigSpaceEx;
	cmdInfo->slot= m_devInfo.nBus;
	cmdInfo->bridge.bus = (ULONG)-1;
	cmdInfo->bridge.hwid= m_devInfo.HardwareId;
	cmdInfo->len	= 4;
	cmdInfo->offset = 0;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		return FALSE;
	}
	
	*brg = cmdInfo->bridge;
	//g_util.ParsePciLocationW((WCHAR*)cmdInfo->data, NULL, &brg->device, &brg->function);

	return TRUE;
}

//-------------------------------------------
// lenָСڴ
//-------------------------------------------
BOOL CChipset::AllocMemory(ULONG len, PMemBlock memBlock)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice || !memBlock || len==0) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdAllocMemory;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	cmdInfo->len = len;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	*memBlock = *cmdInfo;
	if(cmdInfo->inode > 0) {
		memBlock->pnodes = new TMemNode[cmdInfo->inode];
		memcpy(memBlock->pnodes, cmdInfo->nodes, cmdInfo->inode*sizeof(TMemNode));
		ZeroMemory(memBlock->vaddr, memBlock->len);
		return TRUE;
	}
	return FALSE;
}

//-------------------------------------------
// ͷAllocMemoryڴ
//-------------------------------------------
BOOL CChipset::FreeMemory(PMemBlock memBlock)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice || !memBlock) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdFreeMemory;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = *memBlock;
	memcpy(cmdInfo->nodes, memBlock->pnodes, memBlock->inode*sizeof(TMemNode));
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	SAFE_DELETE_ARRAY(memBlock->pnodes);
	ZeroMemory(memBlock, sizeof(TMemBlock));

	return TRUE;
}

//-------------------------------------------
// memBlockָĽ̿ռڴתΪΪڴ
//-------------------------------------------
BOOL CChipset::MapMemory(PMemBlock memBlock)
{
	ULONG ioBuff[IOBUFF_LEN_DW_MAX];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice || !memBlock) { return FALSE; }
	if((memBlock->len / 4096) > (IOBUFF_LEN_DW_MAX - sizeof(TCmdMemoryInfo)/4)) {return FALSE;}

	ioctlInfo->cmd = cmdMapMemory;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = *memBlock;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN_MAX, 
						ioBuff, IOBUFF_LEN_MAX, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	*memBlock = *cmdInfo;
	if(cmdInfo->inode > 0) {
		memBlock->pnodes = new TMemNode[cmdInfo->inode];
		memcpy(memBlock->pnodes, cmdInfo->nodes, cmdInfo->inode*sizeof(TMemNode));
		return TRUE;
	}

	return FALSE;
}

//-------------------------------------------
// memBlockָڴ淴ӳ
//-------------------------------------------
BOOL CChipset::UnmapMemory(PMemBlock memBlock)
{
	ULONG ioBuff[IOBUFF_LEN_DW_MAX];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice || !memBlock) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdUnmapMemory;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = *memBlock;
	memcpy(cmdInfo->nodes, memBlock->pnodes, memBlock->inode*sizeof(TMemNode));
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN_MAX, 
						ioBuff, IOBUFF_LEN_MAX, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	SAFE_DELETE_ARRAY(memBlock->pnodes);
	memBlock->inode = 0;

	return TRUE;
}

//-----------------------------------------------------
// pciBaseָPCIڴռڴӳ䵽̵ڴռ
//-----------------------------------------------------
TMemNode CChipset::MapPciMemory(HANDLE hProcess, ULONG pciBase, ULONG len)
{
	ULONG ioBuff[IOBUFF_LEN_DW_MAX];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG	 bytesRet = 0;
	TMemNode mn = { NULL, 0 };

	if(!m_hDevice) { return mn; }

	ioctlInfo->cmd = cmdMapPciMemory;
	ioctlInfo->hProcess = hProcess;
	cmdInfo->vaddr = (PVOID)pciBase;
	cmdInfo->len   = len;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN_MAX, 
						ioBuff, IOBUFF_LEN_MAX, &bytesRet, NULL)) {
		return mn;
	}

	mn.addr = cmdInfo->vaddr;
	mn.len  = cmdInfo->len;

	return mn;
}

//-------------------------------------------
// memBlockָڴ淴ӳ
//-------------------------------------------
BOOL CChipset::UnmapPciMemory(HANDLE hProcess, ULONG pciBase)
{
	ULONG ioBuff[IOBUFF_LEN_DW_MAX];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdMemoryInfo cmdInfo		= (PCmdMemoryInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice) { return FALSE;	}

	ioctlInfo->cmd = cmdUnmapPciMemory;
	ioctlInfo->hProcess = hProcess;
	cmdInfo->vaddr = (PVOID)pciBase;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN_MAX, 
						ioBuff, IOBUFF_LEN_MAX, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	return TRUE;
}

//-------------------------------------------
// crashָӦõ豸
//-------------------------------------------
BOOL CChipset::AppendCrash(TCrashInfo crash)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdCrashInfo	  cmdInfo	= (PCmdCrashInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdAppendCrash;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = crash;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);

	return TRUE;
}

//-------------------------------------------
// irqָжӦõ豸
//-------------------------------------------
BOOL CChipset::AppendIrq(TCmdIrqInfo irq)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdIrqInfo		  cmdInfo	= (PCmdIrqInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdAppendIrq;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = irq;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);

	return TRUE;
}

//-------------------------------------------
// irqָж޸ĵ豸
//-------------------------------------------
BOOL CChipset::ModifyIrq(TCmdIrqInfo irq)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdIrqInfo		  cmdInfo	= (PCmdIrqInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdModifyIrq;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = irq;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);

	return TRUE;
}

//-------------------------------------------
// irqָжϴ豸ɾ
//-------------------------------------------
BOOL CChipset::RemoveIrq(TCmdIrqInfo irq)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdIrqInfo		  cmdInfo	= (PCmdIrqInfo)ioctlInfo->context;
	ULONG bytesRet = 0;

	if(!m_hDevice) {
		return FALSE;
	}

	ioctlInfo->cmd = cmdRemoveIrq;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = irq;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
						ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}

	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);

	return TRUE;
}

//-------------------------------------------
// tmiָTimerӦõ豸
//-------------------------------------------
BOOL CChipset::AppendTimer(TCmdTimerInfo tmi)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdTimerInfo	  cmdInfo	= (PCmdTimerInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ioctlInfo->cmd = cmdAppendTimer;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = tmi;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}
	
	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	
	return TRUE;
}

//-------------------------------------------
// tmiָTimer豸ɾ
//-------------------------------------------
BOOL CChipset::RemoveTimer(TCmdTimerInfo tmi)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdTimerInfo	  cmdInfo	= (PCmdTimerInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ioctlInfo->cmd = cmdRemoveTimer;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = tmi;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}
	
	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	
	return TRUE;
}

//-------------------------------------------
// ޸tmiָTimer
//-------------------------------------------
BOOL CChipset::ModifyTimer(TCmdTimerInfo tmi)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdTimerInfo	  cmdInfo	= (PCmdTimerInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ioctlInfo->cmd = cmdModifyTimer;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = tmi;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}
	
	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	
	return TRUE;
}

//-------------------------------------------
// tmiָTimerȡ
//-------------------------------------------
BOOL CChipset::CancelTimer(TCmdTimerInfo tmi)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdTimerInfo	  cmdInfo	= (PCmdTimerInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	
	if(!m_hDevice) {
		return FALSE;
	}
	
	ioctlInfo->cmd = cmdCancelTimer;
	ioctlInfo->hProcess = (HANDLE)-1;//OpenProcess(PROCESS_ALL_ACCESS, TRUE, GetCurrentProcessId());
	*cmdInfo = tmi;
	if(!DeviceIoControl(m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN, 
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
		return FALSE;
	}
	
	SAFE_CLOSE_HANDLE(ioctlInfo->hProcess);
	
	return TRUE;
}

/** 
 ***************************************************************************************
 *  SDKݵ¼
 *  @param[in] hEvent¼
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CChipset::ModifyEvt(TCmdNotifyEventInfo evti)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo	ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdNotifyEventInfo	cmdInfo   = (PCmdNotifyEventInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	int   i=0;
	
	if(m_hDevice==NULL) {
		return FALSE;
	}
	
	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdModifyEvt;
	*cmdInfo = evti;
	if(!DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN,
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		return FALSE;
	}
	
	return TRUE;
}

/** 
 ***************************************************************************************
 *  I2Cͨͨ豸
 *  @param[in] i2cI2CϢ
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CChipset::AppendI2C(TCmdI2CInfo &i2c)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo	ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdI2CInfo	cmdInfo   = (PCmdI2CInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	int   i=0;
	
	if(m_hDevice==NULL) {
		return FALSE;
	}
	
	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdAppendI2C;
	*cmdInfo = i2c;
	if(!DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN,
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		return FALSE;
	}

	i2c.i2cHandle = cmdInfo->i2cHandle;
	
	return TRUE;
}

/** 
 ***************************************************************************************
 *  ɾI2Cͨͨ豸
 *  @param[in] i2cI2CϢ
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CChipset::RemoveI2C(TCmdI2CInfo i2c)
{
	ULONG ioBuff[IOBUFF_LEN_DW];
	PIoctlCommandInfo	ioctlInfo = (PIoctlCommandInfo)ioBuff;
	PCmdI2CInfo	cmdInfo   = (PCmdI2CInfo)ioctlInfo->context;
	ULONG bytesRet = 0;
	int   i=0;
	
	if(m_hDevice==NULL) {
		return FALSE;
	}
	
	//ʼ豸
	ZeroMemory(ioBuff, IOBUFF_LEN);
	ioctlInfo->cmd = cmdRemoveI2C;
	*cmdInfo = i2c;
	if(!DeviceIoControl( m_hDevice, IOCTL_COMMAND, ioBuff, IOBUFF_LEN,
		ioBuff, IOBUFF_LEN, &bytesRet, NULL)) {
		return FALSE;
	}
	
	return TRUE;
}