/**
 ***************************************************************************************
 *  Copyright (C) 2009, Techwell
 *  All rights reserved.
 *
 *  @file 		DevBase.h
 *  @brief		豸CDevBase
 *
 *  @version	0.0.0.1
 *  @author		liran
 *  @date		2009/06/25
 ***************************************************************************************
 */

#if !defined(__DEVBASE_H_)
#define __DEVBASE_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Sample.h"

#define IRQ_DMPAP		0				//dma programжʶ
#define IRQ_ERROR		1				//жʶ
#define IRQ_FIELD		2				//żжʶ
#define OPTIMIZE_PERIOD			100
#define OPTIMIZE_PERIOD_PASS	300		//СOPTIMIZE_PERIOD_MAX
#define OPTIMIZE_PERIOD_MAX		1000

#define CAPS_FIELD2		0x00000001		//ԤͲɼʹòͬ
#define CAPS_TVOUT		0x00000020		//ģ
#define CAPS_SWITCH		0x00000040		//һоƬͨ(ֻʹһ)
#define CAPS_4CIF		0x00000080		//4CIFɼ(ֻʹһ)

//豸״̬
typedef enum {
	DEVSTAT_idle,		//ֹͣ״̬
	DEVSTAT_readystop,	//׼ֹͣ
	DEVSTAT_readyrun,	//׼״̬
	DEVSTAT_paused,		//ͣ
	DEVSTAT_running		//״̬
}TDeviceState;

//豸
typedef enum {
	DEV_video,			//Ƶ豸
	DEV_audio,			//Ƶ豸
	DEV_av,				//Ƶ豸
	DEV_unknown			//豸
}TDeviceType;

//豸
typedef struct tagDeviceCapability {
	VOID*			device;
	ULONG			caps;		
	TSubDeviceDesc	ddesc;
}TDeviceCapability, *PDeviceCapability;

class CChipset;

class CDevBase  
{
public:
	CDevBase();
	virtual ~CDevBase();

	BOOL Open(CChipset* pParent, PDeviceCapability devCaps, HANDLE hEvent, int iDev=0);
	void Close();
	CChipset* Parent() { return m_pParent; }

	virtual BOOL CreateChannels();
	virtual BOOL Append2Channel(CChannel* channel);
	virtual int  DeviceType() { return DEV_unknown; }
	virtual PDeviceCapability DeviceCaps() { return m_devCaps; }
	virtual long PrepareRun();
	virtual BOOL Run();
	virtual BOOL Stop();
	virtual long Pause(BOOL enable);
	virtual BOOL EnableDmaIrq(BOOL enable);
	virtual BOOL SetState(TDeviceState dstat){m_state=dstat;return TRUE;}
	virtual BOOL SetStateTo(TDeviceState dstat);
	virtual BOOL UpdateState();
	virtual TDeviceState GetState(){/*CScopedLock autoLock(&m_stateLock); */return m_state;}
	virtual CLock* StateLock(){return &m_stateLock;}
	virtual BOOL HasChannelRunning(ULONG runFlag); 
	virtual BOOL CheckLive(ULONG tkNow);
	virtual VOID NotifyLive(ULONG tkNow) { m_tkLive=tkNow; }
	virtual long HasSignal(){return FALSE;}
	virtual BOOL InsertSample(CSample* pSample);
	virtual BOOL OnData();			//нɼݷַͨ
	virtual long OnIrq();
	virtual BOOL OnDmaError(ULONG status1, ULONG status2, PVOID ctx) { return FALSE; }
	virtual void OnDmaTimeOut() {};
	virtual BOOL ModifyEvt(HANDLE hEvent);

	LONG&		 Optimized() { return m_ifoOptimize.enable; }

	typedef struct tagBufferFull{
		vector<CSample*> splList;
		CChannel*		 chlFull;
		CChannel*		 chlFullTo;
		CChipset*		 chpAlloc;
	}TBufferFull;

	static ULONGLONG timestamp();
	static int& BufferType(){ return m_nBufferType; }
	static int& BufferLen(){ return m_nBufferLen; }
	static BOOL IsBufferFull() { return m_bufFull ? (m_bufFull->chlFullTo!=NULL) : FALSE; }
	static int& AudioPktLen(){ return m_nAudioPktLen; }
	static BOOL SetBufferFull(CChannel* chl) {
		if(m_bufFull == NULL) {	return FALSE; }
		if(m_bufFull->chlFullTo != NULL) { return FALSE; }
		m_bufFull->chlFullTo = chl;
		return TRUE;
	}

protected:
	virtual BOOL OnOpen() { return TRUE; }
	virtual BOOL OnClose();
	virtual BOOL OnAllocResource();
	virtual BOOL OnFreeResource();
	virtual void OnSignalChanged(BOOL bSignal);
	virtual LONG PciOptimize();
	virtual BOOL OnBufferFull(CChannel* chl, BOOL bFull, int nBegin, int nEnd);
	virtual BOOL OnChannel(CChannel* chl) { return TRUE; }
	virtual void RunDmap(UCHAR run) {}

	CChipset*		m_pParent;
	CMemRange*		m_pMem;
	ULONG			m_iDev;
	ULONG			m_iChannel;
	ULONG			m_iDmap;
	ULONG			m_tkLive;		//ʱӣڼ豸Ƿ񻹻
	PShareData		m_Share;		//ڴ
	vector<CChannel*> m_chlList;	//ǰ豸Ҫֵ֧ͨ
	TDeviceState	m_state;		//豸״̬
	TDeviceState	m_stateTo;		//豸е״̬CChannel
	TDeviceState	m_stateOld;		//ոչȥ豸״̬
	CSample*		m_pSample;		//ǰڴ֡
	CLockCS			m_stateLock;	//豸״̬
	vector<CSample*>	m_bufRaw;	//豸DMAʹõbuffer
	PDeviceCapability	m_devCaps;	//豸

	typedef struct tagPciOptimizeInfo{
		long		  enable;
		long		  diff;
	}TPciOptimizeInfo;
	TPciOptimizeInfo  m_ifoOptimize;
	static LONGLONG	  m_tkOptimize;
	static LONG		  m_tkPeriod;

	static TBufferFull* m_bufFull;
	static int		  m_nBufferType;
	static int		  m_nBufferLen;
	static int		  m_nAudioPktLen;
};

#endif // !defined(__DEVBASE_H_)
