//////////////////////////////////////////////////////////////////////
// DmaProgram.cpp: implementation of the Dma program.
// Techwell, 2008.10.10
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "DmaProgram.h"
#include "TW680X.H"

//////////////////////////////////////////////////////////////////////
// class CDmap6802 Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDmap6802::CDmap6802() :
m_phyProg(0), 
m_virProg(0)
{
}

CDmap6802::~CDmap6802()
{
}

void CDmap6802::SetProgAddr(ULONG virProg, ULONG phyProg)
{
	m_virProg = virProg;
	m_phyProg = phyProg;
}

BOOL CDmap6802::AppendRawBuffer(CSample* pSample)
{
	UINT i;
	for(i=0; i<m_splList.size(); i++) {
		if(pSample == m_splList[i]) {
			break;
		}
	}
	if(i == m_splList.size()) {
		m_splList.push_back(pSample);
	}
	return TRUE;
}

BOOL CDmap6802::Build(TaDmaProgInfo* aDpi)
{
	ULONG*	dmap = (ULONG*) m_virProg;
	ULONG	ip = 0;
	int		j;
	UINT	i, k;
	UINT	iLen  = (UINT)(m_splList.size() * (aDpi->lines + 1) << 1);
	UINT	iLoop = (ADMAP_LEN >> 2) / iLen;

	//Ĳظ
	for(k=0; k<iLoop; k++) {
		for(i=0; i<m_splList.size(); i++) {
			for(j=0; j<aDpi->lines; j++) {
				if(j == aDpi->lines-1) {
					dmap[ip++] = DMAP_WRITE | DMAP_SOL | aDpi->line_w | (aDpi->irq ? DMAP_IRQ : 0);
				}
				else {
					dmap[ip++] = DMAP_WRITE | DMAP_SOL | aDpi->line_w;
				}
				if(m_splList[i]->Buffer()->inode == 1) {
					dmap[ip++] = (ULONG)m_splList[i]->Buffer()->pnodes[0].addr + j*aDpi->line_w;
				}
				else {
					dmap[ip++] = (ULONG)m_splList[i]->Buffer()->pnodes[j].addr;
				}
			}
		}

		dmap[ip++] = DMAP_JUMP;
		dmap[ip++] = m_phyProg;
	}

	//ĲJUMP
	iLoop = ((ADMAP_LEN >> 2) % iLen) >> 1;
	for(k=0; k<iLoop; k++) {
		dmap[ip++] = DMAP_JUMP;
		dmap[ip++] = m_phyProg;
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// class CDmap6801 Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDmap6801::CDmap6801() :
m_phyProg(0), m_virProg(0),
m_iDmapLen(0)
{

}

CDmap6801::~CDmap6801()
{

}

void CDmap6801::SetProgAddr(ULONG virProg, ULONG phyProg, ULONG iDmapLen)
{
	PULONG dmap = (PULONG)virProg;
	int	i = 0;

	m_iDmapLen= iDmapLen/4;
	m_virProg = virProg;
	m_phyProg = phyProg;

	if(iDmapLen > 0) {
		while(i < (int)(m_iDmapLen-4)) {
			dmap[i++] = DMAP_SYNCO;
			dmap[i++] = 0;
			dmap[i++] = DMAP_JUMP;
			dmap[i++] = phyProg;
		}
	}
}

//--------------------------------------------------------------------
// DMA program
//--------------------------------------------------------------------
BOOL CDmap6801::Build(TvDmaProgInfo* oDpi, TvDmaProgInfo* eDpi)
{
	ULONG*	dmap = (ULONG*) m_virProg;
	ULONG	ip = 0;
	ULONG	phyWrite; 
	ULONG	lineSize, stride, lenWrite, lenWriten;
	ULONG	page, pageLen;
	ULONG	pageBegin, pageEnd;

	if(dmap == NULL) {
		return FALSE;
	}
	
	//żdma program
	if(eDpi) {
		if(eDpi->memBuf==NULL) {
			dmap[ip++] = DMAP_SYNCE;
			dmap[ip++] = 0;
		}
		else {
			if(eDpi->memBuf->pnodes == NULL) {
				return FALSE;
			}

			ip = 0;
			lineSize = eDpi->width*2;
			stride   = eDpi->pitch;
			if(eDpi->interlace) {
				phyWrite = (ULONG)eDpi->memBuf->pnodes[0].addr + lineSize;
				pageLen	 = eDpi->memBuf->pnodes[0].len - lineSize;
				dmap[ip++] = DMAP_SYNCE;
				dmap[ip++] = 0;
			}
			else {
				phyWrite = (ULONG)eDpi->memBuf->pnodes[0].addr;
				pageLen	 = eDpi->memBuf->pnodes[0].len;
				dmap[ip++] = DMAP_SYNCE;
				dmap[ip++] = 0;
			}
			lenWrite = 0;
			lenWriten= 0;
			page	 = 0;
			pageBegin= phyWrite;
			pageEnd	 = pageBegin + pageLen;

			if(eDpi->memBuf->len < (stride*eDpi->height)) {
				//ڴ治
				return FALSE;
			}

			for(int i=0; i<eDpi->height; i++) {
				if(eDpi->memBuf->inode == 1) {
					//ڴ
					dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
					dmap[ip++] = phyWrite;
					phyWrite  += stride;
				}
				else {
					//ڴ棬ҲǷҳڴ
					if((pageEnd-phyWrite) >= lineSize) {
						dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
						dmap[ip++] = phyWrite;
						phyWrite  += stride;
						
						//lineSizestride
						if(stride > lineSize) {
							while(phyWrite >= pageEnd) {
								page++;
								phyWrite -= pageEnd;
								if(page >= eDpi->memBuf->inode) {
									if(i==eDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = eDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin = (ULONG)eDpi->memBuf->pnodes[page].addr;
								pageEnd	  = pageBegin + pageLen;
								phyWrite += pageBegin;
							}
						}
					}
					else {
						//һҳʣµڴ治дһе
						lenWriten = 0;
						if(pageEnd > phyWrite) {
							lenWrite   = pageEnd-phyWrite;
							dmap[ip++] = DMAP_WRITE | DMAP_SOL | lenWrite;
							dmap[ip++] = phyWrite;
							lenWriten  = lenWrite;
						}
						page++;
						if(page >= eDpi->memBuf->inode) {
							if(i==eDpi->height-1) break;
							return FALSE;
						}
						pageLen	  = eDpi->memBuf->pnodes[page].len;
						if(pageLen==0)  {
							return FALSE;
						}
						pageBegin = (ULONG)eDpi->memBuf->pnodes[page].addr;
						pageEnd	  = pageBegin + pageLen;
						phyWrite  = pageBegin;
						while(lenWriten<lineSize) {
							lenWrite = lineSize-lenWriten;
							if(lenWrite <= (pageEnd-phyWrite)) {
								dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
								dmap[ip++] = phyWrite;
								phyWrite  += lenWrite;
								lenWriten += lenWrite;
							}
							else {
								//һڴҳдһе
								lenWrite = pageEnd - phyWrite;
								dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
								dmap[ip++] = phyWrite;
								lenWriten += lenWrite;

								//ҳ
								page++;
								if(page >= eDpi->memBuf->inode) {
									if(i==eDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = eDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin  = (ULONG)eDpi->memBuf->pnodes[page].addr;
								pageEnd	   = pageBegin + pageLen;
								phyWrite   = pageBegin;
							}
						}
						
						//lineSizestride
						if(stride > lineSize) {
							phyWrite += stride - lineSize;
							while(phyWrite >= pageEnd) {
								page++;
								phyWrite -= pageEnd;
								if(page >= eDpi->memBuf->inode) {
									if(i==eDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = eDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin = (ULONG)eDpi->memBuf->pnodes[page].addr;
								pageEnd	  = pageBegin + pageLen;
								phyWrite += pageBegin;
							}
						}
					}
				}
			}
		}
		dmap[ip++] = DMAP_JUMP | (eDpi->irq ? DMAP_IRQ : 0);
		dmap[ip++] = (oDpi==NULL) ? m_phyProg : (m_phyProg + (ip+1)*4);
	}

	//泡dma program
	if(oDpi){
		if(oDpi->memBuf==NULL) {
			dmap[ip++] = DMAP_SYNCO;
			dmap[ip++] = 0;
		}
		else {
			if(oDpi->memBuf->pnodes == NULL) {
				return FALSE;
			}

 			lineSize = oDpi->width*2;
 			stride   = oDpi->pitch;
			phyWrite = (ULONG)oDpi->memBuf->pnodes[0].addr;
			pageLen	 = oDpi->memBuf->pnodes[0].len;
			lenWrite = 0;
			lenWriten= 0;
			page	 = 0;
			pageBegin= phyWrite;
			pageEnd	 = pageBegin + pageLen;

			if(oDpi->memBuf->len < (stride*oDpi->height)) {
				//ڴ治
				return FALSE;
			}

			dmap[ip++] = DMAP_SYNCO;
			dmap[ip++] = 0;
			
			for(int i=0; i<oDpi->height; i++) {
				if(oDpi->memBuf->inode == 1) {
					//ڴ
					dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
					dmap[ip++] = phyWrite;
					phyWrite  += stride;
				}
				else {
					//ڴ棬ҲǷҳڴ
					if((pageEnd-phyWrite) >= lineSize) {
						dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
						dmap[ip++] = phyWrite;
						phyWrite  += stride;
						
						//lineSizestride
						if(stride > lineSize) {
							while(phyWrite >= pageEnd) {
								page++;
								phyWrite -= pageEnd;
								if(page >= oDpi->memBuf->inode) {
									if(i==oDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = oDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin = (ULONG)oDpi->memBuf->pnodes[page].addr;
								pageEnd	  = pageBegin + pageLen;
								phyWrite += pageBegin;
							}
						}
					}
					else {
						//һҳʣµڴ治дһе
						lenWriten = 0;
						if(pageEnd > phyWrite) {
							lenWrite   = pageEnd-phyWrite;
							dmap[ip++] = DMAP_WRITE | DMAP_SOL | lenWrite;
							dmap[ip++] = phyWrite;
							lenWriten  = lenWrite;
						}
						page++;
						if(page >= oDpi->memBuf->inode) {
							if(i==oDpi->height-1) break;
							return FALSE;
						}
						pageLen	  = oDpi->memBuf->pnodes[page].len;
						if(pageLen==0)  {
							return FALSE;
						}
						pageBegin = (ULONG)oDpi->memBuf->pnodes[page].addr;
						pageEnd	  = pageBegin + pageLen;
						phyWrite  = pageBegin;
						while(lenWriten<lineSize) {
							lenWrite = lineSize-lenWriten;
							if(lenWrite <= (pageEnd-phyWrite)) {
								dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
								dmap[ip++] = phyWrite;
								phyWrite  += lenWrite;
								lenWriten += lenWrite;
							}
							else {
								//һڴҳдһе
								lenWrite = pageEnd - phyWrite;
								dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
								dmap[ip++] = phyWrite;
								lenWriten += lenWrite;

								//ҳ
								page++;
								if(page >= oDpi->memBuf->inode) {
									if(i==oDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = oDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin  = (ULONG)oDpi->memBuf->pnodes[page].addr;
								pageEnd	   = pageBegin + pageLen;
								phyWrite   = pageBegin;
							}
						}
						
						//lineSizestride
						if(stride > lineSize) {
							phyWrite += stride - lineSize;
							while(phyWrite >= pageEnd) {
								page++;
								phyWrite -= pageEnd;
								if(page >= oDpi->memBuf->inode) {
									if(i==oDpi->height-1) break;
									return FALSE;
								}
								pageLen	  = oDpi->memBuf->pnodes[page].len;
								if(pageLen==0)  {
									return FALSE;
								}
								pageBegin = (ULONG)oDpi->memBuf->pnodes[page].addr;
								pageEnd	  = pageBegin + pageLen;
								phyWrite += pageBegin;
							}
						}
					}
				}
			}
		}
		dmap[ip++] = DMAP_JUMP | (oDpi->irq ? DMAP_IRQ : 0);
		dmap[ip++] = m_phyProg;
	}
		
	return TRUE;
}

/** 
 ***************************************************************************************
 *  ָDMA program
 *  @param[in] iField
 *  @param[in] dpiDMA programϢַ
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CDmap6801::Build(int iField, TvDmaProgInfo* dpi)
{
	ULONG*	dmap = (ULONG*) m_virProg;
	ULONG	ip	 = iField*(VDMAP_UNIT/8);
	ULONG	phyWrite; 
	ULONG	lineSize, stride, lenWrite, lenWriten;
	ULONG	page, pageLen;
	ULONG	pageBegin, pageEnd;
	ULONG	jumpTo = ip*4 + VDMAP_UNIT/2;
	BOOL	isOdd = (iField%2)==1;

	if((dmap==NULL) || (dpi==NULL)) { return FALSE;	}
	jumpTo = (jumpTo >= (m_iDmapLen*4)) ? m_phyProg : (jumpTo+m_phyProg);

	dmap[ip++] = isOdd ? DMAP_SYNCE : DMAP_SYNCO;
	dmap[ip++] = 0;
	dmap[ip++] = DMAP_DUMMY;
	dmap[ip++] = 0;

	if((dpi->memBuf!=NULL) && (dpi->memBuf->pnodes!=NULL)) { 
		phyWrite = (ULONG)dpi->memBuf->pnodes[0].addr + dpi->offset;
		pageLen	 = dpi->memBuf->pnodes[0].len - dpi->offset;
		lineSize = dpi->lineSize;
		stride   = dpi->pitch;
		lenWrite = 0;
		lenWriten= 0;
		page	 = 0;
		pageBegin= phyWrite;
		pageEnd	 = pageBegin + pageLen;

		if(dpi->memBuf->len < (stride*dpi->height)) {
			//ڴ治
			return FALSE;
		}

		for(int i=0; i<dpi->height; i++) {
			if(dpi->memBuf->inode == 1) {
				//ڴ
				dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
				dmap[ip++] = phyWrite;
				phyWrite  += stride;
			}
			else {
				//ڴ棬ҲǷҳڴ
				if((pageEnd-phyWrite) >= lineSize) {
					dmap[ip++] = DMAP_WRITE | DMAP_SOL | lineSize;
					dmap[ip++] = phyWrite;
					phyWrite  += stride;
					
					//lineSizestride
					if(stride > lineSize) {
						while(phyWrite >= pageEnd) {
							page++;
							phyWrite -= pageEnd;
							if(page >= dpi->memBuf->inode) {
								if(i==dpi->height-1) break;
								return FALSE;
							}
							pageLen	  = dpi->memBuf->pnodes[page].len;
							if(pageLen==0)  {
								return FALSE;
							}
							pageBegin = (ULONG)dpi->memBuf->pnodes[page].addr;
							pageEnd	  = pageBegin + pageLen;
							phyWrite += pageBegin;
						}
					}
				}
				else {
					//һҳʣµڴ治дһе
					lenWriten = 0;
					if(pageEnd > phyWrite) {
						lenWrite   = pageEnd-phyWrite;
						dmap[ip++] = DMAP_WRITE | DMAP_SOL | lenWrite;
						dmap[ip++] = phyWrite;
						lenWriten  = lenWrite;
					}
					page++;
					if(page >= dpi->memBuf->inode) {
						if(i==dpi->height-1) break;
						return FALSE;
					}
					pageLen	  = dpi->memBuf->pnodes[page].len;
					if(pageLen==0)  {
						return FALSE;
					}
					pageBegin = (ULONG)dpi->memBuf->pnodes[page].addr;
					pageEnd	  = pageBegin + pageLen;
					phyWrite  = pageBegin;
					while(lenWriten<lineSize) {
						lenWrite = lineSize-lenWriten;
						if(lenWrite <= (pageEnd-phyWrite)) {
							dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
							dmap[ip++] = phyWrite;
							phyWrite  += lenWrite;
							lenWriten += lenWrite;
						}
						else {
							//һڴҳдһе
							lenWrite = pageEnd - phyWrite;
							dmap[ip++] = DMAP_WRITE | ((lenWriten==0) ? DMAP_SOL : (DMAP_MOL|DMAP_STARTINGBYTE(lenWriten))) | lenWrite;
							dmap[ip++] = phyWrite;
							lenWriten += lenWrite;

							//ҳ
							page++;
							if(page >= dpi->memBuf->inode) {
								if(i==dpi->height-1) break;
								return FALSE;
							}
							pageLen	  = dpi->memBuf->pnodes[page].len;
							if(pageLen==0)  {
								return FALSE;
							}
							pageBegin  = (ULONG)dpi->memBuf->pnodes[page].addr;
							pageEnd	   = pageBegin + pageLen;
							phyWrite   = pageBegin;
						}
					}
					
					//lineSizestride
					if(stride > lineSize) {
						phyWrite += stride - lineSize;
						while(phyWrite >= pageEnd) {
							page++;
							phyWrite -= pageEnd;
							if(page >= dpi->memBuf->inode) {
								if(i==dpi->height-1) break;
								return FALSE;
							}
							pageLen	  = dpi->memBuf->pnodes[page].len;
							if(pageLen==0)  {
								return FALSE;
							}
							pageBegin = (ULONG)dpi->memBuf->pnodes[page].addr;
							pageEnd	  = pageBegin + pageLen;
							phyWrite += pageBegin;
						}
					}
				}
			}
		}
	}
	dmap[ip++] = DMAP_JUMP|(isOdd ? DMAP_IRQ : 0);
	dmap[ip++] = jumpTo;
		
	return TRUE;
}

/** 
 ***************************************************************************************
 *  Թ/ȡԹָɼ
 *  @param[in] iField
 *  @param[in] bSkipǷ񲻲ɼ˳
 *
 *  @return ɹTRUE򷵻FALSE
 ***************************************************************************************
 */
BOOL CDmap6801::Skip(int iField, BOOL bSkip)
{
	ULONG*	dmap = (ULONG*) m_virProg;
	ULONG	ip	 = iField*(VDMAP_UNIT/8);
	ULONG	jumpTo = ip*4 + VDMAP_UNIT/2;
	BOOL	isOdd  = (iField%2)==1;
	
	if(dmap == NULL) { return FALSE; }
	jumpTo = (jumpTo >= (m_iDmapLen*4)) ? m_phyProg : (jumpTo+m_phyProg);
	
	dmap[ip++] = isOdd ? DMAP_SYNCE : DMAP_SYNCO;
	dmap[ip++] = 0;
	if(bSkip) {
		dmap[ip++] = DMAP_JUMP|(isOdd ? DMAP_IRQ : 0);
		dmap[ip++] = jumpTo;
	}
	else {
		dmap[ip++] = DMAP_DUMMY;
		dmap[ip++] = 0;
	}

	return TRUE;
}

/** 
 ***************************************************************************************
 *  ȡһβжϵBufferţԼbufferǷ(ûSkip)
 *  @param[in]  dmappоƬǰȡdma programַ
 *  @param[out] iEvenһδbufferţ-1ʾûд(skip)
 *  @param[out] iOdd һδbufferţ-1ʾûд(skip)
 *
 *  @return صǰڴbuffer
 ***************************************************************************************
 */
int CDmap6801::LastDmaBuffer(ULONG dmapp, long& iEven, long& iOdd) 
{
	PULONG dmap  = (PULONG)m_virProg;
	ULONG  ip    = 0;
	ULONG  offset= dmapp - m_phyProg;
	int    iIndex= offset/VDMAP_UNIT;
	int	   iLast = iIndex;

	if(iLast == 0) {
		iLast = m_iDmapLen*4 / VDMAP_UNIT;
	}
	iLast--;

	//_trace(TEXT("[Ϣ]dmapp=%08x, iIndex=%d\n")), dmapp, iLast);

	iEven = iOdd = -1;
	ip = (iLast * VDMAP_UNIT) / 4;
	try {
		if(dmap[ip+2] == DMAP_DUMMY) {
			iEven = iLast;
		}
		ip+= (VDMAP_UNIT/2) / 4;
		if(dmap[ip+2] == DMAP_DUMMY) {
			iOdd = iLast;
		}
	}
	catch(...) {
		_trace(TEXT("[쳣]dmapp=%08x, iIndex=%d\n"), dmapp, iLast);
		iIndex = 0;
	}

	return iIndex;
}