//////////////////////////////////////////////////////////////////////
// DmaProgram.h: interface for the Dma Program.
// Techwell, 2008.10.10
//////////////////////////////////////////////////////////////////////

#if !defined(__DMAPROGRAM_H_)
#define __DMAPROGRAM_H_

#define DMAP_LEN		(RAWV_NUM*VDMAP_UNIT)	//DMA program 
#define DMAP_LEN_SKIP	4096
#define VDMAP_UNIT		8192			//һvideo dmap
#define ADMAP_LEN		4096			//DMA program 

/* Type: Instruction
 * Purpose: enumerates all RISC commands
 */
typedef enum
{
	I_WRIT		= 0x09, 
	I_SKIP		= 0x02, 
	I_WRITEC	= 0x05, 
	I_JUMP		= 0x07, 
	I_SYNC		= 0x08, 
	I_WRITE123	= 0x01,  
	I_InLine	= 0x0A,  
	I_SKIP123	= 0x0C, 
	I_WRITE1S23 = 0x0B, 
	I_Reserved	= 0xFF
} TInstruction;

/* Type: SyncCode
 * Purpose: enumerates all sync codes coming out of decoder
*/
typedef enum
{
   SC_FM1 = 0x6, 
   SC_FM3 = 0xE, 
   SC_VRE = 0x4, 
   SC_VRO = 0xC, 
   SC_RESYNC = 1, 
   SC_Reserved = 0xFF
} TSyncCode;

typedef struct tagvDmaProgInfo{
	PMemBlock	memBuf;
	ULONG		offset;
	int			height;
	int			lineSize;
	int			width;
	int			pitch;
	UCHAR		irq;
	UCHAR		interlace;
}TvDmaProgInfo, *PvDmaProgInfo;

typedef struct tagaDmaProgInfo{
	PMemBlock	memBuf;
	UCHAR		irq;
	int			lines;
	int			line_w;
}TaDmaProgInfo, *PaDmaProgInfo;

class CDmap6801
{
public:
	CDmap6801();
	virtual ~CDmap6801();

	void SetProgAddr(ULONG virProg, ULONG phyProg, ULONG iDmapLen=0);
	virtual BOOL Build(TvDmaProgInfo* eDpi, TvDmaProgInfo* oDpi);
	virtual BOOL Build(int iField, TvDmaProgInfo* dpi);
	virtual BOOL Skip(int iField, BOOL bSkip);
	virtual int  LastDmaBuffer(ULONG dmapp, long& iEven, long& iOdd);

protected:

	ULONG m_virProg;	//dma programڴַ
	ULONG m_phyProg;	//dma programڴַ
	ULONG m_iDmapLen;
};

class CDmap6802
{
public:
	CDmap6802();
	virtual ~CDmap6802();

	void SetProgAddr(ULONG virProg, ULONG phyProg);
	virtual BOOL Build(TaDmaProgInfo* aDpi);

	//raw buffer
	BOOL AppendRawBuffer(CSample* pSample);

protected:
	ULONG m_virProg;	//dma programڴַ
	ULONG m_phyProg;	//dma programڴַ
	vector<CSample*> m_splList;
};

#endif // !defined(__DMAPROGRAM_H_)
