/**
 ***************************************************************************************
 *  Copyright (C) 2011, Intersil
 *  All rights reserved.
 *
 *  @file 		Gpio6869.cpp
 *  @brief		GPIOCGpio6869ʵ֣֧TW6869 GPIO
 *
 *  @version	0.0.0.1
 *  @author		liran
 *  @date		2011/05/09
 ***************************************************************************************
 */

#include "stdafx.h"
#include "Gpio6869.h"
#include "TW686X.h"
#include "Thread.h"

//-------------------------------------------------------------------
// GPIO
//-------------------------------------------------------------------
bool CGpio6869::EnableGpioOutput(bool enable, int gpio, int bits)
{
	CScopedLock autoLock(m_lock);

	ULONG regGpio[3] = { TW6864_R_GPIO_REG, TW6869_R_GPIO_REG_1, TW6869_R_GPIO_REG_2 };
	ULONG mask= 0;
	ULONG old = 0;
	int   nGpio = gpio % 16;
	int   nBits = 0;
	int	  nReg = gpio / 16;

	while(bits > 0)
	{
		nBits = min((16-nGpio), bits);
		mask  = ((1 << nBits) - 1) << (nGpio+16);
		old = m_pMem->ind(regGpio[nReg]);
		m_pMem->outd(regGpio[nReg], enable ? (old&~mask) : (old|mask));
		bits -= nBits;
		nGpio = 0;
		nReg++;
	}

	return true;
}

//-------------------------------------------------------------------
// GPIOƽ
//-------------------------------------------------------------------
bool CGpio6869::SetGpioBits(int gpio, int bits, ULONG* value)
{
	CScopedLock autoLock(m_lock);

	ULONG regGpio[3] = { TW6864_R_GPIO_REG, TW6869_R_GPIO_REG_1, TW6869_R_GPIO_REG_2 };
	ULONG mask= 0;
	ULONG old = 0;
	int   nGpio = gpio % 16;
	int   nBits = 0;
	int	  nReg  = gpio / 16;
	ULONG val	= *value;

	while(bits > 0)
	{
		nBits = min((16-nGpio), bits);
		mask= ((1 << nBits) - 1) << nGpio;
		old = m_pMem->ind(regGpio[nReg]) & ~mask;
		m_pMem->outd(regGpio[nReg], old|((val<<nGpio) & 0x0000FFFF));
		val = val >> nBits;
		bits -= nBits;
		nGpio = 0;
		nReg++;
	}

	return true;
}

//-------------------------------------------------------------------
// ȡGPIOƽ
//-------------------------------------------------------------------
bool CGpio6869::GetGpioBits(int gpio, int bits, ULONG* pValue)
{
	CScopedLock autoLock(m_lock);

	ULONG regGpio[3] = { TW6864_R_GPIO_REG, TW6869_R_GPIO_REG_1, TW6869_R_GPIO_REG_2 };
	ULONG mask= 0;
	ULONG old = 0;
	int   nGpio = gpio % 16;
	int   nBits = 0;
	int	  nReg  = gpio / 16;
	ULONG val	= 0;

	while(bits > 0)
	{
		nBits = min((16-nGpio), bits);
		mask= ((1 << nBits) - 1) << nGpio;
		old = (m_pMem->ind(regGpio[nReg]) & mask) >> nGpio;
		val|= (old << (nGpio + nReg*16));
		bits -= nBits;
		nGpio = 0;
		nReg++;
	}

	*pValue = val>>gpio;
	
	return true;
}