//////////////////////////////////////////////////////////////////////////
// I2C.h
// I2C
//////////////////////////////////////////////////////////////////////////

#ifndef __I2C_H_
#define __I2C_H_

#include "Thread.h"

#define I2C_WRITE         0         // write operation
#define I2C_READ          1         // read operation

#define PCI_FREQ          33000000L // frequency for PCI in hz
#define I2CDIV_MAX        15        // I2C maximum divider

#define TIMEOUT           500       // timeout value (500ms) to wait for an I2C operation
                                    // passing 3 values to I2C takes ~450ms

typedef enum { LevelLow, LevelHi } TLevel;
enum {
	I2CERR_OK        = 0,     // no error
	I2CERR_INIT      = 1,     // error in initialization
	I2CERR_MODE      = 2,     // invalid I2C mode (must be either HW or SW)
	I2CERR_NOACK     = 3,     // no ACK received from slave
	I2CERR_TIMEOUT   = 4,     // timeout error
	I2CERR_SCL       = 5,     // unable to change SCL line
	I2CERR_SDA       = 6,     // unable to change SDA line
	I2CERR_NODEVICE	 = 9,	  // the address is error -- for tuner
};

class CI2C
{
protected:
	// define which mode the I2C is selected
	enum I2CMode { I2CMode_None, I2CMode_HW, I2CMode_SW };

	ULONG		_cycle;       // software control of frequency
	int			_errNum;      // error number
	I2CMode		_i2c_mode;    // which mode the I2C is running in

public:
	CI2C();
	virtual ~CI2C(){};

	virtual bool I2CInitHWMode( long freq )					{_i2c_mode = I2CMode_HW; return true;}
	virtual bool I2CHWRead( BYTE address, BYTE *value )		{return true;}
	virtual bool I2CHWWrite2( BYTE address, BYTE value1 )	{return true;}
	virtual bool I2CHWWrite3( BYTE address, BYTE value1, BYTE value2 ) {return true;}
	
	virtual bool I2CSetSync( TLevel ){return true;}
	virtual int  I2CReadSync( void ) {return LevelLow;}

    int 		 I2CGetLastError(void){return _errNum;}

	virtual bool I2CInitSWMode( long freq );
    void 		 I2CSWBitDelay(void);    	// insert delay to simulate frequency
    virtual bool I2CSWWaitForACK(void); 	// wait for ACK from receiver using software
    virtual bool I2CSWStart(void);
    virtual bool I2CSWStop(void);
    virtual bool I2CSWRead(UCHAR *value);
    virtual bool I2CSWWrite(UCHAR value);
    virtual bool I2CSWSendACK(void);
    virtual bool I2CSWSendNACK(void);
    
	virtual bool I2CSWSetSCL(TLevel);
    virtual bool I2CSWSetSDA(TLevel);
    virtual int  I2CSWReadSCL(void) {return LevelLow;}
    virtual int  I2CSWReadSDA(void) {return LevelLow;}

	virtual bool i2cWriteData(int nLength, UCHAR* pData);
	virtual bool i2cReadData(int nWriteBytes, UCHAR *cWrite, int nReadBytes, UCHAR *pData);
	virtual bool i2cWriteDataNoStop(int nLength, UCHAR* pData);
	virtual bool i2cReadData(UCHAR addr, UCHAR offset, int nReadBytes, UCHAR* pData);

	virtual bool i2cTest(UCHAR I2CAdd);
	virtual bool i2cTest24C02();
	virtual bool i2cInit(PVOID pMem) { return true; }
	virtual CLock* i2cLock() { return &_lock; }

	static ULONG TickCount( void );
	static void	 WaitTick(ULONG tick);
	static ULONGLONG _tk_freq;
	CLockMutex _lock;
};

#endif