//////////////////////////////////////////////////////////////////////
// I2C680X.h: interface for the CI2C680X class.
// TW680X I2Cд
// created by liran, 2007.01.19
//////////////////////////////////////////////////////////////////////

#if !defined(__I2C680X_H_)
#define __I2C680X_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "I2C.h"

/////////////////////////////////////////////////////////////////////////////
// CLASS I2C
//
// Description:
//    This class encapsulates the register fields in the I2C register of the
//    Tw68. A complete set of functions are developed to manipulate all the
//    register fields in the I2C for the Tw68.
//
/////////////////////////////////////////////////////////////////////////////

class CI2C680X : public CI2C
{
protected:
	//**************************************************************************
	//	Structures
	//**************************************************************************

	struct I2C_SBUSD
	{
		unsigned int addr_rw:8;
		unsigned int byte1	:8;
		unsigned int byte2	:8;
		unsigned int rdb0	:8;
	} i2cShadow_SBUSD;

	struct I2C_SBUSC   
	{
		unsigned int sbmode	:1;
		unsigned int wren	:1;
		unsigned int noack	:1;
		unsigned int w3b	:1;
		unsigned int bursten:1; 
		unsigned int sync	:1;
		unsigned int scl	:1;
		unsigned int sda	:1;
		unsigned int sbclk	:8;
		unsigned int wdlen	:3;
		unsigned int res19	:1;
		unsigned int rdlen	:3;
		unsigned int res23	:1;
		unsigned int sbrw	:1;
		unsigned int sbdev	:7;
	};
	
	union I2C_SHADOW
	{
		struct I2C_SBUSC SBUSC;
		ULONG value;
	} i2cShadow_SBUSC;

	CMemRange* m_pMem;

public:
	// constructor and destructor
	CI2C680X(CMemRange* pMem) : m_pMem(pMem) {};
	virtual ~CI2C680X() {};

	virtual bool i2cInit(PVOID pMem) { m_pMem=(CMemRange*)pMem; return true; }
	virtual bool I2CInitHWMode( long freq );
	virtual bool I2CHWRead( BYTE address, BYTE *value );
	virtual bool I2CHWWrite2( BYTE address, BYTE value1 );
	virtual bool I2CHWWrite3( BYTE address, BYTE value1, BYTE value2 );
	virtual int  I2CReadSync( void );
	
	virtual bool I2CInitSWMode( long freq );
	virtual bool I2CSWSetSCL( TLevel );
	virtual int  I2CSWReadSCL( void );
	virtual bool I2CSWSetSDA( TLevel );
	virtual int  I2CSWReadSDA( void );

private:
	bool I2CHWWaitUntilDone( int );   // wait until I2C completes operation
	bool I2CHWIsDone( void );         // check interrupt bit for operation done
	bool I2CHWReceivedACK( void );    // check interrupt bit for received ACK
};

#endif // !defined(__I2C680X_H_)
