// I2C6864.cpp: implementation of the CI2C6864 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "I2C6864.h"
#include "TW686X.h"

#define TW6864_I2C_SCLK_BIT			BIT(0)
#define TW6864_I2C_SDA_DATA_BIT		BIT(1)
#define TW6864_I2C_SCLK_CTRL_BIT	BIT(16)  //1:Input,0=Output
#define TW6864_I2C_SDA_CTRL_BIT		BIT(17)  //1:Input,0=Output

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

bool CI2C6864::i2cInit(PVOID pMem)
{ 
	m_pMem=(CMemRange*)pMem; 
	if(m_pMem == NULL) { return false; }

	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word &= ~TW6864_I2C_SCLK_CTRL_BIT;
	control_word &= ~TW6864_I2C_SDA_CTRL_BIT;  
	control_word |= TW6864_I2C_SCLK_BIT;
	control_word |= TW6864_I2C_SDA_DATA_BIT;
	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return true; 
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C6864::I2CSWSetSCL( Level scl )
//  Purpose: Set software SCL value
//  Input:   Level scl - Hi releases the SCL output; Low forces the SCL output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C6864::I2CSWSetSCL( TLevel scl )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word &= ~TW6864_I2C_SCLK_CTRL_BIT; 
	if(scl == LevelHi) {
		control_word |= TW6864_I2C_SCLK_BIT;
	}
	else {
		control_word &= ~TW6864_I2C_SCLK_BIT;
	}

	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return CI2C::I2CSWSetSCL(scl);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C6864::I2CSWSetSDA( Level sda )
//  Purpose: Set software SDA value
//  Input:   Level sda - Hi releases the SDA output; Low forces the SDA output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C6864::I2CSWSetSDA( TLevel sda )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word &= ~TW6864_I2C_SDA_CTRL_BIT; 
	if(sda == LevelHi) {
		control_word |= TW6864_I2C_SDA_DATA_BIT;	
	}
	else {
		control_word &= ~TW6864_I2C_SDA_DATA_BIT;
	}

	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return CI2C::I2CSWSetSDA(sda);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C6864::I2CSWReadSCL( void )
//  Purpose: Read software SCL value
//  Input:   None
//  Output:  None
//  Return:  SCL value
/////////////////////////////////////////////////////////////////////////////
int CI2C6864::I2CSWReadSCL( void )
{
	//control_word |= TW6864_I2C_SCLK_CTRL_BIT; 
	//m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return (m_pMem->ind(TW6864_R_GPIO_REG) & TW6864_I2C_SCLK_BIT);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C6864::I2CSWReadSDA( void )
//  Purpose: Read software SDA value
//  Input:   None
//  Output:  None
//  Return:  SDA value
/////////////////////////////////////////////////////////////////////////////
int CI2C6864::I2CSWReadSDA( void )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word |= TW6864_I2C_SDA_CTRL_BIT; 
	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return ((m_pMem->ind(TW6864_R_GPIO_REG) & TW6864_I2C_SDA_DATA_BIT)>>1);
}