//////////////////////////////////////////////////////////////////////
// I2CuGpio.h: interface for the CI2CuGpio class.
// ʹGPIOʵֵI2C
//////////////////////////////////////////////////////////////////////

#if !defined(__I2CUGPIO_H_)
#define __I2CUGPIO_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Gpio.h"
#include "I2C.h"

class CI2CuGpio : public CI2C
{
protected:
	UCHAR  _scl_gpio;
	UCHAR  _sda_gpio;
	CGpio* _gpio;

public:
	CI2CuGpio() : _scl_gpio(0), _sda_gpio(0), _gpio(NULL) {};
	virtual ~CI2CuGpio() {};

	bool		 Init(CGpio* gpio, UCHAR scl_gpio, UCHAR sda_gpio);
	virtual bool I2CInitHWMode( long freq ) {return false;}
	
	virtual bool I2CSWSetSCL(TLevel);
    virtual bool I2CSWSetSDA(TLevel);
    virtual int  I2CSWReadSCL(void) ;
    virtual int  I2CSWReadSDA(void) ;
};

class CI2CuGpio680X : public CI2CuGpio
{
public:
	CI2CuGpio680X(){};
	virtual ~CI2CuGpio680X() {};

	virtual bool i2cReadData(UCHAR I2CAdd, UCHAR In1, UCHAR In2, UCHAR nLength, UCHAR *pData);
	virtual bool i2cTest24C02();
};

typedef CI2CuGpio680X CI2CuGpio713x;
typedef CI2CuGpio CI2CuGpio686x;

inline bool CI2CuGpio::Init(CGpio* gpio, UCHAR scl_gpio, UCHAR sda_gpio)
{
	_scl_gpio = scl_gpio;
	_sda_gpio = sda_gpio;
	_gpio	  = gpio;
	_gpio->EnableGpioOutput(true, _scl_gpio, 1);
	return true;
}

inline bool CI2CuGpio::I2CSWSetSCL(TLevel level) 
{
	_gpio->SetGpioBits(_scl_gpio, 1, (PULONG)&level);

	return CI2C::I2CSWSetSCL(level);
}

inline bool CI2CuGpio::I2CSWSetSDA(TLevel level) 
{
	_gpio->EnableGpioOutput(true, _sda_gpio, 1);
	_gpio->SetGpioBits(_sda_gpio, 1, (PULONG)&level);

	return CI2C::I2CSWSetSDA(level);
}

inline int CI2CuGpio::I2CSWReadSCL(void)
{
	ULONG val = 0;

	_gpio->GetGpioBits(_scl_gpio, 1, &val);

	return val;
}

inline int CI2CuGpio::I2CSWReadSDA(void)
{
	ULONG val = 0;

	_gpio->EnableGpioOutput(false, _sda_gpio, 1);
	_gpio->GetGpioBits(_sda_gpio, 1, &val);

	return val;
}

//--------------------------------------------------------
// 
// ˺ΪAT88SC0104Cר
//--------------------------------------------------------
inline bool CI2CuGpio680X::i2cReadData(UCHAR I2CAdd, UCHAR In1, UCHAR In2, UCHAR nLength, UCHAR *pData)
{	
	// TW6801ʹGPIOʵI2CͨʱһζȡֽʱCI2C::I2CSWSendACKܹ
	// ʹAT88SC0104CݣSDAøܱ֮ͣԭ
	// ֻһֽһֽڶȡ

	UCHAR data[3] = {I2CAdd, In1, In2};

	for(int i=0; i<nLength; i++) {
		data[2] = In2+i;
		if(!CI2C::i2cReadData(3, data, 1, pData+i)) { return false; }
	}

	return true;
}

//--------------------------------------------------------
// I2Cд24C02Ժ
//--------------------------------------------------------
inline bool CI2CuGpio680X::i2cTest24C02()
{
	int i;
	BYTE data[5] = {0x60, 0x08, 0xB0, 0x01};
	
	//
	// д
	//

	if(!I2CSWStart()) { return false;}

	//д24C02ַ
	if(!I2CSWWrite(0xA0)) { return false; }

	//дsub addr
	if(!I2CSWWrite(0xFC)) { return false; }

	//д
	for(i=0; i<4; i++) {
		if(!I2CSWWrite(data[i])) { return false; }
	}

	if(!I2CSWStop()) { return false; }
	Sleep(40);

	//
	// 
	//

	memset(data, 0, 5);
	for(i=0; i<4; i++) {
		if(!I2CSWStart()) { return false; }

		//д24C02ַ
		if(!I2CSWWrite(0xA0)) { return false; }

		//дsub addr
		if(!I2CSWWrite(0xFC+i)) { return false; }

		if(!I2CSWStart()) { return false; }
		//дaddr
		if(!I2CSWWrite(0xA1)) { return false; }

		//
		if(!I2CSWRead(data+i)) { 
			I2CSWSendNACK();
			return false;
		}
		else {
			I2CSWSendACK();
		}

		if(!I2CSWStop()) { return false; }
	}

	return true;
}

#endif // !defined(__I2CUGPIO_H_)
