/**
 ***************************************************************************************
 *  Copyright (C) 2008, Techwell
 *  All rights reserved.
 *
 *  @file 		PlgvSource.cpp
 *  @brief		豸ԴCPlgvSourceʵ
 *
 *  @version	0.1
 *  @author		liran
 *  @date		2008/10/10
 ***************************************************************************************
 */

#include "stdafx.h"
#include "PlgvSource.h"

#ifdef _DBG
#include "IniFile.h"
#endif

#define IS_WIDTH_CHANGED(flag) (flag&(VSRC_c_pixel_format|VSRC_c_width|VSRC_p_pixel_format|VSRC_p_width))

long TvSourceInfo::default_standard = vsStandardPAL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPlgvSource::CPlgvSource() :
m_dev(NULL)
{
	int iField = VF_Odd;

#ifdef _DBG
	CIniFile iniFile;
	TCHAR path[MAX_PATH];
	g_util.GetModulePath(g_hSelfDll, path);
	lstrcat(path, TEXT("TW68XX.ini"));
	iniFile.Create(path);
	iField = (iniFile.ReadInt(TEXT("SDK"), TEXT("field"), VF_Odd)==VF_Odd) ? VF_Odd : VF_Even;
#endif
	ZeroMemory(&m_info, sizeof(TvSourceInfo));
	m_info.default_standard = vsStandardPAL;
	m_info.field		 = iField;
	m_info.reset		 = DEVSET_devreset|DEVSET_property;
	m_info.frame_run	 = frame_invalid;
	m_info.capture.width = 352;
	m_info.capture.height= 288;
	m_info.capture.bit_count = 16;
	m_info.capture.pitch = m_info.capture.width*m_info.capture.bit_count/8;
	m_info.preview.width = 352;
	m_info.preview.height= 288;
	m_info.preview.bit_count = 16;
	m_info.preview.pitch = m_info.preview.width*m_info.preview.bit_count/8;
	m_info.capture.pixel_format = vpfYUV422Interspersed;
	m_info.preview.pixel_format = vpfYUV422Interspersed;
	m_info.brightness	 = 128;
	m_info.contrast		 = 38;
	m_info.saturation	 = 64;
	m_info.hue			 = 128;
	m_info.sharpness	 = 5;
	m_info.signal_check	 = NOSIGNAL_CHECK*10;
	m_info.vcom			 = 0;
	m_info.vleft		 = -1;
	m_info.vright		 = -1;
	m_info.hdelay		 = -1;
	m_info.vdelay		 = -1;
	m_info.signal		 = FALSE;
	m_info.preview.frame_rate	   = MAX_FRAMERATE(m_info.default_standard);
	m_info.preview.frame_rate_mask = g_util.DownFrameRate(m_info.preview.frame_rate, m_info.preview.frame_rate);
	m_info.capture.frame_rate	   = MAX_FRAMERATE(m_info.default_standard);
	m_info.capture.frame_rate_mask = g_util.DownFrameRate(MAX_FRAMERATE(m_info.default_standard), m_info.capture.frame_rate);
}

CPlgvSource::~CPlgvSource()
{

}

inline long CPlgvSource::BitCount(int iFormat)
{
	switch(iFormat) {
	case vpfRGB15Alpha:
	case vpfRGB16:
	case vpfYUV422Interspersed:
		return 16;
	case vpfYUV420Planar:
	case vpfYUV411Planar:
	case vpfYUV411Interspersed:
		return 12;
	case vpfRGB24:
	case vpfRGB24Alpha:
		return 24;
	}

	return 0;
}

//---------------------------------------------------------------------
// 
//---------------------------------------------------------------------
BOOL CPlgvSource::SetPluginInfo(PPluginInfo plgInfo, ULONGLONG flag)
{
	PvSourceInfo info = (PvSourceInfo)plgInfo;
	
	if(info == NULL) { return FALSE; }

	if((flag & VSRC_standard) && (m_info.standard!=info->standard)) {
		m_info.reset |= DEVSET_standard;
		m_info.standard = info->standard;
		m_info.preview.frame_rate	   = MAX_FRAMERATE(info->standard);
		m_info.preview.frame_rate_mask = g_util.DownFrameRate(m_info.preview.frame_rate, m_info.preview.frame_rate);
		m_info.capture.frame_rate	   = MAX_FRAMERATE(info->standard);
		m_info.capture.frame_rate_mask = g_util.DownFrameRate(MAX_FRAMERATE(info->standard), m_info.capture.frame_rate);
	}

	//ʹ
	if((flag & VSRC_field) && (m_info.field!=info->field)) {
		m_info.reset |= DEVSET_field;
		m_info.field = info->field;
	}

	if((flag & VSRC_frame_run) && (m_info.frame_run != info->frame_run)) {
		m_info.reset |= DEVSET_framerun;
		m_info.frame_run = info->frame_run;
	}

	if((flag & VSRC_c_width) && (m_info.capture.width!=info->capture.width)) {
		m_info.reset |= DEVSET_capturesize;
		m_info.capture.width  = info->capture.width;
	}

	if((flag & VSRC_c_height) && (m_info.capture.height!=info->capture.height)) {
		m_info.reset |= DEVSET_capturesize;
		m_info.capture.height  = info->capture.height;
	}

	if((flag & VSRC_p_width) && (m_info.preview.width!=info->preview.width)) {
		m_info.reset |= DEVSET_previewsize;			
		m_info.preview.width = info->preview.width;
	}
	if((flag & VSRC_p_height) && (m_info.preview.height!=info->preview.height)) {
		m_info.reset |= DEVSET_previewsize;			
		m_info.preview.height = info->preview.height;
	}
	 
	if((flag & VSRC_p_pitch) && (m_info.preview.pitch!=info->preview.pitch)) {
		m_info.reset |= DEVSET_previewsize;			
		m_info.preview.pitch = info->preview.pitch;
	}

	if((flag & VSRC_vleft) && (m_info.vleft!=info->vleft)) {
		m_info.reset |= DEVSET_vleft;			
		m_info.vleft = info->vleft;
	}
	
	if((flag & VSRC_vright) && (m_info.vright!=info->vright)) {
		m_info.reset |= DEVSET_vright;			
		m_info.vright = info->vright;
	}
	
	if((flag & VSRC_hdelay) && (m_info.hdelay!=info->hdelay)) {
		m_info.reset |= DEVSET_hdelay;			
		m_info.hdelay = info->hdelay;
	}
	
	if((flag & VSRC_vdelay) && (m_info.vdelay!=info->vdelay)) {
		m_info.reset |= DEVSET_vdelay;			
		m_info.vdelay = info->vdelay;
	}
	
	if(flag & VSRC_property) {
		if(m_info.brightness!=info->brightness)	{ m_info.reset |= DEVSET_bright; }
		if(m_info.contrast!=info->contrast)		{ m_info.reset |= DEVSET_contrast; }
		if(m_info.saturation!=info->saturation)	{ m_info.reset |= DEVSET_saturation; }
		if(m_info.hue!=info->hue)				{ m_info.reset |= DEVSET_hue; }
		if(m_info.sharpness!=info->sharpness)	{ m_info.reset |= DEVSET_sharpness; }
		m_info.brightness = info->brightness;
		m_info.contrast	  = info->contrast;
		m_info.saturation = info->saturation;
		m_info.hue		  = info->hue;
		m_info.sharpness  = info->sharpness;
	}
	
	if(flag & VSRC_c_frame_rate) {
		PDeviceCapability devCaps = m_dev->DeviceCaps();
		int maxFrameRate = MAX_FRAMERATE(m_info.standard);
		if((devCaps->caps&CAPS_SWITCH) && (devCaps->ddesc.iSwitch>1)) {
			maxFrameRate /= 2;
		}
		m_info.capture.frame_rate = min(info->capture.frame_rate, maxFrameRate);
		m_info.capture.frame_rate_mask = g_util.DownFrameRate(maxFrameRate, m_info.capture.frame_rate);
		m_info.frame_rate_cnt  = 0;
		m_info.reset |= DEVSET_framerate;	
	}

	if(flag & VSRC_p_frame_rate) {
		PDeviceCapability devCaps = m_dev->DeviceCaps();
		int maxFrameRate = MAX_FRAMERATE(m_info.standard);
		if((devCaps->caps&CAPS_SWITCH) && (devCaps->ddesc.iSwitch>1)) {
			maxFrameRate /= 2;
		}
		m_info.preview.frame_rate = min(info->preview.frame_rate, maxFrameRate);
		m_info.preview.frame_rate_mask = g_util.DownFrameRate(maxFrameRate, m_info.preview.frame_rate);
		m_info.frame_rate_cnt  = 0;
		m_info.reset |= DEVSET_framerate;	
	}

	if(flag & VSRC_auto_bright)		m_info.auto_bright = info->auto_bright; 
	if(flag & VSRC_c_pitch)			m_info.capture.pitch	  = info->capture.pitch;
	if(flag & VSRC_default_standard)m_info.default_standard	  = info->default_standard;
	if(flag & VSRC_signal)			m_info.signal			  = info->signal;
	if(flag & VSRC_signal_count)	m_info.signal_count		  = info->signal_count;
	if(flag & VSRC_p_rate_mask)		m_info.preview.frame_rate_mask=info->preview.frame_rate_mask;
	if(flag & VSRC_rate_cnt)		m_info.frame_rate_cnt	  =info->frame_rate_cnt;
	if(flag & VSRC_c_rate_mask)		m_info.capture.frame_rate_mask=info->capture.frame_rate_mask;
	if(flag & VSRC_reset)			m_info.reset = info->reset;
	if(flag & PLGSET_state)			m_info.state = info->state;
	if(flag & VSRC_c_pixel_format)	{
		m_info.capture.pixel_format=info->capture.pixel_format;
		m_info.capture.bit_count = BitCount(m_info.capture.pixel_format);
	}
	if(flag & VSRC_p_pixel_format)	{
		m_info.preview.pixel_format=info->preview.pixel_format;
		m_info.preview.bit_count = BitCount(m_info.preview.pixel_format);
	}
	
	if(IS_WIDTH_CHANGED(flag)) {
		if(!(flag & VSRC_c_pitch)) {
			m_info.capture.pitch = m_info.capture.width * m_info.capture.bit_count / 8;
		}
		if(!(flag & VSRC_p_pitch)) {
			m_info.preview.pitch = m_info.preview.width * m_info.preview.bit_count / 8;
		}
	}

	if(flag & VSRC_vcom) {
		m_info.vcom = info->vcom;
		m_info.reset |= DEVSET_vcom;			
	}
	if(m_info.vcom > 0) {
		if(!(flag & VSRC_c_pitch)) {
			m_info.capture.pitch = m_info.capture.width * m_info.capture.bit_count / 8;
			switch(m_info.vcom) {
			case 1:			// 1/2
				m_info.capture.pitch /= 2;
				break;
			case 2:			// 3/8
				m_info.capture.pitch = m_info.capture.pitch*3/8;
				break;
			case 3:			// 1/4
				m_info.capture.pitch /= 4;
				break;
			}
		}
		if(!(flag & VSRC_p_pitch)) {
			m_info.preview.pitch = m_info.preview.width * m_info.preview.bit_count / 8;
			switch(m_info.vcom) {
			case 1:			// 1/2
				m_info.preview.pitch /= 2;
				break;
			case 2:			// 3/8
				m_info.preview.pitch = m_info.preview.pitch*3/8;
				break;
			case 3:			// 1/4
				m_info.preview.pitch /= 4;
				break;
			}
		}
	}

	if(flag & VSRC_optimize_cif) {
		m_info.reset |= DEVSET_optimize_cif;
		m_info.optimize_cif = info->optimize_cif;
	}
	
	return TRUE;
}

//---------------------------------------------------------------------
// ݴ
//---------------------------------------------------------------------
BOOL CPlgvSource::Transmit(CSample* pSampleIn, CSample* pSampleOut)
{
	CChannel* chl = pSampleIn->GetChannel();
	ULONG	  flag=0, flagNext=0;
	
	if(m_info.signal) {
		flag = chl->IsEnabled(PLG_RunVideoDevice);
		if(!flag) {
			//ʱͨѾҪ
			pSampleIn->ClearFlag(PLG_vsource);
			return TRUE;
		}
	}
	
	if(pSampleIn->GetType() == pktVideo) {
		pSampleIn->SetIndex(m_info.frame_count);
		pSampleIn->SetDataSize(m_info.capture.pitch*m_info.capture.height);
		if(m_info.auto_bright) {
			//ÿӽһȼ
			if((pSampleIn->TimeStamp()-m_tkAutoBright) >= 10000000) {
				AutoBright(pSampleIn->DataPtr(), m_info.capture.width, m_info.capture.height);
			}
		}
	}
	else {
		pSampleIn->SetIndex(m_info.frame_count);
		pSampleIn->SetDataSize(m_info.preview.pitch*m_info.preview.height);
		if(m_info.auto_bright) {
			//ÿӽһȼ
			if((pSampleIn->TimeStamp()-m_tkAutoBright) >= 10000000) {
				AutoBright(pSampleIn->DataPtr(), m_info.preview.width, m_info.preview.height);
			}
		}
	}
	m_info.frame_count++;
	
	//֡뵽ӦĶ
	flagNext |= chl->IsEnabled(PLG_RunVideoDevice);

	if(flagNext) {
		flagNext = PLG_vrawcbk | ((m_info.vcom>0) ? PLG_vdecom : 0);
		pSampleIn->SetFlag(flagNext);
		g_cbkList.push_back(pSampleIn);
	}
	
#if(defined(_DBG))
	if(g_dbgIsPrintFrameRate) {
		__ccount(_cParam(&g_dbgFrameRate[chl->ChannelId()]) TEXT("ͨ%02d -- "), chl->ChannelId());
	}
#endif

	return TRUE;
}

/** 
 ***************************************************************************************
 *  Ƶȼ⣬ÿӽһ
 *  @param[in] yuy2YUY2ʽ֡
 *
 *  @return ɹTRUEFALSE
 ***************************************************************************************
 */
BOOL CPlgvSource::AutoBright(PBYTE yuy2, int width, int height)
{
	ULONG ySum = 0;		//ֵ֮,һULONG㹻720*576Ϊ255صYֵ֮

	for(long h=0; h<height; h++) {
		for(long w=0; w<width; w++) {
			ySum += *yuy2;
			yuy2 += 2;
		}
	}

	ySum /= width*height;
	if(ySum > 0xA0) {
		if(m_info.brightness > B_MAX/5) {
			m_info.brightness = max(m_info.brightness-5, B_MIN);
			m_info.reset |= DEVSET_bright;
		}
		else if(m_info.contrast > C_MAX/5) {
			m_info.contrast = max(m_info.contrast-5, C_MIN);
			m_info.reset |= DEVSET_contrast;
		}
	}
	else if(ySum < 0x40) {
		if(m_info.brightness < B_MAX/2) {
			m_info.brightness = min(m_info.brightness+5, B_MAX);
			m_info.reset |= DEVSET_bright;
		}
		else if(m_info.contrast < C_MAX/2) {
			m_info.contrast = min(m_info.contrast+5, C_MAX);
			m_info.reset |= DEVSET_contrast;
		}
	}

	return TRUE;
}