/**
 ***************************************************************************************
 *  Copyright (C) 2008, Techwell
 *  All rights reserved.
 *
 *  @file 		PlgvSource.h
 *  @brief		豸ԴCPlgvSource
 *
 *  @version	0.1
 *  @author		liran
 *  @date		2007/03/06
 ***************************************************************************************
 */

#if !defined(__PLGVSOURCE_H_)
#define __PLGVSOURCE_H_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Plugin.h"
#include "DevBase.h"

enum {
	DEVSET_none			= 0x00000000,
	DEVSET_previewsize	= 0x00000001, 
	DEVSET_capturesize	= 0x00000002, 
	DEVSET_standard		= 0x00000004, 
	DEVSET_framerate	= 0x00000008,
	DEVSET_framerun		= 0x00000010,
	DEVSET_field		= 0x00000020,
	DEVSET_vcom			= 0x00000040,
	DEVSET_vleft		= 0x00000080,
	DEVSET_vright		= 0x00000100,
	DEVSET_hdelay		= 0x00000200,
	DEVSET_vdelay		= 0x00000400,
	DEVSET_optimize_cif = 0x00000800,

	DEVSET_bright		= 0x00010000,
	DEVSET_contrast		= 0x00020000,
	DEVSET_saturation	= 0x00040000,
	DEVSET_hue			= 0x00080000,
	DEVSET_sharpness	= 0x00100000,

	DEVSET_devreset		= (DEVSET_previewsize|DEVSET_capturesize|DEVSET_standard|
						   DEVSET_framerun|DEVSET_field|DEVSET_vcom|DEVSET_optimize_cif),
	DEVSET_property		= (DEVSET_bright|DEVSET_contrast|DEVSET_saturation|DEVSET_hue|
						   DEVSET_sharpness|DEVSET_hdelay|DEVSET_vdelay|DEVSET_vleft|DEVSET_vright),
};

#define VSRC_default_standard	0x00000002
#define VSRC_reset				0x00000004
#define VSRC_standard			0x00000008
#define VSRC_signal				0x00000010
#define VSRC_brightness			0x00000020
#define VSRC_contrast			0x00000040
#define VSRC_saturation			0x00000080
#define VSRC_hue				0x00000100
#define VSRC_sharpness			0x00000200
#define VSRC_frame_run			0x00000400
#define VSRC_rate_cnt			0x00000800
#define VSRC_signal_count		0x00001000
#define VSRC_p_pixel_format		0x00002000
#define VSRC_p_width			0x00004000
#define VSRC_p_height			0x00008000
#define VSRC_p_pitch			0x00010000
#define VSRC_p_frame_rate		0x00020000
#define VSRC_p_rate_mask		0x00040000
#define VSRC_c_pixel_format		0x00080000
#define VSRC_c_width			0x00100000
#define VSRC_c_height			0x00200000
#define VSRC_c_pitch			0x00400000
#define VSRC_c_frame_rate		0x00800000
#define VSRC_c_rate_mask		0x01000000
#define VSRC_auto_bright		0x02000000
#define VSRC_field				0x04000000 //ʹ
#define VSRC_vcom				0x08000000
#define VSRC_hdelay				0x10000000
#define VSRC_vdelay				0x20000000
#define VSRC_vleft				0x40000000
#define VSRC_vright				0x80000000
#define VSRC_optimize_cif		0x0000000100000000

#define VSRC_property			(VSRC_brightness|VSRC_contrast|VSRC_saturation|VSRC_hue|VSRC_sharpness)

struct tagvSourceInfo : public TPluginInfo {
	static long default_standard;
	long		reset;			//豸ǷҪ
	long		standard;		//Ƶʽ
	long		signal;			//ǷƵź
	long		brightness;		//
	long		contrast;		//Աȶ
	long		saturation;		//Ͷ
	long		hue;			//ɫ
	long		sharpness;		//
	long		frame_run;		//ڲɼ֡
	long		field;			//ɼĳͣVF_Odd -- odd VF_Even -- even //ʹ
	long		frame_rate_cnt; //֡ʿƼ
	ULONG		frame_count;
	struct {
		long	pixel_format;	//Ƶɫʿռʽ
		long	width;			//
		long	height;			//߶
		long	pitch;
		long	frame_rate;		//֡
		ULONG	frame_rate_mask;//֡ʿ
		long	bit_count;
	}preview;
	struct {
		long	pixel_format;	//Ƶɫʿռʽ
		long	width;			//
		long	height;			//߶
		long	pitch;
		long	frame_rate;		//֡
		ULONG	frame_rate_mask;//֡ʿ
		long	bit_count;
	}capture;
	long		auto_bright;	//ƵԶ

	//ҪCDev610Xʵʱлźż
	long		signal_count;	//Ƶźż
	long		signal_check;	//ҪƵźż

	long		vcom;
	int			vleft;			//for tw686x only
	int			vright;			//for tw686x only
	int			hdelay;
	int			vdelay;
	BOOL		optimize_cif;	//for up to 16 CIF capture
};
typedef struct tagvSourceInfo TvSourceInfo, *PvSourceInfo;

class CPlgvSource : public CPlugin  
{
public:
	CPlgvSource();
	virtual ~CPlgvSource();

	void Init(CDevBase* dev){m_dev=dev;}
	CDevBase* Device(){return m_dev;}
	
	virtual ULONG		PluginId(){return PLG_vsource;}
	virtual PPluginInfo GetPluginInfo(){
		if(m_info.standard==0) m_info.standard=m_info.default_standard;
		return &m_info;
	}
	virtual BOOL		SetPluginInfo(PPluginInfo plgInfo, ULONGLONG flag);
	virtual BOOL		Transmit(CSample* pSampleIn, CSample* pSampleOut);
	virtual void		Uninit(){};

protected:
	BOOL AutoBright(PBYTE yuy2, int width, int height);
	long BitCount(int iFormat);

protected:
	CDevBase*	 m_dev;
	TvSourceInfo m_info;
	ULONGLONG	 m_tkAutoBright;
};

#endif // !defined(__VSOURCE_H_)
