//-----------------------------------------------
//Register.cpp
//-----------------------------------------------

#include "StdAfx.h"
#include "MemRange.h"
#include "Register.h"

CRegister::CRegister(CMemRange *pMemRange,ULONG BaseAddr) :
m_pMem(pMemRange),
m_RegAdd(BaseAddr)
{
}

CRegister::~CRegister()
{
}

VOID
CRegister::RegBITSet(ULONG nValue, ULONG npos)
{
	ULONG dw;

	dw= m_pMem->ind(m_RegAdd);
	m_pMem->outd( m_RegAdd, nValue<<npos | (dw&~(1<<npos)));
}

VOID 
CRegister::RegBITGet(ULONG &nValue, ULONG npos)
{
	ULONG dw;

	dw= m_pMem->ind(m_RegAdd);
	dw= (dw>>npos) &0x1;
	nValue= dw;
}

VOID
CRegister::RegBYTESet(ULONG nValue, ULONG npos)
{
	ULONG dw;

	dw= m_pMem->ind(m_RegAdd);
	m_pMem->outd( m_RegAdd, nValue<<npos | dw);
}

VOID 
CRegister::RegBYTEGet(ULONG &nValue, ULONG npos)
{
	ULONG dw;

	dw= m_pMem->ind(m_RegAdd);
	dw= (dw>>npos) &0xFF;
	nValue= dw;
}

VOID
CRegister::RegSet(ULONG nValue, ULONG nWidth, ULONG npos)
{
	ULONG dw;
	ULONG dwk;
	dwk= nWidth;

	dw= m_pMem->ind(m_RegAdd);
	m_pMem->outd( m_RegAdd, nValue <<npos | (dw &(~(((1<<nWidth)-1)<<npos))));
}

VOID
CRegister::RegGet(ULONG &nValue, ULONG npos)
{
	ULONG dw;

	dw= m_pMem->ind(m_RegAdd);
	nValue= dw;
}

VOID
CRegister::RegBYTEAndOrSet(UCHAR msk, UCHAR value)
{
	m_pMem->outb(m_RegAdd, (m_pMem->inb(m_RegAdd) & ~msk) | (value & msk));
}

VOID
CRegister::RegULONGAndOrSet(ULONG msk, ULONG value)
{
	m_pMem->outd(m_RegAdd, (m_pMem->ind(m_RegAdd) & ~msk) | (value & msk));
}


CCompoReg::CCompoReg(CMemRange* pMem, ULONG addrLSB, UCHAR lenLSB,
							 ULONG addrMSB, UCHAR posMSB, UCHAR lenMSB) :
m_pMem(pMem),
m_addrLSB(addrLSB), m_lenLSB(lenLSB),
m_addrMSB(addrMSB), m_posMSB(posMSB), m_lenMSB(lenMSB)
{
}

CCompoReg::operator ULONG()
{
	ULONG LSB = m_pMem->ind(m_addrLSB) & ((1<<m_lenLSB) - 1);
	ULONG MSB = (m_pMem->ind(m_addrMSB) >> m_posMSB) & ((1<<m_lenMSB) - 1);
	return (LSB | (MSB<<m_lenLSB));
}

ULONG CCompoReg::operator =(ULONG value)
{
	ULONG LSB = m_pMem->ind(m_addrLSB);
	ULONG MSB = m_pMem->ind(m_addrMSB);

	LSB &= ~((1<<m_lenLSB) - 1);
	MSB &= ~(((1<<m_lenMSB) - 1) << m_posMSB);

	LSB |= (value & ((1<<m_lenLSB) - 1));
	MSB |= ((value>>m_lenLSB) & ((1<<m_lenMSB)-1)) << m_posMSB;

	m_pMem->outd(m_addrLSB, LSB);
	m_pMem->outd(m_addrMSB, MSB);

	return value;
}
