// SAA7113.cpp: implementation of the CSAA7113 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SAA7113.h"

#define SAA7113_I2CW	0x4A	//0x48
#define SAA7113_I2CR	0x4B	//0x49

//SAA7113豸ȵֵķΧ
#define B_MAX_7113		255
#define B_MIN_7113		0
#define C_MAX_7113		127
#define C_MIN_7113		0
#define S_MAX_7113		127
#define S_MIN_7113		0
#define H_MAX_7113		127
#define H_MIN_7113		-128

#define SAA7113_R_CHIPVERSION	0x00
#define SAA7113_R_HDELAY		0x01
#define SAA7113_R_MUX			0x02	//AICO1
#define SAA7113_R_AICO2			0x03
#define SAA7113_R_AICO3			0x04
#define SAA7113_R_AICO4			0x05
#define SAA7113_R_HSYNC_START	0x06
#define SAA7113_R_HSYNC_STOP	0x07
#define SAA7113_R_SYNC_CTRL		0x08
#define SAA7113_R_LUMA_CTRL		0x09
#define SAA7113_R_BRIGHT		0x0A
#define SAA7113_R_CONTRAST		0x0B
#define SAA7113_R_SATURATION	0x0C
#define SAA7113_R_HUE			0x0D
#define SAA7113_R_CHROM_CTRL	0x0E
#define SAA7113_R_CHROM_GAIN	0x0F
#define SAA7113_R_FORMAT_DELAY	0x10
#define SAA7113_R_OUTPUT_CTRL1	0x11
#define SAA7113_R_RTS_OUTPUT	0x12
#define SAA7113_R_OUTPUT_CTRL2	0x13
#define SAA7113_R_VGATE_START	0x15
#define SAA7113_R_VGATE_STOP	0x16
#define SAA7113_R_VGATE_MSBS	0x17
#define SAA7113_R_VSTATUS		0x1F
#define SAA7113_R_CLKSEL		0x40
#define SAA7113_R_FRAMING_CODE	0x58
#define SAA7113_R_HOFFSET		0x59
#define SAA7113_R_VOFFSET		0x5A
#define SAA7113_R_OFFSET_MSBS	0x5B
#define SAA7113_R_SDID			0x5E
#define SAA7113_R_SLICER_STAT	0x60
#define SAA7113_R_SLICER_STAT1	0x61
#define SAA7113_R_SLICER_STAT2	0x62

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSAA7113::CSAA7113() :
m_i2c(NULL)
{

}

CSAA7113::~CSAA7113()
{

}

void CSAA7113::InitDecoder()
{
	//outb(SAA7113_R_HDELAY,	0x08);
	outb(SAA7113_R_MUX,			0xF0);
	outb(SAA7113_R_AICO2,		0x33);
	outb(SAA7113_R_AICO3,		0x00);
	outb(SAA7113_R_AICO4,		0x00);
	outb(SAA7113_R_HSYNC_START,	0xE9);
	outb(SAA7113_R_HSYNC_STOP,	0x0D);
	outb(SAA7113_R_SYNC_CTRL,	0x98);	//bit[7,6]ʽ
	outb(SAA7113_R_LUMA_CTRL,	0x01);
	outb(SAA7113_R_BRIGHT,		0x60);
	outb(SAA7113_R_CONTRAST,	0x47);
	outb(SAA7113_R_SATURATION,	0x40);
	outb(SAA7113_R_HUE,			0x00);
	outb(SAA7113_R_CHROM_CTRL,	0x01);
	outb(SAA7113_R_CHROM_GAIN,	0x31);
	outb(SAA7113_R_FORMAT_DELAY,0x08);
	outb(SAA7113_R_OUTPUT_CTRL1,0x8C);
	outb(SAA7113_R_RTS_OUTPUT,	0xEA);
	outb(SAA7113_R_OUTPUT_CTRL2,0x0A);
	outb(SAA7113_R_VGATE_START,	0x00);
	outb(SAA7113_R_VGATE_STOP,	0x37);
	outb(SAA7113_R_VGATE_MSBS,	0x01);
	outb(SAA7113_R_CLKSEL,		0x02);
	outb(SAA7113_R_FRAMING_CODE,0x00);
	outb(SAA7113_R_HOFFSET,		0x54);
	outb(SAA7113_R_VOFFSET,		0x07);
	outb(SAA7113_R_OFFSET_MSBS,	0x83);
	outb(SAA7113_R_SDID,		0x00);
}

void CSAA7113::UninitDecoder()
{
	outb(SAA7113_R_RTS_OUTPUT,	0x01);
}

UCHAR  CSAA7113::inb(ULONG Offset)
{
	UCHAR data = 0;
	if(m_i2c==NULL) { return 0; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				if(m_i2c->I2CSWStart()) {
					//дaddr
					if(m_i2c->I2CSWWrite(SAA7113_I2CR)) {
						m_i2c->I2CSWRead(&data);
						m_i2c->I2CSWSendNACK();
					}
				}
			}
		}
	}
	
	m_i2c->I2CSWStop();	
	
	return data;
}

VOID   CSAA7113::outb(ULONG Offset, UCHAR data)
{
	if(m_i2c==NULL) { return; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				//д
				m_i2c->I2CSWWrite(data);
			}
		}
	}
	
	m_i2c->I2CSWStop();
}

USHORT CSAA7113::inw(ULONG Offset)
{
	USHORT data = 0;
	if(m_i2c==NULL) { return 0; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				if(m_i2c->I2CSWStart()) {
					//дaddr
					if(m_i2c->I2CSWWrite(SAA7113_I2CR)) {
						for(int i = 0; i < 2; i++) {
							if(!m_i2c->I2CSWRead(((PUCHAR)&data)+i)) {
								break;
							}
							if(i < 1) {
								m_i2c->I2CSWSendACK();
							}
						}
						m_i2c->I2CSWSendNACK();
					}
				}
			}
		}
	}

	m_i2c->I2CSWStop();	

	return data;
}

VOID   CSAA7113::outw(ULONG Offset, USHORT data)
{
	if(m_i2c==NULL) { return; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				//д
				for(int i=0; i<2; i++) {
					if(!m_i2c->I2CSWWrite(((PUCHAR)&data)[i])) { break; }
				}
			}
		}
	}
	
	m_i2c->I2CSWStop();
}

ULONG  CSAA7113::ind(ULONG Offset)
{
	ULONG data = 0;
	if(m_i2c==NULL) { return 0; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				if(m_i2c->I2CSWStart()) {
					//дaddr
					if(m_i2c->I2CSWWrite(SAA7113_I2CR)) {
						for(int i = 0; i < 4; i++) {
							if(!m_i2c->I2CSWRead(((PUCHAR)&data)+i)) {
								break;
							}
							if(i < 3) {
								m_i2c->I2CSWSendACK();
							}
						}
						m_i2c->I2CSWSendNACK();
					}
				}
			}
		}
	}
	
	m_i2c->I2CSWStop();	
	
	return data;
}

VOID   CSAA7113::outd(ULONG Offset, ULONG data)
{
	if(m_i2c==NULL) { return; }
	if(m_i2c->I2CSWStart()) {
		//дַ
		if(m_i2c->I2CSWWrite(SAA7113_I2CW)) {
			//дsub addr
			if(m_i2c->I2CSWWrite((UCHAR)Offset)) {
				//д
				for(int i=0; i<4; i++) {
					if(!m_i2c->I2CSWWrite(((PUCHAR)&data)[i])) { break; }
				}
			}
		}
	}
	
	m_i2c->I2CSWStop();
}


BOOL CSAA7113::GetSignal(int ch)
{
	BYTE val = inb(SAA7113_R_VSTATUS);

	return (val & 1);
}

void CSAA7113::SetMux(TVideoInput vInput, int ch) 
{
	outb(SAA7113_R_MUX, 0xF0 | (UCHAR)vInput);
}

void CSAA7113::SetStandard(int value, int ch)
{
	switch(value) {
	case vsStandardNone:
		value = 0x07;
		break;
	case vsStandardNTSC:
		value = 0;
		break;
	case vsStandardPAL:
		value = 1;
		break;
	case vsStandardSECAM:
		value = 2;
		break;
	default:
		value = 0x07;
		break;
	}
}

void CSAA7113::SetFormat(int value, int ch)
{
}

void CSAA7113::SetSize(int width, int height, int ch)
{
}

void CSAA7113::SetBright(int value, int ch)
{
	int val = 0;
	g_util.map(B_MIN, B_MAX, value, B_MIN_7113, B_MAX_7113, &val);
	outb(SAA7113_R_BRIGHT, val);
	//	_trace(TEXT("[  ] %03d -> %03d = %02X\n")), value, val, (BYTE)val);
}

void CSAA7113::SetContrast(int value, int ch)
{
	int val = 0;
	g_util.map(C_MIN, C_MAX, value, C_MIN_7113, C_MAX_7113, &val);
	outb(SAA7113_R_CONTRAST, val);
	//	_trace(TEXT("[Աȶ] %03d -> %03d = %02X\n")), value, val, (BYTE)val);
}

void CSAA7113::SetHue(int value, int ch)
{
	int val = 0;
	g_util.map(H_MIN, H_MAX, value, H_MIN_7113, H_MAX_7113, &val);
	outb(SAA7113_R_HUE, val);
	//	_trace(TEXT("[ɫ  ] %03d -> %03d = %02X\n")), value, val, (BYTE)val);
}

void CSAA7113::SetSaturation(int value, int ch)
{
	int val = 0;
	g_util.map(S_MIN, S_MAX, value, S_MIN_7113, S_MAX_7113, &val);
	outb(SAA7113_R_SATURATION, val);
	//	_trace(TEXT("[Ͷ] %03d -> %03d = %02X\n")), value, val, (BYTE)val);
}

void CSAA7113::SetSharpness(int value, int ch)
{
	//int val = 0;
	//g_util.map(SH_MIN, SH_MAX, value, SH_MIN_7113, SH_MAX_7113, &val);
	//	_trace(TEXT("[  ] %03d -> %03d = %02X\n")), value, val, (BYTE)val);
}

