/**
 ***************************************************************************************
 *  Copyright (C) 2009, Techwell
 *  All rights reserved.
 *
 *  @file 		Scaler6801.h
 *  @brief		TW6801оƬScalerʵ֣֧TW6802,TW6805
 *
 *  @version	0.0.0.1
 *  @author		liran
 *  @date		2009/03/31
 ***************************************************************************************
 */

#include "stdafx.h"
#include "TW680X.H"
#include "MemRange.h"
#include "Scaler6801.h"
#include "Register.h"

#define R_CROPHI_offset			0x00
#define R_VDELAYLO_offset		0x04
#define R_VACTIVELO_offset		0x08
#define R_HDELAYLO_offset		0x0C
#define R_HACTIVELO_offset		0x10
#define R_VSCALELO_offset		0x18
#define R_SCALEHI_offset		0x1C
#define R_HSCALELO_offset		0x20

#define RAddr(offset)	(m_baseR+offset)

//===========================================================================
// Tw68 CScaler6801 Class Implementation
//===========================================================================


/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////

CScaler6801::CScaler6801(CMemRange* pMem, TVideoField field ):
   // Since VDelay in hardware is reversed; i.e. odd reg is really even field
   // and vice versa, construct the opposite here
   m_pMem(pMem),
   m_videoFormat(vsStandardPAL),
   m_vScaleLo(0),m_hScaleLo(0),
   m_ScaleHi(0)
{
	if(field == VF_Even) {
		m_baseR = R_F2CROPHI;
		m_pMem->outb(R_F2CNTL, 1);
	}
	else if(field == VF_Odd) {
		m_baseR = R_CROPHI;
		m_pMem->outb(R_F2CNTL, 1);
	}
	else {
		m_baseR = R_CROPHI;
		m_pMem->outb(R_F2CNTL, 0);
	}
}


/////////////////////////////////////////////////////////////////////////////
// Destructor
/////////////////////////////////////////////////////////////////////////////
CScaler6801::~CScaler6801()
{
}

ULONG CScaler6801::RegVScaleLo()
{
	return RAddr(R_VSCALELO_offset);
}

ULONG CScaler6801::RegHScaleLo()
{
	return RAddr(R_HSCALELO_offset);
}

ULONG CScaler6801::RegScaleHi()
{
	return RAddr(R_SCALEHI_offset);
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::VideoFormatChanged( VideoFormat format )
// Purpose: Set which video format is using
// Input:   Video format -
//            Auto format:          VFormat_AutoDetect
//            NTSC (M):             VFormat_NTSC
//            PAL (B, D, G, H, I):  VFormat_PAL_BDGHI
//            PAL (M):              VFormat_PAL_M
//            PAL(N):               VFormat_PAL_N
//            SECAM:                VFormat_SECAM
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::VideoFormatChanged( TVideoStandard format )
{
	m_videoFormat = format;
}


/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::Scale( TRect & clientScr )
// Purpose: Perform scaling
// Input:   TRect & clientScr - rectangle to scale to
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::Scale( TRect & clientScr, int nHDelay, int nVDelay )
{
   // the order of functions calling here is important because some
   // calculations are based on previous results
   SetHActive( clientScr ); 
   SetVActive();
   SetVScale( clientScr );
   SetVDelay( nVDelay);
   SetHDelay( nHDelay);
   SetHScale( clientScr);
}

void CScaler6801::PreScale(TRect & clientScr)
{
	ULONG  hScale=0, vScale=0;
	
	if(clientScr.Width() == 352) { hScale = 0x1FF; }
	else if(clientScr.Width() > 704) {
		hScale = 0x100;
	}
	else { 
		hScale = (0x2C0 * 256 / clientScr.Width()); 
		hScale = max(hScale, 0x100);
	}

	if ( m_videoFormat == vsStandardNTSC ) {
		vScale = 240*256/clientScr.Height();
	}
	else {
		vScale = 288*256/clientScr.Height();
	}
	
	m_hScaleLo= hScale & 0xFF;
	m_ScaleHi&= 0xF0;
	m_ScaleHi|= ((hScale>>8) & 0x0F);

	m_vScaleLo= vScale & 0xFF;
	m_ScaleHi&= 0x0F;
	m_ScaleHi|= (((vScale>>8) & 0x0F) << 4);
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetNTSC320default()
// Purpose: Set default register setting of NTSC 320*240 for Tw9901
// Input:   None
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetNTSC320Default() 
{ 
	CCompoReg regHDelay(m_pMem, RAddr(R_HDELAYLO_offset), 8, RAddr(R_CROPHI_offset), 2, 2);
	CCompoReg regHActive(m_pMem, RAddr(R_HACTIVELO_offset), 8, RAddr(R_CROPHI_offset), 0, 2);
	CCompoReg regHScale(m_pMem, RAddr(R_HSCALELO_offset), 8, RAddr(R_SCALEHI_offset), 0, 4);
	CCompoReg regVDelay(m_pMem, RAddr(R_VDELAYLO_offset), 8, RAddr(R_CROPHI_offset), 6, 2);
	CCompoReg regVActive(m_pMem, RAddr(R_VACTIVELO_offset), 8, RAddr(R_CROPHI_offset), 4, 2);
	CCompoReg regVScale(m_pMem, RAddr(R_VSCALELO_offset), 8, RAddr(R_SCALEHI_offset), 4, 4);

	regHDelay  = 0x07f;
	regHActive = 0x2d0;
	regHScale  = 0x23a;

	regVDelay  = 0x015;
	regVActive = 0x0f0;
	regVScale  = 0x100;

	m_ScaleHi  = 0x00;
	m_vScaleLo = 0x00;
	m_ScaleHi |= (0x01<<4);
	m_hScaleLo = 0x3a;
	m_ScaleHi |= 0x02;
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetHActive( TRect & clientScr )
// Purpose: Set HActive register
// Input:   TRect & clientScr - rectangle to scale to
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetHActive( TRect & clientScr )
{
	CCompoReg regHActive(m_pMem, RAddr(R_HACTIVELO_offset), 8, RAddr(R_CROPHI_offset), 0, 2);
	if((clientScr.Width()<208) || (clientScr.Width()>704)) {
		regHActive = 0x2D0;
	}
	else {
		regHActive = 0x2C8;//0x2DA;//0x300;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetHDelay( void )
// Purpose: Set HDelay register
// Input:   None
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetHDelay( int nHDelay )
{
	CCompoReg regHDelay(m_pMem, RAddr(R_HDELAYLO_offset), 8, RAddr(R_CROPHI_offset), 2, 2);
	regHDelay = (BYTE)nHDelay;//0x04;//0x1C;//0x03C; //0x08
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetHScale( void )
// Purpose: Set HScale register
// Input:   None
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetHScale( TRect & clientScr )
{
	ULONG	  hScale = 0;
	CCompoReg regHScale(m_pMem, RAddr(R_HSCALELO_offset), 8, RAddr(R_SCALEHI_offset), 0, 4);

	if(clientScr.Width() > 704) {
		hScale = 0x100;
	}
	else {
		hScale = (0x2C0 * 256 / clientScr.Width());
		hScale = max(hScale, 0x100);
	}

	regHScale = hScale;
	m_hScaleLo= hScale & 0xFF;
	m_ScaleHi&= 0xF0;
	m_ScaleHi|= ((hScale>>8) & 0x0F);
}  

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetVScale( TRect & clientScr )
// Purpose: Set VScale register
// Input:   TRect & clientScr - rectangle to scale to
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetVScale( TRect & clientScr )
{
	ULONG	  vScale = 0;
	CCompoReg regVScale(m_pMem, RAddr(R_VSCALELO_offset), 8, RAddr(R_SCALEHI_offset), 4, 4);
	
	if ( m_videoFormat == vsStandardNTSC ) {
		vScale = 240*256/clientScr.Height();
	}
	else
	{
		//һֻҪregVScale >= 0x200ͼͻǶ˼Ƶ
		vScale = 288*256/clientScr.Height();
	}

	regVScale = vScale;
	m_vScaleLo= vScale & 0xFF;
	m_ScaleHi&= 0x0F;
	m_ScaleHi|= (((vScale>>8) & 0x0F) << 4);
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetVDelay( void )
// Purpose: Set VDelay register
// Input:   None
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetVDelay( int nVDelay )
{
	CCompoReg regVDelay(m_pMem, RAddr(R_VDELAYLO_offset), 8, RAddr(R_CROPHI_offset), 6, 2);

//     if ( m_videoFormat == vsStandardNTSC )
// 		regVDelay = 0x014;//0x013;    // NTSC
//     else
	regVDelay = (BYTE)nVDelay;//0x018;    // PAL/SECAM
}

/////////////////////////////////////////////////////////////////////////////
// Method:  void CScaler6801::SetVActive( void )
// Purpose: Set VActive register
// Input:   None
// Output:  None
// Return:  None
/////////////////////////////////////////////////////////////////////////////
void CScaler6801::SetVActive( void )
{
	CCompoReg regVActive(m_pMem, RAddr(R_VACTIVELO_offset), 8, RAddr(R_CROPHI_offset), 4, 2);

	if ( m_videoFormat == vsStandardNTSC )
		regVActive = 0x0f0;  
	else
        regVActive = 0x120;//0x1f0; //0x120
}
