/**
 ***************************************************************************************
 *  Copyright (C) 2009, Techwell
 *  All rights reserved.
 *
 *  @file 		Scaler6801.h
 *  @brief		TW6801оƬScaler֧TW6802,TW6805
 *
 *  @version	0.0.0.1
 *  @author		liran
 *  @date		2009/03/31
 ***************************************************************************************
 */

#ifndef __SCALER_H_
#define __SCALER_H_

#define HDROP       HANDLE

typedef POINT TPoint;

//
// class TSize
// ----- -----
//
class  TSize : public tagSIZE {
  public:
    // Constructors
    TSize() {}
    TSize(int dx, int dy) {cx = dx; cy = dy;}
    void Set( int dx, int dy ) { cx = dx; cy = dy; }
};


class  TRect : public tagRECT {
  public:
    // Constructors
    TRect() {}
    TRect( int _left, int _top, int _right, int _bottom );
    TRect( const TPoint& origin, const TSize& extent );
    TRect( const struct tagRECT &orgn );

    void Set( int _left, int _top, int _right, int _bottom );

    // Information/access functions(const and non-const)
    const TPoint& TopLeft() const {return *(TPoint*)&left;}
    int          Width() const {return right-left;}
    int          Height() const {return bottom-top;}
    const TSize  Size() const {return TSize(Width(), Height());}
    bool IsEmpty() const;
    bool IsNull() const;

};

//----------------------------------------------------------------------------
// Inlines
//----------------------------------------------------------------------------
inline void TRect::Set(int _left, int _top, int _right, int _bottom) {
  left = _left;
  top = _top;
  right = _right;
  bottom = _bottom;
}

inline TRect::TRect(int _left, int _top, int _right, int _bottom) {
  Set(_left, _top, _right, _bottom);
}
inline TRect::TRect(const TPoint& orgn, const TSize& extent) {
  Set(orgn.x, orgn.y, orgn.x+extent.cx, orgn.y+extent.cy);
}

inline TRect::TRect( const struct tagRECT &orgn )
{
   Set( orgn.left, orgn.top, orgn.right, orgn. bottom );
}

//
// Return true if the rectangle is empty.
//
inline bool TRect::IsEmpty() const
{
  return bool( left >= right || top >= bottom );
}

//
// Return true if all of the points on the rectangle is 0.
//
inline bool TRect::IsNull() const
{
  return bool( !left && !right && !top && !bottom );
}

/////////////////////////////////////////////////////////////////
//
// for pisces, instantiate as ...
//
// Scaler evenScaler( VF_Even );
// Scaler oddScaler( VF_Odd );
//
/////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// CLASS Scaler
//
// Description:
//    This class encapsulates the register fields in the scaler portion of
//    the Tw68.
//    A complete set of functions are developed to manipulate all the
//    register fields in the scaler registers for the Tw68.
//
// Methods:
//    See below
//
// Note:
//    For Tw68, instantiate as ...
//       Scaler evenScaler( VF_Even );
//       Scaler oddScaler( VF_Odd );
//
/////////////////////////////////////////////////////////////////////////////

class CScaler6801
{
public:
	CScaler6801(CMemRange* pMem, TVideoField );
	~CScaler6801();

	void  VideoFormatChanged( TVideoStandard );
	void  Scale( TRect & rc, int nHDelay, int nVDelay);
	void  PreScale( TRect & );
	void  SetNTSC320Default(); 
	ULONG RegVScaleLo();
	ULONG RegHScaleLo();
	ULONG RegScaleHi();
	ULONG VScaleLo() { return m_vScaleLo; }
	ULONG HScaleLo() { return m_hScaleLo; }
	ULONG ScaleHi()  { return m_ScaleHi; }

	virtual void SetHDelay( int nHDelay );
	virtual void SetVDelay( int nVDelay );

protected:
	CMemRange*	m_pMem;
	ULONG		m_baseR;
	ULONG		m_vScaleLo;
	ULONG		m_hScaleLo;
	ULONG		m_ScaleHi;
	TVideoStandard  m_videoFormat;   // video format

	// member functions to set scaling registers
	virtual void SetHActive( TRect & );
	virtual void SetHScale( TRect & );
	virtual void SetVActive( void );
	virtual void SetVScale( TRect & );
};

#endif //__SCALER_H_








