//////////////////////////////////////////////////////////////////////
// Thread.cpp: implementation of the CThread class.
// ߳
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <windows.h>
#include <process.h>
#include "Thread.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CThread::CThread() :
m_hThread(NULL),
m_bExit(TRUE)
{
	m_hEvt[EVT_EXIT] = CreateEvent(NULL, FALSE, FALSE, NULL);
	m_hEvt[EVT_DATA] = CreateEvent(NULL, FALSE, FALSE, NULL);
}

CThread::~CThread()
{
	Exit();
	SAFE_CLOSE_HANDLE(m_hEvt[EVT_DATA]);
	SAFE_CLOSE_HANDLE(m_hEvt[EVT_EXIT]);
}

//---------------------------------------------------
// ߳
//---------------------------------------------------
BOOL CThread::Create()
{
	unsigned int nThreadID;
	CScopedLock autoLock(&m_lockExit);

	m_bExit = FALSE;
	if(m_hThread || (m_hEvt[EVT_EXIT]==NULL)) {
		return FALSE;
	}

	//ע⣺ҪresetEventܳһ
	//߳ʱEVT_EXTȻЧֱ߳˳
	ResetEvent(m_hEvt[EVT_EXIT]);
	ResetEvent(m_hEvt[EVT_DATA]);
	m_hThread = (HANDLE)_beginthreadex( NULL,
										0,
										ThreadProc,
										this,
										0,
										&nThreadID);
	if(m_hThread==NULL) {
		return FALSE;
	}

	return TRUE;
}

//---------------------------------------------------
// ߳˳
//---------------------------------------------------
void CThread::Exit(ULONG maxWait)
{
	HANDLE hThread = NULL;
	{
		CScopedLock autoLock(&m_lockExit);
		hThread = m_hThread;
		m_bExit = TRUE;
		SetEvent(m_hEvt[EVT_EXIT]);
	}
	if(hThread) {
		if(WaitForSingleObject(hThread, maxWait) == WAIT_TIMEOUT) {
			TerminateThread(hThread, 1);
		}

		//ע⣺ҪresetEventܳһ
		//߳ʱEVT_EXTȻЧֱ߳˳
		CScopedLock autoLock(&m_lockExit);
		ResetEvent(m_hEvt[EVT_EXIT]);
		ResetEvent(m_hEvt[EVT_DATA]);
		
		CloseHandle(m_hThread);
		m_hThread = NULL;
	}
}

//---------------------------------------------------
// ̹߳
//---------------------------------------------------
UINT CThread::ThreadProc(LPVOID pParam)
{
	CThread* pSelf = (CThread*)pParam;

	{
		CScopedLock autoLock(&pSelf->m_lockExit);
		pSelf->OnCreate();
	}

	UINT ret = pSelf->DoWork();

	if(ret==0) {
		ret = pSelf->OnExit();
	}
	else {
		pSelf->OnExit();
	}

	{
		CScopedLock autoLock(&pSelf->m_lockExit);
		//pSelf->m_hThread = NULL; ﲻҪֵӦExitCloseHandle߳̾ɾй©
		pSelf->m_bExit = TRUE;
	}

	return ret;
}

//---------------------------------------------------
// ̹߳
//---------------------------------------------------
UINT CThread::DoWork()
{
	DWORD dwWait = 0;
	BOOL  bExit = FALSE;
	if(!PrepareRun()) {
		return 1;
	}

	while(!bExit) {
		dwWait = WaitForMultipleObjects(EVT_NUM, m_hEvt, FALSE, 30);
		bExit = IsExit();
		if(bExit || (dwWait == (WAIT_OBJECT_0+EVT_EXIT))) {
			break;
		}
		else if(dwWait == WAIT_TIMEOUT) {
//			continue;
		}

		if(Run() != 0) {
			return 2;
		}
	}

	return 0;
}

//---------------------------------------------------
// ߳̾幤׼
//---------------------------------------------------
BOOL CThread::PrepareRun()
{
	return TRUE;
}

//---------------------------------------------------
// ߳̾幤
//---------------------------------------------------
UINT CThread::Run()
{
	return 0;
}

//---------------------------------------------------
// ߳˳׼
//---------------------------------------------------
UINT CThread::OnExit()
{
	return 0;
}
