//////////////////////////////////////////////////////////////////////
// Thread.h: interface for the CThread class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__THREAD_H_)
#define __THREAD_H_

enum {EVT_DATA=0, EVT_EXIT, EVT_NUM};

//------------------
//߳:ʹٽ
//------------------
class CLock
{
public:
	CLock() : m_bAttach(FALSE) {};
	CLock(BOOL bAttach) : m_bAttach(bAttach) {};
	virtual ~CLock(){};

	virtual void  Lock() = 0;
	virtual BOOL  Lock(ULONG time_out) = 0;
	virtual void  Unlock() = 0;
	virtual PVOID Object() = 0;

protected:
	BOOL m_bAttach;
};

class CLockCS : public CLock
{
public:
	CLockCS(CRITICAL_SECTION cs) : CLock(TRUE), m_cs(cs) {}
	CLockCS() : CLock(FALSE) {InitializeCriticalSection(&m_cs);}
	virtual ~CLockCS(){if(!m_bAttach) DeleteCriticalSection(&m_cs);}

	virtual void  Lock(){EnterCriticalSection(&m_cs);}
	virtual BOOL  Lock(ULONG time_out){EnterCriticalSection(&m_cs);return TRUE;}
	virtual void  Unlock(){LeaveCriticalSection(&m_cs);}
	virtual PVOID Object(){return &m_cs;}

protected:
	CRITICAL_SECTION m_cs;
};

class CLockEvt : public CLock
{
public:
	CLockEvt(HANDLE evt) : CLock(TRUE), m_cs(evt) {}
	CLockEvt() : CLock(FALSE) {m_cs = CreateEvent(NULL, FALSE, TRUE, NULL);}
	virtual ~CLockEvt(){if(!m_bAttach) CloseHandle(m_cs);}

	virtual void Lock(){WaitForSingleObject(m_cs, INFINITE);}
	virtual BOOL Lock(ULONG time_out){return WaitForSingleObject(m_cs, time_out)==WAIT_OBJECT_0;}
	virtual void Unlock(){SetEvent(m_cs);}
	virtual PVOID Object(){return m_cs;}

protected:
	HANDLE m_cs;
};

class CLockMutex : public CLock
{
public:
	CLockMutex(HANDLE hMutex) : CLock(TRUE), m_cs(hMutex) {}
	CLockMutex() : CLock(FALSE) {m_cs = CreateMutex(NULL, FALSE, NULL);}
	virtual ~CLockMutex(){if(!m_bAttach) CloseHandle(m_cs);}
	
	virtual void Lock(){WaitForSingleObject(m_cs, INFINITE);}
	virtual BOOL Lock(ULONG time_out){return WaitForSingleObject(m_cs, time_out)==WAIT_OBJECT_0;}
	virtual void Unlock(){ReleaseMutex(m_cs);}
	virtual PVOID Object(){return m_cs;}
	
protected:
	HANDLE m_cs;
};

class CScopedLock
{
public:
	CScopedLock(CLock* lock) : m_lock(lock),m_status(TRUE) {if(m_lock) m_lock->Lock();}
	CScopedLock(CLock* lock, ULONG time_out) : m_lock(lock),m_status(FALSE) {if(m_lock) m_status = m_lock->Lock(time_out);}
	~CScopedLock(){if(m_status && m_lock) m_lock->Unlock();}
	BOOL Status(){return m_status;}

private:
	CLock* m_lock;
	BOOL   m_status;
};

//------------------
//߳
//------------------
class CThread  
{
public:
	CThread();
	virtual ~CThread();

	virtual BOOL Create();
	virtual void Exit(ULONG maxWait=INFINITE);
	virtual void Clear() {}
	HANDLE	DataEvent(){return m_hEvt[EVT_DATA];}

	static UINT CALLBACK ThreadProc(LPVOID pParam);

protected:
	HANDLE	m_hThread;			//߳̾
	HANDLE	m_hEvt[EVT_NUM];	//¼
	BOOL	m_bExit;

	virtual void OnCreate(){};
	virtual BOOL PrepareRun();
	virtual UINT Run();
	virtual UINT DoWork();
	virtual UINT OnExit();
	virtual BOOL IsExit(){CScopedLock autoLock(&m_lockExit); return m_bExit;}

private:
	CLockCS	m_lockExit;
};

#endif // !defined(__THREAD_H_)
