// Util.cpp: implementation of the CUtil class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <windows.h>
#include "TWSDK.h"
#include "Util.h"
#include "FramesTable.h"
#include "deelx.h"

#include <winioctl.h>
#include <setupapi.h>
#include <stdlib.h>

#define IS_NUMBER(n) ((n<=_T('9')) && (n>=_T('0')))

#ifdef UNICODE
#define lstrtoul  wcstoul
#else
#define lstrtoul  strtoul
#endif // !UNICODE

#ifdef UNICODE
#define lsprintf  swprintf
#else
#define lsprintf  sprintf
#endif // !UNICODE

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

int CUtil::map( int srcMin, int srcMax, int  srcVal,
				int dstMin, int dstMax, int* dstVal )
{
    if( srcMin == srcMax || srcVal < srcMin || srcVal > srcMax )
        return FALSE;

    // This is the overall expression used:
    // *dstVal = (srcVal - srcMin)*(dstMax - dstMin) / (srcMax - srcMin) + dstMin;
    // but we need to account for rounding so below we use the modulus
    // operator to find the remainder and increment if necessary.
    int numerator   = ( srcVal - srcMin ) * ( dstMax - dstMin );
    int denominator = srcMax - srcMin;
    int quotient    = numerator / denominator;

    if( 2 * ( numerator % denominator ) >= denominator )
        quotient++;

    *dstVal = quotient + dstMin;

    return TRUE;
}

void CUtil::ReverseBGR888(PBYTE src, PBYTE dst, UINT widthbytes, UINT height)
{
	src = src + widthbytes*(height-1);
	for(UINT y=0; y<height; y++)
	{
		memcpy(dst, src, widthbytes);
		src -= widthbytes;
		dst += widthbytes;
	}
}

void CUtil::ReverseImage(PBYTE src, PBYTE dst, UINT widthbytes, UINT height)
{
	if((UINT)(src-dst) == widthbytes) {
		PBYTE temp = dst+widthbytes;
		src = src + widthbytes*(height-1);
		for(UINT y=0; y<height/2; y++)
		{
			memcpy(dst, src, widthbytes);
			memcpy(src, temp, widthbytes);
			src -= widthbytes;
			dst += widthbytes;
			temp+= widthbytes;
		}
	}
	else if(dst != NULL) {
		src = src + widthbytes*(height-1);
		for(UINT y=0; y<height; y++)
		{
			memcpy(dst, src, widthbytes);
			src -= widthbytes;
			dst += widthbytes;
		}
	}
	else {
		BYTE pixel;
		dst = src + (widthbytes*(height-1));
		for(UINT y=0; y<height/2; y++) {
			for(UINT x=0; x<widthbytes; x++) {
				pixel = *src;
				*src++ = *dst;
				*dst++ = pixel;
			}
			dst -= widthbytes*2;
		}
	}
}

PBYTE CUtil::bgr888_to_YUY2_c(PBYTE src, PBYTE dst, UINT src_stride, 
								 UINT dest_stride, UINT width, UINT height)
{
//#define Y_FROM_RGB(r,g,b) 	((9798 * (r) + 19235 * (g) + 3736 * (b)) >> 15)
//#define U_FROM_BY(b,y) 		(((16122 * ((b) - (y))) >> 15) + 128)
//#define V_FROM_RY(r,y) 		(((25203 * ((r) - (y))) >> 15) + 128)
#define clip(x) (((x) & 0xffffff00) ? ((x) & 0x80000000) ? 0 : 255 : (x))

	int y;
	int Y;
	int b, g, r;

	for (y = height; y; y--) 
	{
		int i;

		for (i = width >> 1; i; i--) 
		{
			b = *src++;
			g = *src++;
			r = *src++;

			Y = (r * 9798 + g * 19235 + b * 3736) >> 15;

			*dst++ = clip(Y);
			*dst++ = clip(((r * (-5505) + g * (-10879) + b * 16384) >> 15) + 128);

//			Y = Y_FROM_RGB(r, g, b);
//
//			*dst++ = clip(Y);
//			*dst++ = clip(U_FROM_BY (b, Y));

			b = *src++;
			g = *src++;
			r = *src++;

			Y = (r * 9798 + g * 19235 + b * 3736) >> 15;

			*dst++ = clip(Y);
			*dst++ = clip(((r * 16384 + g * -13730 + b * -2654) >> 15) + 128);

//			Y = Y_FROM_RGB (r, g, b);
//
//			*dst++ = clip(Y);
//			*dst++ = clip(V_FROM_RY (r, Y));
		}

		src += src_stride;
		dst += dest_stride;
	}

	return dst;
}

BYTE* CUtil::C_RGBtoYUV(BYTE *dest, BYTE *src, DWORD width, DWORD height)
{
	LPBYTE dst = dest;
	DWORD  pxCount = width*height;
	DWORD  w2 = (pxCount % 2) ? (pxCount/2 + 1) : (pxCount/2) ;//>> 1;
	while (w2--)
	{
		int B = *src++;
		int G = *src++;
		int R = *src++;

		int Y = (R * 9798 + G * 19235 + B * 3736) >> 15;
		int U = ((R * (-5505) + G * (-10879) + B * 16384) >> 15) + 128;

	#define clip(x) (((x) & 0xffffff00) ? ((x) & 0x80000000) ? 0 : 255 : (x))

		*dst++ = clip(Y);
		*dst++ = clip(U);

		B = *src++;
		G = *src++;
		R = *src++;

		Y = (R * 9798 + G * 19235 + B * 3736) >> 15;
		int V = ((R * 16384 + G * -13730 + B * -2654) >> 15) + 128;

		*dst++ = clip(Y);
		*dst++ = clip(V);
	}
    return dest;
}

void CUtil::MMXSpecpy(BYTE *pSrc, BYTE *pDrc, DWORD dwCapWidth, DWORD dwWidth, DWORD dwHeight)
{
	unsigned short *pS,*ppS;
	unsigned short *pD,*ppD;
	DWORD i;
	unsigned short ddp1;//,ssp1,ssp2,ddp2;
//	unsigned short dww,temp1;

	
	pS= (unsigned short *)pSrc;
	pD= (unsigned short *)pDrc;
	ppS= (unsigned short *)pS;
	ppD= (unsigned short *)pD;

	for(i= 0; i<dwHeight; i++)
	{
		for(UINT j=0; j<dwWidth; j++)
		{
			ddp1= (*pD);
			if(ddp1 != 0x8008)
			{
				*pS= *pD;
				//*(pS+1)= *(pD+1);
			}

			pS++;
			pD++;
		}

		ppS+= dwCapWidth;
		ppD+= dwWidth;
		pS= ppS;
		pD= ppD;
	}
}

HANDLE CUtil::OpenByInterface(
		GUID* pClassGuid,	// points to the GUID that identifies the interface class
		DWORD instance,		// specifies which instance of the enumerated devices to open
		TDeviceDesc* devInfo,// device information
		TBoardDesc*  cardInfo
		)
{
	HANDLE hDev= NULL;
	HDEVINFO pinfo=NULL;
	DWORD flags = DIGCF_DEVICEINTERFACE | DIGCF_PRESENT ;
	DWORD ReqLen;
	SP_DEVICE_INTERFACE_DATA spData;
	PSP_INTERFACE_DEVICE_DETAIL_DATA pspData;
	SP_DEVINFO_DATA	spDevInfo;
	BOOL Status;
	TCHAR buff[200];

	pinfo= SetupDiGetClassDevs(pClassGuid,NULL,NULL,flags);

	if ( pinfo == INVALID_HANDLE_VALUE )
	{
		return hDev= INVALID_HANDLE_VALUE;
	}

	pspData=NULL;
	spData.cbSize= sizeof(SP_DEVICE_INTERFACE_DATA);
	Status= SetupDiEnumDeviceInterfaces(pinfo,NULL,pClassGuid,instance,&spData);

	if(!Status)
	{
		SetupDiDestroyDeviceInfoList(pinfo);
		return INVALID_HANDLE_VALUE;
	}

	Status= SetupDiGetDeviceInterfaceDetail(pinfo,&spData,NULL,0,&ReqLen,NULL);

	pspData= PSP_INTERFACE_DEVICE_DETAIL_DATA(new char[ReqLen]);
	pspData->cbSize= sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);
	Status= SetupDiGetDeviceInterfaceDetail(pinfo,&spData,pspData,ReqLen,NULL,NULL);

	if(!Status)
	{
		SetupDiDestroyDeviceInfoList(pinfo);
		return INVALID_HANDLE_VALUE;
	}

	if(devInfo) {
		//ȡPCI豸λϢ
		spDevInfo.cbSize = sizeof(SP_DEVINFO_DATA);
		if(SetupDiEnumDeviceInfo(pinfo,instance,&spDevInfo)) {
			//HardwareId
			if(SetupDiGetDeviceRegistryProperty(pinfo, &spDevInfo, SPDRP_HARDWAREID, 
					&ReqLen, (PBYTE)buff, 200, NULL)){
				//ʹʽַеHardwareIdϢ
				CRegexpT <TCHAR> regexp(TEXT("&DEV_...."));
				MatchResult result = regexp.Match(buff);
				TCHAR tmp[10];
				if(result.IsMatched()) {
					lsprintf(tmp, TEXT("%.*s"), 4, buff+result.GetEnd()-4);
					devInfo->HardwareId = lstrtoul(tmp, NULL, 16) << 16;
				}

				//vender id
				CRegexpT <TCHAR> regexp1(TEXT("\\VEN_...."));
				result = regexp1.Match(buff);
				if(result.IsMatched()) {
					lsprintf(tmp, TEXT("%.*s"), 4, buff+result.GetEnd()-4);
					devInfo->HardwareId |= lstrtoul(tmp, NULL, 16);
				}
			}

			//slotϢ
			if(SetupDiGetDeviceRegistryProperty(pinfo, &spDevInfo, SPDRP_UI_NUMBER, 
					&ReqLen, (PBYTE)buff, 200, NULL)){
				cardInfo->nSlot = *(int*)buff;
			}
			else {
				cardInfo->nSlot = -1;
			}

			if(SetupDiGetDeviceRegistryProperty(pinfo, &spDevInfo, SPDRP_LOCATION_INFORMATION, 
					&ReqLen, (PBYTE)buff, 200, NULL)){
				ParsePciLocation(buff, (ULONG*)&devInfo->nBus, (ULONG*)&devInfo->DeviceNumber, (ULONG*)&devInfo->Function);
				if(cardInfo->nSlot == -1) {
					cardInfo->nSlot = devInfo->nBus;
				}
			}
		}
	}
	
	hDev = CreateFile(
			pspData->DevicePath,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL
			);

	SetupDiDestroyDeviceInfoList(pinfo);
	delete[] pspData;
	pspData= NULL;

	return hDev;
}

//ʹʽַеbusϢ
BOOL CUtil::ParsePciLocation(TCHAR* strLocation, ULONG* bus, ULONG* device, ULONG* function)
{
	//ʹʽַеbusϢ
	CRegexpT <TCHAR> regexp(TEXT("\\b\\d+"));
	TCHAR  tmp[10];
	MatchResult result = regexp.Match(strLocation);
	
	if(!result.IsMatched()) {
		return FALSE;
	}
	
	//bus
	if(bus) {
		lsprintf(tmp, TEXT("%.*s"), result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*bus = lstrtoul(tmp, NULL, 10);
	}
	
	// get next
	result = regexp.Match(strLocation, result.GetEnd());
	if(!result.IsMatched()) {
		return FALSE;
	}
	
	//豸
	if(device) {
		lsprintf(tmp, TEXT("%.*s"), result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*device = lstrtoul(tmp, NULL, 10);
	}
	
	// get next
	result = regexp.Match(strLocation, result.GetEnd());
	if(!result.IsMatched()) {
		return FALSE;
	}
	
	//ܺ
	if(function) {
		lsprintf(tmp, TEXT("%.*s"), result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*function = lstrtoul(tmp, NULL, 10);
	}
	
	return TRUE;
}

//ʹʽַеbusϢ
BOOL CUtil::ParsePciLocationW(WCHAR* strLocation, ULONG* bus, ULONG* device, ULONG* function)
{
	//ʹʽַеbusϢ
	CRegexpT <WCHAR> regexp(L"\\b\\d+");
	WCHAR  tmp[10];
	MatchResult result = regexp.Match(strLocation);
	
	if(!result.IsMatched()) {
		return FALSE;
	}

	//bus
	if(bus) {
		swprintf(tmp, L"%.*s", result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*bus = wcstol(tmp, NULL, 10);
	}
	
	// get next
	result = regexp.Match(strLocation, result.GetEnd());
	if(!result.IsMatched()) {
		return FALSE;
	}

	//豸
	if(device) {
		swprintf(tmp, L"%.*s", result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*device = wcstol(tmp, NULL, 10);
	}
	
	// get next
	result = regexp.Match(strLocation, result.GetEnd());
	if(!result.IsMatched()) {
		return FALSE;
	}

	//ܺ
	if(function) {
		swprintf(tmp, L"%.*s", result.GetEnd() - result.GetStart(), strLocation + result.GetStart());
		*function = wcstol(tmp, NULL, 10);
	}

	return TRUE;
}

#if( !defined(WIN64) && !defined(_VC6))
void CUtil::MMXDecCPY(BYTE *Dest, BYTE *Src, DWORD nBytes)
{
	DWORD temeax= 0;
	DWORD temebx= 0;
    __asm {
		mov     eax, temeax
		mov     ebx, temebx
        mov     esi, dword ptr[Src]
        mov     edi, dword ptr[Dest]
        mov     ecx, nBytes
        shr     ecx, 1                      // nBytes / 64

CopyLoop:
		mov     ax, word ptr[esi]
		sub     ax, 800ah
		test    ax, 0ffffh
		je     cpynd
		mov     word ptr[edi], ax
;EndCopyLoop:
		emms
cpynd:
		add     esi,2
		add     edi,2
        loop CopyLoop
    }
}


void CUtil::memcpyMMX(void *Dest, void *Src, size_t nBytes)
{
    __asm {
        mov     esi, dword ptr[Src]
        mov     edi, dword ptr[Dest]
        mov     ecx, nBytes
        shr     ecx, 6                      // nBytes / 64
align 8
CopyLoop:
        movq    mm0, qword ptr[esi]
        movq    mm1, qword ptr[esi+8*1]
        movq    mm2, qword ptr[esi+8*2]
        movq    mm3, qword ptr[esi+8*3]
        movq    mm4, qword ptr[esi+8*4]
        movq    mm5, qword ptr[esi+8*5]
        movq    mm6, qword ptr[esi+8*6]
        movq    mm7, qword ptr[esi+8*7]
        movq    qword ptr[edi], mm0
        movq    qword ptr[edi+8*1], mm1
        movq    qword ptr[edi+8*2], mm2
        movq    qword ptr[edi+8*3], mm3
        movq    qword ptr[edi+8*4], mm4
        movq    qword ptr[edi+8*5], mm5
        movq    qword ptr[edi+8*6], mm6
        movq    qword ptr[edi+8*7], mm7
        add     esi, 64
        add     edi, 64
        loop CopyLoop
        mov     ecx, nBytes
        and     ecx, 63
        cmp     ecx, 0
        je EndCopyLoop
EndCopyLoop:
		emms
    }
}
    
void CUtil::memcpySSE(void* Dest, void* Src, size_t nBytes)
{
    __asm
    {

		mov     esi, dword ptr[Src]
		mov     edi, dword ptr[Dest]
		mov     ecx, nBytes
		shr     ecx, 7
	align 4
	MemcpySSE_Loop:
		; movaps should be slightly more efficient
		; as the data is 16 bit aligned
		movaps  xmm0, [esi]
		movaps  xmm1, [esi+16*1]
		movaps  xmm2, [esi+16*2]
		movaps  xmm3, [esi+16*3]
		movaps  xmm4, [esi+16*4]
		movaps  xmm5, [esi+16*5]
		movaps  xmm6, [esi+16*6]
		movaps  xmm7, [esi+16*7]
		movntps [edi], xmm0
		movntps [edi+16*1], xmm1
		movntps [edi+16*2], xmm2
		movntps [edi+16*3], xmm3
		movntps [edi+16*4], xmm4
		movntps [edi+16*5], xmm5
		movntps [edi+16*6], xmm6
		movntps [edi+16*7], xmm7
		add     esi, 128
		add     edi, 128
		dec ecx
		jnz MemcpySSE_Loop

		mov     ecx, nBytes
		and     ecx, 127
		cmp     ecx, 0
		je MemcpySSE_End
align 8
CopyLoop2:
        mov dl, byte ptr[esi] 
        mov byte ptr[edi], dl
        inc esi
        inc edi
        dec ecx
        jne near CopyLoop2
	MemcpySSE_End:
	}
}

void CUtil::memcpy_sse(void *dst, void *src, int nQWORDs)
{   
#define   CACHEBLOCK   1024   //һжQWORDs   
	//޸Ĵֵпʵָߵٶ   
	int   n=((int)(nQWORDs/CACHEBLOCK))*CACHEBLOCK;   
	int   m=nQWORDs-n;   
	if   (n)   
	{   
		_asm   //ȿ   
		{   
			mov   esi,   src   
			mov   edi,   dst   
			mov   ecx,   n //Ҫٸ   
			lea   esi,   [esi+ecx*8]   
			lea   edi,   [edi+ecx*8]   
			neg   ecx   
mainloop:   
			mov   eax,   CACHEBLOCK   /   16   
prefetchloop:   
			mov   ebx,   [esi+ecx*8]   //Ԥѭ   
			mov   ebx,   [esi+ecx*8+64] //Ԥѭ   
			add   ecx,   16   
			dec   eax   
			jnz   prefetchloop   
			sub   ecx,   CACHEBLOCK   
			mov   eax,   CACHEBLOCK   /   8   
writeloop:   
			movq   mm0,   qword   ptr   [esi+ecx*8         ]   
			movq   mm1,   qword   ptr   [esi+ecx*8+8   ]   
			movq   mm2,   qword   ptr   [esi+ecx*8+16]   
			movq   mm3,   qword   ptr   [esi+ecx*8+24]   
			movq   mm4,   qword   ptr   [esi+ecx*8+32]   
			movq   mm5,   qword   ptr   [esi+ecx*8+40]   
			movq   mm6,   qword   ptr   [esi+ecx*8+48]   
			movq   mm7,   qword   ptr   [esi+ecx*8+56]   
			
			movntq   qword   ptr   [edi+ecx*8       ],   mm0   
			movntq   qword   ptr   [edi+ecx*8+8   ],   mm1   
			movntq   qword   ptr   [edi+ecx*8+16],   mm2   
			movntq   qword   ptr   [edi+ecx*8+24],   mm3   
			movntq   qword   ptr   [edi+ecx*8+32],   mm4   
			movntq   qword   ptr   [edi+ecx*8+40],   mm5   
			movntq   qword   ptr   [edi+ecx*8+48],   mm6   
			movntq   qword   ptr   [edi+ecx*8+56],   mm7   
			add   ecx,   8   
			dec   eax   
			jnz   writeloop   
			or   ecx,   ecx   
			jnz   mainloop   
		}   
	}   
	if   (m)   
	{   
		_asm   
		{   
			mov   esi,   src   
			mov   edi,   dst   
			mov   ecx,   m   
			mov   ebx,   nQWORDs   
			lea   esi,   [esi+ebx*8]   
			lea   edi,   [edi+ebx*8]   
			neg   ecx   
copyloop:   
			prefetchnta   [esi+ecx*8+512]     //Ԥ   
			movq   mm0,   qword   ptr   [esi+ecx*8       ]   
			movq   mm1,   qword   ptr   [esi+ecx*8+8   ]   
			movq   mm2,   qword   ptr   [esi+ecx*8+16]   
			movq   mm3,   qword   ptr   [esi+ecx*8+24]   
			movq   mm4,   qword   ptr   [esi+ecx*8+32]   
			movq   mm5,   qword   ptr   [esi+ecx*8+40]   
			movq   mm6,   qword   ptr   [esi+ecx*8+48]   
			movq   mm7,   qword   ptr   [esi+ecx*8+56]   
			
			movntq   qword   ptr   [edi+ecx*8       ],   mm0   
			movntq   qword   ptr   [edi+ecx*8+8   ],   mm1   
			movntq   qword   ptr   [edi+ecx*8+16],   mm2   
			movntq   qword   ptr   [edi+ecx*8+24],   mm3   
			movntq   qword   ptr   [edi+ecx*8+32],   mm4   
			movntq   qword   ptr   [edi+ecx*8+40],   mm5   
			movntq   qword   ptr   [edi+ecx*8+48],   mm6   
			movntq   qword   ptr   [edi+ecx*8+56],   mm7   
			add   ecx,   8   
			jnz   copyloop   
			sfence   
			emms   
		}   
	}   
	else   
	{   
		_asm   
		{   
			sfence   
			emms   
		}   
	}   
} 

//352
void CUtil::DownScale(BYTE *pIn,BYTE *pOut, int width, int height, int stride)
{
	int nHeight = height / 2;
	__asm{
		MOV ESI,DWORD PTR SS:[pIn]
		MOV EDI,DWORD PTR SS:[pOut]
		MOV ECX,nHeight
Loop1:
		MOV EAX,stride
Loop2:
		MOVQ MM0,QWORD PTR DS:[ESI]
		MOVQ MM1,QWORD PTR DS:[ESI+8]
		MOVQ MM2,QWORD PTR DS:[ESI+0x10]
		MOVQ MM3,QWORD PTR DS:[ESI+0x18]
		ADD ESI,0x20
		PSLLQ MM0,8
		PSLLQ MM1,8
		PSLLQ MM2,8
		PSLLQ MM3,8
		PSHUFW MM0,MM0,0xB4
		MOVQ MM4,QWORD PTR DS:[ESI]
		PSHUFW MM1,MM1,0xB4
		MOVQ MM5,QWORD PTR DS:[ESI+8]
		PSHUFW MM2,MM2,0xB4
		MOVQ MM6,QWORD PTR DS:[ESI+0x10]
		PSHUFW MM3,MM3,0xB4
		MOVQ MM7,QWORD PTR DS:[ESI+0x18]
		PSRLQ MM0,0x18
		PSRLQ MM1,0x18
		PSRLQ MM2,0x18
		PSRLQ MM3,0x18
		PUNPCKLDQ MM0,MM1
		PUNPCKLDQ MM2,MM3
		PSLLQ MM4,8
		PSLLQ MM5,8
		PSLLQ MM6,8
		PSLLQ MM7,8
		MOVQ QWORD PTR DS:[EDI],MM0
		MOVQ QWORD PTR DS:[EDI+8],MM2
		ADD ESI,0x20
		PSHUFW MM4,MM4,0xB4
		PSHUFW MM5,MM5,0xB4
		PSHUFW MM6,MM6,0xB4
		PSHUFW MM7,MM7,0xB4
		PSRLQ MM4,0x18
		PSRLQ MM5,0x18
		PSRLQ MM6,0x18
		PSRLQ MM7,0x18
		PUNPCKLDQ MM4,MM5
		PUNPCKLDQ MM6,MM7
		MOVQ QWORD PTR DS:[EDI+0x10],MM4
		MOVQ QWORD PTR DS:[EDI+0x18],MM6
		ADD EDI,0x20
		SUB EAX,0x40
		JNZ near Loop2
		ADD ESI,stride
		SUB ECX,1
		JNZ near Loop1
	emms
	}
}
#endif

LPCTSTR CUtil::GetModulePath(HANDLE hModule, LPTSTR path, int len)
{
	if(path == NULL)
		return NULL;

	GetModuleFileName((HMODULE)hModule, path, len);

	int   nPathLen = lstrlen(path);

	for(int i=nPathLen-1; i>=0; i--)
	{
		if(path[i] == '\\')
		{
			path[i+1] = 0;
			break;
		}
	}

	return path;
}

void CUtil::DownHalfHeight(BYTE* pIn, BYTE *pOut, int height, int stride, int dst_pitch) 
{
	int nHalfH = height/2;
	if(dst_pitch == 0) dst_pitch = stride;
	for(int i=0; i<nHalfH; i++) {
		memcpy(pOut, pIn, stride);
		pIn += stride*2;
		pOut+= dst_pitch;
	}
}

DWORD CUtil::DownFrameRate(int from, int to)
{
	DWORD dwRet = 0xFFFFFFFF;

	switch(from) {
	case 1:	{
		dwRet = FRAME_TABLE_01_01;
		break;
	}
	case 2:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_02_01;	break;
		case 2:	dwRet = FRAME_TABLE_02_02;	break;
		}
		break;
	}
	case 3:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_03_01;	break;
		case 2:	dwRet = FRAME_TABLE_03_02;	break;
		case 3:	dwRet = FRAME_TABLE_03_03;	break;
		}
		break;
	}
	case 4:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_04_01;	break;
		case 2:	dwRet = FRAME_TABLE_04_02;	break;
		case 3:	dwRet = FRAME_TABLE_04_03;	break;
		case 4:	dwRet = FRAME_TABLE_04_04;	break;
		}
		break;
	}
	case 5:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_05_01;	break;
		case 2:	dwRet = FRAME_TABLE_05_02;	break;
		case 3:	dwRet = FRAME_TABLE_05_03;	break;
		case 4:	dwRet = FRAME_TABLE_05_04;	break;
		case 5:	dwRet = FRAME_TABLE_05_05;	break;
		}
		break;
	}
	case 6:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_06_01;	break;
		case 2:	dwRet = FRAME_TABLE_06_02;	break;
		case 3:	dwRet = FRAME_TABLE_06_03;	break;
		case 4:	dwRet = FRAME_TABLE_06_04;	break;
		case 5:	dwRet = FRAME_TABLE_06_05;	break;
		case 6:	dwRet = FRAME_TABLE_06_06;	break;
		}
		break;
	}
	case 7:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_07_01;	break;
		case 2:	dwRet = FRAME_TABLE_07_02;	break;
		case 3:	dwRet = FRAME_TABLE_07_03;	break;
		case 4:	dwRet = FRAME_TABLE_07_04;	break;
		case 5:	dwRet = FRAME_TABLE_07_05;	break;
		case 6:	dwRet = FRAME_TABLE_07_06;	break;
		case 7:	dwRet = FRAME_TABLE_07_07;	break;
		}
		break;
	}
	case 8:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_08_01;	break;
		case 2:	dwRet = FRAME_TABLE_08_02;	break;
		case 3:	dwRet = FRAME_TABLE_08_03;	break;
		case 4:	dwRet = FRAME_TABLE_08_04;	break;
		case 5:	dwRet = FRAME_TABLE_08_05;	break;
		case 6:	dwRet = FRAME_TABLE_08_06;	break;
		case 7:	dwRet = FRAME_TABLE_08_07;	break;
		case 8:	dwRet = FRAME_TABLE_08_08;	break;
		}
		break;
	}
	case 9:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_09_01;	break;
		case 2:	dwRet = FRAME_TABLE_09_02;	break;
		case 3:	dwRet = FRAME_TABLE_09_03;	break;
		case 4:	dwRet = FRAME_TABLE_09_04;	break;
		case 5:	dwRet = FRAME_TABLE_09_05;	break;
		case 6:	dwRet = FRAME_TABLE_09_06;	break;
		case 7:	dwRet = FRAME_TABLE_09_07;	break;
		case 8:	dwRet = FRAME_TABLE_09_08;	break;
		case 9:	dwRet = FRAME_TABLE_09_09;	break;
		}
		break;
	}
	case 10:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_10_01;	break;
		case 2:	dwRet = FRAME_TABLE_10_02;	break;
		case 3:	dwRet = FRAME_TABLE_10_03;	break;
		case 4:	dwRet = FRAME_TABLE_10_04;	break;
		case 5:	dwRet = FRAME_TABLE_10_05;	break;
		case 6:	dwRet = FRAME_TABLE_10_06;	break;
		case 7:	dwRet = FRAME_TABLE_10_07;	break;
		case 8:	dwRet = FRAME_TABLE_10_08;	break;
		case 9:	dwRet = FRAME_TABLE_10_09;	break;
		case 10:dwRet = FRAME_TABLE_10_10;	break;
		}
		break;
	}
	case 11:	{
		switch(to) {
		case 1:	dwRet = FRAME_TABLE_11_01;	break;
		case 2:	dwRet = FRAME_TABLE_11_02;	break;
		case 3:	dwRet = FRAME_TABLE_11_03;	break;
		case 4:	dwRet = FRAME_TABLE_11_04;	break;
		case 5:	dwRet = FRAME_TABLE_11_05;	break;
		case 6:	dwRet = FRAME_TABLE_11_06;	break;
		case 7:	dwRet = FRAME_TABLE_11_07;	break;
		case 8:	dwRet = FRAME_TABLE_11_08;	break;
		case 9:	dwRet = FRAME_TABLE_11_09;	break;
		case 10:dwRet = FRAME_TABLE_11_10;	break;
		case 11:dwRet = FRAME_TABLE_11_11;	break;
		}
		break;
	}
	case 12:	{
		switch(to) {
		case 1: dwRet = FRAME_TABLE_12_01;	break;
		case 2:	dwRet = FRAME_TABLE_12_02;	break;
		case 3:	dwRet = FRAME_TABLE_12_03;	break;
		case 4:	dwRet = FRAME_TABLE_12_04;	break;
		case 5:	dwRet = FRAME_TABLE_12_05;	break;
		case 6:	dwRet = FRAME_TABLE_12_06;	break;
		case 7:	dwRet = FRAME_TABLE_12_07;	break;
		case 8:	dwRet = FRAME_TABLE_12_08;	break;
		case 9:	dwRet = FRAME_TABLE_12_09;	break;
		case 10:dwRet = FRAME_TABLE_12_10;	break;
		case 11:dwRet = FRAME_TABLE_12_11;	break;
		case 12:dwRet = FRAME_TABLE_12_12;	break;
		}
		break;
	}
	case 13:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_13_01; break;
		case 2:	 dwRet = FRAME_TABLE_13_02; break;
		case 3:	 dwRet = FRAME_TABLE_13_03; break;
		case 4:	 dwRet = FRAME_TABLE_13_04; break;
		case 5:	 dwRet = FRAME_TABLE_13_05; break;
		case 6:  dwRet = FRAME_TABLE_13_06;	break;
		case 7:  dwRet = FRAME_TABLE_13_07;	break;
		case 8:  dwRet = FRAME_TABLE_13_08;	break;
		case 9:  dwRet = FRAME_TABLE_13_09;	break;
		case 10: dwRet = FRAME_TABLE_13_10;	break;
		case 11: dwRet = FRAME_TABLE_13_11;	break;
		case 12: dwRet = FRAME_TABLE_13_12;	break;
		case 13: dwRet = FRAME_TABLE_13_13;	break;
		}
		break;
	}
	case 14:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_14_01; break;
		case 2:	 dwRet = FRAME_TABLE_14_02; break;
		case 3:	 dwRet = FRAME_TABLE_14_03; break;
		case 4:	 dwRet = FRAME_TABLE_14_04; break;
		case 5:	 dwRet = FRAME_TABLE_14_05; break;
		case 6:  dwRet = FRAME_TABLE_14_06;	break;
		case 7:  dwRet = FRAME_TABLE_14_07;	break;
		case 8:  dwRet = FRAME_TABLE_14_08;	break;
		case 9:  dwRet = FRAME_TABLE_14_09;	break;
		case 10: dwRet = FRAME_TABLE_14_10;	break;
		case 11: dwRet = FRAME_TABLE_14_11;	break;
		case 12: dwRet = FRAME_TABLE_14_12;	break;
		case 13: dwRet = FRAME_TABLE_14_13;	break;
		case 14: dwRet = FRAME_TABLE_14_14;	break;
		}
		break;
	}
	case 15:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_15_01; break;
		case 2:	 dwRet = FRAME_TABLE_15_02; break;
		case 3:	 dwRet = FRAME_TABLE_15_03; break;
		case 4:	 dwRet = FRAME_TABLE_15_04; break;
		case 5:	 dwRet = FRAME_TABLE_15_05; break;
		case 6:  dwRet = FRAME_TABLE_15_06;	break;
		case 7:  dwRet = FRAME_TABLE_15_07;	break;
		case 8:  dwRet = FRAME_TABLE_15_08;	break;
		case 9:  dwRet = FRAME_TABLE_15_09;	break;
		case 10: dwRet = FRAME_TABLE_15_10;	break;
		case 11: dwRet = FRAME_TABLE_15_11;	break;
		case 12: dwRet = FRAME_TABLE_15_12;	break;
		case 13: dwRet = FRAME_TABLE_15_13;	break;
		case 14: dwRet = FRAME_TABLE_15_14;	break;
		case 15: dwRet = FRAME_TABLE_15_15;	break;
		}
		break;
	}
	case 16:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_16_01; break;
		case 2:	 dwRet = FRAME_TABLE_16_02; break;
		case 3:	 dwRet = FRAME_TABLE_16_03; break;
		case 4:	 dwRet = FRAME_TABLE_16_04; break;
		case 5:	 dwRet = FRAME_TABLE_16_05; break;
		case 6:  dwRet = FRAME_TABLE_16_06;	break;
		case 7:  dwRet = FRAME_TABLE_16_07;	break;
		case 8:  dwRet = FRAME_TABLE_16_08;	break;
		case 9:  dwRet = FRAME_TABLE_16_09;	break;
		case 10: dwRet = FRAME_TABLE_16_10;	break;
		case 11: dwRet = FRAME_TABLE_16_11;	break;
		case 12: dwRet = FRAME_TABLE_16_12;	break;
		case 13: dwRet = FRAME_TABLE_16_13;	break;
		case 14: dwRet = FRAME_TABLE_16_14;	break;
		case 15: dwRet = FRAME_TABLE_16_15;	break;
		case 16: dwRet = FRAME_TABLE_16_16;	break;
		}
		break;
	}
	case 17:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_17_01; break;
		case 2:	 dwRet = FRAME_TABLE_17_02; break;
		case 3:	 dwRet = FRAME_TABLE_17_03; break;
		case 4:	 dwRet = FRAME_TABLE_17_04; break;
		case 5:	 dwRet = FRAME_TABLE_17_05; break;
		case 6:  dwRet = FRAME_TABLE_17_06;	break;
		case 7:  dwRet = FRAME_TABLE_17_07;	break;
		case 8:  dwRet = FRAME_TABLE_17_08;	break;
		case 9:  dwRet = FRAME_TABLE_17_09;	break;
		case 10: dwRet = FRAME_TABLE_17_10;	break;
		case 11: dwRet = FRAME_TABLE_17_11;	break;
		case 12: dwRet = FRAME_TABLE_17_12;	break;
		case 13: dwRet = FRAME_TABLE_17_13;	break;
		case 14: dwRet = FRAME_TABLE_17_14;	break;
		case 15: dwRet = FRAME_TABLE_17_15;	break;
		case 16: dwRet = FRAME_TABLE_17_16;	break;
		case 17: dwRet = FRAME_TABLE_17_17;	break;
		}
		break;
	}
	case 18:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_18_01; break;
		case 2:	 dwRet = FRAME_TABLE_18_02; break;
		case 3:	 dwRet = FRAME_TABLE_18_03; break;
		case 4:	 dwRet = FRAME_TABLE_18_04; break;
		case 5:	 dwRet = FRAME_TABLE_18_05; break;
		case 6:  dwRet = FRAME_TABLE_18_06;	break;
		case 7:  dwRet = FRAME_TABLE_18_07;	break;
		case 8:  dwRet = FRAME_TABLE_18_08;	break;
		case 9:  dwRet = FRAME_TABLE_18_09;	break;
		case 10: dwRet = FRAME_TABLE_18_10;	break;
		case 11: dwRet = FRAME_TABLE_18_11;	break;
		case 12: dwRet = FRAME_TABLE_18_12;	break;
		case 13: dwRet = FRAME_TABLE_18_13;	break;
		case 14: dwRet = FRAME_TABLE_18_14;	break;
		case 15: dwRet = FRAME_TABLE_18_15;	break;
		case 16: dwRet = FRAME_TABLE_18_16;	break;
		case 17: dwRet = FRAME_TABLE_18_17;	break;
		case 18: dwRet = FRAME_TABLE_18_18;	break;
		}
		break;
	}
	case 19:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_19_01; break;
		case 2:	 dwRet = FRAME_TABLE_19_02; break;
		case 3:	 dwRet = FRAME_TABLE_19_03; break;
		case 4:	 dwRet = FRAME_TABLE_19_04; break;
		case 5:	 dwRet = FRAME_TABLE_19_05; break;
		case 6:  dwRet = FRAME_TABLE_19_06;	break;
		case 7:  dwRet = FRAME_TABLE_19_07;	break;
		case 8:  dwRet = FRAME_TABLE_19_08;	break;
		case 9:  dwRet = FRAME_TABLE_19_09;	break;
		case 10: dwRet = FRAME_TABLE_19_10;	break;
		case 11: dwRet = FRAME_TABLE_19_11;	break;
		case 12: dwRet = FRAME_TABLE_19_12;	break;
		case 13: dwRet = FRAME_TABLE_19_13;	break;
		case 14: dwRet = FRAME_TABLE_19_14;	break;
		case 15: dwRet = FRAME_TABLE_19_15;	break;
		case 16: dwRet = FRAME_TABLE_19_16;	break;
		case 17: dwRet = FRAME_TABLE_19_17;	break;
		case 18: dwRet = FRAME_TABLE_19_18;	break;
		case 19: dwRet = FRAME_TABLE_19_19;	break;
		}
		break;
	}
	case 20:{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_20_01; break;
		case 2:	 dwRet = FRAME_TABLE_20_02; break;
		case 3:	 dwRet = FRAME_TABLE_20_03; break;
		case 4:	 dwRet = FRAME_TABLE_20_04; break;
		case 5:	 dwRet = FRAME_TABLE_20_05; break;
		case 6:  dwRet = FRAME_TABLE_20_06;	break;
		case 7:  dwRet = FRAME_TABLE_20_07;	break;
		case 8:  dwRet = FRAME_TABLE_20_08;	break;
		case 9:  dwRet = FRAME_TABLE_20_09;	break;
		case 10: dwRet = FRAME_TABLE_20_10;	break;
		case 11: dwRet = FRAME_TABLE_20_11;	break;
		case 12: dwRet = FRAME_TABLE_20_12;	break;
		case 13: dwRet = FRAME_TABLE_20_13;	break;
		case 14: dwRet = FRAME_TABLE_20_14;	break;
		case 15: dwRet = FRAME_TABLE_20_15;	break;
		case 16: dwRet = FRAME_TABLE_20_16;	break;
		case 17: dwRet = FRAME_TABLE_20_17;	break;
		case 18: dwRet = FRAME_TABLE_20_18;	break;
		case 19: dwRet = FRAME_TABLE_20_19;	break;
		case 20: dwRet = FRAME_TABLE_20_20;	break;
		}
		break;
	}
	case 21:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_21_01; break;
		case 2:	 dwRet = FRAME_TABLE_21_02; break;
		case 3:	 dwRet = FRAME_TABLE_21_03; break;
		case 4:	 dwRet = FRAME_TABLE_21_04; break;
		case 5:	 dwRet = FRAME_TABLE_21_05; break;
		case 6:  dwRet = FRAME_TABLE_21_06;	break;
		case 7:  dwRet = FRAME_TABLE_21_07;	break;
		case 8:  dwRet = FRAME_TABLE_21_08;	break;
		case 9:  dwRet = FRAME_TABLE_21_09;	break;
		case 10: dwRet = FRAME_TABLE_21_10;	break;
		case 11: dwRet = FRAME_TABLE_21_11;	break;
		case 12: dwRet = FRAME_TABLE_21_12;	break;
		case 13: dwRet = FRAME_TABLE_21_13;	break;
		case 14: dwRet = FRAME_TABLE_21_14;	break;
		case 15: dwRet = FRAME_TABLE_21_15;	break;
		case 16: dwRet = FRAME_TABLE_21_16;	break;
		case 17: dwRet = FRAME_TABLE_21_17;	break;
		case 18: dwRet = FRAME_TABLE_21_18;	break;
		case 19: dwRet = FRAME_TABLE_21_19;	break;
		case 20: dwRet = FRAME_TABLE_21_20;	break;
		case 21: dwRet = FRAME_TABLE_21_21;	break;
		}
		break;
	}
	case 22:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_22_01; break;
		case 2:	 dwRet = FRAME_TABLE_22_02; break;
		case 3:	 dwRet = FRAME_TABLE_22_03; break;
		case 4:	 dwRet = FRAME_TABLE_22_04; break;
		case 5:	 dwRet = FRAME_TABLE_22_05; break;
		case 6:  dwRet = FRAME_TABLE_22_06;	break;
		case 7:  dwRet = FRAME_TABLE_22_07;	break;
		case 8:  dwRet = FRAME_TABLE_22_08;	break;
		case 9:  dwRet = FRAME_TABLE_22_09;	break;
		case 10: dwRet = FRAME_TABLE_22_10;	break;
		case 11: dwRet = FRAME_TABLE_22_11;	break;
		case 12: dwRet = FRAME_TABLE_22_12;	break;
		case 13: dwRet = FRAME_TABLE_22_13;	break;
		case 14: dwRet = FRAME_TABLE_22_14;	break;
		case 15: dwRet = FRAME_TABLE_22_15;	break;
		case 16: dwRet = FRAME_TABLE_22_16;	break;
		case 17: dwRet = FRAME_TABLE_22_17;	break;
		case 18: dwRet = FRAME_TABLE_22_18;	break;
		case 19: dwRet = FRAME_TABLE_22_19;	break;
		case 20: dwRet = FRAME_TABLE_22_20;	break;
		case 21: dwRet = FRAME_TABLE_22_21;	break;
		case 22: dwRet = FRAME_TABLE_22_22;	break;
		}
		break;
	}
	case 23:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_23_01; break;
		case 2:	 dwRet = FRAME_TABLE_23_02; break;
		case 3:	 dwRet = FRAME_TABLE_23_03; break;
		case 4:	 dwRet = FRAME_TABLE_23_04; break;
		case 5:	 dwRet = FRAME_TABLE_23_05; break;
		case 6:  dwRet = FRAME_TABLE_23_06;	break;
		case 7:  dwRet = FRAME_TABLE_23_07;	break;
		case 8:  dwRet = FRAME_TABLE_23_08;	break;
		case 9:  dwRet = FRAME_TABLE_23_09;	break;
		case 10: dwRet = FRAME_TABLE_23_10;	break;
		case 11: dwRet = FRAME_TABLE_23_11;	break;
		case 12: dwRet = FRAME_TABLE_23_12;	break;
		case 13: dwRet = FRAME_TABLE_23_13;	break;
		case 14: dwRet = FRAME_TABLE_23_14;	break;
		case 15: dwRet = FRAME_TABLE_23_15;	break;
		case 16: dwRet = FRAME_TABLE_23_16;	break;
		case 17: dwRet = FRAME_TABLE_23_17;	break;
		case 18: dwRet = FRAME_TABLE_23_18;	break;
		case 19: dwRet = FRAME_TABLE_23_19;	break;
		case 20: dwRet = FRAME_TABLE_23_20;	break;
		case 21: dwRet = FRAME_TABLE_23_21;	break;
		case 22: dwRet = FRAME_TABLE_23_22;	break;
		case 23: dwRet = FRAME_TABLE_23_23;	break;
		}
		break;
	}
	case 24:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_24_01; break;
		case 2:	 dwRet = FRAME_TABLE_24_02; break;
		case 3:	 dwRet = FRAME_TABLE_24_03; break;
		case 4:	 dwRet = FRAME_TABLE_24_04; break;
		case 5:	 dwRet = FRAME_TABLE_24_05; break;
		case 6:  dwRet = FRAME_TABLE_24_06;	break;
		case 7:  dwRet = FRAME_TABLE_24_07;	break;
		case 8:  dwRet = FRAME_TABLE_24_08;	break;
		case 9:  dwRet = FRAME_TABLE_24_09;	break;
		case 10: dwRet = FRAME_TABLE_24_10;	break;
		case 11: dwRet = FRAME_TABLE_24_11;	break;
		case 12: dwRet = FRAME_TABLE_24_12;	break;
		case 13: dwRet = FRAME_TABLE_24_13;	break;
		case 14: dwRet = FRAME_TABLE_24_14;	break;
		case 15: dwRet = FRAME_TABLE_24_15;	break;
		case 16: dwRet = FRAME_TABLE_24_16;	break;
		case 17: dwRet = FRAME_TABLE_24_17;	break;
		case 18: dwRet = FRAME_TABLE_24_18;	break;
		case 19: dwRet = FRAME_TABLE_24_19;	break;
		case 20: dwRet = FRAME_TABLE_24_20;	break;
		case 21: dwRet = FRAME_TABLE_24_21;	break;
		case 22: dwRet = FRAME_TABLE_24_22;	break;
		case 23: dwRet = FRAME_TABLE_24_23;	break;
		case 24: dwRet = FRAME_TABLE_24_24;	break;
		}
		break;
	}
	case 25:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_25_01; break;
		case 2:	 dwRet = FRAME_TABLE_25_02; break;
		case 3:	 dwRet = FRAME_TABLE_25_03; break;
		case 4:	 dwRet = FRAME_TABLE_25_04; break;
		case 5:	 dwRet = FRAME_TABLE_25_05; break;
		case 6:  dwRet = FRAME_TABLE_25_06;	break;
		case 7:  dwRet = FRAME_TABLE_25_07;	break;
		case 8:  dwRet = FRAME_TABLE_25_08;	break;
		case 9:  dwRet = FRAME_TABLE_25_09;	break;
		case 10: dwRet = FRAME_TABLE_25_10;	break;
		case 11: dwRet = FRAME_TABLE_25_11;	break;
		case 12: dwRet = FRAME_TABLE_25_12;	break;
		case 13: dwRet = FRAME_TABLE_25_13;	break;
		case 14: dwRet = FRAME_TABLE_25_14;	break;
		case 15: dwRet = FRAME_TABLE_25_15;	break;
		case 16: dwRet = FRAME_TABLE_25_16;	break;
		case 17: dwRet = FRAME_TABLE_25_17;	break;
		case 18: dwRet = FRAME_TABLE_25_18;	break;
		case 19: dwRet = FRAME_TABLE_25_19;	break;
		case 20: dwRet = FRAME_TABLE_25_20;	break;
		case 21: dwRet = FRAME_TABLE_25_21;	break;
		case 22: dwRet = FRAME_TABLE_25_22;	break;
		case 23: dwRet = FRAME_TABLE_25_23;	break;
		case 24: dwRet = FRAME_TABLE_25_24;	break;
		case 25: dwRet = FRAME_TABLE_25_25;	break;
		}
		break;
	}
	case 26:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_26_01; break;
		case 2:	 dwRet = FRAME_TABLE_26_02; break;
		case 3:	 dwRet = FRAME_TABLE_26_03; break;
		case 4:	 dwRet = FRAME_TABLE_26_04; break;
		case 5:	 dwRet = FRAME_TABLE_26_05; break;
		case 6:  dwRet = FRAME_TABLE_26_06;	break;
		case 7:  dwRet = FRAME_TABLE_26_07;	break;
		case 8:  dwRet = FRAME_TABLE_26_08;	break;
		case 9:  dwRet = FRAME_TABLE_26_09;	break;
		case 10: dwRet = FRAME_TABLE_26_10;	break;
		case 11: dwRet = FRAME_TABLE_26_11;	break;
		case 12: dwRet = FRAME_TABLE_26_12;	break;
		case 13: dwRet = FRAME_TABLE_26_13;	break;
		case 14: dwRet = FRAME_TABLE_26_14;	break;
		case 15: dwRet = FRAME_TABLE_26_15;	break;
		case 16: dwRet = FRAME_TABLE_26_16;	break;
		case 17: dwRet = FRAME_TABLE_26_17;	break;
		case 18: dwRet = FRAME_TABLE_26_18;	break;
		case 19: dwRet = FRAME_TABLE_26_19;	break;
		case 20: dwRet = FRAME_TABLE_26_20;	break;
		case 21: dwRet = FRAME_TABLE_26_21;	break;
		case 22: dwRet = FRAME_TABLE_26_22;	break;
		case 23: dwRet = FRAME_TABLE_26_23;	break;
		case 24: dwRet = FRAME_TABLE_26_24;	break;
		case 25: dwRet = FRAME_TABLE_26_25;	break;
		case 26: dwRet = FRAME_TABLE_26_26;	break;
		}
		break;
	}
	case 27:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_27_01; break;
		case 2:	 dwRet = FRAME_TABLE_27_02; break;
		case 3:	 dwRet = FRAME_TABLE_27_03; break;
		case 4:	 dwRet = FRAME_TABLE_27_04; break;
		case 5:	 dwRet = FRAME_TABLE_27_05; break;
		case 6:  dwRet = FRAME_TABLE_27_06;	break;
		case 7:  dwRet = FRAME_TABLE_27_07;	break;
		case 8:  dwRet = FRAME_TABLE_27_08;	break;
		case 9:  dwRet = FRAME_TABLE_27_09;	break;
		case 10: dwRet = FRAME_TABLE_27_10;	break;
		case 11: dwRet = FRAME_TABLE_27_11;	break;
		case 12: dwRet = FRAME_TABLE_27_12;	break;
		case 13: dwRet = FRAME_TABLE_27_13;	break;
		case 14: dwRet = FRAME_TABLE_27_14;	break;
		case 15: dwRet = FRAME_TABLE_27_15;	break;
		case 16: dwRet = FRAME_TABLE_27_16;	break;
		case 17: dwRet = FRAME_TABLE_27_17;	break;
		case 18: dwRet = FRAME_TABLE_27_18;	break;
		case 19: dwRet = FRAME_TABLE_27_19;	break;
		case 20: dwRet = FRAME_TABLE_27_20;	break;
		case 21: dwRet = FRAME_TABLE_27_21;	break;
		case 22: dwRet = FRAME_TABLE_27_22;	break;
		case 23: dwRet = FRAME_TABLE_27_23;	break;
		case 24: dwRet = FRAME_TABLE_27_24;	break;
		case 25: dwRet = FRAME_TABLE_27_25;	break;
		case 26: dwRet = FRAME_TABLE_27_26;	break;
		case 27: dwRet = FRAME_TABLE_27_27;	break;
		}
		break;
	}
	case 28:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_28_01; break;
		case 2:	 dwRet = FRAME_TABLE_28_02; break;
		case 3:	 dwRet = FRAME_TABLE_28_03; break;
		case 4:	 dwRet = FRAME_TABLE_28_04; break;
		case 5:	 dwRet = FRAME_TABLE_28_05; break;
		case 6:  dwRet = FRAME_TABLE_28_06;	break;
		case 7:  dwRet = FRAME_TABLE_28_07;	break;
		case 8:  dwRet = FRAME_TABLE_28_08;	break;
		case 9:  dwRet = FRAME_TABLE_28_09;	break;
		case 10: dwRet = FRAME_TABLE_28_10;	break;
		case 11: dwRet = FRAME_TABLE_28_11;	break;
		case 12: dwRet = FRAME_TABLE_28_12;	break;
		case 13: dwRet = FRAME_TABLE_28_13;	break;
		case 14: dwRet = FRAME_TABLE_28_14;	break;
		case 15: dwRet = FRAME_TABLE_28_15;	break;
		case 16: dwRet = FRAME_TABLE_28_16;	break;
		case 17: dwRet = FRAME_TABLE_28_17;	break;
		case 18: dwRet = FRAME_TABLE_28_18;	break;
		case 19: dwRet = FRAME_TABLE_28_19;	break;
		case 20: dwRet = FRAME_TABLE_28_20;	break;
		case 21: dwRet = FRAME_TABLE_28_21;	break;
		case 22: dwRet = FRAME_TABLE_28_22;	break;
		case 23: dwRet = FRAME_TABLE_28_23;	break;
		case 24: dwRet = FRAME_TABLE_28_24;	break;
		case 25: dwRet = FRAME_TABLE_28_25;	break;
		case 26: dwRet = FRAME_TABLE_28_26;	break;
		case 27: dwRet = FRAME_TABLE_28_27;	break;
		case 28: dwRet = FRAME_TABLE_28_28;	break;
		}
		break;
	}
	case 29:	{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_29_01; break;
		case 2:	 dwRet = FRAME_TABLE_29_02; break;
		case 3:	 dwRet = FRAME_TABLE_29_03; break;
		case 4:	 dwRet = FRAME_TABLE_29_04; break;
		case 5:	 dwRet = FRAME_TABLE_29_05; break;
		case 6:  dwRet = FRAME_TABLE_29_06;	break;
		case 7:  dwRet = FRAME_TABLE_29_07;	break;
		case 8:  dwRet = FRAME_TABLE_29_08;	break;
		case 9:  dwRet = FRAME_TABLE_29_09;	break;
		case 10: dwRet = FRAME_TABLE_29_10;	break;
		case 11: dwRet = FRAME_TABLE_29_11;	break;
		case 12: dwRet = FRAME_TABLE_29_12;	break;
		case 13: dwRet = FRAME_TABLE_29_13;	break;
		case 14: dwRet = FRAME_TABLE_29_14;	break;
		case 15: dwRet = FRAME_TABLE_29_15;	break;
		case 16: dwRet = FRAME_TABLE_29_16;	break;
		case 17: dwRet = FRAME_TABLE_29_17;	break;
		case 18: dwRet = FRAME_TABLE_29_18;	break;
		case 19: dwRet = FRAME_TABLE_29_19;	break;
		case 20: dwRet = FRAME_TABLE_29_20;	break;
		case 21: dwRet = FRAME_TABLE_29_21;	break;
		case 22: dwRet = FRAME_TABLE_29_22;	break;
		case 23: dwRet = FRAME_TABLE_29_23;	break;
		case 24: dwRet = FRAME_TABLE_29_24;	break;
		case 25: dwRet = FRAME_TABLE_29_25;	break;
		case 26: dwRet = FRAME_TABLE_29_26;	break;
		case 27: dwRet = FRAME_TABLE_29_27;	break;
		case 28: dwRet = FRAME_TABLE_29_28;	break;
		case 29: dwRet = FRAME_TABLE_29_29;	break;
		}
		break;
	}
	case 30:{
		switch(to) {
		case 1:	 dwRet = FRAME_TABLE_30_01; break;
		case 2:	 dwRet = FRAME_TABLE_30_02; break;
		case 3:	 dwRet = FRAME_TABLE_30_03; break;
		case 4:	 dwRet = FRAME_TABLE_30_04; break;
		case 5:	 dwRet = FRAME_TABLE_30_05; break;
		case 6:  dwRet = FRAME_TABLE_30_06;	break;
		case 7:  dwRet = FRAME_TABLE_30_07;	break;
		case 8:  dwRet = FRAME_TABLE_30_08;	break;
		case 9:  dwRet = FRAME_TABLE_30_09;	break;
		case 10: dwRet = FRAME_TABLE_30_10;	break;
		case 11: dwRet = FRAME_TABLE_30_11;	break;
		case 12: dwRet = FRAME_TABLE_30_12;	break;
		case 13: dwRet = FRAME_TABLE_30_13;	break;
		case 14: dwRet = FRAME_TABLE_30_14;	break;
		case 15: dwRet = FRAME_TABLE_30_15;	break;
		case 16: dwRet = FRAME_TABLE_30_16;	break;
		case 17: dwRet = FRAME_TABLE_30_17;	break;
		case 18: dwRet = FRAME_TABLE_30_18;	break;
		case 19: dwRet = FRAME_TABLE_30_19;	break;
		case 20: dwRet = FRAME_TABLE_30_20;	break;
		case 21: dwRet = FRAME_TABLE_30_21;	break;
		case 22: dwRet = FRAME_TABLE_30_22;	break;
		case 23: dwRet = FRAME_TABLE_30_23;	break;
		case 24: dwRet = FRAME_TABLE_30_24;	break;
		case 25: dwRet = FRAME_TABLE_30_25;	break;
		case 26: dwRet = FRAME_TABLE_30_26;	break;
		case 27: dwRet = FRAME_TABLE_30_27;	break;
		case 28: dwRet = FRAME_TABLE_30_28;	break;
		case 29: dwRet = FRAME_TABLE_30_29;	break;
		case 30: dwRet = FRAME_TABLE_30_30;	break;
		}
		break;
	}
	}

	return dwRet;
}
