//////////////////////////////////////////////////////////////////////////
// dbg.h
// Ϣ
//////////////////////////////////////////////////////////////////////////

#ifndef __DBG_H__
#define __DBG_H__

#ifdef _DBG
void __trace(LPCTSTR fmt, ...);
void __traced(LPCTSTR file, int line, LPCTSTR fmt, ...);
#define _assert(x) {if(!(x)) _asm{int 0x03}}
#define _verify(x) {if(!(x)) _asm{int 0x03}}
#else
#define _assert(x)
#define _verify(x) x
#endif

#ifdef _DBG
#define _trace  __trace
#define _traced __traced
#else
inline void __trace(LPCTSTR fmt, ...) { }
inline void __traced(LPCTSTR file, int line, LPCTSTR fmt, ...) { }
#define _trace  1 ? (void)0 : __trace
#define _traced 1 ? (void)0 : __traced
#endif

//
#ifdef _DBG

typedef struct tagCount {
	ULONG tStart;
	ULONG nCount;
	ULONG nInterval;
}TCount, *PCount;
#define _cParam(cnt) cnt,
void __cstart(PCount cnt, ULONG interval);
void __ccount(PCount cnt, LPCTSTR fmt, ...);
void __ccountd(PCount cnt, LPCTSTR file, int line, LPCTSTR fmt, ...);
#define _cObj(obj)		TCount obj
#define _cObjs(objs, n)	TCount objs[n]
#define _cstart			__cstart
#define _ccount			__ccount
#define _ccountd		__ccountd

#define DbgCount(cnt, nm) {		\
	static LONG _cnt##cnt=0;			\
	static ULONG _tkLast##cnt = 0;		\
	ULONG _tkNow##cnt = GetTickCount();	\
	_cnt##cnt++;						\
	if((_tkNow##cnt - _tkLast##cnt) >= 1000) {\
		_trace(TEXT("%s=%d\n"), nm, _cnt##cnt);	\
		_tkLast##cnt = _tkNow##cnt;			\
		_cnt##cnt = 0;					\
	}									\
}

#define DbgInterval(diff, nm) {			\
	static ULONG _tkLast##diff = 0;		\
	ULONG _tkNow##diff = GetTickCount();\
	_trace(TEXT("%s=%d\n"), nm, (_tkNow##diff-_tkLast##diff));	\
	_tkLast##diff = _tkNow##diff;		\
}

#define DbgCountRun(rct, cnt, ret) 		\
	static int _iRun##rct = 0;			\
	if(_iRun##rct<cnt) _iRun##rct++;	\
	else {ret}							\

#else

#define _cParam(cnt)
inline void __cstart(_cParam(cnt) ULONG interval) { }
inline void __ccount(_cParam(cnt) LPCTSTR fmt, ...) { }
inline void __ccountd(_cParam(cnt) LPCTSTR file, int line, LPCTSTR fmt, ...) { }
#define _cObj(obj)		
#define _cObjs(objs, n)	
#define _cstart  1 ? (void)0 : __cstart
#define _ccount  1 ? (void)0 : __ccount
#define _ccountd 1 ? (void)0 : __ccountd
#define DbgCount(cnt, nm)
#define DbgInterval(diff, nm)
#define DbgCountRun(rct, cnt, ret)

#endif

#endif // __DBG_H__